/*
** Copyright (c) 1998, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished  -
** rights reserved under the Copyright Laws of the United States.
**
*/

#include "conform.h"
#include "vertex.h"
#include "vmglide.h"

static char *test_name = "alpha-combine";
static char *test_description = "Test alpha combine unit";

// Tolerance for auto compares
// 5 bits - six bits fails on Voodoo II
#define TOLERANCE (1.0f/32.0f)

// local data
typedef struct {
  GrVertex verts[4];  // Temporary vertex storage
  GrTexInfo info;
  GrChipID_t curtmu;
  int debug;
} test_data;

// Prototypes
void setup_prim(conform_state *state);
void texture_gen(test_data *local_ptr);
void word_texture(FxU16 *tmp, int height, int width);
TestResult check_frame(conform_state *state,
                       FxU32 *screen,
                       GrCombineFunction_t function,
                       GrCombineFactor_t factor,
                       GrCombineLocal_t local,
                       GrCombineOther_t other,
                       FxBool invert);

#define NUM_COMBINE 566

struct {
  GrCombineFunction_t function;
  GrCombineFactor_t factor;
  GrCombineLocal_t local;
  GrCombineOther_t other;
  FxBool invert;
} combine[NUM_COMBINE] = {
  // ZERO, NONE
  GR_COMBINE_FUNCTION_ZERO,
  GR_COMBINE_FACTOR_NONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // ZERO, NONE, INVERT
  GR_COMBINE_FUNCTION_ZERO,
  GR_COMBINE_FACTOR_NONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_NONE,
  FXTRUE,

  // LOCAL, NONE
  GR_COMBINE_FUNCTION_LOCAL,
  GR_COMBINE_FACTOR_NONE,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  GR_COMBINE_FUNCTION_LOCAL,
  GR_COMBINE_FACTOR_NONE,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  GR_COMBINE_FUNCTION_LOCAL,
  GR_COMBINE_FACTOR_NONE,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // LOCAL, ONE, TEXTURE
  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // LOCAL_ALPHA, NONE
  GR_COMBINE_FUNCTION_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_NONE,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  GR_COMBINE_FUNCTION_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_NONE,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  GR_COMBINE_FUNCTION_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_NONE,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  GR_COMBINE_FUNCTION_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_NONE,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  GR_COMBINE_FUNCTION_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_NONE,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  GR_COMBINE_FUNCTION_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_NONE,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // SCALE_OTHER, ONE
  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  // SCALE_OTHER, ZERO
  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_ZERO,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // {SCALE,BLEND}_OTHER, LOCAL
  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {SCALE,BLEND}_OTHER, OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {SCALE,BLEND}_OTHER, LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {SCALE,BLEND}_OTHER, TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {SCALE,BLEND}_OTHER, ONE_MINUS_LOCAL
  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {SCALE,BLEND}_OTHER, ONE_MINUS_OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {SCALE,BLEND}_OTHER, ONE_MINUS_LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {SCALE,BLEND}_OTHER, ONE_MINUS_TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_OTHER,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL, ONE
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL, ZERO
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ZERO,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL, LOCAL
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL, OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL, LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL, TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL, ONE_MINUS_LOCAL
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL, ONE_MINUS_OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL, ONE_MINUS_LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL, ONE_MINUS_TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL_ALPHA, ONE
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL_ALPHA, ZERO
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ZERO,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL_ALPHA, LOCAL
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL_ALPHA, OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL_ALPHA, LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL_ALPHA, TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL_ALPHA, ONE_MINUS_LOCAL
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL_ALPHA, ONE_MINUS_OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL_ALPHA, ONE_MINUS_LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_ADD_LOCAL_ALPHA, ONE_MINUS_TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL, ONE
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL, ZERO
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ZERO,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL, LOCAL
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL, OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL, LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL, TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL, ONE_MINUS_LOCAL
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL, ONE_MINUS_OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL, ONE_MINUS_LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL, ONE_MINUS_TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {BLEND, SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL}, ONE
  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  // {BLEND, SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL}, ZERO
  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ZERO,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // {BLEND, SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL}, LOCAL
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {BLEND, SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL}, OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {BLEND, SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL}, LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {BLEND, SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL}, TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {BLEND, SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL}, ONE_MINUS_LOCAL
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {BLEND, SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL}, ONE_MINUS_OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {BLEND, SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL}, ONE_MINUS_LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {BLEND, SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL}, ONE_MINUS_TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA, ONE
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA, ZERO
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ZERO,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA, LOCAL
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA, OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA, LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA, TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA, ONE_MINUS_LOCAL
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA, ONE_MINUS_OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA, ONE_MINUS_LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA, ONE_MINUS_TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {BLEND_LOCAL, SCALE_MINUS_LOCAL_ADD_LOCAL}, ONE
  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  // {BLEND_LOCAL, SCALE_MINUS_LOCAL_ADD_LOCAL}, ZERO
  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ZERO,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // {BLEND_LOCAL, SCALE_MINUS_LOCAL_ADD_LOCAL}, LOCAL
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // {BLEND_LOCAL, SCALE_MINUS_LOCAL_ADD_LOCAL}, OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {BLEND_LOCAL, SCALE_MINUS_LOCAL_ADD_LOCAL}, LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // {BLEND_LOCAL, SCALE_MINUS_LOCAL_ADD_LOCAL}, TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,
  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // {BLEND_LOCAL, SCALE_MINUS_LOCAL_ADD_LOCAL}, TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {BLEND_LOCAL, SCALE_MINUS_LOCAL_ADD_LOCAL}, ONE_MINUS_LOCAL
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // {BLEND_LOCAL, SCALE_MINUS_LOCAL_ADD_LOCAL}, ONE_MINUS_OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // {BLEND_LOCAL, SCALE_MINUS_LOCAL_ADD_LOCAL}, ONE_MINUS_LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // {BLEND_LOCAL, SCALE_MINUS_LOCAL_ADD_LOCAL}, ONE_MINUS_TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_BLEND_LOCAL,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA, ONE
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  // SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA, ZERO
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ZERO,
  GR_COMBINE_LOCAL_NONE,  
  GR_COMBINE_OTHER_NONE,
  FXFALSE,

  // SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA, LOCAL
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA, OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA, LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA, TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA, ONE_MINUS_LOCAL
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA, ONE_MINUS_OTHER_ALPHA
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA, ONE_MINUS_LOCAL_ALPHA
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  // SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA, ONE_MINUS_TEXTURE_ALPHA
  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_CONSTANT,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_ITERATED,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_CONSTANT,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_ITERATED,
  FXFALSE,

  GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA,
  GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
  GR_COMBINE_LOCAL_DEPTH,  
  GR_COMBINE_OTHER_TEXTURE,
  FXFALSE,
};

int init_conform(conform_state *state, int local_argc, char **local_argv)
{
  test_data *local_ptr;
  int num_tmu;

  state->test_name = test_name;
  state->test_description = test_description;

  if((state->local_data = malloc(sizeof(test_data))) == NULL) {
    // Could not allocate data for local memory
    log_perror(state,"Could not allocate memory for local data");
    return(0);
  } 

  local_ptr = (test_data *)state->local_data;
  local_ptr->debug = 0;

  // pick a random TMU
  num_tmu = nrand((int)(state->num_tmu));

  // print out test dependent args
  ++local_argv; // skip argv[0] (the test name)
  while(--local_argc) {
    if(!_stricmp(*local_argv,"-tmu") || // TMU to use
       !_stricmp(*local_argv,"/tmu")) {
      ++local_argv;
      if(!sscanf(*local_argv,"%d",&num_tmu)) {
        // bad argument
      } 
      --local_argc;
    } else if(!_stricmp(*local_argv,"-debug") || // enable debug output
              !_stricmp(*local_argv,"/debug")) {
      local_ptr->debug = 1;
    } else {
      fprintf(stderr, "Invalid argument '%s' - ignored.\n", *local_argv);
    }
    ++local_argv;
  }

  // Convert numeric TMU to argument for glide
  if(num_tmu >= state->num_tmu) {
    num_tmu = 0;
  }
  switch(num_tmu) {
  case 1:
    local_ptr->curtmu = GR_TMU1;
    break;
  case 2:
    local_ptr->curtmu = GR_TMU2;
    break;
  default:
    local_ptr->curtmu = GR_TMU0;
    break;
  }

  // don't use any visuals with aux buffers
  state->num_aux_bufs = 0;

  // use 100x100 image area
  state->conform_width = 100;
  state->conform_height = 100;

  // tiling not supported
  state->tile = 0;

  // xycheck/xylimit not supported.
  state->xycheck = 0;
  state->xylimit = 0;

  // setup image file names
  // this is manditory
  frame_name(test_name, state);
  img_file_type(IMG_P6,state);

  // Initialization complete
  return(-1);
}

void close_conform(conform_state *state)
{
  // Free up local data
  free(state->local_data);
}

int do_conform(conform_state *state)
{
  test_data *local_ptr;
  GrColor_t clearcolor;
  int num_tmu;
  int comb;
  FxU32 *screen;
  char outstr[128];

  local_ptr = (test_data *)state->local_data;
  
  // if multiple reps - randomize
  if(state->num_reps != 1) {
    // randomize TMU
    num_tmu = nrand(state->num_tmu);
    switch(num_tmu) {
    case 1:
      local_ptr->curtmu = GR_TMU1;
      break;
    case 2:
      local_ptr->curtmu = GR_TMU2;
      break;
    default:
      local_ptr->curtmu = GR_TMU0;
      break;
    }
  }
  
  // write record of TMU to loog file
  switch(local_ptr->curtmu) {
  case 1:
    log_message(state,"Using TMU 1");
    break;
  case 2:
    log_message(state,"Using TMU 2");
    break;
  default:
    log_message(state,"Using TMU 0");
    break;
  }
  
  // allocate screen for automated compare
  if((screen = (FxU32 *)malloc(state->conform_width *
                               state->conform_height *
                               sizeof(FxU32))) == NULL) {
    // fail
    log_message(state,"Could not allocate memory buffer for screen image");
  }

  for(comb = 0; comb < NUM_COMBINE; comb++) {
    // clear front buffer to lightgrey
    clearcolor = rgbacolor(state, 0x2f2f2fff);
    grRenderBuffer(GR_BUFFER_FRONTBUFFER);
    grColorMask(FXTRUE,FXFALSE);
    grDepthMask(FXFALSE);
    grCoordinateSpace(GR_WINDOW_COORDS);
    grClipWindow(state->conform_x,
                 state->conform_y,
                 state->conform_x+state->conform_width,
                 state->conform_y+state->conform_height);
    
    grViewport(state->conform_x, state->conform_y, 
               state->conform_width, state->conform_height);
    
    setup_prim(state);
    
    grBufferClear(clearcolor,0x0000,0x00000000);
    
    // Init vertex data
    local_ptr->verts[0].x = 18.0f;
    local_ptr->verts[0].y = 18.0f;
    local_ptr->verts[0].ooz = 0.0f;
    local_ptr->verts[1].x = 18.0f + 64.0f;
    local_ptr->verts[1].y = 18.0f;
    local_ptr->verts[1].ooz = 65535.0f;
    local_ptr->verts[2].x = 18.0f + 64.0f;
    local_ptr->verts[2].y = 18.0f + 64.0f;
    local_ptr->verts[2].ooz = 65535.0f;
    local_ptr->verts[3].x = 18.0f;
    local_ptr->verts[3].y = 18.0f + 64.0f;
    local_ptr->verts[3].ooz = 0.0f;
    
    // Init constant color to white & 1/2 alpha
    grConstantColorValue(rgbacolor(state,0xffffff7f));
    
    // Init Interpolated colors
    local_ptr->verts[0].a = 255.0f;
    local_ptr->verts[0].r = 0.0f;
    local_ptr->verts[0].g = 0.0f;
    local_ptr->verts[0].b = 0.0f;
    local_ptr->verts[1].a = 0.0f;
    local_ptr->verts[1].r = 0.0f;
    local_ptr->verts[1].g = 0.0f;
    local_ptr->verts[1].b = 0.0f;
    local_ptr->verts[2].a = 0.0f;
    local_ptr->verts[2].r = 255.0f;
    local_ptr->verts[2].g = 255.0f;
    local_ptr->verts[2].b = 255.0f;
    local_ptr->verts[3].a =255.0f;
    local_ptr->verts[3].r = 255.0f;
    local_ptr->verts[3].g = 255.0f;
    local_ptr->verts[3].b = 255.0f;
    
    // Choose alphacombine
    grAlphaCombine(combine[comb].function,
                   combine[comb].factor,
                   combine[comb].local,
                   combine[comb].other,
                   combine[comb].invert);

    // Draw square
    grDrawTriangle(&(local_ptr->verts[0]), 
                   &(local_ptr->verts[1]), 
                   &(local_ptr->verts[2]));
    
    grDrawTriangle(&(local_ptr->verts[0]), 
                   &(local_ptr->verts[2]), 
                   &(local_ptr->verts[3]));
    
    end_frame(END_OP_SAVE_TO_MEM, 
              state, 
              (void **)&screen, 
              END_BUFFER_FRONTBUFFER);

    // print out frame number
    sprintf(outstr,"frame = %d",comb);
    log_message(state,outstr);

    if (!log_status(state,
                    check_frame(state, screen,
                                combine[comb].function,
                                combine[comb].factor,
                                combine[comb].local,
                                combine[comb].other,
                                combine[comb].invert),
                    "screen vs. emulated")) {
      free(screen);
      return(TEST_FAIL);
    }
  }

  free(screen);

  return(TEST_PASS);
}

void help_conform(conform_state *state)
{
  fprintf(stderr,"   -tmu <n>   texture unit to test\n");
}


void setup_prim(conform_state *state)
{
  test_data *local_ptr;
  GrChipID_t passtmu;
  FxU32 baseaddr;
  FxU16 *buffer;
  int i;

  local_ptr = (test_data *)state->local_data;

  grVertexLayout(GR_PARAM_XY,  GR_VERTEX_X_OFFSET << 2, GR_PARAM_ENABLE);  
  grVertexLayout(GR_PARAM_Q , GR_VERTEX_OOW_OFFSET << 2, GR_PARAM_ENABLE);
  grVertexLayout(GR_PARAM_Z, GR_VERTEX_OOZ_OFFSET << 2, GR_PARAM_ENABLE);
  grVertexLayout(GR_PARAM_A,  GR_VERTEX_A_OFFSET << 2, GR_PARAM_ENABLE);  
  grVertexLayout(GR_PARAM_RGB,  GR_VERTEX_R_OFFSET << 2, GR_PARAM_ENABLE);  
  // set for max 3 TMUs for now
  grVertexLayout(GR_PARAM_ST0, GR_VERTEX_SOW_TMU0_OFFSET << 2,
                 GR_PARAM_ENABLE);
  grVertexLayout(GR_PARAM_Q0, GR_VERTEX_OOW_TMU0_OFFSET << 2,
                 GR_PARAM_ENABLE);
  grVertexLayout(GR_PARAM_ST1, GR_VERTEX_SOW_TMU1_OFFSET << 2,
                 GR_PARAM_ENABLE);
  grVertexLayout(GR_PARAM_Q1, GR_VERTEX_OOW_TMU1_OFFSET << 2,
                 GR_PARAM_ENABLE);
  grVertexLayout(GR_PARAM_ST2, GR_VERTEX_SOW_TMU2_OFFSET << 2,
                 GR_PARAM_ENABLE);
  grVertexLayout(GR_PARAM_Q2, GR_VERTEX_OOW_TMU2_OFFSET << 2,
                 GR_PARAM_ENABLE);

  // Choose matching colorcombine
  grColorCombine(GR_COMBINE_FUNCTION_LOCAL,
                 GR_COMBINE_FACTOR_NONE,
                 GR_COMBINE_LOCAL_CONSTANT,
                 GR_COMBINE_OTHER_NONE,
                 FXFALSE);        
    
  grAlphaBlendFunction(GR_BLEND_SRC_ALPHA,
                       GR_BLEND_ZERO,
                       GR_BLEND_ONE,
                       GR_BLEND_ZERO);

  grAlphaTestReferenceValue(0x00);
  grAlphaTestFunction(GR_CMP_ALWAYS);

  // Setup other TMUs to pass thru
  for(i = 0; i< local_ptr->curtmu; i++) {
    switch(i) {
    case 1:
      passtmu = GR_TMU1;
      break;
    case 2:
      passtmu = GR_TMU2;
      break;
    default:
      passtmu = GR_TMU0;
      break;
    }
    grTexCombine(passtmu,
                 GR_COMBINE_FUNCTION_SCALE_OTHER,
                 GR_COMBINE_FACTOR_ONE,
                 GR_COMBINE_FUNCTION_SCALE_OTHER,
                 GR_COMBINE_FACTOR_ONE,
                 FXFALSE, FXFALSE );
  }

  grTexCombine(local_ptr->curtmu,
               GR_COMBINE_FUNCTION_LOCAL,
               GR_COMBINE_FACTOR_NONE,
               GR_COMBINE_FUNCTION_LOCAL,
               GR_COMBINE_FACTOR_ONE,
               FXFALSE, FXFALSE );
  
  grTexClampMode(local_ptr->curtmu, 
                 GR_TEXTURECLAMP_WRAP, GR_TEXTURECLAMP_WRAP);

  grTexFilterMode(local_ptr->curtmu,
                  GR_TEXTUREFILTER_POINT_SAMPLED,
                  GR_TEXTUREFILTER_POINT_SAMPLED );

  grTexMipMapMode(local_ptr->curtmu,
                  GR_MIPMAP_NEAREST,
                  FXFALSE );

  grTexLodBiasValue(local_ptr->curtmu, 0.0f);

  // Init texture
  if((buffer = (FxU16 *)malloc(64*64*2)) == NULL) {
    // fail
    log_message(state,"Could not allocate memory buffer for texture map");
  }
  
  local_ptr->info.data = buffer;
  texture_gen(local_ptr);
  
  // get random texture address
  baseaddr = texture_base(state, local_ptr->curtmu, GR_MIPMAPLEVELMASK_BOTH, &(local_ptr->info));
  
  // download texture
  grTexDownloadMipMap(local_ptr->curtmu,
                      baseaddr,
                      GR_MIPMAPLEVELMASK_BOTH,
                      &(local_ptr->info));
  
  grTexSource(local_ptr->curtmu,
              baseaddr,
              GR_MIPMAPLEVELMASK_BOTH,
              &(local_ptr->info));

  free(buffer);
}

void texture_gen(test_data *local_ptr)
{
  int width;
  int height;
  int i;

  height = 64;
  width = 64;
  local_ptr->info.aspectRatioLog2 = GR_ASPECT_LOG2_1x1;
  
  local_ptr->info.format = GR_TEXFMT_ALPHA_INTENSITY_88;
  word_texture((FxU16 *)local_ptr->info.data,
               height, width);
  
  local_ptr->info.smallLodLog2 = GR_LOD_LOG2_64;
  local_ptr->info.largeLodLog2 = GR_LOD_LOG2_64;

  // setup vertex data for all TMUs
  for(i = 0; i < GRVERTEX_NUM_TMU; i++) {
    local_ptr->verts[0].tmuvtx[i].sow = 0.0f;
    local_ptr->verts[0].tmuvtx[i].tow =0.0f;
    local_ptr->verts[0].oow = 1.0f;  
    local_ptr->verts[0].tmuvtx[i].oow = 1.0f;
    
    local_ptr->verts[1].tmuvtx[i].sow = 256.0f;
    local_ptr->verts[1].tmuvtx[i].tow = 0.0f;
    local_ptr->verts[1].oow = 1.0f;  
    local_ptr->verts[1].tmuvtx[i].oow = 1.0f;
    
    local_ptr->verts[2].tmuvtx[i].sow = 256.0f;
    local_ptr->verts[2].tmuvtx[i].tow = 256.0f;
    local_ptr->verts[2].oow = 1.0f;  
    local_ptr->verts[2].tmuvtx[i].oow = 1.0f;
    
    local_ptr->verts[3].tmuvtx[i].sow = 0.0f;
    local_ptr->verts[3].tmuvtx[i].tow = 256.0f;
    local_ptr->verts[3].oow = 1.0f;  
    local_ptr->verts[3].tmuvtx[i].oow = 1.0f;
  }
}

void word_texture(FxU16 *tmp, int height, int width)
{
  int i,j;

  for(i=0; i<height; i++) {
    for(j=0; j<width ; j++) {
      *tmp++ = (j*(256/width))|((i*(256/height))<<8);
    }
  }
}

TestResult check_frame(conform_state *state,
                      FxU32 *screen,
                      GrCombineFunction_t function,
                      GrCombineFactor_t factor,
                      GrCombineLocal_t local,
                      GrCombineOther_t other,
                      FxBool invert)
{
  FxU32 *tmp, *screen_top;
  TestResult status = TEST_PASS;
  int failed;
  int x, y;
  char outstr[128];
  Color4D expected_color;
  float expected_alpha;
  test_data *local_ptr = (test_data *)state->local_data;


  failed = 0;
  screen_top = tmp = screen;
  tmp += 18 +(18 * state->conform_width);
  // go to first active pixel
  for(y = 0; y < 64; y++) {
    for(x = 0; x < 64; x++) {
      vmgrAlphaCombine(function,
                       factor,
                       local,
                       other,
                       invert,
                       &expected_alpha,
                       ((float)(63-x))/63.0f, // iterated (inverse x)
                       0.5f, // constant
                       ((float)(y))/63.0f, // texture (inverse y)
                       ((float)x)/63.0f); // zbuffer (x)

      // This test only works because we're using gray ramps.     
      expected_color.red = expected_alpha;
      expected_color.green = expected_alpha;
      expected_color.blue = expected_alpha;
      expected_color.alpha = 0.0f;
      if (!compare_color(state, *tmp, &expected_color)) {
        sprintf(outstr, "@(%x %x), actual alpha %x != expected %x",x, y,
                (int)(expected_color.green*255.5f),
                (int)(expected_alpha*255.5f));
        log_message(state,outstr);
        status = TEST_FAIL;

        if (local_ptr->debug) {
          *tmp = 0x00000fff; // debug, mark mismatched pixels
        }
      }
      ++tmp;      
    }
    // skip border
    tmp += state->conform_width-64;
  }

  if (local_ptr->debug) {
    char filename[128];
    sprintf(filename, "%s/emu-%s-%02d", 
            state->output_dir, test_name, state->framecnt-1);
    save_image_to_file(state, filename, screen_top,
                       state->conform_width, state->conform_height,
                       FXTRUE);
  }


  return(status);
}
