/*
** Copyright (c) 2000, 3dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/

#include "C2DPerfView.h"
#include "ControlPanelPPob.h"
#include "hrm_user_settings.h"

#include <LCaption.h>
#include <LPopupButton.h>
#include <LMenu.h>


/*______________________________________________________________________________

	Constructor

*/

C2DPerfView::C2DPerfView()
{
	
}


/*______________________________________________________________________________

	Constructor (using a Stream, that's the one that will be used)

*/

C2DPerfView::C2DPerfView(
	LStream *			inStream)
	: LView ( inStream )
{
	StartListening();
}


/*______________________________________________________________________________

	Destructor

*/

C2DPerfView::~C2DPerfView()
{
}


#pragma mark -




/*______________________________________________________________________________

	SetPCIHardware

*/

void
C2DPerfView::SetPCIHardware(
	CPCIHardware *		inBoard)
{
	Boolean		NoErr;
	
	if ( inBoard != 0 )
	{
		mBoard = inBoard;

		/*	Load 2D perf views objects	*/
		NoErr = Load2DPerfView();
	}

}


/*______________________________________________________________________________

	Load2DPerfView

*/

Boolean
C2DPerfView::Load2DPerfView()
{
	Boolean 		result = true;
	
	// Gather the informations
	result = SetPopupButtonContentFromHRM (perf2D_QDAccel, HRM_2D_AccelMode);
	//if (!result)		return false;
	result = SetPopupButtonContentFromHRM (perf2D_FontCaching, HRM_2D_FontCaching);
	//if (!result)		return false;
	result = SetPopupButtonContentFromHRM (perf2D_PictCaching, HRM_2D_PictCaching);
	//if (!result)		return false;
	result = SetPopupButtonContentFromHRM (perf2D_QTAccel, HRM_QT_AccelMode);
	//if (!result)		return false;
	result = SetPopupButtonContentFromHRM (perf2D_MPEGAccel, HRM_QT_MPEGMode);
	//if (!result)		return false;
	
	
	
	if (mVVersion == 3)
	{
		/*	Disable some popupmenu 	*/
		LPopupButton * theFCacheButton = (LPopupButton*) this->FindPaneByID(perf2D_FontCaching);
		theFCacheButton->SetCurrentMenuItem( kDisabled );
		theFCacheButton->Disable();
		
		LPopupButton * thePCacheButton = (LPopupButton*) this->FindPaneByID(perf2D_PictCaching);
		thePCacheButton->SetCurrentMenuItem( kDisabled );
		thePCacheButton->Disable();
		
		LPopupButton * theQTButton = (LPopupButton*) this->FindPaneByID(perf2D_QTAccel);
		theQTButton->SetCurrentMenuItem( kDisabled );
		theQTButton->Disable();
		
		LPopupButton * theMPEGButton = (LPopupButton*) this->FindPaneByID(perf2D_MPEGAccel);
		theMPEGButton->SetCurrentMenuItem( kDisabled );
		theMPEGButton->Disable();
		
		/*	Disable some option in the PopupMenu	*/
		LPopupButton * theQDButton = (LPopupButton*) this->FindPaneByID(perf2D_QDAccel);
		theQDButton->SetCurrentMenuItem( kBasic );
		MenuHandle menuH = theQDButton->GetMacMenuH();
		::DisableItem(menuH, 3);
	}
	
	
	return result;
}



/*______________________________________________________________________________

	FinisCreateSelf
	
*/

void
C2DPerfView::FinishCreateSelf()
{
	LPopupButton * theQDButton = (LPopupButton*) this->FindPaneByID(perf2D_QDAccel);
	if (theQDButton != nil)		theQDButton->AddListener(this);
	
	LPopupButton * theFCacheButton = (LPopupButton*) this->FindPaneByID(perf2D_FontCaching);
	if (theFCacheButton != nil)		theFCacheButton->AddListener(this);
	
	LPopupButton * thePCacheButton = (LPopupButton*) this->FindPaneByID(perf2D_PictCaching);
	if (thePCacheButton != nil)		thePCacheButton->AddListener(this);
	
	LPopupButton * theQTButton = (LPopupButton*) this->FindPaneByID(perf2D_QTAccel);
	if (theQTButton != nil)		theQTButton->AddListener(this);
	
	LPopupButton * theMPEGButton = (LPopupButton*) this->FindPaneByID(perf2D_MPEGAccel);
	if (theMPEGButton != nil)		theMPEGButton->AddListener(this);
	
	

}



/*______________________________________________________________________________

	ListenToMessage
	
*/

void					
C2DPerfView::ListenToMessage(
	MessageT 			inMessage, 
	void *				ioParam)
{
	bool		res;
	UInt32 		longParam = *(UInt32*)ioParam;
	
	if ( mBoard )
	{
		switch(inMessage) {
			case msg_2DPerf_QD:
			{
				res = mBoard->SetSettingUL(HRM_2D_AccelMode, *((UInt32*)ioParam));
				break;
			}
			
			case msg_2DPerf_FontCach:
			{
				res = mBoard->SetSettingUL(HRM_2D_FontCaching, *((UInt32*)ioParam));
				break;
			}
			
			case msg_2DPerf_PictCach:
			{
				res = mBoard->SetSettingUL(HRM_2D_PictCaching, *((UInt32*)ioParam));
				break;
			}
			
			case msg_2DPerf_QT:
			{
				res = mBoard->SetSettingUL(HRM_QT_AccelMode, *((UInt32*)ioParam));
				break;
			}
			
			case msg_2DPerf_MPEG:
			{	
				res = mBoard->SetSettingUL(HRM_QT_MPEGMode, *((UInt32*)ioParam));
				break;
			}
			
			default:
				break;
		}
	}
}


/*______________________________________________________________________________

	SetPopupButtonContentFromHRM
*/

Boolean
C2DPerfView::SetPopupButtonContentFromHRM(
	PaneIDT		inPane,
	char *		inSettingName)
{
	Boolean		res = true;
	UInt32 		outputSetting;
	
	LPopupButton * 	thePopupButton = (LPopupButton*) this->FindPaneByID(inPane);
	
	if ( mBoard )
	{
		res = mBoard->GetSettingUL(inSettingName, &outputSetting);
	}
	
	if ( res ) thePopupButton->SetCurrentMenuItem( outputSetting );
	
	return res;
}


/*______________________________________________________________________________

	SetVoodooVersion

*/

void
C2DPerfView::SetVoodooVersion(
	int	 		inValue)
{
	mVVersion = inValue;
}

