/*
** Copyright (c) 2000, 3dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/

#include "C3DPerfView.h"

#include "COpenGLView.h"
#include "CGlideView.h"
#include "CRaveView.h"

#include "ControlPanelPPob.h"

#include <LCaption.h>
#include <LPopupGroupBox.h>
#include <LPushButton.h>
#include <LPopupButton.h>
#include <LMultiPanelView.h>


/*______________________________________________________________________________

	Constructor

*/

C3DPerfView::C3DPerfView()
{
	
}


/*______________________________________________________________________________

	Constructor (using a Stream, that's the one that will be used)

*/

C3DPerfView::C3DPerfView(
	LStream *			inStream)
	: LView ( inStream )
{
	StartListening();
}


/*______________________________________________________________________________

	Destructor

*/

C3DPerfView::~C3DPerfView()
{
}


#pragma mark -



/*______________________________________________________________________________

	FinishCreateSelf

*/
void
C3DPerfView::FinishCreateSelf()
{
	// Load the PopuGroupBox and its MultiPanelView for the 3DPerf View
	//LPopupGroupBox*		perf3DPGB = (LPopupGroupBox*) this->FindPaneByID(perf3D_PopupGroupBox);
	//if (perf3DPGB != nil)	perf3DPGB->AddListener(this);	
			
	LMultiPanelView*	perf3DMPV = (LMultiPanelView*) this->FindPaneByID(perf3D_MultiPanelView);
	if (perf3DMPV != nil)		perf3DMPV->CreateAllPanels();
	
	LPushButton*		theTestButton = (LPushButton*) this->FindPaneByID(perf3D_TestButton);
	if (theTestButton != nil) theTestButton->AddListener(this);
}


/*______________________________________________________________________________

	SetPCIHardware

*/

void
C3DPerfView::SetPCIHardware(
	CPCIHardware *		inBoard)
{
	if ( inBoard != 0 ) 
	{
		mBoard = inBoard;
		
		Load3DPerfView();
	}

}


/*______________________________________________________________________________

	Load3DPerfView
*/

Boolean
C3DPerfView::Load3DPerfView()
{
	Boolean	result = true;
	
	mMultiPanelView = dynamic_cast<LMultiPanelView*> (this->FindPaneByID(perf3D_MultiPanelView));
	mMultiPanelView->SwitchToPanel(index_3DPerf_Glide);
	
	COpenGLView* theOpenGLView = (COpenGLView*) this->FindPaneByID(perf3D_OpenGLView);
	theOpenGLView->SetPCIHardware( mBoard );
	
	CGlideView* theGlideView = (CGlideView*) this->FindPaneByID(perf3D_GlideView);
	theGlideView->SetVoodooVersion( mVVersion );
	theGlideView->SetPCIHardware( mBoard );
	
	CRaveView* theRaveView = (CRaveView*) this->FindPaneByID(perf3D_RaveView);
	theRaveView->SetPCIHardware( mBoard );
	
	return result;
}


/*______________________________________________________________________________

	ListenToMessage

*/

void					
C3DPerfView::ListenToMessage(
	MessageT 			inMessage, 
	void *				ioParam)
{
	UInt32 longParam = *(UInt32*)ioParam;
	
	switch (inMessage) {
		case msg_3DPerf_PGB:
			mMultiPanelView->SwitchToPanel(longParam);
			break;
		
		case msg_3DPerf_TestButton:
			int i =0;
			break;
		
		default:
			break;
	}

}


/*______________________________________________________________________________

	SetVoodooVersion

*/

void
C3DPerfView::SetVoodooVersion(
	int	 		inValue)
{
	mVVersion = inValue;
}


