/*
** Copyright (c) 2000, 3dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/


#include "CCheckRom.h"
#include "CControlPanelApp.h"
#include "CFlashRomWindow.h"
#include "ControlPanelPPob.h"
#include "3dfx.h"

#include <stdio.h>
#include <String.h>
#include <Types.h>
#include <Resources.h>
#include <DriverFamilyMatching.h>

#define		kMajor				4
#define		kMinor				3
#define		kStage				2
#define		kRev				1

#define		kVoodooName			"Voodoo"	

static List_t RomUList = { 0, 0, 0 };


/*______________________________________________________________________________

	Constructor 	[CAboutWindow class]

*/

CCheckRom::CCheckRom(
	CPCIHardware *			inBoard,
	LCommander *			inCommander)
{
	mBoard = inBoard;
	mCommander = inCommander;
}


/*______________________________________________________________________________

	Destructor

*/

CCheckRom::~CCheckRom()
{
}

#pragma mark -


/*______________________________________________________________________________

	CheckRomProcess

*/

Boolean
CCheckRom::CheckRomProcess()
{
	OSErr		err = noErr;
	List_t *	theRomUList;
	Handle		theHandle = nil;
	
	theRomUList = SetRomUList();
	
	if (err == noErr)
	{
		/*	Gather all informations from all RomU	*/
		GetAllRomUInfo(theRomUList);
		
		/*	Check if a RomU from the list is newer for the installed HW	*/
		
		theHandle = CheckNewerRom(theRomUList);
		
		if ( theHandle != nil )
		{
			CFlashRomWindow* mFlashRomWindow = (CFlashRomWindow*) LWindow::CreateWindow(FlashRom_Window, mCommander);
			//ThrowIfNil_(mFlashRomWindow);
			if (mFlashRomWindow != nil)
			{
				mFlashRomWindow->Hide();
				mFlashRomWindow->SetControlPanel( (CControlPanel*)mCommander );
				mFlashRomWindow->SetPCIHardware( mBoard );			
				mFlashRomWindow->CenterWindowOnDisplay();
				mFlashRomWindow->SetRomUHdl( theHandle );
				mFlashRomWindow->Show();
				return 0;
			}
		}
		else
		{
			CControlPanel * theCP = (CControlPanel*) mCommander;
			if (theCP != nil) theCP->Show();
		}
		
		return 1;
	}
	
}


/*______________________________________________________________________________

	SetRomUList

*/

List_t *
CCheckRom::SetRomUList()
{
	OSErr 		err = noErr;
	int			NbrOfRsrc, currentRsrcID;
	int			currentResFile;
	RomURsrc_t	*theRomURsrc;
	Handle		theHandle;

	/* initialize the RomUList */
	NewList(&RomUList);
	
	currentResFile = CurResFile();
	UseResFile(currentResFile);
	NbrOfRsrc = Count1Resources(rom_ResourceType);
	
	err = ResError();
	if (err != noErr)
		return nil;
	
	currentRsrcID = 1;
	while (currentRsrcID <= NbrOfRsrc)
	{
		theRomURsrc = (RomURsrc_t *)NewPtr(sizeof(*theRomURsrc));
		if (!theRomURsrc)
			break;
			
		theHandle = Get1IndResource(rom_ResourceType, currentRsrcID);
		err = ResError();
		if (err != noErr)
		{
			DisposePtr ((Ptr) theRomURsrc);
			break;
		}
		theRomURsrc->desc.RomUHdl = theHandle;
		theRomURsrc->desc.RomUIndex = currentRsrcID++;
		theRomURsrc->desc.IsNewer = 0;
	
		/*	Add the new Handle to the RomURsrc list	*/
		AddTail (&RomUList, &theRomURsrc->node);
	}
	
	return (&RomUList);
}




/*______________________________________________________________________________

	GetAllRomUInfo

*/

void
CCheckRom::GetAllRomUInfo(
	List_t *		inRomUList)
{
	Node_t 			*node;
	RomURsrc_t		*currentRomURsrc;

	/*	Search DriverDescription structr inside each Handle	*/
	
	for(node = inRomUList->head; node->succ; node = node->succ) 
  	{
 		&currentRomURsrc->node = node;

		FindRomUInfo(currentRomURsrc);
	}
		
}

/*______________________________________________________________________________

	FindRomUInfo

*/

Boolean
CCheckRom::FindRomUInfo(
	RomURsrc_t*	inRomURsrc)
{
	Boolean			FindIt = false;
	OSErr 			err;
	char 			*theMark, *theTag, *temp;
	char			*RomUModel;
	long			resourceSize;	
	SInt32			theTagLength, theRomModelLength, compteur = 0;
	
	inRomURsrc->desc.IsNewer = 0;
	inRomURsrc->desc.RomUModel[0] = 0;
	inRomURsrc->desc.RomUVersion = 0;
	theTagLength = strlen( BeginDriverDescriptionStruct );
	theTag = ( char* ) malloc ( theTagLength*sizeof(theTag) );
	strcpy(theTag, BeginDriverDescriptionStruct);
	temp = theTag;
	
	//LoadResource(currentRomUHdl);
	HLock(Handle(inRomURsrc->desc.RomUHdl));
		
	theMark = *inRomURsrc->desc.RomUHdl;
	resourceSize = GetMaxResourceSize(inRomURsrc->desc.RomUHdl);

	/* Find the DriverDescription structure */
	while (resourceSize >= 0 && !FindIt)
	{
		if (*theMark == *temp)
		{
			*temp++;
			compteur++;
			if (compteur == theTagLength)
				FindIt = true;
		}
		else
		{
			compteur = 0;
			temp = theTag;	
		}
		*theMark++;
		resourceSize--;
	}
	
	if (FindIt)
	{
		/*	Gather ROM VERSION	*/
		inRomURsrc->desc.RomUVersion = *((UInt32*)theMark);
		theMark += 4;
//		inRomURsrc->desc.RomUVersion = *theMark++ << 24;
//		inRomURsrc->desc.RomUVersion += *theMark++ << 16;
//		inRomURsrc->desc.RomUVersion += *theMark++ << 8;
//		inRomURsrc->desc.RomUVersion += *theMark++;
		
		/*	Gather BOARD VERSION	*/
		theMark = MoveMarkby(theMark, 2);
		inRomURsrc->desc.RomUModel[0] = *theMark++;
		inRomURsrc->desc.RomUModel[1] = *theMark;
		
		
		/*
		resourceSize -= 3;
		theMark = MoveMarkby(theMark, 2);
		
		theRomModelLength = *theMark;
		theMark = MoveMarkby(theMark, 1);
		strncpy(inRomURsrc->desc.RomUModel, theMark, theRomModelLength);
		inRomURsrc->desc.RomUModel[theRomModelLength] = '\0';
		
		resourceSize -= theRomModelLength;
		theMark = MoveMarkby(theMark, theRomModelLength);
			
		resourceSize -= 1;
		theMark = MoveMarkby(theMark, 1);
		
		compteur = 4;
		while (compteur > 0 && resourceSize >= 0)
		{
			*theMark++;
			
			
			switch(compteur)
			{
				case kMajor:
					inRomURsrc->desc.RomUMajor = *theMark;
					break;
				case kMinor:
					inRomURsrc->desc.RomUMinor = *theMark;
					break;
				case kStage:
					inRomURsrc->desc.RomUStage = *theMark;
					break;
				case kRev:
					inRomURsrc->desc.RomURev = *theMark;
					break;
				default:
					break;
			}
			compteur--;
		}*/
		
	}
	
	/*	free memory	*/
	free(theTag);
	
	return FindIt;
}



/*______________________________________________________________________________

	MoveMarkby

*/

char *
CCheckRom::MoveMarkby(
	char *		inMark,
	int			increment)
{
	while (increment)
	{
		*inMark++;
		increment--;
	}
	return inMark;
}



/*______________________________________________________________________________

	CheckNewerRom

*/

Handle
CCheckRom::CheckNewerRom(
	List_t *		inRomUList)
{
	Str255			installedModel;
	int				installedSerie, resVersion;
	Node_t 			*node;
	UInt16			resultat = 0;
	RomURsrc_t		*currentRomURsrc;
	short			instMajor, instMinor, instStage, instRev;
	/*	Get the Voodoo series installed (4 or 5) */
	mBoard->GetModelName( installedModel );
	
	installedSerie = GetVoodooSerie( (char*) &installedModel[1] );

	/*	Search DriverDescription structr inside each Handle	*/
	
	for(node = inRomUList->head; node->succ; node = node->succ) 
  	{
 		&currentRomURsrc->node = node;
		
		/*	Check if it is the good HW ROM */
		if (installedSerie == (currentRomURsrc->desc.RomUModel[1]))
		{
			
			/*	Check if it is a newer ROM Version */
			resVersion = IsNewerRomVersion(currentRomURsrc);
			if ( resVersion > 0 )
			{
				currentRomURsrc->desc.IsNewer = 1;
//				free(myModelName);
				resultat += 1;
				return currentRomURsrc->desc.RomUHdl;
			}
		}
	}
	
//	free(myModelName);
	
	return nil;
	
	
}


/*______________________________________________________________________________

	GetVoodooSerie

*/

int
CCheckRom::GetVoodooSerie(
	char *		inModelName)
{
	UInt16		theStringStart;
	UInt16		theStringLenght;
	
	
	theStringStart = strlen( inModelName );
	while ( --theStringStart >= 0 && inModelName[theStringStart] != kVoodooName[0] )
		;
	
	if ( theStringStart >= 0 )
	{
		theStringLenght = strlen( kVoodooName );
		if ( !strncmp( kVoodooName, &inModelName[theStringStart], theStringLenght ) )
		{
			return (int) inModelName[ theStringLenght + theStringStart ];
		}
	}
	return 0;
}



/*______________________________________________________________________________

	IsNewerRomVersion

*/

int
CCheckRom::IsNewerRomVersion(
	RomURsrc_t*		inRomURsrc)
{
	Boolean 	isNewer = false;
	UInt8		myMajor, myMinor, myStage, myRev;
	UInt32      theCurrentVersion, theNewROMVersion;
	
	/*	Compare the RomVersion ! */
	mBoard->GetROMVersion( &myMajor, &myMinor, &myStage, &myRev);
	
	theCurrentVersion = (myMajor << 24) + (myMinor << 16) + (myStage << 8) + myRev;
	theNewROMVersion = inRomURsrc->desc.RomUVersion;
	
	//theNewROMVersion = 0x01038000;
	
	return (theCurrentVersion < theNewROMVersion ? 1 : 0);


	return 0;
}
















