/*
** Copyright (c) 2000, 3dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/

#include "CDisplayView.h"
#include "UDisplayMode.h"
#include "ControlPanelPPob.h"
#include "CControlPanelApp.h"

#include <LCaption.h>
#include <LPopupButton.h>
#include <LPushButton.h>
#include <String.h>
#include <Gestalt.h>
#include <LWindow.h>
#include <LString.h>

#include "URequestVideo.h"

#include <Dialogs.h>
#include <ROMDefs.h>
#include <Devices.h>
#include <Errors.h>
#include <Memory.h>
#include <Palettes.h>
#include <Displays.h>
#include <Video.h>
#include <QuickDraw.h>
#include <ConditionalMacros.h>	//DL
#include <CodeFragments.h>	//DL

static List_t *theDMList;

/*______________________________________________________________________________

	Constructor

*/

CDisplayView::CDisplayView()
{
	
}


/*______________________________________________________________________________

	Constructor (using a Stream, that's the one that will be used)

*/

CDisplayView::CDisplayView(
	LStream *			inStream)
	: LView ( inStream )
{
	NeedToChange = true;
	FirstChange = true;
}


/*______________________________________________________________________________

	Destructor

*/

CDisplayView::~CDisplayView()
{
}


#pragma mark -


/*______________________________________________________________________________

	FinishCreateSelf

*/

void
CDisplayView::FinishCreateSelf()
{
	PBResolution = (LPopupButton*) this->FindPaneByID(display_PButtResolution);
	PBRefresh = (LPopupButton*) this->FindPaneByID(display_PButtRefreshRate);
	PBDepth = (LPopupButton*) this->FindPaneByID(display_PButtDepth);
		
	LPushButton* theButton = (LPushButton*) FindPaneByID(display_ButtCancel);
	if (theButton != nil) 
	{
		theButton->AddListener(this);
		theButton->Disable();
	}
	
	if (PBResolution != nil) PBResolution->AddListener(this);
	if (PBRefresh != nil) PBRefresh->AddListener(this);
	if (PBDepth != nil) PBDepth->AddListener(this);
	
	//StartListening();
	
}


/*______________________________________________________________________________

	ListenToMessage

*/

void					
CDisplayView::ListenToMessage(
	MessageT 			inMessage, 
	void *				ioParam)
{
	UInt32 longParam = *(UInt32*)ioParam;
	
	switch (inMessage) {
		
		case msg_DispResolution:
		{
			if (NeedToChange)
			{
				EnableCancelButton();
				ChangeDisplayResolution();
				// Update the Refresh & Depth PopupButton 
				NeedToChange = false;
				SetDisplayPopupButton(GetCurrentDisplay());
				NeedToChange = true;
			}	
			break;
		}
		case msg_DispRefreshRate:
		{
			if (NeedToChange)
			{
				EnableCancelButton();
				ChangeDisplayRefresh();
				// Update the Depth PopupButton 
				NeedToChange = false;
				SetDisplayPopupButton(GetCurrentDisplay());
				NeedToChange = true;
			}
			break;
		}
		case msg_DispDepth:
		{
			if (NeedToChange)
			{
				EnableCancelButton();
				ChangeDisplayDepth();
				
				NeedToChange = false;
				SetDisplayPopupButton(GetCurrentDisplay());
				NeedToChange = true;
			}
			break;
		}
		
		case msg_DispCancel:
		{
			RVSetVideoAsScreenPrefs();
			NeedToChange = false;
			SetDisplayPopupButton(GetCurrentDisplay());
			FirstChange = true;
			LPushButton* theCancel = (LPushButton*) FindPaneByID(display_ButtCancel);
			if (theCancel != nil)
				theCancel->Disable();
			NeedToChange = true;
			
			Refresh();
			break;
		}
		
		default:
			break;
	}

}

/*______________________________________________________________________________

	EnableCancelButton

*/

void
CDisplayView::EnableCancelButton()
{
	if (FirstChange)
	{
		LPushButton* theCancel = (LPushButton*) FindPaneByID(display_ButtCancel);
		if (theCancel != nil)
			theCancel->Enable();
		FirstChange = false;
	}
}


/*______________________________________________________________________________

	SetPCIHardware

*/

void
CDisplayView::SetPCIHardware(
	CPCIHardware *		inBoard)
{
	if ( inBoard != 0 ) {
		
		mBoard = inBoard;
				
	}

	/*	Now that we have the mBoard ... we can load all the PopupButton	*/
	LoadDisplay_View();
	
}


/*______________________________________________________________________________

	SetControlPanelPtr

*/

void
CDisplayView::SetControlPanelPtr(
	CControlPanel *		inControlPanel)
{
	if ( inControlPanel != 0 ) {
		
		mControlPanel = inControlPanel;
				
	}

}


/*______________________________________________________________________________

	LoadView_Display
*/

Boolean
CDisplayView::LoadDisplay_View()
{
	
	
	Boolean 		result = true;
	Node_t 			*node;
	CPDisplay_t 	*currentCPDM;
	CPDisplay_t 	*lastCPDM = nil;
	LStr255 		ResoStr;
	unsigned long	inData;
	unsigned short	inMode;
	
	NeedToChange = false;
	
	/* Gather all the different display modes */
	theDMList = MakeDisplayModeList(mBoard->GetRefNum());
	
	/* Load the Resolution PopupButton */
	MakeCPDisplayPBIndex();
	
	/* Set up the PBResolution Items */
  	for(node = theDMList->head; node->succ; node = node->succ) 
  	{
 		&currentCPDM->node = node;	
		
		// temporary debug lines
		
		ResoStr = ResoStr.Remove(0,255);
		ResoStr = ResoStr.Append((long)currentCPDM->desc.Horizontal);
 		ResoStr = ResoStr.Append(kCHAR_BY);
 		ResoStr = ResoStr.Append((long)currentCPDM->desc.Vertical);
 		
 		short test = mBoard->GetRefNum();
 		if (currentCPDM->desc.mygdRefNum == test)
		{ 		
	 		if (currentCPDM->desc.Horizontal != lastCPDM->desc.Horizontal ||
	 					currentCPDM->desc.Vertical != lastCPDM->desc.Vertical)
	 			PBResolution->AppendMenu((const unsigned char *) ResoStr, true);
		}
 		lastCPDM = currentCPDM;
    }   
    
	currentDisplayMode = GetCurrentDisplay();
	
	if (!SetDisplayPopupButton(currentDisplayMode))
	{
		UpdateDisplay_RefreshRate(1);
		UpdateDisplay_Depth(1,1);
	}
	
	NeedToChange = true;
	return result;
}





/*______________________________________________________________________________

	MakeCPDisplayPBIndex
	
*/

void
CDisplayView::MakeCPDisplayPBIndex()
{
	SInt16				count1, count2, count3;
	Node_t				*node;
	CPDisplay_t			*currentCPDM;
	CPDisplay_t			*lastCPDM = nil;
	DisplayPBIndex_t	*theDPBI;
	Boolean 			theFirst = true;
	
	/* Initialize theDPBI */
	count1 = count2 = count3 = 0;
	
	/* Find the mycsData matching in the CPDList */
	for(node = theDMList->head; node->succ; node = node->succ) 
  	{
 		&currentCPDM->node = node;
		if (currentCPDM->desc.mygdRefNum != lastCPDM->desc.mygdRefNum)
		{
			count1 = count2 = count3 = 0;
			lastCPDM = (CPDisplay_t*) nil;
		}	
	 	if (currentCPDM->desc.Horizontal != lastCPDM->desc.Horizontal || 
			currentCPDM->desc.Vertical != lastCPDM->desc.Vertical)
 		{
 			count1++;
 			count2 = 0;
 			theFirst = true;
 		}
 		
		if (theFirst || currentCPDM->desc.RefreshRate != lastCPDM->desc.RefreshRate)
		{
			count2++;
			count3 = 0;
			theFirst = false;
		}
		
		if (currentCPDM->desc.Depth != lastCPDM->desc.Depth)
			count3++;
 		
		currentCPDM->desc.resIndex = count1;
		currentCPDM->desc.refIndex = count2;
		currentCPDM->desc.depthIndex = count3;	
		
		lastCPDM = currentCPDM;
		
 	}
}




/*______________________________________________________________________________

	UpdateDisplay_RefreshRate
*/

void
CDisplayView::UpdateDisplay_RefreshRate(SInt16 theReso)
{
	Node_t 			*node;
	CPDisplay_t 	*currentCPDM;
	CPDisplay_t 	*lastCPDM = nil;
	double_t		refreshRate;		
	LStr255 		RefreshStr;
	SInt16			position = 0;
	StringPtr		orig;
	Boolean			theSuccess = false;
	Boolean			theFirst = true;
	
	
	/* Update the RefreshRate PopupButton */
	PBRefresh->DeleteAllMenuItems();
	
	
	/* Walk list of resolution-refresh nodes. */
  	for(node = theDMList->head; node->succ; node = node->succ) 
  	{
 		&currentCPDM->node = node;
		
		
		
		if (currentCPDM->desc.mygdRefNum == mBoard->GetRefNum())
		{ 		
	 		if (currentCPDM->desc.Horizontal != lastCPDM->desc.Horizontal ||
	 					currentCPDM->desc.Vertical != lastCPDM->desc.Vertical)
	 		{
	 			position++;	
	 			theFirst = true;
			}
			if (position == theReso)
	 		{

	 			if (currentCPDM->desc.RefreshRate != lastCPDM->desc.RefreshRate || theFirst)
	 			{
					theFirst = false;
					theSuccess = true;
					//RefreshStr = RefreshStr.Remove(0,255);
					refreshRate = Fix2X (currentCPDM->desc.RefreshRate);
					
					LStr255		originalStr(refreshRate, FIXEDDECIMAL, 1);
					
					//clean_originalStr(originalStr);
					
					//RefreshStr = RefreshStr.Append(currentCPDM->desc.RefreshRate);
	 				originalStr = originalStr.Append(" Hz");
	 				
	 				PBRefresh->AppendMenu((const unsigned char *) originalStr, true);
	 			}
	 		}
	 		lastCPDM = currentCPDM;
	 	}
     } 

	
	if ( !theSuccess )
     	PBRefresh->AppendMenu(kOptimalRefresh);

}



/*______________________________________________________________________________

	UpdateDisplay_Depth
*/

void
CDisplayView::UpdateDisplay_Depth(SInt16 theReso, SInt16 theRefresh)
{
	Node_t 			*node;
	CPDisplay_t 	*currentCPDM;
	CPDisplay_t 	*lastCPDM = nil;
	LStr255 		DepthStr;
	SInt16			posReso = 0;
	SInt16			posRefresh = 0;
	Boolean			theFirst = true;
		
	/* Update the RefreshRate PopupButton */
	PBDepth->DeleteAllMenuItems();
	
	/* Walk list of resolution-refresh nodes. */
  	for(node = theDMList->head; node->succ; node = node->succ) 
  	{
 		&currentCPDM->node = node;
		if (currentCPDM->desc.mygdRefNum == mBoard->GetRefNum())
		{ 		
		 	
			if (currentCPDM->desc.Horizontal != lastCPDM->desc.Horizontal || 
						currentCPDM->desc.Vertical != lastCPDM->desc.Vertical)
	 		{
	 			posReso++;	
	 			theFirst = true;
			}
			if (posReso == theReso)
			{
				if (theFirst  || currentCPDM->desc.RefreshRate != lastCPDM->desc.RefreshRate)
				{
					posRefresh++;
					theFirst = false;
				}
			}
			if ((posReso == theReso) && (posRefresh == theRefresh))
	 		{
	 			if (currentCPDM->desc.Depth != lastCPDM->desc.Depth)
	 			{
	 				DepthStr = DepthStr.Remove(0, 255);
	 				switch (currentCPDM->desc.Depth)
	 				{
	 					case 8:
	 					{
	 						DepthStr.Assign(display_Depth_STR, k8Bits);
	 						break;
	 					}
	 					case 16:
	 					{
	 						DepthStr.Assign(display_Depth_STR, k16Bits);
	 						break;
	 					}
	 					case 32:
	 					{
	 						DepthStr.Assign(display_Depth_STR, k32Bits);
	 						break;
	 					}
	 					default:
	 						break;
	 				}
					PBDepth->AppendMenu((const unsigned char *) DepthStr, true);
	 			}
	 		}
	 		lastCPDM = currentCPDM;
	 	}
    }   

}

#pragma mark -

/*______________________________________________________________________________

	ChangeDisplayResolution

*/

void 
CDisplayView::ChangeDisplayResolution()
{
	CPDisplay_t *		currentCPDM;
	CPDisplay_t *		lastCPDM = nil;
	CPDisplayDesc_t	*	RequestedCPDM;
	CPDisplay_t *		temp;
	Node_t *			node;
	DisplayPBIndex_t *	currentDPBIndex;
	SInt16				posReso = 0;
	Boolean				Findit = false;
	
	
	/* Get new resolution */
	RequestedCPDM = (CPDisplayDesc_t *)NewPtr(sizeof(*RequestedCPDM));
	
	currentDisplayMode = GetCurrentDisplay();
	
	RequestedCPDM->RefreshRate = currentDisplayMode->RefreshRate;
	RequestedCPDM->Depth = currentDisplayMode->Depth;
	
	currentDPBIndex = GetDisplayPBIndexSelected();
	for(node = theDMList->head; node->succ; node = node->succ) 
  	{
 		&currentCPDM->node = node;
		if (currentCPDM->desc.mygdRefNum == mBoard->GetRefNum())
		{ 		
	 	
			if ((currentCPDM->desc.Horizontal != lastCPDM->desc.Horizontal) || 
						(currentCPDM->desc.Vertical != lastCPDM->desc.Vertical))
	 			posReso++;	
				
			if (posReso == currentDPBIndex->theReso) 
	 		{
	 			RequestedCPDM->Horizontal = currentCPDM->desc.Horizontal;
	 			RequestedCPDM->Vertical = currentCPDM->desc.Vertical;
	 			Findit = true;
	 			break;
	 		}
	 		lastCPDM = currentCPDM;
		}
 		
    } 
	
	if (Findit)
	{
		/* Look for best match with currentCPDM and new resolution */
		if (!FindCPDMRequested(RequestedCPDM))
			RequestedCPDM = FindBestDepthRefreshMatch(RequestedCPDM);
		
		/* Set the new resolution */
		SetNewResolution(RequestedCPDM);
		
	}
	
	/* Center the control panel in the window */
	mControlPanel->CenterWindowOnDisplay();
		
	/* Free some memory */
	DisposePtr((Ptr)RequestedCPDM);
	DisposePtr((Ptr)currentDPBIndex);
}



/*______________________________________________________________________________

	ChangeDisplayDepth

*/

void 
CDisplayView::ChangeDisplayDepth()
{
	CPDisplay_t *		currentCPDM;
	CPDisplay_t *		lastCPDM = nil;
	CPDisplayDesc_t	*	RequestedCPDM;
	CPDisplay_t *		temp;
	Node_t *			node;
	DisplayPBIndex_t *	currentDPBIndex;
	SInt16				posDepth = 0;
	Boolean				Findit = false;
	
	
	/* Get new resolution */
	RequestedCPDM = (CPDisplayDesc_t *)NewPtr(sizeof(*RequestedCPDM));
	currentDisplayMode = GetCurrentDisplay();
	
	RequestedCPDM->Horizontal = currentDisplayMode->Horizontal;
	RequestedCPDM->Vertical = currentDisplayMode->Vertical;
	RequestedCPDM->RefreshRate = currentDisplayMode->RefreshRate;
	
	currentDPBIndex = GetDisplayPBIndexSelected();
	for(node = theDMList->head; node->succ; node = node->succ) 
  	{
 		&currentCPDM->node = node;
		if (currentCPDM->desc.mygdRefNum == mBoard->GetRefNum())
		{ 		
		 	
			if (RequestedCPDM->Horizontal == currentCPDM->desc.Horizontal &&
				RequestedCPDM->Vertical == currentCPDM->desc.Vertical &&
				RequestedCPDM->RefreshRate == currentCPDM->desc.RefreshRate)
	 		{
	 			posDepth++;	
				
				if (posDepth == currentDPBIndex->theDepth) 
	 			{
	 				RequestedCPDM->Depth = currentCPDM->desc.Depth;
	 				Findit = true;
	 				break;
	 			}
	 		}
    	}
    } 
	
	if (Findit)
	{
		/* Look for best match with currentCPDM and new resolution */
		if (!FindCPDMRequested(RequestedCPDM))
			RequestedCPDM = FindBestDepthRefreshMatch(RequestedCPDM);
		
		/* Set the new resolution */
		SetNewResolution(RequestedCPDM);
		
	}
	
	/* Free some memory */
	DisposePtr((Ptr)RequestedCPDM);
	DisposePtr((Ptr)currentDPBIndex);
}



/*______________________________________________________________________________

	ChangeDisplayRefresh

*/

void 
CDisplayView::ChangeDisplayRefresh()
{
	CPDisplay_t *		currentCPDM;
	CPDisplay_t *		lastCPDM = nil;
	CPDisplayDesc_t	*	RequestedCPDM;
	CPDisplay_t *		temp;
	Node_t *			node;
	DisplayPBIndex_t *	currentDPBIndex;
	SInt16				posReso = 0;
	SInt16				posRefresh = 0;
	Boolean				Findit = false;
	
	
	/* Get new resolution */
	RequestedCPDM = (CPDisplayDesc_t *)NewPtr(sizeof(*RequestedCPDM));
	currentDisplayMode = GetCurrentDisplay();
	
	RequestedCPDM->Horizontal = currentDisplayMode->Horizontal;
	RequestedCPDM->Vertical = currentDisplayMode->Vertical;
	RequestedCPDM->Depth = currentDisplayMode->Depth;
	
	currentDPBIndex = GetDisplayPBIndexSelected();
	
	for(node = theDMList->head; node->succ; node = node->succ) 
  	{
 		&currentCPDM->node = node;
		if (currentCPDM->desc.mygdRefNum == mBoard->GetRefNum())
		{ 		
	 	
			if (currentCPDM->desc.Horizontal != lastCPDM->desc.Horizontal || 
					currentCPDM->desc.Vertical != lastCPDM->desc.Vertical)
	 		{
	 			posReso++;
	 			posRefresh = 0;
	 		}
	 		
			if (currentCPDM->desc.RefreshRate != lastCPDM->desc.RefreshRate)
				posRefresh++;

				
	 		if (posReso == currentDPBIndex->theReso &&
	 				posRefresh == currentDPBIndex->theRefresh)
	 		{
	 			RequestedCPDM->RefreshRate = currentCPDM->desc.RefreshRate;
	 			Findit = true;
	 			break;
	 		}
	 		lastCPDM = currentCPDM;
	 	}
 		
    } 
	if (Findit)
	{
		/* Look for best match with currentCPDM and new resolution */
		if (!FindCPDMRequested(RequestedCPDM))
			RequestedCPDM = FindBestDepthMatch(RequestedCPDM);
		
		/* Set the new resolution */
		SetNewResolution(RequestedCPDM);
		
	}
	
	/* Free some memory */
	DisposePtr((Ptr)RequestedCPDM);
	DisposePtr((Ptr)currentDPBIndex);
}


/*______________________________________________________________________________

	FindCPDMRequested

*/

Boolean
CDisplayView::FindCPDMRequested(CPDisplayDesc_t * theCPDM)
{
	CPDisplay_t *		currentCPDM;
	Node_t *			node;
		
	/* Walk list of nodes to find theCPDM */
  	for(node = theDMList->head; node->succ; node = node->succ) 
  	{
 		&currentCPDM->node = node;
		if (currentCPDM->desc.mygdRefNum == mBoard->GetRefNum())
		{ 		
	 	
			if (theCPDM->Horizontal == currentCPDM->desc.Horizontal &&
				theCPDM->Vertical == currentCPDM->desc.Vertical &&
				theCPDM->RefreshRate == currentCPDM->desc.RefreshRate &&
				theCPDM->Depth == currentCPDM->desc.Depth)
				{
					return true;
				}
				
		}
    } 
    return false;  

}



/*______________________________________________________________________________

	FindBestDepthRefreshMatch

*/

CPDisplayDesc_t * 
CDisplayView::FindBestDepthRefreshMatch(CPDisplayDesc_t * RequestedCPDM)
{	
	CPDisplay_t *		currentCPDM;
	CPDisplay_t *		lastCPDM;
	Node_t *			node;
	SInt16				posReso = 0;
	Boolean				FindResol = false;
	Boolean				FindDepth = false;
	Boolean				FindRefresh = false;
	Boolean				JustChange = false;
	
	for(node = theDMList->head; node->succ; node = node->succ) 
  	{
 		&currentCPDM->node = node;
		
		
			/* Find the requested resolution */
			if ((currentCPDM->desc.Horizontal == RequestedCPDM->Horizontal) && 
						(currentCPDM->desc.Vertical == RequestedCPDM->Vertical) &&
						(currentCPDM->desc.mygdRefNum == mBoard->GetRefNum()))
	 		{
	 			FindResol = true;
	 			JustChange = false;
	 			/* Find the best Depth */
	 			if (currentCPDM->desc.Depth == RequestedCPDM->Depth)
	 			{
	 				RequestedCPDM->Depth = currentCPDM->desc.Depth;
	 				FindDepth = true;
	 			}
	 			if (!FindDepth && (currentCPDM->desc.Depth > RequestedCPDM->Depth))
	 			{
	 				RequestedCPDM->Depth = lastCPDM->desc.Depth;
	 				FindDepth = true;
	 			}
	 			
	 			/* Find the best Refresh Rate */	
	 			if (FindDepth)
	 			{
	 				if (currentCPDM->desc.RefreshRate == RequestedCPDM->RefreshRate)
	 				{
	 					FindRefresh = true;
	 					RequestedCPDM->RefreshRate = currentCPDM->desc.RefreshRate;
	 				}
	 				if (!FindRefresh && (currentCPDM->desc.RefreshRate > RequestedCPDM->RefreshRate))
	 				{
	 					RequestedCPDM->RefreshRate = lastCPDM->desc.RefreshRate;
	 					FindRefresh = true;
	 				}
	 			}
	 		}
	 		else
	 			JustChange = true;
	 			
	 		if (&currentCPDM->node == theDMList->tailPred || JustChange)
	 		{
	 			if (FindResol && !FindDepth)
	 			{
	 				RequestedCPDM->Depth = lastCPDM->desc.Depth;
	 				FindDepth = true;
	 			}
	 			
	 			if (FindResol && FindDepth && !FindRefresh)
				{
				 	RequestedCPDM->RefreshRate = lastCPDM->desc.RefreshRate;
				 	FindRefresh = true;
				}
				if (FindResol && FindDepth && FindRefresh)
					return RequestedCPDM; 
	 		}
	 		lastCPDM = currentCPDM;
	 	
 		
    }
    
    return RequestedCPDM; 
}


/*______________________________________________________________________________

	FindBestDepthMatch

*/

CPDisplayDesc_t * 
CDisplayView::FindBestDepthMatch(CPDisplayDesc_t * RequestedCPDM)
{	
	CPDisplay_t *		currentCPDM;
	CPDisplay_t *		lastCPDM;
	Node_t *			node;
	SInt16				posReso = 0;
	Boolean				FindResol = false;
	Boolean				FindDepth = false;
	Boolean				FindRefresh = false;
	Boolean				JustChange = false;
	
	for(node = theDMList->head; node->succ; node = node->succ) 
  	{
 		&currentCPDM->node = node;
		
			/* Find the requested resolution */
			if ((currentCPDM->desc.Horizontal == RequestedCPDM->Horizontal) && 
					(currentCPDM->desc.Vertical == RequestedCPDM->Vertical) &&
					(currentCPDM->desc.RefreshRate == RequestedCPDM->RefreshRate) &&
					(currentCPDM->desc.mygdRefNum == mBoard->GetRefNum()))
	 		{
	 			FindResol = true;
	 			JustChange = false;
	 			/* Find the best Depth */
	 			if (currentCPDM->desc.Depth == RequestedCPDM->Depth)
	 			{
	 				RequestedCPDM->Depth = currentCPDM->desc.Depth;
	 				FindDepth = true;
	 			}
	 			if (!FindDepth && (currentCPDM->desc.Depth > RequestedCPDM->Depth))
	 			{
	 				RequestedCPDM->Depth = lastCPDM->desc.Depth;
	 				FindDepth = true;
	 			}
	 			
	 		}
	 		else
	 			JustChange = true;
	 			
	 		if (&currentCPDM->node == theDMList->tailPred || JustChange)
	 		{
	 			if (FindResol && !FindDepth)
	 			{
	 				RequestedCPDM->Depth = lastCPDM->desc.Depth;
	 				FindDepth = true;
	 			}
	 			
				if (FindResol && FindDepth && FindRefresh)
					return RequestedCPDM; 
	 		}
	 		lastCPDM = currentCPDM;
	 
 		
    }
    
    return RequestedCPDM; 
}



/*______________________________________________________________________________

	GetDisplayPBIndexSelected

*/

DisplayPBIndex_t * 
CDisplayView::GetDisplayPBIndexSelected()
{
	DisplayPBIndex_t * selectedDPBIndex;
	
	selectedDPBIndex = (DisplayPBIndex_t *)NewPtr(sizeof(*selectedDPBIndex));
	
	
	selectedDPBIndex->theReso = PBResolution->GetCurrentMenuItem();
	selectedDPBIndex->theRefresh = PBRefresh->GetCurrentMenuItem();
	selectedDPBIndex->theDepth = PBDepth->GetCurrentMenuItem();
	
	return selectedDPBIndex;
}


/*______________________________________________________________________________

	SetDisplayPBIndexSelected

*/

void 
CDisplayView::SetDisplayPBIndexSelected(DisplayPBIndex_t * theDPBI)
{	
	
	PBResolution->SetCurrentMenuItem(theDPBI->theReso);
	UpdateDisplay_RefreshRate(theDPBI->theReso);
	PBRefresh->SetCurrentMenuItem(theDPBI->theRefresh);
	UpdateDisplay_Depth(theDPBI->theReso, theDPBI->theRefresh);
	PBDepth->SetCurrentMenuItem(theDPBI->theDepth);
			
}

/*______________________________________________________________________________

	GetCurrentDisplay

*/

CPDisplayDesc_t	*
CDisplayView::GetCurrentDisplay()
{
	CPDisplayDesc_t	*				Response;
	Boolean							displayMgrPresent;
	DMDisplayModeListIteratorUPP	myModeIteratorProc = nil;	// for DM2.0 searches
	SpBlock							spBlock;
	Boolean							suppliedGDevice;	
	long							value = 0;
	GDHandle						walkDevice = nil;			// for everybody
	GDHandle						theDevice;
	
	unsigned long		displayMgrVersion;
	OSErr				error = paramErr;
	VDSwitchInfoRec		switchInfo;
	AuxDCEHandle		theDCE;
	VDSwitchInfoRec		videoMode;		
	
	Response = (CPDisplayDesc_t *)NewPtr(sizeof(*Response));
	
	
	Response->mycsData = 0;
	Response->mycsMode = 0;
	
	Gestalt(gestaltDisplayMgrAttr,&value);
	displayMgrPresent=value&(1<<gestaltDisplayMgrPresent);
	displayMgrPresent=displayMgrPresent && (SVersion(&spBlock)==noErr);	// need slot manager
	if (displayMgrPresent)												// and Display Manager
	{	
		theDevice  = ::GetDeviceList();
	
		/*	Find the right screen ! */
		while (theDevice != nil) {
			if ( (*theDevice)->gdRefNum == mBoard->GetRefNum() ) 
				walkDevice = theDevice;
	  		theDevice = ::GetNextDevice(theDevice);
		}
		//walkDevice = DMGetFirstScreenDevice (dmOnlyActiveDisplays);			// for everybody
		suppliedGDevice = false;
		
		myModeIteratorProc = NewDMDisplayModeListIteratorProc(ModeListIterator);	// for DM2.0 searches
	
		// Note that we are hosed if somebody changes the gdevice list behind our backs while we are iterating....
		// ...now do the loop if we can start
		if( walkDevice && myModeIteratorProc) // start the search
		{
			Gestalt(gestaltDisplayMgrVers, (long*)&displayMgrVersion);
			if (displayMgrVersion >= 0x00020000)
			{	// get the info the DM 2.0 way
				error = DMGetDisplayMode(walkDevice, &switchInfo);
				if (noErr == error)
				{
					Response->mycsData = switchInfo.csData;
					Response->mycsMode = switchInfo.csMode;
					Response->mygdRefNum = mBoard->GetRefNum();
					Response = FindCPDMbyDataMode(Response);
				}
			}
		}
	}
	return Response;
}


/*______________________________________________________________________________

	FindCPDMbyDataMode
	
*/

CPDisplayDesc_t *
CDisplayView::FindCPDMbyDataMode(CPDisplayDesc_t *theCPDM)
{
	SInt16				posReso = 0;
	SInt16				posRefresh = 0;
	SInt16				posDepth = 0;
	Node_t				*node;
	CPDisplay_t			*currentCPDM = nil;
	CPDisplay_t			*lastCPDM = nil;
	
	if (theCPDM->mycsData != 0 && theCPDM->mycsMode != 0)
	{
		/* Find the mycsData matching in the CPDList */
		for(node = theDMList->head; node->succ; node = node->succ) 
  		{
 			&currentCPDM->node = node;
			if (currentCPDM->desc.mygdRefNum == mBoard->GetRefNum())
			{ 		
	 	
				if (currentCPDM->desc.Horizontal != lastCPDM->desc.Horizontal || 
					currentCPDM->desc.Vertical != lastCPDM->desc.Vertical)
	 			{
	 				posReso++;
	 				posRefresh = 0;
	 			}
	 		
				if (currentCPDM->desc.RefreshRate != lastCPDM->desc.RefreshRate)
				{
					posRefresh++;
					posDepth = 0;
				}
				if (currentCPDM->desc.Depth != lastCPDM->desc.Depth)
					posDepth++;
	 		
	 			if (currentCPDM->desc.mycsData == theCPDM->mycsData &&
					currentCPDM->desc.mycsMode == theCPDM->mycsMode)
				{
					theCPDM->Horizontal = currentCPDM->desc.Horizontal;
					theCPDM->Vertical = currentCPDM->desc.Vertical;
					theCPDM->RefreshRate = currentCPDM->desc.RefreshRate;
					theCPDM->Depth = currentCPDM->desc.Depth;
					return theCPDM;
				}
				lastCPDM = currentCPDM;	
			}
 			
 		}
 	}
 	return nil; 
 }

/*______________________________________________________________________________

	SetDisplayPopupButton
	
*/

Boolean
CDisplayView::SetDisplayPopupButton(CPDisplayDesc_t *theCPDM)
{
	SInt16				count1, count2, count3;
	Node_t				*node;
	CPDisplay_t			*currentCPDM;
	CPDisplay_t			*lastCPDM = nil;
	DisplayPBIndex_t	*theDPBI;
	
	theDPBI = (DisplayPBIndex_t *)NewPtr(sizeof(*theDPBI));
	
	
	/* Initialize theDPBI */
	count1 = count2 = count3 = 0;
	
	/* Find the mycsData matching in the CPDList */
	for(node = theDMList->head; node->succ; node = node->succ) 
  	{
 		&currentCPDM->node = node;
		if (currentCPDM->desc.mygdRefNum == mBoard->GetRefNum())
		{ 		
		 	
		 	if (theCPDM->Horizontal == currentCPDM->desc.Horizontal &&
		 		theCPDM->Vertical == currentCPDM->desc.Vertical &&
		 		theCPDM->RefreshRate == currentCPDM->desc.RefreshRate &&
		 		theCPDM->Depth == currentCPDM->desc.Depth)
		 	{
		 		theDPBI->theReso = currentCPDM->desc.resIndex;
				theDPBI->theRefresh = currentCPDM->desc.refIndex;
				theDPBI->theDepth = currentCPDM->desc.depthIndex;
				SetDisplayPBIndexSelected (theDPBI);
				return true;
		 	}
		 	
		 	
		 	/*if (currentCPDM->desc.Horizontal != lastCPDM->desc.Horizontal || 
				currentCPDM->desc.Vertical != lastCPDM->desc.Vertical)
	 		{
	 			count1++;
	 			count2 = 0;
	 		}
	 		
			if (currentCPDM->desc.RefreshRate != lastCPDM->desc.RefreshRate)
			{
				count2++;
				count3 = 0;
			}
			
			if (currentCPDM->desc.Depth != lastCPDM->desc.Depth)
				count3++;
	 		
	 		if (currentCPDM->desc.mycsData == theCPDM->mycsData &&
				currentCPDM->desc.mycsMode == theCPDM->mycsMode)
			{
				theDPBI->theReso = count1;
				theDPBI->theRefresh = count2;
				theDPBI->theDepth = count3;
				SetDisplayPBIndexSelected (theDPBI);
				return true;
			}	*/
			
			lastCPDM = currentCPDM;
		}

 		
 	}
 	return false; 
}


/*______________________________________________________________________________

	SetNewResolution
	
*/

Boolean
CDisplayView::SetNewResolution(CPDisplayDesc_t	* theCPDM)
{
	VideoRequestRec	requestRec;
	VideoRequestRec	originalRec;
	
	requestRec.screenDevice = nil;		// any screen
	
	requestRec.reqBitDepth = theCPDM->Depth;
	requestRec.reqHorizontal = theCPDM->Horizontal;
	requestRec.reqVertical = theCPDM->Vertical;
	requestRec.refreshRate = theCPDM->RefreshRate;
	
	requestRec.displayMode = nil;		// Must init to nil
	requestRec.depthMode = nil;			// Must init to nil
	requestRec.requestFlags = 1<<kAllValidModesBit;
	
	// make the request and set it if we have it ...
		
	//Get current setting
	GDHandle	theDevice  = ::GetDeviceList();
		while (theDevice != nil) 
		{
			if ( (*theDevice)->gdRefNum == mBoard->GetRefNum() ) 
			{
				originalRec.screenDevice = theDevice;		// this screen
				requestRec.screenDevice = theDevice;
			}
		  	theDevice = ::GetNextDevice(theDevice);
		}			
	RVRequestVideoSetting(&requestRec);
	
	if (requestRec.screenDevice != nil)		// make sure we found a device...possible if there are no "safe" video modes
	{
		GDHandle	theDevice  = ::GetDeviceList();
		while (theDevice != nil) 
		{
			if ( (*theDevice)->gdRefNum == mBoard->GetRefNum() ) 
			{
				originalRec.screenDevice = theDevice;		// this screen
				requestRec.screenDevice = theDevice;
			}
		  	theDevice = ::GetNextDevice(theDevice);
		}			
		RVGetCurrentVideoSetting(&originalRec);		
		
		//requestRec.refreshRate = theCPDM->RefreshRate;
		// Set/Confirm/Reset the request
		RVSetVideoRequest (&requestRec);
		
		if (noErr != RVConfirmVideoRequest (&requestRec))
		{
			RVSetVideoRequest (&originalRec);
			return false;
		}
		return true;		
	}
	return false;
}


	
