/*
** Copyright (c) 1999, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/

#include "CEndUpdateRom.h"
#include "ControlPanelPPob.h"
#include "CFlashRomWindow.h"
#include "CControlPanel.h"

#include <UDesktop.h>
/*______________________________________________________________________________

	Constructor

*/

CEndUpdateRom::CEndUpdateRom()
{
	mFlashRomWindow = nil;
}


/*______________________________________________________________________________

	Constructor (using a Stream, that's the one that will be used)

*/

CEndUpdateRom::CEndUpdateRom(
	LStream *			inStream)
	: LWindow ( inStream )
{
}


/*______________________________________________________________________________

	Destructor

*/

CEndUpdateRom::~CEndUpdateRom()
{
}


#pragma mark -

/*______________________________________________________________________________

	FinishCreateSelf

*/

void 
CEndUpdateRom::FinishCreateSelf()
{	
	LPushButton* theButton = (LPushButton*) FindPaneByID(flashRom_Close);
	if (theButton != nil)		theButton->AddListener(this);
	
	theButton = (LPushButton*) FindPaneByID(flashRom_Restart);
	if (theButton != nil)		theButton->AddListener(this);
}



/*______________________________________________________________________________

	ListenToMessage

*/


void 
CEndUpdateRom::ListenToMessage(
	MessageT 	inMessage, 
	void *		ioParam)
{
	UInt32 		longParam = *(UInt32*)ioParam;

	switch(inMessage) 
	{
		case flashRom_Close:
			mFlashRomWindow->DoFlashQuit();
			break;
			
		case flashRom_Restart:
			mFlashRomWindow->DoFlashRestart();
			break;
			
		default:
			break;
	}
}

#pragma mark -





/*______________________________________________________________________________

	CenterWindowOnDisplay

*/

void
CEndUpdateRom::CenterWindowOnDisplay()
{	
	GDHandle	theDevice  = ::GetDeviceList();

	while (theDevice != nil) {
	
		if ( (*theDevice)->gdRefNum == mBoard->GetRefNum() ) {
		
			Rect 	r = (**theDevice).gdRect;
			long	x,y;
			
			x = r.right - r.left;
			y = r.bottom - r.top;
			
			x = ( x - mFrameSize.width) / 2 + r.left;
			y = ( y - mFrameSize.height) / 2 + r.top;
			
	  		MoveWindowTo( x, y );
	  		mFrameLocation.h = y;
  			mFrameLocation.v = x;
	  	}

  		theDevice = ::GetNextDevice(theDevice);
	}
}




/*______________________________________________________________________________

	SetCFlashRomWindow

*/

void
CEndUpdateRom::SetCFlashRomWindow(
	CFlashRomWindow *		inFlashRomWindow)
{	
	mFlashRomWindow = inFlashRomWindow;
}


/*______________________________________________________________________________

	SetControlPanel

*/

void
CEndUpdateRom::SetControlPanel(
	CControlPanel *		inCPanel)
{	
	if ( inCPanel != 0 )
	{
		mCPanel = inCPanel;
	}

}

/*______________________________________________________________________________

	SetCPCIHardware

*/

void
CEndUpdateRom::SetPCIHardware(
	CPCIHardware *		inBoard)
{	
	if ( inBoard != 0 )
	{
		mBoard = inBoard;
	}

}


/*______________________________________________________________________________

	ClickInDrag

*/

void
CEndUpdateRom::ClickInDrag(
	const EventRecord	&inMacEvent)
{
	GDHandle	theDevice  = ::GetDeviceList();

	while (theDevice != nil) {
	
		if ( (*theDevice)->gdRefNum == mBoard->GetRefNum() ) {
		
			Rect 	r = (**theDevice).gdRect;		
												// Save old bounds (content region
												//	is in global coords)
			Rect	bounds = (*((WindowPeek)mMacWindowP)->contRgn)->rgnBBox;
			
			//Rect	dragRect = (**(GetGrayRgn())).rgnBBox;
			
			Rect	dragRect;
			
			dragRect.left = ((&inMacEvent)->where.h - bounds.left) + r.left;
			dragRect.top = ((&inMacEvent)->where.v - bounds.top) + 20 + r.top;
			dragRect.right = r.right - ((bounds.left + mFrameSize.width) - (&inMacEvent)->where.h);
			dragRect.bottom = r.bottom - ((bounds.top + mFrameSize.height) - (&inMacEvent)->where.v);

			::MacInsetRect(&dragRect, 4, 4);
			UDesktop::DragDeskWindow(this, inMacEvent, dragRect);
			
				// DragWindow will move the window. However, send AppleEvent to
				// set the bounds (but don't execute it), so that Script engines
				// can record the action.
				
												// Compare old and new bounds
			if (!::MacEqualRect(&bounds,
							    &(*((WindowPeek)mMacWindowP)->contRgn)->rgnBBox)) {
							
				Point	newPosition =
							topLeft((*((WindowPeek)mMacWindowP)->contRgn)->rgnBBox);
							
												// Send but don't execute
												//   SetPosition AppleEvent
				SendAESetPosition(newPosition, ExecuteAE_No);
				
				AdjustUserBounds();				// Moving Window changes user
												//   bounds for zooming
												
				
			}
					

	  	}
  			
  		theDevice = ::GetNextDevice(theDevice);
	}

}


