/*
** Copyright (c) 1999, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/

#include "CFlashRomWindow.h"
#include "ControlPanelPPob.h"
#include "CUpdateRom.h"
#include "CEndUpdateRom.h"
#include "CControlPanel.h"
#include "URomFlasher.h"

#include <LPushButton.h>
#include <LCheckBox.h>
#include <LDialogBox.h>
#include <LCaption.h>
#include <UModalDialogs.h>

#include <LWindow.h>
#include <AERegistry.h>

#include <UDesktop.h>

/*______________________________________________________________________________

	Constructor

*/

CFlashRomWindow::CFlashRomWindow()
{
	
}


/*______________________________________________________________________________

	Constructor (using a Stream, that's the one that will be used)

*/

CFlashRomWindow::CFlashRomWindow(
	LStream *			inStream)
	: LWindow ( inStream )
{
	mCPanel = nil;
}


/*______________________________________________________________________________

	Destructor

*/

CFlashRomWindow::~CFlashRomWindow()
{
}


#pragma mark -

/*______________________________________________________________________________

	FinishCreateSelf

*/

void 
CFlashRomWindow::FinishCreateSelf()
{	
	LPushButton* theFlashButton = (LPushButton*) FindPaneByID(flashRom_Flash);
	if (theFlashButton != nil)		theFlashButton->AddListener(this);
	
	LPushButton* theCancelButton = (LPushButton*) FindPaneByID(flashRom_Cancel);
	if (theCancelButton != nil)		theCancelButton->AddListener(this);
	
	LCheckBox* theAskAgain = (LCheckBox*) FindPaneByID(flashRom_AskAgain);
	if (theAskAgain != nil)			theAskAgain->AddListener(this);
}



/*______________________________________________________________________________

	ListenToMessage

*/


void 
CFlashRomWindow::ListenToMessage(
	MessageT 	inMessage, 
	void *		ioParam)
{
	UInt32 		longParam = *(UInt32*)ioParam;

	switch(inMessage) 
	{
		case flashRom_Cancel:
		{
			if (mCPanel != nil)
				mCPanel->Show();
			DoClose();
			break;
		}
		
		case flashRom_Flash:
			DoFlashRom();
			break;
		
		case flashRom_AskAgain:
		{
			break;
		}
		
		default:
			break;
	}
}

#pragma mark -



/*______________________________________________________________________________

	InitFlashRomWindows

*/

void
CFlashRomWindow::InitFlashRomWindows()
{	

	mUpdateRom = (CUpdateRom*) LWindow::CreateWindow(UpdateRom_Window, this);
	ThrowIfNil_(mUpdateRom);
	mUpdateRom->SetCFlashRomWindow(this);
	mUpdateRom->SetPCIHardware(mBoard);
	mUpdateRom->CenterWindowOnDisplay();
	mUpdateRom->Hide();
	
	
	mEndUpdateRom = (CEndUpdateRom*) LWindow::CreateWindow(EndUpdateRom_Window, this);
	ThrowIfNil_(mEndUpdateRom);
	mEndUpdateRom->SetControlPanel(mCPanel);
	mEndUpdateRom->SetCFlashRomWindow(this);
	mEndUpdateRom->SetPCIHardware(mBoard);
	mEndUpdateRom->CenterWindowOnDisplay();
	mEndUpdateRom->Hide();
	
	
}



/*______________________________________________________________________________

	SetPCIHardware

*/

void
CFlashRomWindow::SetPCIHardware(
	CPCIHardware *		inBoard)
{	
	if ( inBoard != 0 )
	{
		mBoard = inBoard;
	}
	
	InitFlashRomWindows();
}


/*______________________________________________________________________________

	GetPCIHardware

*/


CPCIHardware *	
CFlashRomWindow::GetPCIHardware()
{	
	return	mBoard;
}


/*______________________________________________________________________________

	SetControlPanel

*/

void
CFlashRomWindow::SetControlPanel(
	CControlPanel *		inCPanel)
{	
	if ( inCPanel != 0 )
	{
		mCPanel = inCPanel;
	}

}


/*______________________________________________________________________________

	SettheDone

*/

void
CFlashRomWindow::SettheDone(
	Boolean 		inBoolean)
{	
	theDone = inBoolean;
}


/*______________________________________________________________________________

	SetRomUHdl

*/

void	
CFlashRomWindow::SetRomUHdl(
	Handle					inHandle)
{	
	mRomUHdl = inHandle;
}	

/*______________________________________________________________________________

	GetRomUHdl

*/

Handle	
CFlashRomWindow::GetRomUHdl()
{	
	return mRomUHdl;
}	

/*______________________________________________________________________________

	DoFlashRom

*/

Boolean
CFlashRomWindow::DoFlashRom()
{	
	Hide();
	
	Boolean	success = true;
	theDone = false;
	
	/*	Show the progess bar window while do the flash	*/
	StartFlashRom();
	
	/*	Flashing is over ... lets show the final window */
	EndFlashRomProcess();


	return success;
}



/*______________________________________________________________________________

	StartFlashRom

*/

void
CFlashRomWindow::StartFlashRom()
{	
	FLASHROMERROR theSuccess;
	TRomInfo theRom;
	TRomBuffer * theRomBuffer;
	
	//mUpdateRom->Show();
	Handle		theRomHandle;
	
	theRomHandle = GetRomUHdl();
	if ( !theRomHandle )
	{
		// no Rom Handle !!
		return;
	}
	
	HLock( theRomHandle );
	theRomBuffer = (TRomBuffer*) NewPtr( sizeof( TRomBuffer ) );

	if ( theRomBuffer )
	{
		theRom.mRegEntryID = mBoard->GetRegEntryID();
		theRomBuffer->mBufferSize = 64 * 1024;
		BlockMove( *theRomHandle, &theRomBuffer->mBuffer, theRomBuffer->mBufferSize );
		theRom.mBuffer = theRomBuffer;
		theSuccess = FlashROM( &theRom );

		DisposePtr( (Ptr) theRomBuffer );
	}
	
	HUnlock( theRomHandle );
}



/*______________________________________________________________________________

	EndFlashRomProcess

*/

void
CFlashRomWindow::EndFlashRomProcess()
{	
	//mUpdateRom->Hide();
	
	mEndUpdateRom->Show();	
}



/*______________________________________________________________________________

	CenterWindowOnDisplay

*/

void
CFlashRomWindow::CenterWindowOnDisplay()
{	
	GDHandle	theDevice  = ::GetDeviceList();

	while (theDevice != nil) {
	
		if ( (*theDevice)->gdRefNum == mBoard->GetRefNum() ) {
		
			Rect 	r = (**theDevice).gdRect;
			long	x,y;
			
			x = r.right - r.left;
			y = r.bottom - r.top;
			
			x = ( x - mFrameSize.width) / 2 + r.left;
			y = ( y - mFrameSize.height) / 2 + r.top;
			
	  		MoveWindowTo( x, y );
	  		mFrameLocation.h = y;
  			mFrameLocation.v = x;
	  	}

  		theDevice = ::GetNextDevice(theDevice);
	}
}



/*______________________________________________________________________________

	ClickInDrag

*/

void
CFlashRomWindow::ClickInDrag(
	const EventRecord	&inMacEvent)
{
	GDHandle	theDevice  = ::GetDeviceList();

	while (theDevice != nil) {
	
		if ( (*theDevice)->gdRefNum == mBoard->GetRefNum() ) {
		
			Rect 	r = (**theDevice).gdRect;		
												// Save old bounds (content region
												//	is in global coords)
			Rect	bounds = (*((WindowPeek)mMacWindowP)->contRgn)->rgnBBox;
			
			//Rect	dragRect = (**(GetGrayRgn())).rgnBBox;
			
			Rect	dragRect;
			
			dragRect.left = ((&inMacEvent)->where.h - bounds.left) + r.left;
			dragRect.top = ((&inMacEvent)->where.v - bounds.top) + 20 + r.top;
			dragRect.right = r.right - ((bounds.left + mFrameSize.width) - (&inMacEvent)->where.h);
			dragRect.bottom = r.bottom - ((bounds.top + mFrameSize.height) - (&inMacEvent)->where.v);

			::MacInsetRect(&dragRect, 4, 4);
			UDesktop::DragDeskWindow(this, inMacEvent, dragRect);
			
				// DragWindow will move the window. However, send AppleEvent to
				// set the bounds (but don't execute it), so that Script engines
				// can record the action.
				
												// Compare old and new bounds
			if (!::MacEqualRect(&bounds,
							    &(*((WindowPeek)mMacWindowP)->contRgn)->rgnBBox)) {
							
				Point	newPosition =
							topLeft((*((WindowPeek)mMacWindowP)->contRgn)->rgnBBox);
							
												// Send but don't execute
												//   SetPosition AppleEvent
				SendAESetPosition(newPosition, ExecuteAE_No);
				
				AdjustUserBounds();				// Moving Window changes user
												//   bounds for zooming
												
				
			}
					

	  	}
  			
  		theDevice = ::GetNextDevice(theDevice);
	}

}



/*______________________________________________________________________________

	DoFlashQuit

*/

void
CFlashRomWindow::DoFlashQuit()
{
	
	mUpdateRom->DoClose();
	mEndUpdateRom->DoClose();

	ObeyCommand(cmd_Quit, nil);

}



/*______________________________________________________________________________

	DoClose

*/

void
CFlashRomWindow::DoFlashRestart()
{
//	char theFinderSig[] = "FNDR";
	OSType fndrCreator = 'MACS';
	AEDesc theFinderAddr;
	AppleEvent theShutDown;
	AppleEvent theReply;
	OSErr theErr;
	
	//DoClose();

	theErr = ::AECreateDesc( typeApplSignature, (Ptr) &fndrCreator, sizeof(OSType), &theFinderAddr);
	if ( theErr == noErr )
	{
		theErr = ::AECreateAppleEvent( kAEFinderEvents, kAERestart, &theFinderAddr,
								kAutoGenerateReturnID, kAnyTransactionID, &theShutDown);
		if ( theErr == noErr )
		{
			theErr = ::AESend( &theShutDown, &theReply, kAENoReply + kAECanSwitchLayer + kAEAlwaysInteract,
										kAENormalPriority, kAEDefaultTimeout, 0, 0);
		}
	}
}











