/*
** Copyright (c) 1999, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/

#include "CAboutWindow.h"
#include "CUpdateDrv.h"
#include "CGETDrvThread.h"
#include "ControlPanelPPob.h"
#include "UHqx.h"
#include "UFullPath.h"
#include "UFinderLaunch.h"

#include <UInternet.h>
#include <stdio.h>
#include <LURL.h>
#include <UInternetConfig.h>
#include <UModalDialogs.h>

#include <LHTTPMessage.h>
#include <LHTTPConnection.h>
#include <LTextEditView.h>
#include <LChasingArrows.h>
#include <String.h>
#include <LString.h>
#include <Folders.h>
#include <Files.h>

#include <Types.h>
#include <Windows.h>
#include <Dialogs.h>
#include <Gestalt.h>
#include <TextUtils.h>
#include <Errors.h>
#include <Memory.h>
#include <ErrMgr.h>
#include <PLStringFuncs.h>
#include <CursorCtl.h>
#include <stdlib.h>

#define kOutputBufferSize 	(8*1024)

FILE		*gSrcFile;
FSSpec 		targetfile;


static OSErr MySource(void* buffer, long *count, long param) {
	//SpinCursor(16);
	*count =  fread(buffer, 1, *count, gSrcFile);
	return noErr;
}



static OSErr MyNameFilter(StringPtr name, short *vol, long *dir, long param) {
	FInfo info;
	
	PLstrcpy(name, targetfile.name);
	*vol = targetfile.vRefNum;
	*dir = targetfile.parID;
	/*
	if (replace_names) return noErr;
	if (HGetFInfo(*vol, *dir, name, &info) == noErr) {
		char** pathname;
		Str255 tname;
		long n;
		if ((pathname = GetFullPath(name, *vol, *dir)) != NULL) {
			n = GetHandleSize((Handle) pathname);
			if (n > 255) n = 255;
			tname[0] = n;
			BlockMoveData(*pathname, tname+1, n);
			ParamText(tname, NULL, NULL, NULL);
			DisposeHandle((Handle) pathname);
		} else ParamText(name, NULL, NULL, NULL);
		InitCursor();
		if (Alert(129, NULL) == ok) {
			return HDelete(*vol, *dir, name);
		} else return userCanceledErr;
	}*/
	return noErr;
}


/*______________________________________________________________________________

	Constructor

*/

CGETDrvThread::CGETDrvThread(CUpdateDrv * inUpdateDrvWindow)
		: LThread(false, thread_DefaultStack, threadOption_Default, nil),
			mUpdateDrvWindow(inUpdateDrvWindow)
{
}


/*______________________________________________________________________________

	Destructor

*/

CGETDrvThread::~CGETDrvThread()
{

}

#pragma mark -


/*______________________________________________________________________________

	FinishCreateSelf

*/

void
CGETDrvThread::FinishCreateSelf()
{
}


	
/*______________________________________________________________________________

	ListenToMessage

*/

void
CGETDrvThread::ListenToMessage(
	MessageT inMessage, 
	void *ioParam)
{
	LStr255 addon1, addon2;
	SProgressMessage * theMsg = (SProgressMessage *) ioParam;
	mCaptionText.Remove(0, 255);
	
	switch (inMessage) {
		case msg_RetrieveItemFailed:
			mCaptionText.Assign(UpdateDrvStatus_STR, kFailedRetrieve);
			mCaptionText.Append((StringPtr)theMsg->currentItem);
			SetCaptionDescriptor(updatedrv_Status, mCaptionText);
			break;

		case msg_RetrieveItemSuccess:
			mCaptionText.Assign(UpdateDrvStatus_STR, kFinishRetrieve);
			mCaptionText.Append((StringPtr)theMsg->currentItem);
			SetCaptionDescriptor(updatedrv_Status, mCaptionText);
			break;
		
		case msg_OpeningConnection:
			mCaptionText.Assign(UpdateDrvStatus_STR, kOpenConnection);
			mCaptionText.Append((StringPtr)theMsg->currentItem);
			SetCaptionDescriptor(updatedrv_Status, mCaptionText);
			break;

		case msg_Connected:
			mCaptionText.Assign(UpdateDrvStatus_STR, kConnectedTo);
			mCaptionText.Append((StringPtr)theMsg->currentItem);
			SetCaptionDescriptor(updatedrv_Status, mCaptionText);
			break;
			
		case msg_Disconnected:
			SetCaptionDescriptor(UpdateDrvStatus_STR, updatedrv_Status, kConnectionTerminated);
			break;

		case msg_ReceivingData:
			addon1.Remove(0, 255); 
			addon2.Remove(0, 255);
			addon1.Assign(UpdateDrvStatus_STR, kOf);
			addon2.Assign(UpdateDrvStatus_STR, kKiloUnit);
			mCaptionText.Assign(UpdateDrvStatus_STR, kReceivedK);
				
			if (theMsg->totalBytes) 
			{
				float temp1 = theMsg->completedBytes;
				float temp2 = theMsg->totalBytes;
				UInt32 thePercentage = (temp1/temp2) * 100;
				if (thePercentage > 100)
					thePercentage = 100;
				UInt32 theTotalK = (theMsg->totalBytes/1024);
				
				mCaptionText.Append((long)thePercentage);
				mCaptionText.Append(addon1);
				mCaptionText.Append((long)theTotalK);
				mCaptionText.Append(addon2);
			} 
			else 
			{
				mCaptionText.Append((long)theMsg->completedBytes/1024);
				mCaptionText.Append(addon2);		
			}
			SetCaptionDescriptor(updatedrv_Status, mCaptionText);
			break;
		
		default:
			break;
	}
}

#pragma mark -


/*______________________________________________________________________________

	Run

*/

void *
CGETDrvThread::Run()
{
	OSErr 		err = noErr;
	char 		statusMessage[256];
	SInt32 		theResponseCode;
	FSSpec		theHQXFSSpec;
	
	mUpdateDrvWindow->IsGETThreadActive = true;
	
	LPushButton* 	theCancelButt	= (LPushButton*) mUpdateDrvWindow->FindPaneByID(updatedrv_Cancel);
	LPushButton* 	theDldButt		= (LPushButton*) mUpdateDrvWindow->FindPaneByID(updatedrv_Download);
	LChasingArrows* theArrows 	 	= (LChasingArrows*) mUpdateDrvWindow->FindPaneByID('carr');
	
	mProgress = (LCaption*) mUpdateDrvWindow->FindPaneByID(updatedrv_Status);
		
	LStr255 theResource (UPDATEDRV_RESOURCE);
	theResource.Append(mUpdateDrvWindow->GetHqxName());	
	
	
	if (theArrows != nil)	theArrows->Show();
	
	LHTTPResponse theResponse;
	LHTTPConnection connection(*this);
	connection.AddListener(this);

	
	try {
		
		LHTTPMessage theMessage;
		
		BuildMessage(theMessage);
		
		SetCaptionDescriptor(UpdateDrv_STR, updatedrv_Line1, kNowDownloading);
		
		if (theDldButt != nil)	theDldButt->Disable();
							
		connection.RequestResource(
				UPDATEDRV_HOST,
				theResource,
				theMessage,
				theResponse,
				(EHTTPMethod)0);
		
		theResponseCode = (SInt32) (&theResponse)->GetResponseCode();

		if (theResponseCode == 200)
		{
			/*	Lets save the file on the HD */
			SetCaptionDescriptor(UpdateDrv_STR, updatedrv_Line1, kNowSaving);
			SetCaptionDescriptor(UpdateDrvStatus_STR, updatedrv_Status, kSavingFile);
			
			err = SaveDownloadDrivers(&theResponse);
			if (err != noErr)
			{
				/*	Error while saving	*/
				SetCaptionDescriptor(UpdateDrv_STR, updatedrv_Line1, kErrorSaving);		

				mCaptionText.Assign(UpdateDrvStatus_STR, kIdle);
				mCaptionText.Append(kOpenBracket);
				mCaptionText.Append(err);
				mCaptionText.Append(kCloseBracket);
				SetCaptionDescriptor(updatedrv_Status, mCaptionText);

				if (theDldButt != nil)	theDldButt->Enable();
			}
			else
			{
				/*	Lets decode the hqx file ! */
				SetCaptionDescriptor(UpdateDrv_STR, updatedrv_Line1, kNowDecompressing);
				SetCaptionDescriptor(UpdateDrvStatus_STR, updatedrv_Status, kDecompressingFile);
				
				err = DecompressDownloadDrivers();
				if (err != noErr)
				{
					/*	Error while decompressing	*/
					SetCaptionDescriptor(UpdateDrv_STR, updatedrv_Line1, kErrorDecompressing);

					mCaptionText.Assign(UpdateDrvStatus_STR, kIdle);
					mCaptionText.Append(kOpenBracket);
					mCaptionText.Append(err);
					mCaptionText.Append(kCloseBracket);
					SetCaptionDescriptor(updatedrv_Status, mCaptionText);

					if (theDldButt != nil)	theDldButt->Enable();
				}
				else
				{
					/*	We can launch the decompressed HQX file (i.e. the SEA file)	*/
					err = FinderLaunch(1, &targetfile);
					
					if (err != noErr)
					{
						/*	Error while decompressing the SEA file (not that bad)	*/
						SetCaptionDescriptor(UpdateDrv_STR, updatedrv_Line1, kErrorDecompressing);

						mCaptionText.Assign(UpdateDrvStatus_STR, kIdle);
						mCaptionText.Append(kOpenBracket);
						mCaptionText.Append(err);
						mCaptionText.Append(kCloseBracket);
						SetCaptionDescriptor(updatedrv_Status, mCaptionText);
					}
					else
					{					
						/*	Everything went just fine !	*/
						
						if (theCancelButt != nil)	theCancelButt->SetDescriptor("\pClose");
						
						SetCaptionDescriptor(UpdateDrv_STR, updatedrv_Line1, kAllGood);
						SetCaptionDescriptor(UpdateDrvStatus_STR, updatedrv_Status, kIdle);
					}
					
					/*	We can now delete the HQX file	*/
					theDrvFile->GetSpecifier(theHQXFSSpec);
					err = FSpDelete(&theHQXFSSpec);
					
				}
			}
		}
		else
		{
			/*	Error while downloading	*/
			SetCaptionDescriptor(UpdateDrv_STR, updatedrv_Line1, kErrorDownloading);

			mCaptionText.Assign(UpdateDrvStatus_STR, kIdle);
			mCaptionText.Append(kOpenBracket);
			mCaptionText.Append(theResponseCode);
			mCaptionText.Append(kCloseBracket);
			SetCaptionDescriptor(updatedrv_Status, mCaptionText);

						
			if (theDldButt != nil)	theDldButt->Enable();
		}
		
		
	}
	catch (LHTTPResponse & theResponse) {
		//SysBeep(30);
		//SetUpdateDrvResponse(&theResponse);
		DisplayResponseError();
	}
	catch (...) {
		//Could/should display error messags here
		//SysBeep(30);
		DisplayResponseError();
	}

	//if (theCancel != nil)	theCancel->Enable();
	if (theArrows != nil)	theArrows->Hide();
	connection.RemoveListener(this);

	mUpdateDrvWindow->IsGETThreadActive = false;
	return nil;

}


/*______________________________________________________________________________

	BuildMessage

*/

void
CGETDrvThread::BuildMessage(LHTTPMessage& theMessage)
{
	theMessage.SetUserName("");
	theMessage.SetPassword("");
	theMessage.SetUserAgent(USER_AGENT);
}


/*______________________________________________________________________________

	SetCaptionDescriptor

*/

void
CGETDrvThread::SetCaptionDescriptor(
	PaneIDT			inCaptionID,
	LStr255			inTextDescriptor)
{
	LCaption* 		theCaption 	 = (LCaption*) mUpdateDrvWindow->FindPaneByID(inCaptionID);
	
	if (theCaption != nil)	
			theCaption->SetDescriptor(inTextDescriptor);
}
		
/*______________________________________________________________________________

	SetCaptionDescriptor

*/

void
CGETDrvThread::SetCaptionDescriptor(
	ResIDT			inSTRID,
	PaneIDT			inCaptionID,
	UInt16			inTextDescriptor)
{
	LCaption* 		theCaption 	 = (LCaption*) mUpdateDrvWindow->FindPaneByID(inCaptionID);	
	
	mCaptionText.Remove(0, 255);
	mCaptionText.Assign(inSTRID, inTextDescriptor);
	
	if (theCaption != nil)	
			theCaption->SetDescriptor(mCaptionText);
}

/*______________________________________________________________________________

	DisplayProgress

*/

void
CGETDrvThread::DisplayProgress(char * theStatus)
{
	c2pstr(theStatus);
	mProgress->SetDescriptor((unsigned char *)theStatus);
}


/*______________________________________________________________________________

	DisplayError

*/

void
CGETDrvThread::DisplayResponseError()
{
	SetCaptionDescriptor(UpdateDrv_STR, updatedrv_Line1, kErrorDownloading);
	SetCaptionDescriptor(UpdateDrvStatus_STR, updatedrv_Status, kIdle);
}



/*______________________________________________________________________________

	SetDownloadFileName

*/

OSErr 
CGETDrvThread::SetDownloadFileName()
{
	OSErr		err = noErr;
	SInt16		desktopVol = -1;
	SInt32		desktopDir;
	FSSpec		theDesktopSpec;
	
	
	// Get the location of the desktop folder
	err = FindFolder(kOnSystemDisk, kDesktopFolderType, 
				kDontCreateFolder, &desktopVol, &desktopDir);
	if (err != noErr)
		return err;

	// Make a spec to it so we can open the thing.
	err = FSMakeFSSpec(desktopVol, desktopDir, 
			mUpdateDrvWindow->GetHqxName(), &theDesktopSpec);
	if (err == fnfErr)
	{
		err = noErr;
		FSpCreateResFile(&theDesktopSpec, kDRVFILE_CREATOR, kDRVFILE_TYPE, smSystemScript);	
	}
	else if (err != noErr)
	{
		return err;
	}
	
	theDrvFile = new LFile();
	
	theDrvFile->SetSpecifier(theDesktopSpec);
	
	return err;
}

/*______________________________________________________________________________

	SaveDownloadDrivers

*/

OSErr 
CGETDrvThread::SaveDownloadDrivers(
	LHTTPResponse * 		theResponse)
{
	OSErr		err = noErr;
	
	err = SetDownloadFileName();
	if (err != noErr)
		return err;
	
	// Get the text out of the message and clean it up 
	LDynamicBuffer * theDisplay = theResponse->GetReturnMessage()->GetMessageBody();

	theDisplay->BufferToFile(*theDrvFile);
		
	delete theDisplay;

	return err;
}

/*______________________________________________________________________________

	DecompressDownloadDrivers

*/

OSErr 
CGETDrvThread::DecompressDownloadDrivers()
{
	OSErr		err	= noErr;
	char 		*output_buffer = NULL;
	FSSpec		theFileNameSpec;
	Handle 		fullPath;
	short 		fullPathLength;
	char		theFileName[256];
	
	/* Create the decoding-output buffer	*/
	if (output_buffer == NULL) 
	{
		output_buffer = (char*) NewPtr(kOutputBufferSize);
		if ((err = MemError()) != noErr)
		 	goto misc_abort;
	}
	
	/* Get the FullPath of the hqx file for fopen	*/
	theDrvFile->GetSpecifier(theFileNameSpec);
	
	err = GetFullPath(theFileNameSpec.vRefNum,
							theFileNameSpec.parID,
							theFileNameSpec.name,
							&fullPathLength,
							&fullPath);
	
	if (err != noErr)
		goto misc_abort;
	
	strncpy (theFileName, *fullPath, fullPathLength);
	theFileName[fullPathLength] = '\0';
	
	gSrcFile = fopen(theFileName, "r");
	
	if (gSrcFile == NULL) 
		goto misc_abort;
		
	setvbuf(gSrcFile, output_buffer, _IOFBF, kOutputBufferSize);
	
	/* Create the FSSpec for the Decompressed file	*/
	err = FSMakeFSSpec(theFileNameSpec.vRefNum, 
						theFileNameSpec.parID, 
						mUpdateDrvWindow->GetDriverName(), 
						&targetfile);
						
	if (err == fnfErr) 
		err = noErr;
	if (err != noErr)
		goto misc_abort;
	
	/*	Starting decoding	*/
	err = HQXDecode(MySource, MyNameFilter, true, true, 0);
	if (err != noErr) 
		goto misc_abort;

	fclose(gSrcFile);
	DisposePtr((Ptr) output_buffer);
			
			
	return err;
	
misc_abort:
	if (gSrcFile != NULL) fclose(gSrcFile);
	if (output_buffer != NULL) DisposePtr((Ptr) output_buffer);
	return err;
}
