/*
** Copyright (c) 1999, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/

#include "CControlPanelApp.h"
#include "CAboutWindow.h"
#include "CUpdateDrv.h"
#include "CGETVerThread.h"
#include "ControlPanelPPob.h"

#include <UInternet.h>
#include <stdio.h>
#include <LURL.h>
#include <UInternetConfig.h>
#include <UModalDialogs.h>

#include <LHTTPMessage.h>
#include <LHTTPConnection.h>
#include <LTextEditView.h>
#include <LChasingArrows.h>
#include <String.h>
#include <LString.h>
#include <stdlib.h>

#define		kMajorVersion	3
#define		kMediumVersion	2
#define		kMinorVersion	1

/*______________________________________________________________________________

	Constructor

*/

CHTTPSendMsgThread::CHTTPSendMsgThread(CUpdateDrv * inUpdateDrvWindow)
		: LThread(false, thread_DefaultStack, threadOption_Default, nil),
			mUpdateDrvWindow(inUpdateDrvWindow)
{
}


/*______________________________________________________________________________

	Destructor

*/

CHTTPSendMsgThread::~CHTTPSendMsgThread()
{

}

#pragma mark -


/*______________________________________________________________________________

	FinishCreateSelf

*/

void
CHTTPSendMsgThread::FinishCreateSelf()
{
}


	
/*______________________________________________________________________________

	ListenToMessage

*/

void
CHTTPSendMsgThread::ListenToMessage(
	MessageT inMessage, 
	void *ioParam)
{
	LStr255 addon;
	SProgressMessage * theMsg = (SProgressMessage *) ioParam;
	mCaptionText.Remove(0, 255);
	
	switch (inMessage) {
		case msg_RetrieveItemFailed:
			mCaptionText.Assign(UpdateDrvStatus_STR, kFailedRetrieve);
			mCaptionText.Append((StringPtr)theMsg->currentItem);
			SetCaptionDescriptor(updatedrv_Status, mCaptionText);
			break;

		case msg_RetrieveItemSuccess:
			mCaptionText.Assign(UpdateDrvStatus_STR, kFinishRetrieve);
			mCaptionText.Append((StringPtr)theMsg->currentItem);
			SetCaptionDescriptor(updatedrv_Status, mCaptionText);
			break;
		
		case msg_OpeningConnection:
			mCaptionText.Assign(UpdateDrvStatus_STR, kOpenConnection);
			mCaptionText.Append((StringPtr)theMsg->currentItem);
			SetCaptionDescriptor(updatedrv_Status, mCaptionText);
			break;

		case msg_Connected:
			mCaptionText.Assign(UpdateDrvStatus_STR, kConnectedTo);
			mCaptionText.Append((StringPtr)theMsg->currentItem);
			SetCaptionDescriptor(updatedrv_Status, mCaptionText);
			break;
			
		case msg_Disconnected:
			SetCaptionDescriptor(UpdateDrvStatus_STR, updatedrv_Status, kConnectionTerminated);
			break;

		case msg_ReceivingData:
			if (theMsg->totalBytes) 
			{
				float temp1 = theMsg->completedBytes;
				float temp2 = theMsg->totalBytes;
				UInt32 thePercentage = (temp1/temp2) * 100;
				if (thePercentage > 100)
					SetCaptionDescriptor(UpdateDrvStatus_STR, updatedrv_Status, kVersionReceived);
				else
					SetCaptionDescriptor(UpdateDrvStatus_STR, updatedrv_Status, kReceivingVersion);
			} 
			else 
			{
				addon.Remove(0, 255);
				addon.Assign(UpdateDrvStatus_STR, kKiloUnit);
				mCaptionText.Assign(UpdateDrvStatus_STR, kReceivedK);
				mCaptionText.Append((long)theMsg->completedBytes/1024);
				mCaptionText.Append(addon);
				SetCaptionDescriptor(updatedrv_Status, mCaptionText);
			}
			break;
		
		default:
			break;
	}
}

#pragma mark -


/*______________________________________________________________________________

	Run

*/

void *
CHTTPSendMsgThread::Run()
{
	
	mUpdateDrvWindow->IsHTTPThreadActive = true;
	
	mProgress = (LCaption*) mUpdateDrvWindow->FindPaneByID(updatedrv_Status);
	LChasingArrows* theArrows = (LChasingArrows*) mUpdateDrvWindow->FindPaneByID('carr');
	
	/*	Set the resource information */
	(&mVersionFile)->Assign(UpdateFile_STR, kVersionFileName);
	LStr255 theResource (UPDATEDRV_RESOURCE);
	theResource.Append(mVersionFile);
	
	if (theArrows != nil)	theArrows->Show();
	
	LHTTPResponse theResponse;
	LHTTPConnection connection(*this);
	connection.AddListener(this);

	try {
		
		LHTTPMessage theMessage;
		
		BuildMessage(theMessage);
						
		connection.RequestResource(
				UPDATEDRV_HOST,
				theResource,
				theMessage,
				theResponse,
				(EHTTPMethod)0);
				
		SetUpdateDrvResponse(&theResponse);
		
	}
	catch (LHTTPResponse & theResponse) {
		//SysBeep(30);
		SetUpdateDrvResponse(&theResponse);
	}
	catch (...) {
		//Could/should display error messags here
		//SysBeep(30);
		DisplayResponseError();
	}

	if (theArrows != nil)	theArrows->Hide();
	connection.RemoveListener(this);
	
	mUpdateDrvWindow->IsHTTPThreadActive = false;
	return nil;

}


/*______________________________________________________________________________

	BuildMessage

*/

void
CHTTPSendMsgThread::BuildMessage(LHTTPMessage& theMessage)
{
	theMessage.SetUserName("");
	theMessage.SetPassword("");
	theMessage.SetUserAgent(USER_AGENT);
}




/*______________________________________________________________________________

	DisplayProgress

*/

void
CHTTPSendMsgThread::DisplayProgress(char * theStatus)
{
	c2pstr(theStatus);
	mProgress->SetDescriptor((unsigned char *)theStatus);
}


/*______________________________________________________________________________

	DisplayError

*/

void
CHTTPSendMsgThread::DisplayResponseError()
{
	
	SetCaptionDescriptor(UpdateDrv_STR, updatedrv_Line1, kErrorConnecting);
	SetCaptionDescriptor(UpdateDrvStatus_STR, updatedrv_Status, kIdle);
}


/*______________________________________________________________________________

	SetUpdateDrvResponse

*/

void 
CHTTPSendMsgThread::SetUpdateDrvResponse(
	LHTTPResponse * 		theResponse)
{
	//Set the result code field
	SInt32 theResponseCode = theResponse->GetResponseCode();
	char statusMessage[256];
	
	if (theResponseCode == 200)
	{
		LStr255  theVersion, theDriverName, theHqxName;
	
		/* Get the text out of the message and clean it up */
		LDynamicBuffer * theDisplay = theResponse->GetReturnMessage()->GetMessageBody();
	
		/* Get the needed informations */
		//Grab at most 32000 chars from the buffer to display
		StHandleBlock theText(Size_Zero);
		ThrowIfNot_(theDisplay->BufferToHandle(theText, 32000));
		
		GetInformationfromResponse(theText, "DRIVERNAME", &theDriverName);
		mUpdateDrvWindow->SetDriverName(theDriverName);
		
		GetInformationfromResponse(theText, "HQXNAME", &theHqxName);
		mUpdateDrvWindow->SetHqxName(theHqxName);
		
		GetInformationfromResponse(theText, "VERSION", &theVersion);
		mUpdateDrvWindow->SetLastVersion(theVersion);
		
			
		if (IsNewerVersion(&theVersion))
		{
			SetCaptionDescriptor(UpdateDrv_STR, updatedrv_Line1, kNewerVersionFound);
			SetCaptionDescriptor(UpdateDrvStatus_STR, updatedrv_Status, kIdle);
			
			LPushButton* theDldButton = (LPushButton*) mUpdateDrvWindow->FindPaneByID(updatedrv_Download);
			if (theDldButton != nil)	theDldButton->Enable();
		}
		else
		{
			SetCaptionDescriptor(UpdateDrv_STR, updatedrv_Line1, kNoNewerVersionFound);
			SetCaptionDescriptor(UpdateDrvStatus_STR, updatedrv_Status, kIdle);
		}
		
		delete theDisplay;
	}
	else
	{
		SetCaptionDescriptor(UpdateDrv_STR, updatedrv_Line1, kErrorCheckingVersion);
		
		mCaptionText.Remove(0, 255);
		mCaptionText.Assign(UpdateDrvStatus_STR, kIdle);
		mCaptionText.Append(" [");
		mCaptionText.Append(theResponseCode);
		mCaptionText.Append("]");
		SetCaptionDescriptor(updatedrv_Status, mCaptionText);	
	}
	
}

/*______________________________________________________________________________

	SetCustomCaptionDescriptor

*/

void
CHTTPSendMsgThread::SetCaptionDescriptor(
	PaneIDT			inCaptionID,
	LStr255			inTextDescriptor)
{
	LCaption* 		theCaption 	 = (LCaption*) mUpdateDrvWindow->FindPaneByID(inCaptionID);
	
	if (theCaption != nil)	
			theCaption->SetDescriptor(inTextDescriptor);
}


/*______________________________________________________________________________

	SetCaptionDescriptor

*/

void
CHTTPSendMsgThread::SetCaptionDescriptor(
	ResIDT			inSTRID,
	PaneIDT			inCaptionID,
	UInt16			inTextDescriptor)
{
	LCaption* 		theCaption 	 = (LCaption*) mUpdateDrvWindow->FindPaneByID(inCaptionID);	
	
	mCaptionText.Remove(0, 255);
	mCaptionText.Assign(inSTRID, inTextDescriptor);
	
	if (theCaption != nil)	
			theCaption->SetDescriptor(mCaptionText);
}

/*______________________________________________________________________________

	GetInformationfromResponse

*/

Boolean 
CHTTPSendMsgThread::GetInformationfromResponse(
	Handle  			theTextHandle, 
	const char * 		inTagName,
	LStr255	*			outValue)
{
	Boolean 	FindIt = false;
	char 		*theMark, *theTag, *temp;
	SInt32		theMarkLength, theTagLength;
	
	strcpy(theTag, inTagName);
	temp = theTag;
	
	theMark = *theTextHandle;
	theMarkLength = strlen(theMark);
	
	/* Find the inTagName inside the MessageBody */
	while ((theMarkLength >= 0) && !FindIt)
	{
		if (*theMark == *temp)
		{
			*temp++; *theMark++;
			if (*temp == '\0')
				if (*theMark == '=')
					FindIt = true;
				else
				{
					temp = theTag;
					theMarkLength--;
				}
			else
				theMarkLength--;
		}	
		else
		{
			*theMark++;
			theMarkLength--;
		}
	}
	
	/* if we found it, get the value then */
	if (FindIt)
	{
		*theMark++;
		theMarkLength = theMarkLength -2;
		while ((theMarkLength >= 0) && (*theMark != '>'))
		{
			outValue->Append(*theMark);
			*theMark++;
			theMarkLength--;
		}
		if (*theMark == '>')
			return true;
		else
			return false;
	}
	else
		return false;
}


/*______________________________________________________________________________

	IsNewerVersion

*/

Boolean 
CHTTPSendMsgThread::IsNewerVersion(
	LStr255 *		inVersion)
{
	//return (*inVersion == CURRENT_DRIVER_VERSION) ? false : true;
	UInt32	currentVersion = 0;
	UInt32	serverVersion = 0;

	currentVersion = ExtrudeVersion( CURRENT_DRIVER_VERSION, (int) strlen( CURRENT_DRIVER_VERSION ) );
	serverVersion = ExtrudeVersion( (char*) inVersion->TextPtr(), inVersion->Length());
	
	return (serverVersion > currentVersion ? 1 : 0);
}

/*______________________________________________________________________________

	ExtrudeVersion

*/

UInt32 
CHTTPSendMsgThread::ExtrudeVersion(
	char *		inVersion,
	int			inVersionLength)
{
	UInt32  theTempVersion = 0;
	Boolean IsFirstDigit = true;
	SInt8	VersionLength;
	UInt8	digit = 1;
	UInt8 	value = 0;
	UInt8	position = 0;
	UInt8	stage = 3;
	
	while( inVersionLength >= 0  )
	{
		if ( inVersion[position] == '.' || inVersionLength == 0)
		{
			switch(stage)
			{
				case kMajorVersion:
					theTempVersion += value << 16;
					break;
				case kMediumVersion:
					theTempVersion += value << 8;
					break;
				case kMinorVersion:
					theTempVersion += value;
					break;
				default:
					break;
			}
			digit = 1;
			value = 0;
			stage--;
			position++;
			inVersionLength--;
		}
		else
			if (!IsFirstDigit)
				digit *= 10;
			else
				IsFirstDigit = false;
		
		value += atoi((char*) &inVersion[position++])*digit;
		inVersionLength--;
	}
	
	
	if (stage != 0)
	{
		value = 0;
		while (stage != 0)
		{
			switch(stage)
			{
				case kMajorVersion:
					theTempVersion += value << 16;
					break;
				case kMediumVersion:
					theTempVersion += value << 8;
					break;
				case kMinorVersion:
					theTempVersion += value;
					break;
				default:
					break;
			}
			
			stage--;
		}
	}
	
	return theTempVersion;
	
}
