/*
** Copyright (c) 2000, 3dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/

#include "CGlideView.h"
#include <LCaption.h>
#include <LPopupButton.h>
#include "hrm_user_settings.h"
#include "CPopupButton.h"

/*______________________________________________________________________________

	Constructor

*/

CGlideView::CGlideView()
{
}


/*______________________________________________________________________________

	Constructor (using a Stream, that's the one that will be used)

*/

CGlideView::CGlideView(
	LStream *			inStream)
	: LView ( inStream )
{
}


/*______________________________________________________________________________

	Destructor

*/

CGlideView::~CGlideView()
{
	StopListening();
}


#pragma mark -





/*______________________________________________________________________________

	FinishCreateSelf

*/

void
CGlideView::FinishCreateSelf()
{
	mAASampling = (LPopupButton*) this->FindPaneByID(perf3D_AAliasing);
	if (mAASampling != nil)		mAASampling->AddListener(this);
	
	mAAKey = (CKeyComboField*) this->FindPaneByID(perf3D_AAliasingKey);
	if (mAAKey != nil)		mAAKey->AddListener(this);

	mGamma = (LSlider*) this->FindPaneByID(perf3D_GammaTbl);
	if (mGamma != nil)		mGamma->AddListener(this);

	mVSync = (LPopupButton*) this->FindPaneByID(perf3D_VSync);
	if (mVSync != nil)		mVSync->AddListener(this);
	
	m32Bit = (LPopupButton*) this->FindPaneByID(perf3D_32Bit);
	if (m32Bit != nil)		m32Bit->AddListener(this);
	
	StartListening();
}





/*______________________________________________________________________________

	ListenToMessage

*/

void
CGlideView::ListenToMessage(
	MessageT 			inMessage, 
	void *				ioParam)
{
	UInt32 		longParam = *(UInt32*)ioParam;
	Boolean 	res;
	
	if ( mBoard )
	{
		switch(inMessage)
		{
			case msg_3DPerf_AAliasing:
				res = mBoard->SetSettingUL(HRM_GlobalGlideAntiAliasing, *((UInt32*)ioParam));
				break;

			case perf3D_AAliasingKey:
				res = mBoard->SetSettingUL(HRM_GlobalGlideAntiAliasingKey, *((UInt32*)ioParam));
				break;

			case msg_3DPerf_GammaTbl:
			{
				float theFloat = 1.0 + ( longParam * 0.1 );			
				res = mBoard->SetSettingF(HRM_GlobalGlideGamma, theFloat);
				break;
			}
			
			case msg_3DPerf_VSync:
				res = mBoard->SetSettingUL(HRM_GlobalGlideVSyncEnable, longParam);
				break;
				
			case perf3D_32Bit:
				res = mBoard->SetSettingUL(HRM_GlobalGlideForce32bit, longParam);
				break;
			
			default:
				break;
		}
	}
}





/*______________________________________________________________________________

	SetPCIHardware

*/

void
CGlideView::SetPCIHardware(
	CPCIHardware *		inBoard)
{
	if ( inBoard != 0 ) {
		mBoard = inBoard;
		
		LoadGlide_View();
	}

}


/*______________________________________________________________________________

	LoadGlide_View

*/

Boolean
CGlideView::LoadGlide_View()
{
	UInt32 		theValue;
	float 		theFloat;
	Boolean 	result = true;
	
	mBoard->GetSettingUL(HRM_GlobalGlideAntiAliasing, &theValue);
	mAASampling->SetCurrentMenuItem( theValue );
	
	mBoard->GetSettingUL(HRM_GlobalGlideAntiAliasingKey, &theValue);
	mAAKey->SetValue( theValue );
	
	mBoard->GetSettingF(HRM_GlobalGlideGamma, &theFloat);
	theValue = (theFloat - 1.0) * 10;
	mGamma->SetValue( theValue );
	
	mBoard->GetSettingUL(HRM_GlobalGlideVSyncEnable, &theValue);
	mVSync->SetCurrentMenuItem( theValue );
	
	mBoard->GetSettingUL(HRM_GlobalGlideForce32bit, &theValue);
	m32Bit->SetCurrentMenuItem( theValue );
	
		
	if (mVVersion == 3)
	{
		/*	Disable some popupmenu 	*/
		CPopupButton * theVSyncPopupButton = (CPopupButton*) FindPaneByID(perf3D_VSync);
		if (theVSyncPopupButton != nil)	
		{
			theVSyncPopupButton->SetCurrentMenuItem( kDisabled );
			theVSyncPopupButton->Disable();
		}
		
		LEditText * theAAKeyEditText = (LEditText*) FindPaneByID(perf3D_AAliasingKey);
		if (theAAKeyEditText != nil)
		{
			theAAKeyEditText->Disable();
		}
		
		LSlider * theGammaSlider = (LSlider*) FindPaneByID(perf3D_GammaTbl);
		if (theGammaSlider != nil)		
		{
			theGammaSlider->Disable();
		}
		
		LPopupButton * the32BitPopupButton = (LPopupButton*) FindPaneByID(perf3D_32Bit);
		if (the32BitPopupButton != nil)
		{
			the32BitPopupButton->SetCurrentMenuItem( kDisabled );
			the32BitPopupButton->Disable();
		}
		
		/*	Disable some option in the PopupMenu	*/
		CPopupButton * theAAPopupButton = (CPopupButton*) FindPaneByID(perf3D_AAliasing);
		if (theAAPopupButton != nil)
		{
			theAAPopupButton->SetCurrentMenuItem( kSingleChip );
		 	MenuHandle menuH = theAAPopupButton->GetMacMenuH();
			::DisableItem(menuH, 2);
			::DisableItem(menuH, 3);
			::DisableItem(menuH, 4);
		}
	}
	else if (mVVersion == 4)
	{
		/*	Disable some option in the PopupMenu	*/
		CPopupButton * theAAPopupButton = (CPopupButton*) FindPaneByID(perf3D_AAliasing);
		if (theAAPopupButton != nil)
		{
			SInt16 theAAValue = theAAPopupButton->GetCurrentMenuItem();
			
			if ( theAAValue < 2 || theAAValue > 3 )
				theAAPopupButton->SetCurrentMenuItem( kDefault3DAA );
			
		 	MenuHandle menuH = theAAPopupButton->GetMacMenuH();
			::DisableItem(menuH, 1);
			::DisableItem(menuH, 4);
		}
	}
	else if (mVVersion == 5)
	{
		mBoard->GetSettingUL(HRM_GlobalGlideAntiAliasing, &theValue);
		if (theValue == 0)
			mAASampling->SetCurrentMenuItem( kDefault3DAA );
		
		mBoard->GetSettingF(HRM_GlobalGlideGamma, &theFloat);
		theValue = (theFloat - 1.0) * 10;
		if (theValue == 0)
			mGamma->SetValue( kDefault3DGamma );
		
		mBoard->GetSettingUL(HRM_GlobalGlideVSyncEnable, &theValue);
		if (theValue == 0)
			mVSync->SetCurrentMenuItem( kDefault3DVSync );
		
		mBoard->GetSettingUL(HRM_GlobalGlideForce32bit, &theValue);
		if (theValue == 0)
			m32Bit->SetCurrentMenuItem( kDefault3D32Bit );
	}
		
	
	
	return result;
}


/*______________________________________________________________________________

	SetVoodooVersion

*/

void
CGlideView::SetVoodooVersion(
	int	 		inValue)
{
	mVVersion = inValue;
}