/*
** Copyright (c) 2000, 3dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/


#include "CKeyComboField.h"

#include <stdio.h>
#include <PP_keyCodes.h>

/*______________________________________________________________________________

	Constructor

*/

CKeyComboField::CKeyComboField()
{
}


/*______________________________________________________________________________

	Constructor (using a Stream, that's the one that will be used)

*/

CKeyComboField::CKeyComboField(
	LStream *			inStream)
	: LEditText ( inStream )
{
}


/*______________________________________________________________________________

	Destructor

*/

CKeyComboField::~CKeyComboField()
{
}


#pragma mark -





/*________________________________________________________________________________________ 

	 GetValue

	return the Key Combination value.
*/
SInt32
CKeyComboField::GetValue() const
{
	return mValue;
}





/*________________________________________________________________________________________ 

	 SetValue

	Set contents of Key Combination value.
*/
void
CKeyComboField::SetValue(
	SInt32				inValue)
{
	mValue = inValue;
	BuildCharStr(inValue, mText);
	SetText((Ptr) mText + 1, mText[0]);
	Refresh();
	BroadcastMessage( mValueMessage, &mValue);
	SwitchTarget(mSuperCommander);
}





/*________________________________________________________________________________________ 

	 ClickSelf

	Respond to Click inside an KeyComboField

*/
void
CKeyComboField::ClickSelf(
	const SMouseDownEvent	&/*inMouseDown*/)
{
	if (!IsTarget()) {
		SwitchTarget(this);
	}
	
	if (IsTarget()) {
		SelectAll();
	}
}





/*________________________________________________________________________________________ 

	 HandleKeyPress

	Handle key stroke directed at an KeyComboField
	Return true if the KeyComboField handles the keystroke

*/
Boolean
CKeyComboField::HandleKeyPress(
	const EventRecord&	inKeyEvent)
{
	Boolean				keyHandled = false;
	EKeyStatus			theKeyStatus = keyStatus_PassUp;
	UInt16				theKey;
	UInt16				theKeyboardCode = (inKeyEvent.message & keyCodeMask) >> 8;
	UInt32				theValue;
	unsigned long		theState = 0;
	UInt32				theTransData;
	
	KeyMapByteArray		theKeyMap;
	GetKeys( (UInt32*) theKeyMap );
	
	theTransData = ::GetScriptManagerVariable(smKCHRCache);
	theKeyboardCode &= 0xFF7F;
	theKey = ::KeyTranslate((Ptr)theTransData,theKeyboardCode,&theState);
	
	UInt32 theFound = theKeyMap[theKeyboardCode >> 3] & (1L << (theKeyboardCode & 0x7));
	
#if MUST_HAVE_MODIFIER	
	if (inKeyEvent.modifiers & (cmdKey | optionKey | controlKey)) {
		theKeyStatus = keyStatus_Input;
		keyHandled = true;
	} else
#else
	if ( !(inKeyEvent.modifiers & (cmdKey)) ) {
		keyHandled = true;
		theKeyStatus = keyStatus_Input;
	}
#endif		
	if (theKey == char_Backspace) {
		theKeyStatus = keyStatus_TEDelete;
		keyHandled = true;
	} else if ( theKey == 0x10 ) {
		theKeyStatus = keyStatus_Input;
	}
	StFocusAndClipIfHidden	focus(this);
	
	switch (theKeyStatus) {
	
		case keyStatus_Input:
			theValue =    ((inKeyEvent.modifiers & 0xFF00) << 16)
						| (inKeyEvent.message & 0xFFFFFF);
//			if ( keyUnused( theValue ))
				SetValue( theValue);
			break;
			
		case keyStatus_TEDelete:
			theValue = 0;
			SetValue( theValue);
			break;
			
		case keyStatus_PassUp:
			keyHandled = LCommander::HandleKeyPress(inKeyEvent);
			break;
	}
	
	return keyHandled;
}





/*________________________________________________________________________________________ 

	 BuildCharStr

	Creates a Char String from the state of the modifiers, keyboard code and charCode
	to be displayed in KeyComboField.
*/
void
CKeyComboField::BuildCharStr(
	UInt32				inValue,
	StringPtr			inText)
{
	UInt32				theModifierKeys = (inValue >> 16) & 0xFF00;
	UInt16				theKey = inValue & charCodeMask;
	UInt16				theKeyboardCode = (inValue & keyCodeMask) >> 8;
	char *				theKeyString;
	UInt16				theStringPos;
	char				theChar[] = " ";
	unsigned long		theState = 0;

	theStringPos = 1;
	if ( inValue == 0 ) {
		theStringPos += sprintf( (char*) &inText[theStringPos], "Not Set");
	}
	else
	{
		if ( theModifierKeys & controlKey)
			theStringPos += sprintf( (char*) &inText[theStringPos], "Ctrl-");
		if ( theModifierKeys & cmdKey)
			theStringPos += sprintf( (char*) &inText[theStringPos], "Cmd-");
		if ( theModifierKeys & optionKey)
			theStringPos += sprintf( (char*) &inText[theStringPos], "Opt-");
		if ( theModifierKeys & shiftKey)
			theStringPos += sprintf( (char*) &inText[theStringPos], "Shift-");

		UInt32 theTransData = ::GetScriptManagerVariable(smKCHRCache);
		theKeyboardCode = 0xFF7F & theKeyboardCode;
		theKey = ::KeyTranslate((Ptr)theTransData,theKeyboardCode,&theState);

		switch (theKey) {
			case 0x20:
				theKeyString = "Space";
				break;
			case 0x1C:
				theKeyString = "Left";
				break;
			case 0x1D:
				theKeyString = "Right";
				break;
			case 0x1E:
				theKeyString = "Up";
				break;
			case 0x1F:
				theKeyString = "Down";
				break;
			case 0x1B:
				theKeyString = "Esc";
				break;
			case 0x08:
				theKeyString = "Delete";	// actually can be clear or escape
				break;
			case 0x0D:
				theKeyString = "Return";
				break;
			case 0x03:
				theKeyString = "Enter";
				break;
			case 0x0B:
				theKeyString = "PgUp";
				break;
			case 0x0C:
				theKeyString = "PgDn";
				break;
			case 0x01:
				theKeyString = "Home";
				break;
			case 0x04:
				theKeyString = "End";
				break;
			case 0x05:
				theKeyString = "Help";
				break;
			case 0x7f:
				theKeyString = "FDel";
				break;
			case 0x10:   /* An F-Key */
				switch (theKeyboardCode)
					{
					case 0x7a:
						theKeyString = "F1";
						break;
					case 0x78:
						theKeyString = "F2";
						break;
					case 0x63:
						theKeyString = "F3";
						break;
					case 0x76:
						theKeyString = "F4";
						break;
					case 0x60:
						theKeyString = "F5";
						break;
					case 0x61:
						theKeyString = "F6";
						break;
					case 0x62:
						theKeyString = "F7";
						break;
					case 0x64:
						theKeyString = "F8";
						break;
					case 0x65:
						theKeyString = "F9";
						break;
					case 0x6d:
						theKeyString = "F10";
						break;
					case 0x67:
						theKeyString = "F11";
						break;
					case 0x6f:
						theKeyString = "F12";
						break;
					case 0x69:
						theKeyString = "F13";
						break;
					case 0x6b:
						theKeyString = "F14";
						break;
					case 0x71:
						theKeyString = "F15";
						break;
					}
				break;
		
			default:
				theChar[0] = theKey;
				theKeyString = theChar;
				break;
		};
		theStringPos += sprintf( (char*) &inText[theStringPos], "%s", theKeyString);
	}

	inText[0] = theStringPos;
}



