/*
** Copyright (c) 2000, 3dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/

#include "COpenGLView.h"
#include <LCaption.h>
#include <LPopupButton.h>
#include "hrm_user_settings.h"





/*______________________________________________________________________________

	Constructor

*/

COpenGLView::COpenGLView()
{
}





/*______________________________________________________________________________

	Constructor (using a Stream, that's the one that will be used)

*/

COpenGLView::COpenGLView(
	LStream *			inStream)
	: LView ( inStream )
{
}





/*______________________________________________________________________________

	Destructor

*/

COpenGLView::~COpenGLView()
{
	StopListening();
}


#pragma mark -





/*______________________________________________________________________________

	FinishCreateSelf
	
*/

void
COpenGLView::FinishCreateSelf()
{
	mAASampling = (LPopupButton*) this->FindPaneByID(perf3D_AAliasing);
	if (mAASampling != nil)		mAASampling->AddListener(this);
	
	mAAKey = (CKeyComboField*) this->FindPaneByID(perf3D_AAliasingKey);
	if (mAAKey != nil)		mAAKey->AddListener(this);

	mGamma = (LSlider*) this->FindPaneByID(perf3D_GammaTbl);
	if (mGamma != nil)		mGamma->AddListener(this);

	mVSync = (LPopupButton*) this->FindPaneByID(perf3D_VSync);
	if (mVSync != nil)		mVSync->AddListener(this);
	
	StartListening();
}





/*______________________________________________________________________________

	ListenToMessage
	
*/

void					
COpenGLView::ListenToMessage(
	MessageT 			inMessage, 
	void *				ioParam)
{
	UInt32 		longParam = *(UInt32*)ioParam;
	Boolean 	res;
	
	if ( mBoard )
	{
		switch(inMessage)
		{
			case msg_3DPerf_AAliasing:
				res = mBoard->SetSettingUL(HRM_GlobalOpenGLAntiAliasing, longParam);
				break;

			case perf3D_AAliasingKey:
				res = mBoard->SetSettingUL(HRM_GlobalOpenGLAntiAliasingKey, longParam);
				break;

			case msg_3DPerf_GammaTbl:
			{
				float theFloat = 1.0 + ( longParam * 0.1 );			
				res = mBoard->SetSettingF(HRM_GlobalOpenGLGamma, theFloat);
				break;
			}
			
			case msg_3DPerf_VSync:
				res = mBoard->SetSettingUL(HRM_GlobalOpenGLVSyncEnable, longParam);
				break;
			
			default:
				break;
		}
	}
}





/*______________________________________________________________________________

	SetPCIHardware

*/

void
COpenGLView::SetPCIHardware(
	CPCIHardware *		inBoard)
{
	if ( inBoard != 0 ) {
		mBoard = inBoard;
		
		LoadOpenGLView();
	}

}

/*______________________________________________________________________________

	LoadOpenGLView

*/

Boolean
COpenGLView::LoadOpenGLView()
{
	UInt32 		theValue;
	float 		theFloat;
	Boolean 	result = true;
	
	mBoard->GetSettingUL(HRM_GlobalOpenGLAntiAliasing, &theValue);
	mAASampling->SetCurrentMenuItem( theValue );
	
	mBoard->GetSettingUL(HRM_GlobalOpenGLAntiAliasingKey, &theValue);
	mAAKey->SetValue( theValue );
	
	mBoard->GetSettingF(HRM_GlobalOpenGLGamma, &theFloat);
	theValue = (theFloat - 1.0) * 10;
	mGamma->SetValue( theValue );
	
	mBoard->GetSettingUL(HRM_GlobalOpenGLVSyncEnable, &theValue);
	mVSync->SetCurrentMenuItem( theValue );
	
	return result;

}
