/*
** Copyright (c) 2000, 3dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/


#include "CProfileView.h"
#include "hrm_user_settings.h"
#include "ControlPanelPPob.h"
#include "C3DPerfView.h"
#include "C2DPerfView.h"
#include "CGlideView.h"

#include <stdio.h>
#include <string.h>

#include <LCaption.h>
#include <LPushButton.h>
#include <LPopupButton.h>


/*______________________________________________________________________________

	Constructor

*/

CProfileView::CProfileView()
{
	
}


/*______________________________________________________________________________

	Constructor (using a Stream, that's the one that will be used)

*/

CProfileView::CProfileView(
	LStream *			inStream)
	: LView ( inStream )
{
	
}


/*______________________________________________________________________________

	Destructor

*/

CProfileView::~CProfileView()
{
}


#pragma mark -

/*______________________________________________________________________________

	FinishCreateSelf

*/

void
CProfileView::FinishCreateSelf()
{
		
	LPushButton* theButton = (LPushButton*) FindPaneByID(profile_RestoreDefault);
	if (theButton != nil) 
	{
		theButton->AddListener(this);
	}
	
}

/*______________________________________________________________________________

	ListenToMessage
	
*/

void					
CProfileView::ListenToMessage(
	MessageT 			inMessage, 
	void *				ioParam)
{
	bool		res;
	UInt32 		longParam = *(UInt32*)ioParam;
	
	if ( mBoard )
	{
		switch(inMessage) {
			case msg_ProfRestore:
			{
				RestoreFactoryDefaults();
				break;
			}
			
			default:
				break;
		}
	}
}




/*______________________________________________________________________________

	RestoreFactoryDefaults
	
*/

void					
CProfileView::RestoreFactoryDefaults()
{
	mGlideView = (CGlideView*) m3DPerfView->FindPaneByID(perf3D_GlideView);
		
	/*		Restore all the 2D/3D factory default settings 	*/
	if (mVVersion != 3)
	{
		LPopupButton * theQDButton = (LPopupButton*) m2DPerfView->FindPaneByID(perf2D_QDAccel);
		theQDButton->SetCurrentMenuItem( kDefault2DQDAcce );
		mBoard->SetSettingUL( HRM_2D_AccelMode, kDefault2DQDAcce );
		
		LPopupButton * theFCacheButton = (LPopupButton*) m2DPerfView->FindPaneByID(perf2D_FontCaching);
		theFCacheButton->SetCurrentMenuItem( kDefault2DFontCache );
		mBoard->SetSettingUL(HRM_2D_FontCaching, kDefault2DFontCache);
		
		LPopupButton * thePCacheButton = (LPopupButton*) m2DPerfView->FindPaneByID(perf2D_PictCaching);
		thePCacheButton->SetCurrentMenuItem( kDefault2DPictCache );
		mBoard->SetSettingUL(HRM_2D_PictCaching, kDefault2DPictCache);
		
		LPopupButton * theQTButton = (LPopupButton*) m2DPerfView->FindPaneByID(perf2D_QTAccel);
		theQTButton->SetCurrentMenuItem( kDefault2DQT );
		mBoard->SetSettingUL(HRM_QT_AccelMode, kDefault2DQT);

		LPopupButton * theMPEGButton = (LPopupButton*) m2DPerfView->FindPaneByID(perf2D_MPEGAccel);
		theMPEGButton->SetCurrentMenuItem( kDefault2DMPEG );
		mBoard->SetSettingUL(HRM_QT_MPEGMode, kDefault2DMPEG);
		
		
		LPopupButton *			mAASampling;
		mAASampling = (LPopupButton*) mGlideView->FindPaneByID(perf3D_AAliasing);
		mAASampling->SetCurrentMenuItem( kDefault3DAA );
		mBoard->SetSettingUL(HRM_GlobalGlideAntiAliasing, kDefault3DAA);
		
		CKeyComboField *		mAAKey;
		mAAKey = (CKeyComboField*) mGlideView->FindPaneByID(perf3D_AAliasingKey);
		mAAKey->SetValue( kDefault3DAAKey );
		mBoard->SetSettingUL(HRM_GlobalGlideAntiAliasingKey, kDefault3DAAKey);
		
		LSlider *				mGamma;
		mGamma = (LSlider*) mGlideView->FindPaneByID(perf3D_GammaTbl);
		mGamma->SetValue( kDefault3DGamma );
		mBoard->SetSettingF(HRM_GlobalGlideGamma, kDefault3DGamma);
		
		LPopupButton *			mVSync;
		mVSync = (LPopupButton*) mGlideView->FindPaneByID(perf3D_VSync);
		mVSync->SetCurrentMenuItem( kDefault3DVSync );
		mBoard->SetSettingUL(HRM_GlobalGlideVSyncEnable, kDefault3DVSync);
		
		LPopupButton *			m32Bit;
		m32Bit = (LPopupButton*) mGlideView->FindPaneByID(perf3D_32Bit);
		m32Bit->SetCurrentMenuItem( kDefault3D32Bit );
		mBoard->SetSettingUL(HRM_GlobalGlideForce32bit, kDefault3D32Bit);
	}
	else
	{
		LPopupButton * theQDButton = (LPopupButton*) m2DPerfView->FindPaneByID(perf2D_QDAccel);
		theQDButton->SetCurrentMenuItem( kBasic );
		mBoard->SetSettingUL( HRM_2D_AccelMode, kBasic );
		
	}
}




/*______________________________________________________________________________

	SetPCIHardware

*/

void
CProfileView::SetPCIHardware(
	CPCIHardware *		inBoard)
{
	if ( inBoard != 0 ) {

		mBoard = inBoard;
		
		/*	Load the profile view captions */
		LoadProfileView();
	}

}


/*______________________________________________________________________________

	Set2DPerfView

*/

void
CProfileView::Set2DPerfView(
	C2DPerfView *		in2DView)
{
	if ( in2DView != 0 ) {

		m2DPerfView = in2DView;
	}
}


/*______________________________________________________________________________

	Set3DPerfView

*/

void
CProfileView::Set3DPerfView(
	C3DPerfView *		in3DView)
{
	if ( in3DView != 0 ) {

		m3DPerfView = in3DView;
	}
}


/*______________________________________________________________________________

	SetVoodooVersion

*/

void
CProfileView::SetVoodooVersion(
	int	 		inValue)
{
	mVVersion = inValue;
}



/*______________________________________________________________________________

	SetPCIHardware

*/
void
CProfileView::LoadProfileView()
{
	Str255		temp;
	long		theValue;
		
	mBoard->GetModelName( temp );
	LCaption* theCaption = (LCaption*) FindPaneByID( profile_CaptName );
	theCaption->SetDescriptor( temp );
			
	mBoard->GetBusTypeName( temp );
	theCaption = (LCaption*) FindPaneByID( profile_CaptBus );
	theCaption->SetDescriptor( temp );
			
	mBoard->GetSettingUL(HRM_GraphicsClockFreq, (unsigned long *) &theValue);
	theCaption = (LCaption*) FindPaneByID( profile_CaptClock );
	sprintf( (char*) &temp[1], "%d.%02d MHz", theValue/100, theValue-(100*(theValue/100)) );
	temp[0] = strlen( (char*) &temp[1] );
	theCaption->SetDescriptor( temp );
		
	mBoard->GetChipName( temp );
	theCaption = (LCaption*) FindPaneByID( profile_CaptCSeries );
	theCaption->SetDescriptor( temp );
		
	theValue = mBoard->GetNumberOfChips();
	theCaption = (LCaption*) FindPaneByID( profile_CaptCNum );
	if ( theValue == 1 )
		theCaption->SetValue( theValue );
	else
		theCaption->SetDescriptor( "\p2 (SLI)" );
		
	theValue = mBoard->GetMemorySize();
	theCaption = (LCaption*) FindPaneByID( profile_CaptMemory );
	sprintf( (char*) &temp[1], "%d MB", theValue );
	temp[0] = strlen( (char*) &temp[1] );
	theCaption->SetDescriptor( temp );
		
	mBoard->GetVersion( temp );
	theCaption = (LCaption*) FindPaneByID( profile_CaptVersion );
	theCaption->SetDescriptor( temp );
}




