/*
** Copyright (c) 1999, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/

#include "CAboutWindow.h"
#include "CUpdateDrv.h"
#include "CGETVerThread.h"
#include "CGETDrvThread.h"
#include "ControlPanelPPob.h"

#include <UInternet.h>
#include <LMailMessage.h>
#include <stdio.h>
#include <LURL.h>
#include <UInternetConfig.h>
#include <UModalDialogs.h>

#include <LMultiPanelView.h>
#include <LPushButton.h>

/*______________________________________________________________________________

	Constructor

*/

CUpdateDrv::CUpdateDrv()
{
	/* Set all the parameters for the http download */
	
}

/*______________________________________________________________________________

	Constructor (using a Stream, that's the one that will be used)

*/

CUpdateDrv::CUpdateDrv(
	LStream *			inStream)
	: LWindow ( inStream )
{
	IsHTTPThreadActive = false;
	IsGETThreadActive = false;
}


/*______________________________________________________________________________

	Destructor

*/

CUpdateDrv::~CUpdateDrv()
{
}


/*______________________________________________________________________________

	FinishCreateSelf

*/

void
CUpdateDrv::FinishCreateSelf()
{
	LStr255	CaptionText;
	
	LPushButton * theButton = (LPushButton*) FindPaneByID(updatedrv_Download);
	if (theButton != nil)		theButton->AddListener(this);
	
	theButton = (LPushButton*) FindPaneByID(updatedrv_Cancel);
	if (theButton != nil)		theButton->AddListener(this);
	
	mProgress = (LCaption*) FindPaneByID(updatedrv_Status);
	
	/*	Set the text inside the UpdateDrv Window */
	
	CaptionText.Assign(UpdateDrv_STR, kChecking4Version);
	LCaption*	theText = (LCaption*) FindPaneByID(updatedrv_Line1);
	if (theText != nil)	theText->SetDescriptor(CaptionText);
	
	CheckDriverVersion();
	
}


/*______________________________________________________________________________

	ListenToMessage

*/


void 
CUpdateDrv::ListenToMessage(
	MessageT 	inMessage, 
	void *		ioParam)
{
	UInt32 		longParam = *(UInt32*)ioParam;

	switch(inMessage) 
	{
		case updatedrv_Download:
		{
			DownloadDrvNow();
			break;
		}
		
		case updatedrv_Cancel:
		{
			/* Need to kill the thread before closing the window, if needed */
			if (IsHTTPThreadActive)
				theHTTPThread->DeleteThread();
			if (IsGETThreadActive)
				theGETDrvThread->DeleteThread();
				
			DoClose();
			break;
		}
		
		default:
			break;
	}
}

#pragma mark -


/*______________________________________________________________________________

	CheckDriverVersion

*/


void 
CUpdateDrv::CheckDriverVersion()
{
	theHTTPThread = new CHTTPSendMsgThread(this);
	
	theHTTPThread->Resume();

}



/*______________________________________________________________________________

	SetLastVersion

*/


void 
CUpdateDrv::SetLastVersion(LStr255 & inLastVersion)
{
	mLastVersion = inLastVersion;
}


/*______________________________________________________________________________

	SetHqxName

*/


void 
CUpdateDrv::SetHqxName(LStr255 & inHqxName)
{
	mHqxName = inHqxName;
}



/*______________________________________________________________________________

	SetDriverName

*/


void 
CUpdateDrv::SetDriverName(LStr255 & inDriverName)
{
	mDriverName = inDriverName;
}


/*______________________________________________________________________________

	GetDriverName

*/


LStr255 & 
CUpdateDrv::GetDriverName()
{
	return mDriverName;
}


/*______________________________________________________________________________

	GetHqxName

*/


LStr255 & 
CUpdateDrv::GetHqxName()
{
	return mHqxName;
}


/*______________________________________________________________________________

	DownloadDrvNow

*/


void 
CUpdateDrv::DownloadDrvNow()
{
	theGETDrvThread = new CGETDrvThread(this);
	
	theGETDrvThread->Resume();

}


