/*
** Copyright (c) 1999, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/

#ifndef __UTILITIES_ROM_FLASHER__
#define __UTILITIES_ROM_FLASHER__

#include <NameRegistry.h>
#include <minihwc.h>

#ifdef __cplusplus
extern "C" {
#endif

#define kMaxBufferSize (128 * 1024)

#define P6FENCE	__sync()

typedef unsigned long BOOL;
typedef BOOL *LPBOOL;
typedef unsigned long DWORD;
typedef unsigned short WORD;
typedef unsigned char BYTE;
typedef DWORD *LPDWORD;
typedef WORD *LPWORD;
typedef BYTE *LPBYTE;

/* ROM Manufacturer ID */
typedef enum {
  MANID_AMD_TI = 0x01,			/* AMD & TI 29010 have same ID's */
  MANID_ATMEL = 0x1F,
  MANID_SST = 0xBF,
  MAMID_UNKNOWN = -1
} MANIDT;

/* ROM Device ID */
typedef enum {
  AMD_TI_29010 = 0x20,
  ATMEL_29010 = 0xD5,
  ATMEL_49F010 = 0x17,
  SST_29EE010 = 0x07,
  SST_39SF010 = 0xB5,
  SST_39VF512 = 0xD4,
  DEVID_UNKNOWN = -1
} DEVIDT;

/* Platform ID */
typedef enum {
  PLATID_INTEL = 0,
  PLATID_OPENFIRMWARE = 1,
  PLATID_UNKNOWN = -1
} PLATIDT;

/* Platform ID */
typedef enum {
  kFRE_noErr = 0,
  kFRE_unkownError = -1,
  kFRE_unkownROMDevice = -2,
  kFRE_cannotFlashApcROM = -3,
  kFRE_cannotFlashAppleROM = -4,
  kFRE_cannotFlashUnkownROM = -5,
  kFRE_badROMHeader = -6,
  kFRE_FailedDuringProgramming = -7
} FLASHROMERROR;


typedef struct TRomBuffer{
  FxU8         mBuffer[kMaxBufferSize];
  FxU32        mBufferSize;
} TRomBuffer;

typedef struct TRomInfo{
  MANIDT        mManufacturer;					/* Manufacturer ID from ROM */
  DEVIDT        mDevice;						/* Device ID from ROM */
  WORD          mSectorSize;					/* Sector size in BYTES of EEPROM */
  PLATIDT       mPlatform;
  FxU32         mHwBaseAddress;
  FxBool        mRomOpen;
  FxU32         mOriginalRomBaseAddr;
  FxU32         mOriginalMiscInit1;
  FxU32         mRomBaseAddr;
  RegEntryID *	mRegEntryID;
  hwcRegInfo	mRegInfo;
  TRomBuffer *  mBuffer;
} TRomInfo;


FLASHROMERROR		FlashROM(
							TRomInfo *			ioRomInfo);

void				ReadVoodooMapping(
							TRomInfo *		inRomInfo);

void				EnableVoodoo(
							 TRomInfo *			inRomInfo);

void				ReadRomInfo(
							TRomInfo *			ioRomInfo);

#ifdef __cplusplus
}
#endif

#endif /* __UTILITIES_ROM_FLASHER__ */
