/*
** Copyright  1994 - 1998, Conix Graphics.
** All Rights Reserved.
** 
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Conix Graphics;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Conix Graphics.
** 
*/

#ifndef _GLT_FIXED_COLOR_H
#define _GLT_FIXED_COLOR_H

/*
** The following fixed point color format is used to
** allow for the calculation (c^2) to not be greater
** than the overflow bits.
**
** This format is a bit odd in that the integer part is
** only used for place holders during a calculation.
** Normalized colors are scaled such that an incoming
** value of 1.0 fills the color bits.
**
** Here is what it looks like in a word:
** 	 +----------------+----------------+
** 	 +   overflow     +   color bits   +
** 	 +----------------+----------------+
*/

typedef unsigned long GLTuColor;

#define GLT_FIXED_COLOR_BITS             16
#define GLT_FIXED_COLOR_BITS_MINUS_ONE   15
#define GLT_FIXED_COLOR_EXACTLY_ONE      0x0000FFFF
#define GLT_FIXED_COLOR_ONE_INV          1.52590218967e-5f

#define GLT_FIXED_COLOR_INDEX_MAX        255

#define gltByteToFixedColor(c)           (((GLint) ((c) & 0x7F) << 9) | ((GLint) ((c) & 0x7F) << 2) | ((GLint) ((c) & 0x7F) >> 5))
#define gltUbyteToFixedColor(c)          (((GLint) (c) << 8) | ((GLint) (c)))
#define gltShortToFixedColor(c)          (((GLint) ((c) & 0x7FFF) << 1) | ((GLint) ((c) & 0x7FFF) >> 14))
#define gltUshortToFixedColor(c)         ((GLint) (c))
#define gltIntToFixedColor(c)            ((GLint) ((c) & 0x7FFFFFFF) >> 15)
#define gltUintToFixedColor(c)           ((GLint) ((c) >> 16))
#define gltFloatToFixedColor(c)          ((c) * (GLTcolor) GLT_FIXED_COLOR_EXACTLY_ONE)

#define gltIndexToFixedColor(c)          ((c) * 257)
#define gltFixedColorToIndex(c)          ((c) * 3.89105058366e-3)

#define gltFixedColorToByte(c)           ((GLint) (c) >> 9)
#define gltFixedColorToUbyte(c)          ((GLint) (c) >> 8)
#define gltFixedColorToShort(c)          ((GLint) (c) >> 1)
#define gltFixedColorToInt(c)            (((GLint) (c) << 15) | ((GLint) (c) >> 1))
#define gltFixedColorToFloat(fixed)      ((GLTcolor) (fixed) * GLT_FIXED_COLOR_ONE_INV)

#define gltFixedColorAdd(f1,f2,f3)       (f1) = (f2) + (f3)
#define gltFixedColorSub(f1,f2,f3)       (f1) = (f2) - (f3)
#define gltFixedColorMult(f1,f2,f3)      (f1) = (((f2) * (f3)) >> GLT_FIXED_COLOR_BITS)
#define gltFixedColorDiv(f1,f2,f3)       (f1) = (((f2) / (f3)) << GLT_FIXED_COLOR_BITS)

#endif /* _GLT_FIXED_COLOR_H */
