/*
** Copyright  1994 - 1998, Conix Graphics.
** All Rights Reserved.
** 
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Conix Graphics;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Conix Graphics.
** 
*/

#ifndef _GLT_FIXED_COORD_H
#define _GLT_FIXED_COORD_H

/*
** For the fixed point arithmetic to be accurate, the following breakdown
** is needed.  The drawing area defined by the GL extends a screens
** amount to the left and right and above and below the normal screen
** area.  This gives an area 9 times the screen area.  To make things
** simpler, the larger of the screen width/height is used.
**
** For DDX, the maximum window size is 2048x2048.  The largest dimension
** is 2048, requiring a GL rendering range of -2048 to +4095 (inclusive).
** After biasing, the range is +10240 to +16383 (inclusive).  This can
** be represented in 14 bits.  We give ourselves one bit extra for
** comfort, and to deal with a value that is exactly on the furthest
** positive edge of the viewport.  This value, 32768 needs 15 bits to
** represent.  Please note that all of the numbers being used will be
** positive, so the sign bit can be used to hold the extra bit.
** implementation does not use the sign bit.
**
** For iteration to work properly, it needs to handle a maximal change in
** one coordinate with a minimal unit change in the other.  For instance,
** a line drawn from (-2048,0) to (4095,1) should render properly.  For
** this to work right, the change in y across this line must be representable
** in the fixed point number.  This requires a fraciton of 1/(3*2048) thus
** necessitating 13 bits.
**
** So, 16 bits are needed for the integer portion, 13 bits are required
** for the iteration to work, thus consuming 29 bits out of 32 bits in a
** long word.  Thus, 3 bits of subpixel precision remain.
**
** Here is what it looks like in a word:
** 	 +-+-----------+--------+------------+
** 	 +S+ integer   + subpix + fraction   +
** 	 +-+-----------+--------+------------+
*/

#define GLT_FIXED_COORD_INT_BITS           16
#define GLT_FIXED_COORD_FRACTION_BITS      16
#define GLT_FIXED_COORD_SUBPIXEL_BITS      3
#define GLT_FIXED_COORD_ALMOST_HALF        0x00007FFF
#define GLT_FIXED_COORD_EXACTLY_HALF       0x00008000
#define GLT_FIXED_COORD_ALMOST_ONE         0x0000FFFF
#define GLT_FIXED_COORD_EXACTLY_ONE        0x00010000
#define GLT_FIXED_COORD_ONE_INV            1.52587890625e-5f

#define gltSignedFloatToFixedCoord(flt)  (gltFloor((flt) * (GLTcoord) GLT_FIXED_COORD_EXACTLY_ONE + GLT_HALF))
#define gltFloatToFixedCoord(flt)        ((GLTfixedCoord) ((flt) * (GLTcoord) GLT_FIXED_COORD_EXACTLY_ONE + GLT_HALF))
#define gltIntToFixedCoord(i)            ((i) << GLT_FIXED_COORD_FRACTION_BITS)

#define gltFixedCoordToFloat(fixed)      ((GLTcoord) (fixed) * GLT_FIXED_COORD_ONE_INV)
#define gltFixedCoordToInt(fixed)        ((fixed) >> GLT_FIXED_COORD_FRACTION_BITS)
#define gltFixedCoordFrac(fixed)         ((fixed) & GLT_FIXED_COORD_ALMOST_ONE)
#define gltFixedCoordAdd(f1,f2,f3)       (f1) = (f2) + (f3)
#define gltFixedCoordSub(f1,f2,f3)       (f1) = (f2) - (f3)
#define gltFixedCoordMult(f1,f2,f3)      (f1) = (((f2) * (f3)) >> GLT_FIXED_COORD_FRACTION_BITS)
#define gltFixedCoordDiv(f1,f2,f3)       (f1) = (((f2) / (f3)) << GLT_FIXED_COORD_FRACTION_BITS)


#endif /* _GLT_FIXED_COORD_H */
