inline float clampF(float x)
{
  double zero = 0.0, maxf = 1.0;
  double _x;
  
  /* Use doubles to prevent compiler from using frsp all
     over the place.  Note the careful use of 255.0f so
     that the compiler generates a single precision multiply. */
  _x = x;
  _x = __fsel(_x,_x,zero);
  _x = __fsel(maxf - _x, _x, maxf);
  return (float)_x;
}

#pragma mark -                       
/*
________________________________________________________________________________________

      glrReadPixels
________________________________________________________________________________________

*/

// values for endian mode

#define EM_SWAP_BYTES 0
#define EM_SWAP_SHORTS 1
#define EM_NO_SWAP 2


inline void _h3ReadDepthPixelsGeneric(
	GLint endian_mode,			// must be a constant
	GLint output_type,			// must be a constant

	GLint inX, 
	GLint inY, 
	GLsizei inWidth, 
	GLsizei inHeight, 
	GLsizei buffHeight,
	FxU32 stride,
	FxU8 *depthBuffer, 
	GLvoid *outPixels, 
	GLint out_row_bytes,
	GLfloat depth_bias, 
	GLfloat depth_scale)
{
	FxU32 x, y;
	FxU16 depthPixel, *rowPixels;
	FxFloat depthFloat;
	void * outRow;
	
	for(y = 0; y < inHeight; y++) {
		outRow = ((FxU8 *)outPixels + y * out_row_bytes);
		
		rowPixels = (FxU16 *)(depthBuffer + stride * (buffHeight - y - inY - 1) + inX * sizeof(FxU16));
		for(x = 0; x < inWidth; x++) {
			switch(endian_mode){
			case EM_SWAP_BYTES:
				depthPixel = __lhbrx(&rowPixels[x],0);
				break;
			case EM_NO_SWAP:
				depthPixel = rowPixels[x];
				break;
			case EM_SWAP_SHORTS:
				// to do: verify that this is correct if inX is odd
      			depthPixel = rowPixels[x ^ 1];
				break;
			}
			depthFloat = ((FxFloat)depthPixel) * (1.0f / 65535.0f);
			depthFloat = depthFloat * depth_scale + depth_bias;
			if(depthFloat < 0.0f) depthFloat = 0.0f;
			if(depthFloat > 1.0f) depthFloat = 1.0f;
			switch(output_type){
			case GL_BYTE:
				((GLbyte *)outRow)[x] = depthFloat * (GLuint)0x0000007f;
				break;
			case GL_UNSIGNED_BYTE:
				((GLubyte *)outRow)[x] = depthFloat * (GLuint)0x000000ff;
				break;
			case GL_SHORT:
				((GLshort *)outRow)[x] = depthFloat * (GLuint)0x00007fff;
				break;
			case GL_UNSIGNED_SHORT:
				((GLushort *)outRow)[x] = depthFloat * (GLuint)0x0000ffff;
				break;
			case GL_INT:
				((GLint *)outRow)[x] = depthFloat * (GLuint)0x7fffffff;
				break;
			case GL_UNSIGNED_INT:
				((GLuint *)outRow)[x] = depthFloat * (GLuint)0xffffffff;
				break;
			case GL_FLOAT:
				((GLfloat *)outRow)[x] = depthFloat;
				break;
			}
		}
	}
}

inline void _h3ReadPixelsGeneric16(
    GLint pixel_mode,           // must be a constant
	GLint endian_mode,			// must be a constant
	GLint output_type,			// must be a constant
    GLint output_format,        // must be a constant
    
	GLint inX, 
	GLint inY, 
	GLsizei inWidth, 
	GLsizei inHeight, 
	GLsizei buffHeight,
	FxU32 stride,
	FxU16 *colorBuffer,
	GLvoid *outPixels, 
	GLint out_row_bytes,
	GLDPixelStore *pixelStore,
	GLDPixelTransfer *pixelTransfer,
	GLDPixelMap *pixelMap)
{
	FxU32 x, y;
	FxU16 colorPixel, *rowPixels;
	FxU32 redI, greenI, blueI, alphaI;
	FxFloat redF, greenF, blueF, alphaF;
	void * outRow;
	
	for(y = 0; y < inHeight; y++) {
		outRow = ((FxU8 *)outPixels + y * out_row_bytes);
		
		rowPixels = (FxU16 *)(colorBuffer + stride * (buffHeight - y - inY - 1) + inX * sizeof(FxU16));
		for(x = 0; x < inWidth; x++) {
			switch(endian_mode){
			case EM_SWAP_BYTES:
				colorPixel = __lhbrx(&rowPixels[x],0);
				break;
			case EM_NO_SWAP:
				colorPixel = rowPixels[x];
				break;
			case EM_SWAP_SHORTS:
				// to do: verify that this is correct if inX is odd
      			colorPixel = rowPixels[x ^ 1];
				break;
			}
			
			switch(pixel_mode) {
			case GR_PIXFMT_RGB_565:
			case GR_PIXFMT_AA_2_RGB_565:
			case GR_PIXFMT_AA_4_RGB_565:
              /* Get integer values */
			  redI   = (colorPixel >> 11) & 0x1f;
			  greenI = (colorPixel >> 5) & 0x3f;
              blueI  = (colorPixel >> 0) & 0x1f;
              alphaI = 0;
              
              /* Convert to float between 0.0 and 1.0 */
              redF   = ((FxFloat)redI)   * (1.0f / 31.0f);
              greenF = ((FxFloat)greenI) * (1.0f / 63.0f);
              blueF  = ((FxFloat)blueI)  * (1.0f / 31.0f);
              alphaF = alphaI;
              break;
              
            case GR_PIXFMT_ARGB_1555:
            case GR_PIXFMT_AA_2_ARGB_1555:
            case GR_PIXFMT_AA_4_ARGB_1555:

              /* Get integer values */
			  redI   = (colorPixel >> 10) & 0x1f;
			  greenI = (colorPixel >> 5)  & 0x1f;
              blueI  = (colorPixel >> 0)  & 0x1f;
              alphaI = (colorPixel >> 15) & 0x01;
              
              /* Convert to float between 0.0 and 1.0 */
              redF   = ((FxFloat)redI)   * (1.0f / 31.0f);
              greenF = ((FxFloat)greenI) * (1.0f / 31.0f);
              blueF  = ((FxFloat)blueI)  * (1.0f / 31.0f);
              alphaF = alphaI;
			  break;
			}			

            /* Apply scale & bias */
			redF   = redF   * pixelTransfer->red_scale   + pixelTransfer->red_bias;
			greenF = greenF * pixelTransfer->green_scale + pixelTransfer->green_bias;
			blueF  = blueF  * pixelTransfer->blue_scale  + pixelTransfer->blue_bias;
			alphaF = alphaF * pixelTransfer->alpha_scale + pixelTransfer->alpha_bias;
			
			/* Clamp values between 0 and 1 */
			redF   = clampF(redF);
			greenF = clampF(greenF);
			blueF  = clampF(blueF);
			alphaF = clampF(alphaF);
			
			/* Possibly do mapping */
			if(pixelMap->map_color) {			  
			  redI   = (redF   * pixelMap->r_to_r_size);
			  greenI = (greenF * pixelMap->g_to_g_size);
			  blueI  = (blueF  * pixelMap->b_to_b_size);
			  alphaI = (alphaF * pixelMap->a_to_a_size);
			  redF   = pixelMap->r_to_r_table[redI];
			  greenF = pixelMap->g_to_g_table[greenI];
			  blueF  = pixelMap->b_to_b_table[blueI];
			  alphaF = pixelMap->a_to_a_table[alphaI];
			}
			
			/* This code currently is not dealing with GL_UNPACK_SWAP_BYTES */
			switch(output_format) {
			case GL_RED:
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x] = redF * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x] = redF * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x] = redF * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x] = redF * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x] = redF * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x] = redF * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x] = redF;
			    break;
			  }
			  break;

			case GL_GREEN:
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x] = greenF * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x] = greenF * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x] = greenF * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x] = greenF * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x] = greenF * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x] = greenF * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x] = greenF;
			    break;
			  }
			  break;

			case GL_BLUE:
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x] = blueF * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x] = blueF * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x] = blueF * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x] = blueF * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x] = blueF * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x] = blueF * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x] = blueF;
			    break;
			  }
			  break;

			case GL_ALPHA:
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x] = alphaF * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x] = alphaF * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x] = alphaF * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x] = alphaF * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x] = alphaF * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x] = alphaF * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x] = alphaF;
			    break;
			  }
			  break;
			  
			case GL_RGB:
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x*3+0] = redF   * (GLuint)0x0000007f;
			    ((GLbyte *)outRow)[x*3+1] = greenF * (GLuint)0x0000007f;
			    ((GLbyte *)outRow)[x*3+2] = blueF  * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x*3+0] = redF   * (GLuint)0x000000ff;
			    ((GLubyte *)outRow)[x*3+1] = greenF * (GLuint)0x000000ff;
			    ((GLubyte *)outRow)[x*3+2] = blueF  * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x*3+0] = redF   * (GLuint)0x00007fff;
			    ((GLshort *)outRow)[x*3+1] = greenF * (GLuint)0x00007fff;
			    ((GLshort *)outRow)[x*3+2] = blueF  * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x*3+0] = redF   * (GLuint)0x0000ffff;
			    ((GLushort *)outRow)[x*3+1] = greenF * (GLuint)0x0000ffff;
			    ((GLushort *)outRow)[x*3+2] = blueF  * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x*3+0] = redF   * (GLuint)0x7fffffff;
			    ((GLint *)outRow)[x*3+1] = greenF * (GLuint)0x7fffffff;
			    ((GLint *)outRow)[x*3+2] = blueF  * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x*3+0] = redF   * (GLuint)0xffffffff;
			    ((GLuint *)outRow)[x*3+1] = greenF * (GLuint)0xffffffff;
			    ((GLuint *)outRow)[x*3+2] = blueF  * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x*3+0] = redF;
			    ((GLfloat *)outRow)[x*3+1] = greenF;
			    ((GLfloat *)outRow)[x*3+2] = blueF;
			    break;
			  }
			  break;
			
			case GL_RGBA:
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x*4+0] = redF   * (GLuint)0x0000007f;
			    ((GLbyte *)outRow)[x*4+1] = greenF * (GLuint)0x0000007f;
			    ((GLbyte *)outRow)[x*4+2] = blueF  * (GLuint)0x0000007f;
			    ((GLbyte *)outRow)[x*4+3] = alphaF * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x*4+0] = redF   * (GLuint)0x000000ff;
			    ((GLubyte *)outRow)[x*4+1] = greenF * (GLuint)0x000000ff;
			    ((GLubyte *)outRow)[x*4+2] = blueF  * (GLuint)0x000000ff;
			    ((GLubyte *)outRow)[x*4+3] = alphaF * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x*4+0] = redF   * (GLuint)0x00007fff;
			    ((GLshort *)outRow)[x*4+1] = greenF * (GLuint)0x00007fff;
			    ((GLshort *)outRow)[x*4+2] = blueF  * (GLuint)0x00007fff;
			    ((GLshort *)outRow)[x*4+3] = alphaF * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x*4+0] = redF   * (GLuint)0x0000ffff;
			    ((GLushort *)outRow)[x*4+1] = greenF * (GLuint)0x0000ffff;
			    ((GLushort *)outRow)[x*4+2] = blueF  * (GLuint)0x0000ffff;
			    ((GLushort *)outRow)[x*4+3] = alphaF * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x*4+0] = redF   * (GLuint)0x7fffffff;
			    ((GLint *)outRow)[x*4+1] = greenF * (GLuint)0x7fffffff;
			    ((GLint *)outRow)[x*4+2] = blueF  * (GLuint)0x7fffffff;
			    ((GLint *)outRow)[x*4+3] = alphaF * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x*4+0] = redF   * (GLuint)0xffffffff;
			    ((GLuint *)outRow)[x*4+1] = greenF * (GLuint)0xffffffff;
			    ((GLuint *)outRow)[x*4+2] = blueF  * (GLuint)0xffffffff;
			    ((GLuint *)outRow)[x*4+3] = alphaF * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x*4+0] = redF;
			    ((GLfloat *)outRow)[x*4+1] = greenF;
			    ((GLfloat *)outRow)[x*4+2] = blueF;
			    ((GLfloat *)outRow)[x*4+3] = alphaF;
			    break;
			  }
			  break;
			
			case GL_LUMINANCE:
			  redF = clampF(redF + greenF + blueF);			  
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x] = redF * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x] = redF * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x] = redF * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x] = redF * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x] = redF * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x] = redF * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x] = redF;
			    break;
			  }
			  break;
			
			case GL_LUMINANCE_ALPHA:
			  redF = clampF(redF + greenF + blueF);
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x*2+0] = redF   * (GLuint)0x0000007f;
			    ((GLbyte *)outRow)[x*2+1] = alphaF * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x*2+0] = redF   * (GLuint)0x000000ff;
			    ((GLubyte *)outRow)[x*2+1] = alphaF * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x*2+0] = redF   * (GLuint)0x00007fff;
			    ((GLshort *)outRow)[x*2+1] = alphaF * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x*2+0] = redF   * (GLuint)0x0000ffff;
			    ((GLushort *)outRow)[x*2+1] = alphaF * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x*2+0] = redF   * (GLuint)0x7fffffff;
			    ((GLint *)outRow)[x*2+1] = alphaF * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x*2+0] = redF   * (GLuint)0xffffffff;
			    ((GLuint *)outRow)[x*2+1] = alphaF * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x*2+0] = redF;
			    ((GLfloat *)outRow)[x*2+1] = alphaF;
			    break;
			  }
			  break;
			}			
		}
	}
}


inline void _h3ReadPixelsGeneric32(
	GLint output_type,			// must be a constant
    GLint output_format,        // must be a constant
    
	GLint inX, 
	GLint inY, 
	GLsizei inWidth, 
	GLsizei inHeight, 
	GLsizei buffHeight,
	FxU32 stride,
	FxU32 *colorBuffer,
	GLvoid *outPixels, 
	GLint out_row_bytes,
	GLDPixelStore *pixelStore,
	GLDPixelTransfer *pixelTransfer,
	GLDPixelMap *pixelMap)
{
	FxU32 x, y;
	FxU32 colorPixel, *rowPixels;
	FxU32 redI, greenI, blueI, alphaI;
	FxFloat redF, greenF, blueF, alphaF;
	void * outRow;
	
	for(y = 0; y < inHeight; y++) {
		outRow = ((FxU8 *)outPixels + y * out_row_bytes);
		
		rowPixels = (FxU32 *)(colorBuffer + stride * (buffHeight - y - inY - 1) + inX * sizeof(FxU32));
		for(x = 0; x < inWidth; x++) {
		    colorPixel = rowPixels[x];
			
            /* Get integer values */
            redI   = (colorPixel >> 16) & 0xff;
			greenI = (colorPixel >>  8) & 0xff;
            blueI  = (colorPixel >>  0) & 0xff;
            alphaI = (colorPixel >> 24) & 0xff;
              
            /* Convert to float between 0.0 and 1.0 */
            redF   = ((FxFloat)redI)   * (1.0f / 255.0f);
            greenF = ((FxFloat)greenI) * (1.0f / 255.0f);
            blueF  = ((FxFloat)blueI)  * (1.0f / 255.0f);
            alphaF = ((FxFloat)alphaI) * (1.0f / 255.0f);

            /* Apply scale & bias */
			redF   = redF   * pixelTransfer->red_scale   + pixelTransfer->red_bias;
			greenF = greenF * pixelTransfer->green_scale + pixelTransfer->green_bias;
			blueF  = blueF  * pixelTransfer->blue_scale  + pixelTransfer->blue_bias;
			alphaF = alphaF * pixelTransfer->alpha_scale + pixelTransfer->alpha_bias;
			
			/* Clamp values between 0 and 1 */
			redF   = clampF(redF);
			greenF = clampF(greenF);
			blueF  = clampF(blueF);
			alphaF = clampF(alphaF);
			
			/* Possibly do mapping */
			if(pixelMap->map_color) {			  
			  redI   = (redF   * pixelMap->r_to_r_size);
			  greenI = (greenF * pixelMap->g_to_g_size);
			  blueI  = (blueF  * pixelMap->b_to_b_size);
			  alphaI = (alphaF * pixelMap->a_to_a_size);
			  redF   = pixelMap->r_to_r_table[redI];
			  greenF = pixelMap->g_to_g_table[greenI];
			  blueF  = pixelMap->b_to_b_table[blueI];
			  alphaF = pixelMap->a_to_a_table[alphaI];
			}
			
			/* This code currently is not dealing with GL_UNPACK_SWAP_BYTES */
			switch(output_format) {
			case GL_RED:
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x] = redF * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x] = redF * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x] = redF * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x] = redF * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x] = redF * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x] = redF * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x] = redF;
			    break;
			  }
			  break;

			case GL_GREEN:
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x] = greenF * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x] = greenF * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x] = greenF * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x] = greenF * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x] = greenF * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x] = greenF * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x] = greenF;
			    break;
			  }
			  break;

			case GL_BLUE:
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x] = blueF * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x] = blueF * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x] = blueF * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x] = blueF * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x] = blueF * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x] = blueF * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x] = blueF;
			    break;
			  }
			  break;

			case GL_ALPHA:
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x] = alphaF * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x] = alphaF * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x] = alphaF * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x] = alphaF * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x] = alphaF * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x] = alphaF * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x] = alphaF;
			    break;
			  }
			  break;
			  
			case GL_RGB:
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x*3+0] = redF   * (GLuint)0x0000007f;
			    ((GLbyte *)outRow)[x*3+1] = greenF * (GLuint)0x0000007f;
			    ((GLbyte *)outRow)[x*3+2] = blueF  * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x*3+0] = redF   * (GLuint)0x000000ff;
			    ((GLubyte *)outRow)[x*3+1] = greenF * (GLuint)0x000000ff;
			    ((GLubyte *)outRow)[x*3+2] = blueF  * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x*3+0] = redF   * (GLuint)0x00007fff;
			    ((GLshort *)outRow)[x*3+1] = greenF * (GLuint)0x00007fff;
			    ((GLshort *)outRow)[x*3+2] = blueF  * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x*3+0] = redF   * (GLuint)0x0000ffff;
			    ((GLushort *)outRow)[x*3+1] = greenF * (GLuint)0x0000ffff;
			    ((GLushort *)outRow)[x*3+2] = blueF  * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x*3+0] = redF   * (GLuint)0x7fffffff;
			    ((GLint *)outRow)[x*3+1] = greenF * (GLuint)0x7fffffff;
			    ((GLint *)outRow)[x*3+2] = blueF  * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x*3+0] = redF   * (GLuint)0xffffffff;
			    ((GLuint *)outRow)[x*3+1] = greenF * (GLuint)0xffffffff;
			    ((GLuint *)outRow)[x*3+2] = blueF  * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x*3+0] = redF;
			    ((GLfloat *)outRow)[x*3+1] = greenF;
			    ((GLfloat *)outRow)[x*3+2] = blueF;
			    break;
			  }
			  break;
			
			case GL_RGBA:
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x*4+0] = redF   * (GLuint)0x0000007f;
			    ((GLbyte *)outRow)[x*4+1] = greenF * (GLuint)0x0000007f;
			    ((GLbyte *)outRow)[x*4+2] = blueF  * (GLuint)0x0000007f;
			    ((GLbyte *)outRow)[x*4+3] = alphaF * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x*4+0] = redF   * (GLuint)0x000000ff;
			    ((GLubyte *)outRow)[x*4+1] = greenF * (GLuint)0x000000ff;
			    ((GLubyte *)outRow)[x*4+2] = blueF  * (GLuint)0x000000ff;
			    ((GLubyte *)outRow)[x*4+3] = alphaF * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x*4+0] = redF   * (GLuint)0x00007fff;
			    ((GLshort *)outRow)[x*4+1] = greenF * (GLuint)0x00007fff;
			    ((GLshort *)outRow)[x*4+2] = blueF  * (GLuint)0x00007fff;
			    ((GLshort *)outRow)[x*4+3] = alphaF * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x*4+0] = redF   * (GLuint)0x0000ffff;
			    ((GLushort *)outRow)[x*4+1] = greenF * (GLuint)0x0000ffff;
			    ((GLushort *)outRow)[x*4+2] = blueF  * (GLuint)0x0000ffff;
			    ((GLushort *)outRow)[x*4+3] = alphaF * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x*4+0] = redF   * (GLuint)0x7fffffff;
			    ((GLint *)outRow)[x*4+1] = greenF * (GLuint)0x7fffffff;
			    ((GLint *)outRow)[x*4+2] = blueF  * (GLuint)0x7fffffff;
			    ((GLint *)outRow)[x*4+3] = alphaF * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x*4+0] = redF   * (GLuint)0xffffffff;
			    ((GLuint *)outRow)[x*4+1] = greenF * (GLuint)0xffffffff;
			    ((GLuint *)outRow)[x*4+2] = blueF  * (GLuint)0xffffffff;
			    ((GLuint *)outRow)[x*4+3] = alphaF * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x*4+0] = redF;
			    ((GLfloat *)outRow)[x*4+1] = greenF;
			    ((GLfloat *)outRow)[x*4+2] = blueF;
			    ((GLfloat *)outRow)[x*4+3] = alphaF;
			    break;
			  }
			  break;
			
			case GL_LUMINANCE:
			  redF = clampF(redF + greenF + blueF);			  
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x] = redF * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x] = redF * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x] = redF * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x] = redF * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x] = redF * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x] = redF * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x] = redF;
			    break;
			  }
			  break;
			
			case GL_LUMINANCE_ALPHA:
			  redF = clampF(redF + greenF + blueF);
			  switch(output_type) {
			  case GL_BYTE:
			    ((GLbyte *)outRow)[x*2+0] = redF   * (GLuint)0x0000007f;
			    ((GLbyte *)outRow)[x*2+1] = alphaF * (GLuint)0x0000007f;
			    break;
			  case GL_UNSIGNED_BYTE:
			    ((GLubyte *)outRow)[x*2+0] = redF   * (GLuint)0x000000ff;
			    ((GLubyte *)outRow)[x*2+1] = alphaF * (GLuint)0x000000ff;
			    break;
			  case GL_SHORT:
			    ((GLshort *)outRow)[x*2+0] = redF   * (GLuint)0x00007fff;
			    ((GLshort *)outRow)[x*2+1] = alphaF * (GLuint)0x00007fff;
			    break;
			  case GL_UNSIGNED_SHORT:
			    ((GLushort *)outRow)[x*2+0] = redF   * (GLuint)0x0000ffff;
			    ((GLushort *)outRow)[x*2+1] = alphaF * (GLuint)0x0000ffff;
			    break;
			  case GL_INT:
			    ((GLint *)outRow)[x*2+0] = redF   * (GLuint)0x7fffffff;
			    ((GLint *)outRow)[x*2+1] = alphaF * (GLuint)0x7fffffff;
			    break;
			  case GL_UNSIGNED_INT:
			    ((GLuint *)outRow)[x*2+0] = redF   * (GLuint)0xffffffff;
			    ((GLuint *)outRow)[x*2+1] = alphaF * (GLuint)0xffffffff;
			    break;
			  case GL_FLOAT:
			    ((GLfloat *)outRow)[x*2+0] = redF;
			    ((GLfloat *)outRow)[x*2+1] = alphaF;
			    break;
			  }
			  break;
			}			
		}
	}
}
