/*________________________________________________________________________________________
** 
** Copyright (c) 1999, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**________________________________________________________________________________________
**
**  Description: glide stubs
**
** 
**
*/

#include "glr_glide.h"
#include "glr_debug.h"

//#define __NO_TRACE__ 1

#define TRACE(x) DEBUG_PRINTF(#x)


/*________________________________________________________________________________________
**
**
**	Glide Procs Trace Definitions
**
**________________________________________________________________________________________
*/


#ifndef GLIDE3_ALPHA
  void glrDrawPlanarPolygon( int nverts, const int ilist[], const GrVertex vlist[] ){TRACE(glrDrawPlanarPolygon);}
  void glrDrawPlanarPolygonVertexList( int nverts, const GrVertex vlist[] ){TRACE(glrDrawPlanarPolygonVertexList);}
  void glrDrawPolygon( int nverts, const int ilist[], const GrVertex vlist[] ){TRACE(glrDrawPolygon);}
  void glrDrawPolygonVertexList( int nverts, const GrVertex vlist[] ){TRACE(glrDrawPolygonVertexList);}
#endif /* !GLIDE3_ALPHA */

#ifdef GLIDE3
  void glrDrawPoint( void *pt ){TRACE(glrDrawPoint);}
  void glrDrawLine( void *v1, void *v2 ){TRACE(glrDrawLine);}
  void glrDrawTriangle( void *a, void *b, void *c ){TRACE(glrDrawTriangle);}
#else
  void glrDrawPoint( const GrVertex *pt ){TRACE(glrDrawPoint);}
  void glrDrawLine( const GrVertex *v1, const GrVertex *v2 ){TRACE(glrDrawLine);}
  void glrDrawTriangle( const GrVertex *a, const GrVertex *b, const GrVertex *c ){TRACE(glrDrawTriangle);}
#endif

/*
** buffer management
*/
void glrBufferClear( GrColor_t color, GrAlpha_t alpha, FxU16 depth ){TRACE(glrBufferClear);}

#ifndef GLIDE3_ALPHA
  int glrBufferNumPending( void ){ return 0; }
#endif /* !GLIDE3_ALPHA */
void glrBufferSwap( int swap_interval ){TRACE(glrBufferSwap);}
void glrRenderBuffer( GrBuffer_t buffer ){TRACE(glrRenderBuffer);}

/*
** error management
*/
void glrErrorSetCallback( GrErrorCallbackFnc_t fnc ){TRACE(glrErrorSetCallback);}

/*
** SST routines
*/
#ifdef GLIDE3
  void glrFinish(void){TRACE(glrFinish);}
#endif

#if defined(GLIDE3) && defined(GLIDE3_ALPHA)
  void glrFlush(void){TRACE(glrFlush);}
#endif

#ifndef GLIDE3_ALPHA
  void glrSstIdle(void){TRACE(glrSstIdle);}
  FxU32 glrSstVideoLine( void ){ return 0; }
  FxBool glrSstVRetraceOn( void ){ return false; }
  FxBool glrSstIsBusy( void ){ TRACE(glrSstIsBusy); return false; }
#endif /* !GLIDE3_ALPHA */

FxBool glrSstWinOpen(
          FxU32                hWnd,
          GrScreenResolution_t screen_resolution,
          GrScreenRefresh_t    refresh_rate,
          GrColorFormat_t      color_format,
          GrOriginLocation_t   origin_location,
          int                  nColBuffers,
          int                  nAuxBuffers){ TRACE(glrSstWinOpen); return 0; }
void glrSstWinClose( void ){TRACE(glrSstWinClose);}
FxBool glrSstControl( FxU32 code ){TRACE(glrSstControl);}

#ifndef GLIDE3_ALPHA
  FxBool glrSstQueryHardware( GrHwConfiguration *hwconfig ){TRACE(glrSstQueryHardware);}
  FxBool glrSstQueryBoards( GrHwConfiguration *hwconfig ){TRACE(glrSstQueryBoards); }
#endif /* !GLIDE3_ALPHA */
void glrSstOrigin(GrOriginLocation_t  origin){TRACE(glrSstOrigin);}
void glrSstSelect( int which_sst ){TRACE(glrSstSelect);}

#ifndef GLIDE3_ALPHA
  FxU32 glrSstScreenHeight( void ){}
  FxU32 glrSstScreenWidth( void ){}
  FxU32 glrSstStatus( void ){TRACE(glrSstStatus);}
#endif /* !GLIDE3_ALPHA */

/*
**  Drawing Statistics
*/
#ifndef GLIDE3_ALPHA
  void glrSstPerfStats(GrSstPerfStats_t *pStats){}
  void glrSstResetPerfStats(void){}
#endif /* !GLIDE3_ALPHA */
void glrResetTriStats(){}
void glrTriStats(FxU32 *trisProcessed, FxU32 *trisDrawn){}

/*
** Glide configuration and special effect maintenance functions
*/
void glrAlphaBlendFunction(
                     GrAlphaBlendFnc_t rgb_sf,   GrAlphaBlendFnc_t rgb_df,
                     GrAlphaBlendFnc_t alpha_sf, GrAlphaBlendFnc_t alpha_df
                     ){TRACE(glrAlphaBlendFunction);}
void glrAlphaCombine(
               GrCombineFunction_t function, GrCombineFactor_t factor,
               GrCombineLocal_t local, GrCombineOther_t other,
               FxBool invert
               ){}
void glrAlphaControlsITRGBLighting( FxBool enable ){}
void glrAlphaTestFunction( GrCmpFnc_t function ){TRACE(glrAlphaTestFunction);}
void glrAlphaTestReferenceValue( GrAlpha_t value ){TRACE(glrAlphaTestReferenceValue);}
void glrChromakeyMode( GrChromakeyMode_t mode ){}
void glrChromakeyValue( GrColor_t value ){}
void glrClipWindow( FxU32 minx, FxU32 miny, FxU32 maxx, FxU32 maxy ){TRACE(glrClipWindow);}
void glrColorCombine(
               GrCombineFunction_t function, GrCombineFactor_t factor,
               GrCombineLocal_t local, GrCombineOther_t other,
               FxBool invert ){TRACE(glrColorCombine);}
void glrColorMask( FxBool rgb, FxBool a ){TRACE(glrColorMask);}
void glrCullMode( GrCullMode_t mode ){}
void glrConstantColorValue( GrColor_t value ){}
void glrConstantColorValue4( float a, float r, float g, float b ){}
void glrDepthBiasLevel( FxI16 level ){}
void glrDepthBufferFunction( GrCmpFnc_t function ){TRACE(glrDepthBufferFunction);}
void glrDepthBufferMode( GrDepthBufferMode_t mode ){TRACE(glrDepthBufferMode);}
void glrDepthMask( FxBool mask ){TRACE(glrDepthMask);}
void glrDisableAllEffects( void ){}
void glrDitherMode( GrDitherMode_t mode ){TRACE(glrDitherMode);}
void glrFogColorValue( GrColor_t fogcolor ){TRACE(glrFogColorValue);}
void glrFogMode( GrFogMode_t mode ){TRACE(glrFogMode);}
void glrFogTable( const GrFog_t ft[] ){TRACE(glrFogTable);}
void glrGammaCorrectionValue( float value ){}

#if defined(GLIDE3) && defined(GLIDE3_ALPHA)
  void glrLoadGammaTable( FxU32 nentries, FxU32 *red, FxU32 *green, FxU32 *blue){}
#endif

#ifdef DIAG_BUILD
  void glrSplashCb(float x, float y, float width, float height, FxU32 frame,
        void (*fn)(int)){}
#endif

void glrSplash(float x, float y, float width, float height, FxU32 frame){}

#ifdef GLIDE3
  FxBool glrGet( FxU32 pname, FxU32 plength, FxI32 *params ){}
  const char * grGetString( FxU32 pname ){}
  FxBool glrReset( FxU32 what ){}
  GrProc grGetProcAddress( char *procName ){}
  void glrEnable( GrEnableMode_t mode ){}
  void glrDisable( GrEnableMode_t mode ){}
  void glrCoordinateSpace( GrCoordinateSpaceMode_t mode ){}
  void glrDepthRange( FxFloat n, FxFloat f ){}
  void glrViewport( FxI32 x, FxI32 y, FxI32 width, FxI32 height ){}
#endif

/*
** texture mapping control functions
*/
FxU32 glrTexCalcMemRequired(
                     GrLOD_t lodmin, GrLOD_t lodmax,
                     GrAspectRatio_t aspect, GrTextureFormat_t fmt){}
FxU32 glrTexTextureMemRequired( FxU32     evenOdd,
                                 GrTexInfo *info   ){}
FxU32 glrTexMinAddress( GrChipID_t tmu ){}

FxU32 glrTexMaxAddress( GrChipID_t tmu ){}


#if defined(GLIDE3) && defined(GLIDE3_ALPHA)
  void glrTexNCCTable( GrNCCTable_t table ){}
#else
  void glrTexNCCTable( GrChipID_t tmu, GrNCCTable_t table ){}
#endif

void glrTexSource( GrChipID_t tmu,
             FxU32      startAddress,
             FxU32      evenOdd,
             GrTexInfo  *info ){TRACE(glrTexSource);}
void glrTexClampMode(
               GrChipID_t tmu,
               GrTextureClampMode_t s_clampmode,
               GrTextureClampMode_t t_clampmode
               ){TRACE(glrTexClampMode);}
void glrTexCombine(
             GrChipID_t tmu,
             GrCombineFunction_t rgb_function,
             GrCombineFactor_t rgb_factor, 
             GrCombineFunction_t alpha_function,
             GrCombineFactor_t alpha_factor,
             FxBool rgb_invert,
             FxBool alpha_invert
             ){TRACE(glrTexCombine);}

#ifndef GLIDE3_ALPHA
  void glrTexCombineFunction(
                     GrChipID_t tmu,
                     GrTextureCombineFnc_t fnc
                     ){TRACE(glrTexCombineFunction);}
#endif

void glrTexDetailControl(
                   GrChipID_t tmu,
                   int lod_bias,
                   FxU8 detail_scale,
                   float detail_max
                   ){}
void glrTexFilterMode(
                GrChipID_t tmu,
                GrTextureFilterMode_t minfilter_mode,
                GrTextureFilterMode_t magfilter_mode
                ){TRACE(glrTexFilterMode);}

void glrTexLodBiasValue(GrChipID_t tmu, float bias ){}
void glrTexDownloadMipMap( GrChipID_t tmu,
                     FxU32      startAddress,
                     FxU32      evenOdd,
                     GrTexInfo  *info ){}
void glrTexDownloadMipMapLevel( GrChipID_t        tmu,
                          FxU32             startAddress,
                          GrLOD_t           thisLod,
                          GrLOD_t           largeLod,
                          GrAspectRatio_t   aspectRatio,
                          GrTextureFormat_t format,
                          FxU32             evenOdd,
                          void              *data ){}
void glrTexDownloadMipMapLevelPartial( GrChipID_t        tmu,
                                 FxU32             startAddress,
                                 GrLOD_t           thisLod,
                                 GrLOD_t           largeLod,
                                 GrAspectRatio_t   aspectRatio,
                                 GrTextureFormat_t format,
                                 FxU32             evenOdd,
                                 void              *data,
                                 int               start,
                                 int               end ){}

void ConvertAndDownloadRle( GrChipID_t        tmu,
                        FxU32             startAddress,
                        GrLOD_t           thisLod,
                        GrLOD_t           largeLod,
                        GrAspectRatio_t   aspectRatio,
                        GrTextureFormat_t format,
                        FxU32             evenOdd,
                        FxU8              *bm_data,
                        long              bm_h,
                        FxU32             u0,
                        FxU32             v0,
                        FxU32             width,
                        FxU32             height,
                        FxU32             dest_width,
                        FxU32             dest_height,
                        FxU16             *tlut){}
void glrCheckForRoom(FxI32 n){}

#if defined(GLIDE3) && defined(GLIDE3_ALPHA)
  void glrTexDownloadTable( GrTexTable_t type, 
                    void         *data ){}
  void glrTexDownloadTablePartial( GrTexTable_t type, 
                           void         *data,
                           int          start,
                           int          end ){}
#else
  void glrTexDownloadTable( GrChipID_t   tmu,
                    GrTexTable_t type, 
                    void         *data ){}
  void glrTexDownloadTablePartial( GrChipID_t   tmu,
                           GrTexTable_t type, 
                           void         *data,
                           int          start,
                           int          end ){}
#endif

void glrTexMipMapMode( GrChipID_t     tmu, 
                 GrMipMapMode_t mode,
                 FxBool         lodBlend ){TRACE(glrTexMipMapMode);}
void glrTexMultibase( GrChipID_t tmu,
                FxBool     enable ){}
void glrTexMultibaseAddress( GrChipID_t       tmu,
                       GrTexBaseRange_t range,
                       FxU32            startAddress,
                       FxU32            evenOdd,
                       GrTexInfo        *info ){}

/*
** utility texture functions
*/

#ifndef GLIDE3_ALPHA
  GrMipMapId_t guTexAllocateMemory(
                    GrChipID_t tmu,
                    FxU8 odd_even_mask,
                    int width, int height,
                    GrTextureFormat_t fmt,
                    GrMipMapMode_t mm_mode,
                    GrLOD_t smallest_lod, GrLOD_t largest_lod,
                    GrAspectRatio_t aspect,
                    GrTextureClampMode_t s_clamp_mode,
                    GrTextureClampMode_t t_clamp_mode,
                    GrTextureFilterMode_t minfilter_mode,
                    GrTextureFilterMode_t magfilter_mode,
                    float lod_bias,
                    FxBool trilinear
                    ){}
  FxBool guTexChangeAttributes(
                      GrMipMapId_t mmid,
                      int width, int height,
                      GrTextureFormat_t fmt,
                      GrMipMapMode_t mm_mode,
                      GrLOD_t smallest_lod, GrLOD_t largest_lod,
                      GrAspectRatio_t aspect,
                      GrTextureClampMode_t s_clamp_mode,
                      GrTextureClampMode_t t_clamp_mode,
                      GrTextureFilterMode_t minFilterMode,
                      GrTextureFilterMode_t magFilterMode
                      ){}
  void guTexCombineFunction(
                     GrChipID_t tmu,
                     GrTextureCombineFnc_t fnc
                     ){}
  GrMipMapId_t guTexGetCurrentMipMap( GrChipID_t tmu ){}
  GrMipMapInfo * guTexGetMipMapInfo( GrMipMapId_t mmid ){}
  FxU32 guTexMemQueryAvail( GrChipID_t tmu ){}
  void guTexMemReset( void ){}
  void guTexDownloadMipMap(
                    GrMipMapId_t mmid,
                    const void *src,
                    const GuNccTable *table
                    ){}
  void guTexDownloadMipMapLevel(
                         GrMipMapId_t mmid,
                         GrLOD_t lod,
                         const void **src
                         ){}void guTexSource( GrMipMapId_t id ){}
#endif /* !GLIDE3_ALPHA */

/*
** linear frame buffer functions
*/
FxBool glrLfbLock( GrLock_t type, GrBuffer_t buffer, GrLfbWriteMode_t writeMode,
           GrOriginLocation_t origin, FxBool pixelPipeline, 
           GrLfbInfo_t *info ){}
FxBool glrLfbUnlock( GrLock_t type, GrBuffer_t buffer ){}
void glrLfbConstantAlpha( GrAlpha_t alpha ){}
void glrLfbConstantDepth( FxU16 depth ){}
void glrLfbWriteColorSwizzle(FxBool swizzleBytes, FxBool swapWords){}
void glrLfbWriteColorFormat(GrColorFormat_t colorFormat){}

#if defined(GLIDE3) && defined(GLIDE3_ALPHA)
  FxBool glrLfbWriteRegion( GrBuffer_t dst_buffer, 
                  FxU32 dst_x, FxU32 dst_y, 
                  GrLfbSrcFmt_t src_format, 
                  FxU32 src_width, FxU32 src_height, 
                  FxBool pixelPipeline,
                  FxI32 src_stride, void *src_data ){}
#else
  FxBool glrLfbWriteRegion( GrBuffer_t dst_buffer, 
                  FxU32 dst_x, FxU32 dst_y, 
                  GrLfbSrcFmt_t src_format, 
                  FxU32 src_width, FxU32 src_height, 
                  FxI32 src_stride, void *src_data ){}
#endif

FxBool glrLfbReadRegion( GrBuffer_t src_buffer,
                 FxU32 src_x, FxU32 src_y,
                 FxU32 src_width, FxU32 src_height,
                 FxU32 dst_stride, void *dst_data ){}


/*
**  Antialiasing Functions
*/

#ifndef GLIDE3_ALPHA
  void glrAADrawLine(const GrVertex *v1, const GrVertex *v2){}
  void glrAADrawPoint(const GrVertex *pt ){}
  void glrAADrawPolygon(const int nverts, const int ilist[], const GrVertex vlist[]){}
  void glrAADrawPolygonVertexList(const int nverts, const GrVertex vlist[]){}
#endif /* !GLIDE3_ALPHA */
void glrAADrawTriangle(
                 const GrVertex *a, const GrVertex *b, const GrVertex *c,
                 FxBool ab_antialias, FxBool bc_antialias, FxBool ca_antialias
                 ){}

/*
** glide management functions
*/
void glrGlideInit( void ){TRACE(glrGlideInit);}
void glrGlideShutdown( void ){}

#ifndef GLIDE3_ALPHA
  void glrGlideGetVersion( char version[80] ){}
#endif /* !GLIDE3_ALPHA */
void glrGlideGetState( GrState *state ){}
void glrGlideSetState( const GrState *state ){}

#ifndef GLIDE3_ALPHA
  void glrGlideShamelessPlug(const FxBool on){}
  void glrHints(GrHint_t hintType, FxU32 hintMask){}
#endif /* !GLIDE3_ALPHA */

#ifdef GLIDE3
  void glrVertexLayout(FxU32 param, FxI32 offset, FxU32 mode){}
  void glrDrawVertexArray(FxU32 mode, FxU32 Count, void *pointers){}
  void glrDrawVertexArrayLinear(FxU32 mode, FxU32 Count, void *pointers, FxU32 stride){}
  void glrDrawVertexArrayLinear(FxU32 mode, FxU32 Count, void *pointers, FxU32 stride){}
#endif


/*________________________________________________________________________________________
**
**
**	Glide Utils Procs Trace Definitions
**
**________________________________________________________________________________________
*/


/*
** rendering functions
*/

#ifndef GLIDE3_ALPHA
void glruAADrawTriangleWithClip( const GrVertex *a, const GrVertex
                         *b, const GrVertex *c)
{}

void glruDrawTriangleWithClip(
                       const GrVertex *a,
                       const GrVertex *b,
                       const GrVertex *c
                       ){}

void glruDrawPolygonVertexListWithClip( int nverts, const GrVertex vlist[] ){}

/*
** hi-level rendering utility functions
*/
void glruAlphaSource( GrAlphaSource_t mode ){}
void glruColorCombineFunction( GrColorCombineFnc_t fnc ){}
int glruEncodeRLE16( void *dst, 
               void *src, 
               FxU32 width, 
               FxU32 height ){}
FxU16 * glruTexCreateColorMipMap( void ){}
#endif /* !GLIDE3_ALPHA */

#ifdef GLIDE3
void FX_CALL 
glruGammaCorrectionRGB( FxFloat red, FxFloat green, FxFloat blue ){}
#endif

/*
** fog stuff
*/
float glruFogTableIndexToW( int i ){}
void glruFogGenerateExp( GrFog_t fogtable[], float density ){TRACE(glruFogGenerateExp);}
void glruFogGenerateExp2( GrFog_t fogtable[], float density ){TRACE(glruFogGenerateExp2);}
void glruFogGenerateLinear(GrFog_t fogtable[],
                    float nearZ, float farZ ){TRACE(glruFogGenerateLinear);}

/*
** endian stuff
*/
#ifndef GLIDE3_ALPHA
FxU32 glruEndianSwapWords( FxU32 value ){}
FxU16 glruEndianSwapBytes( FxU16 value ){}
#endif /* !GLIDE3_ALPHA */

/*
** hi-level texture manipulation tools.
*/
FxBool glru3dfGetInfo( const char *filename, Gu3dfInfo *info ){}
FxBool glru3dfLoad( const char *filename, Gu3dfInfo *data ){}
