//: sst1Blts.c
//: alt.drivers inc.
//: Glenn Nissen


#include "r3Core.h"


//////////////////
//              //
//  Blt_16to16  //
//              //
//////////////////

void 
Blt_16to16(UInt16* dst,
           UInt16* src,
           UInt32 x,
           UInt32 y,
           UInt32 width,
           UInt32 height,
           UInt32 pitch)
{
	UInt32 i;
	UInt32 j;
	
	dst += x + y * (TNSL_GLIDE_PITCH >> 1);

	for(j = 0; j < height; ++j) 
	{
		for(i = 0; i < width; ++i) 
		{
			dst[i] = src[i];
		}

		src += pitch / 2;
		dst += (TNSL_GLIDE_PITCH >> 1);
	}
}


//////////////////
//              //
//  Blt_32to32  //
//              //
//////////////////

void 
Blt_32to32(UInt32* dst,
           UInt32* src,
           UInt32 x,
           UInt32 y,
           UInt32 width,
           UInt32 height,
           UInt32 pitch)
{
	UInt32 ii;
	UInt32 jj;
	
	dst += x + y * TNSL_GLIDE_PITCH / 2;

	for(jj = 0; jj < height; ++jj) {
		for(ii = 0; ii < width; ++ii) {
			dst[ii] = src[ii];
		}

		src += pitch / 4;
		dst += TNSL_GLIDE_PITCH / 2;
	}
}


/////////////////
//             //
//  Blt_8to32  //
//             //
/////////////////

void 
Blt_8to32(UInt32* dst,
          UInt8*  src,
          UInt32* tbl,
          UInt32 x,
          UInt32 y,
          UInt32 width,
          UInt32 height,
          UInt32 pitch)
{
	UInt32 ii;
	UInt32 jj;
	
	dst += x + y * TNSL_GLIDE_PITCH / 2;

	for(jj = 0; jj < height; ++jj) {
		for(ii = 0; ii < width; ++ii) {
			dst[ii] = tbl[src[ii]];
		}

		src += pitch;
		dst += TNSL_GLIDE_PITCH / 2;
	}
}


/////////////////
//             //
//  Blt_4to32  //
//             //
/////////////////

void 
Blt_4to32(UInt32* dst,
          UInt8*  src,
          UInt32* tbl,
          UInt32 x,
          UInt32 y,
          UInt32 width,
          UInt32 height,
          UInt32 pitch)
{
	UInt32 ii;
	UInt32 jj;
	
	dst += x + y * TNSL_GLIDE_PITCH / 2;

	for(jj = 0; jj < height; ++jj) {
		for(ii = 0; ii < width / 2; ++ii) {
			dst[ii * 2 + 1] = tbl[src[ii] & 0x0F];
			dst[ii * 2 + 0] = tbl[src[ii] >> 4];
		}

		src += pitch;
		dst += TNSL_GLIDE_PITCH / 2;
	}
}


////////////////////////////////
//                            //
//  BltMipMap_16ARGBto16ARGB  //
//                            //
////////////////////////////////

void 
BltMipMap_16ARGBto16ARGB(void*     ptr,
                         TQAImage* images,
                         UInt32    levels)
{
	UInt32  dstAddr;
	UInt32  srcAddr;
	UInt32  ii;
	UInt32  jj;
	UInt32  kk;
	
	dstAddr = (UInt32) ptr;

	for(kk = 0; kk < levels; ++kk) {
		srcAddr = (UInt32) images[kk].pixmap;// + images[kk].rowBytes * images[kk].height;

		if(images[kk].width >= 4) {
			UInt32* dstPtr;
			UInt32* srcPtr;
		
			for(jj = 0; jj < images[kk].height; ++jj) {
				//srcAddr -= images[kk].rowBytes;
				
				dstPtr = (UInt32*) dstAddr;
				srcPtr = (UInt32*) srcAddr;
				
				for(ii = 0; ii < images[kk].width / 4; ++ii) {
					*dstPtr++ = *srcPtr++;
					*dstPtr++ = *srcPtr++;
				}
				
				dstAddr += images[kk].width * 2;
				srcAddr += images[kk].rowBytes;
			}
		} else {
			UInt16* dstPtr;
			UInt16* srcPtr;
		
			for(jj = 0; jj < images[kk].height; ++jj) {
				//srcAddr -= images[kk].rowBytes;
				
				dstPtr = (UInt16*) dstAddr;
				srcPtr = (UInt16*) srcAddr;
				
				for(ii = 0; ii < images[kk].width; ++ii) {
					*dstPtr++ = *srcPtr++;
				}
				
				dstAddr += images[kk].width * 2;
				srcAddr += images[kk].rowBytes;
			}
		}
	}
}


//////////////////////////////
//                          //
//  BltMipMap_32RGBto16RGB  //
//                          //
//////////////////////////////

void 
BltMipMap_32RGBto16RGB(void*     ptr,
                       TQAImage* images,
                       UInt32    levels)
{
	UInt16* dst;
	UInt32* src;
	
	UInt32 ii;
	UInt32 jj;
	UInt32 kk;

	dst = (UInt16*) ptr;

	for(kk = 0; kk < levels; ++kk) {
		src = (UInt32*) images[kk].pixmap;// + images[kk].rowBytes * images[kk].height / 4;

		for(jj = 0; jj < images[kk].height; ++jj) {
			//src -= images[kk].rowBytes / 4;
			for(ii = 0; ii < images[kk].width; ++ii) {	
				UInt16 r = (UInt16) ((src[ii] & 0xF80000) >> 9);
				UInt16 g = (UInt16) ((src[ii] & 0x00F800) >> 6);
				UInt16 b = (UInt16) ((src[ii] & 0x0000F8) >> 3);
			
				dst[ii] = r | g | b | 0x8000;
			}

			dst += images[kk].width;
			src += images[kk].rowBytes / 4;
		}
	}
}


////////////////////////////////
//                            //
//  BltMipMap_32ARGBto16ARGB  //
//                            //
////////////////////////////////

void 
BltMipMap_32ARGBto16ARGB(void*     ptr,
                         TQAImage* images,
                         UInt32    levels)
{
	UInt16* dst;
	UInt32* src;
	
	UInt32 ii;
	UInt32 jj;
	UInt32 kk;

	dst = (UInt16*) ptr;

	for(kk = 0; kk < levels; ++kk) {
		src = (UInt32*) images[kk].pixmap;// + images[kk].rowBytes * images[kk].height / 4;

		for(jj = 0; jj < images[kk].height; ++jj) {
			//src -= images[kk].rowBytes / 4;
			for(ii = 0; ii < images[kk].width; ++ii) {
				dst[ii] = (((src[ii] & 0xF0000000) >> 16) | 
                   ((src[ii] & 0x00F00000) >> 12) | 
                   ((src[ii] & 0x0000F000) >> 8) | 
                   ((src[ii] & 0x000000F0) >> 4));
			}

			dst += images[kk].width;
			src += images[kk].rowBytes / 4;
		}
	}
}


//////////////////////////
//                      //
//  BltMipMap_CL8toCL8  //
//                      //
//////////////////////////

void 
BltMipMap_CL8toCL8(void*     ptr,
                   TQAImage* images,
                   UInt32    levels)
{
	UInt8* dst;
	UInt8* src;
	
	UInt32 ii;
	UInt32 jj;
	UInt32 kk;

	dst = (UInt8*) ptr;

	for(kk = 0; kk < levels; ++kk) {
		src = (UInt8*) images[kk].pixmap;// + images[kk].rowBytes * images[kk].height;

		for(jj = 0; jj < images[kk].height; ++jj) {
			//src -= images[kk].rowBytes;
			for(ii = 0; ii < images[kk].width; ++ii) {
				dst[ii] = src[ii];
			}

			dst += images[kk].width;
			src += images[kk].rowBytes;
		}
	}
}


//////////////////////////
//                      //
//  BltMipMap_CL4toCL8  //
//                      //
//////////////////////////

void
BltMipMap_CL4toCL8(void*     ptr,
                   TQAImage* images,
                   UInt32    levels)
{
	UInt8* dst;
	UInt8* src;
	
	UInt32 ii;
	UInt32 jj;
	UInt32 kk;

	dst = (UInt8*) ptr;

	for(kk = 0; kk < levels; ++kk) {
		src = (UInt8*) images[kk].pixmap;// + images[kk].rowBytes * images[kk].height;

		for(jj = 0; jj < images[kk].height; ++jj) {
			//src -= images[kk].rowBytes;
			for(ii = 0; ii < images[kk].width / 2; ii++) {
				dst[ii * 2 + 1] = src[ii] & 0x0F;
				dst[ii * 2 + 0] = src[ii] >> 4;
			}

			dst += images[kk].width;
			src += images[kk].rowBytes;
		}
	}
}


////////////////////////////////
//                            //
//  BltMipMap_32ARGBtoAlpha8  //
//                            //
////////////////////////////////

void 
BltMipMap_32ARGBtoAlpha8(void*     ptr,
                         TQAImage* images,
                         UInt32    levels)
{
	UInt8*  dst;
	UInt32* src;
	
	UInt32 ii;
	UInt32 jj;
	UInt32 kk;

	dst = (UInt8*) ptr;

	for(kk = 0; kk < levels; ++kk) {
		src = (UInt32*) images[kk].pixmap;// + images[kk].rowBytes * images[kk].height / 4;

		for(jj = 0; jj < images[kk].height; ++jj) {
			//src -= images[kk].rowBytes / 4;
			for(ii = 0; ii < images[kk].width; ++ii) {
				dst[ii] = src[ii] >> 24;
			}

			dst += images[kk].width;
			src += images[kk].rowBytes / 4;
		}
	}
}


