/*
** Copyright (c) 1999, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
**
*/


#include <Types.h>
#include <Memory.h>
#include <Quickdraw.h>
#include <Fonts.h>
#include <Events.h>
#include <Menus.h>
#include <Windows.h>
#include <TextEdit.h>
#include <Dialogs.h>
#include <OSUtils.h>
#include <ToolUtils.h>
#include <SegLoad.h>
#include <Sound.h>
#include <Folders.h>
#include <Notification.h>

#include <hdwr_res_mgr.h>
//#include <GraphicsPrivHwc.h>


/* Prototypes */
void Initialize(void);

void DisplayMessage(
  unsigned char *       theMessage);


int main(void)
{
  hrmDispatcherPtr      theHrmDispatcher;
  FSSpec                theNewFragFile;
  short                 theNewCFragFileRefNum;
  Handle                theNewCFragFileH = 0;
  OSErr                 theSuccess;
  
  Initialize();

  /* look for the original HRM in the extension folder and bind the hrmUpdate symbol */
  {
    SInt16              theVRef;
    SInt32              theDirID;
    FSSpec              theFile;

    theSuccess = FindFolder(kOnSystemDisk, kExtensionFolderType,
					            false, &theVRef, &theDirID);

    if (theSuccess == noErr) {
      CFragConnectionID  theHRMLib; 
      Str255             theError;

      FSMakeFSSpec(theVRef, theDirID, "\p3Dfx Hardware Resource Mgr", &theFile);

      theSuccess = GetDiskFragment( &theFile, 0, kCFragGoesToEOF,
                                    (StringPtr)"\p3DfxHrdwResMgr", kLoadCFrag,
                                         &theHRMLib, (Ptr *) &theHrmDispatcher,
                                                                       theError);
      
    } else {
      DisplayMessage( "\pCouldn't find the HRM library in Extension Folder");
    }
  }


  
  /* look for the new HRM in the HRM_Install folder */
  if ( theSuccess == 0 ) {
    ProcessSerialNumber theCurApp;
    ProcessInfoRec      theAppInfo;
    FSSpec              theAppFile;

    if ( GetCurrentProcess( &theCurApp ) == noErr ) {

      theAppInfo.processInfoLength = sizeof(ProcessInfoRec);
      theAppInfo.processName = NULL;
      theAppInfo.processAppSpec = &theAppFile;
		
      theSuccess = GetProcessInformation( &theCurApp, &theAppInfo);
		
      if ( theSuccess == noErr ) {

        theSuccess = FSMakeFSSpec( theAppFile.vRefNum, theAppFile.parID,
		                                    "\p3Dfx HRM update", &theNewFragFile );
        if ( theSuccess != 0 ) {
          DisplayMessage( "\pCouldn't find the update file");
        }
      } else {
        DisplayMessage( "\pCouldn't obtain current process informations");
      }
    } else {
      DisplayMessage( "\pCouldn't identify current process");
    }
  }
  
  /* load the new HRM into system memory so it won't go away when we close the installer */
  if ( theSuccess == 0 ) {
  
    theSuccess = FSpOpenDF( &theNewFragFile, fsRdPerm, &theNewCFragFileRefNum);
    
    if ( theSuccess == 0 ) {
      long              theSize;
      
      theSuccess = GetEOF( theNewCFragFileRefNum, &theSize );
      
      if ( theSuccess == 0 ) {

        theNewCFragFileH = NewHandleSys( theSize );
        
        if ( theNewCFragFileH != 0 ) {
          
          theSuccess = SetFPos( theNewCFragFileRefNum, fsFromStart, 0 );
            
          if ( theSuccess == 0 ) {

            HLock( theNewCFragFileH );
            theSuccess = FSRead( theNewCFragFileRefNum, &theSize, *theNewCFragFileH );
          } else {
            DisplayMessage( "\pCouldn't read update file");
          }
        } else {
          DisplayMessage( "\pNot enough memory to perform the update!");
        }
      } else {
        DisplayMessage( "\pUpdate file is corrupted");
      }
    } else {
      DisplayMessage( "\pCouldn't open update file");
    }
  }

  /* update the HRM with the new one */
  if ( theSuccess == 0 ) {
    hrmCommand_t            theCmd;
       
    theCmd.resourceSelector = HRM_SELF;
    theCmd.methodSelector = HRM_UPDATE;
    theCmd.u.updateData.opaqueData =  (Ptr) theNewCFragFileH;

    theSuccess = (*theHrmDispatcher)( &theCmd ); 

    if ( theSuccess != 0) {
      if ( theSuccess > -100 ) {
        DisplayMessage( "\pCouldn't save data");
      } else {
        switch( theSuccess ) {
          case -100:
            DisplayMessage( "\pthe HRM is still in use, cannot update!");
            break;
            
          case -101:
            DisplayMessage( "\pthe memory manager is still in use, cannot update!");
            break;
            
          case -102:
            DisplayMessage( "\pthe overlay surface is still in use, cannot update!");
            break;
          
          default:
            DisplayMessage( "\pCurrent HRM refuse the update");
            break;
        }

      }
    }

  }

  /* let's check everything is ok */
  if ( theSuccess == 0 ) {
    hrmCommand_t            theCmd;
       
    theCmd.resourceSelector = HRM_SELF;
    theCmd.methodSelector = HRM_GET_RESOURCE_INFO;
    theCmd.u.resInfo.ioResourceTag = HRM_SELF;

    theSuccess = (*theHrmDispatcher)( &theCmd );
    
    if ( theSuccess != 0 ) {
      DisplayMessage( "\pAfter update, HRM is no longer responding!");
    }
  }

  return 0;	
}



void DisplayMessage(
  unsigned char *       theMessage)
{
  NMRecPtr              theRecPtr;
  char *                theTextPtr;
  long                  i;
  OSErr                 theSuccess;

  theTextPtr = NewPtrSys( sizeof( NMRec ) + 64 );
  theRecPtr = (NMRec*) theTextPtr;
  theTextPtr += sizeof( NMRec );
	
  i = theMessage[0];
  while ( i >= 0 ) {
    theTextPtr[i] = theMessage[i];
    i--;
  }
	
  theRecPtr->nmResp     = 0;
  theRecPtr->qType      = nmType;
  theRecPtr->nmRefCon   = 0;
  theRecPtr->nmSound    = (Handle) -1;
  theRecPtr->nmStr      = (unsigned char*) theTextPtr;
  theRecPtr->nmMark     = 0;
  theRecPtr->nmIcon     = 0;

  theSuccess = NMInstall( theRecPtr );
	
  DisposePtr( theTextPtr );
}


void Initialize(void)
{
  /* Initialize all the needed managers. */
  InitGraf(&qd.thePort);
  InitFonts();
  InitWindows();
  InitMenus();
  TEInit();
  InitDialogs(nil);
  InitCursor();

}

