/*
** Copyright (c) 1999, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** <1>    05.20.99  Stephane Huaulme ...
**
*/

#ifndef __HRM_PRIV_H__
#define __HRM_PRIV_H__

#include "hdwr_res_mgr.h"
#include "hrm_fifo.h"
#include "hrm_mem.h"
#include "hrm_lists.h"
#include "hrm_arbitration.h"
#define __MACERRORS__
#include <minihwc.h>
#include <CodeFragments.h>
#include <Devices.h>
#include <Displays.h>
#include <DriverServices.h>
#include "AGP.h"

#define kHrmVersion 0001   /* this version number should be incremented for each
                              changes made to the Hardware Resource Manager
                              (including any changes to any of the resources) */
                          

#define kHrmMaxNumberOfTargets 8

typedef struct mmBlockHeader_t mmBlockHeader_t;

struct hrmRegistration_t
{
  long
    inDrvrRefNum,
    inFrameBufferBaseAddress,
    inFrameBufferSize;
  hwcBoardInfo
    inBoardInfo;
};

typedef struct hrmVideoInfo_s
{
  VDSwitchInfoRec oldVideoMode;
  FxBool oldModeValid;
} hrmVideoInfo;

struct hrmBoard_s
{
  FxU32
    drvrRefNum,
    frameBufferBaseAddress,
    prefs;                      // first four bytes of nvram
        
  
  FxU32 nextContextID;
  FxU32 lastContextID;
  
  FxBool         agpEnabled;
  FxBool         agpAllocated;
  AGPAddressPair agpAddress;
  FxU32          agpSize;
  
  Handle drvrGDeviceHandle;
  
//    frameBufferSize;
  List_t
    memAreas;
  hwcBoardInfo
    boardInfo;
  mmBlock_t 
    *fifoMemBlock;
  hrmFifoInfo
    fifoInfo;
  hrmVideoInfo
    videoInfo;
  FxBool
    exclusiveMode;
  FxBool
    supportsAGP;
  FxU32
    swizzleOffset[4];
  FxU32
    prefsIdx;                   // index of prefs for each board (board number)
  SInt32
    arbiterDepth;
  QHdr
    arbiterQueue;
};


extern hrmBoard_t gHdwrData[];
extern CFragConnectionID gSysCFragConnectionID; 
extern GDHandle _hrmFindGDeviceHandle(short inRefNum);


typedef struct hrmUpdateGlobalsBucket_t
{
  hrmBoard_t
    hdwrData[kHrmMaxNumberOfTargets];
} hrmUpdateGlobalsBucket_t;

typedef struct hrmExtensionTableEntry_s
{
  char *extName;
  void *extValue;
} hrmExtensionTableEntry_t;

typedef struct hrmExtensionNode_s
{
  Node_t node;
  long   count;
  hrmExtensionTableEntry_t *table;
} hrmExtensionNode_t;




extern hrmBoard_t * hrmIdentifyTarget( short inRefNum );
long hrm_FindPrefIdx(const GDHandle inDisplay);



#endif /* __HRM_PRIV_H__ */