/*
 *
 *	NQDAcceleration.h - 
 *
 *	Native QuickDraw Acceleration interface
 *
 *  Copyright:   1994 by Apple Computer, Inc., all rights reserved.
 *
 *
 */

#ifndef __NQDAcceleration__

#define __NQDAcceleration__

/****************************************************************************
 *																		   	*
 *                        Substitutes for Instrinsic Types					*
 *																		   	*
 ****************************************************************************/	

typedef char              Int8;           /* Signed 8-bit integer     */
typedef short             Int16;          /* Signed 16-bit integer    */
typedef long              Int32;          /* Signed 32-bit integer    */
typedef double long       Int64;          /* Signed 32-bit integer    */


/****************************************************************************
 *																		   	*
 *  Fixed Point Constants:													*
 *																		   	*
 ****************************************************************************/	

#define kBitsPerFract   16			/* Number of bits in the fractional part of a fixed-point value */

#define kFractOne			((Fixed) 0x00010000)	/* 1.0 in Fixed Point */
#define kFractMinusOne		((Fixed) 0xFFFF0000)	/* -1.0 in Fixed Point */
#define kFractOneHalf 		((Fixed) kFractOne>>1) 	/* 1/2 in Fixed Point = (Fixed 1) >> 1 */


/****************************************************************************
 *																		   	*
 *  Fixed Point Macros:														*
 * 																		   	*
 *  Macros used to manipulate Signed and Unsigned Fixed Point numbers.		*
 *  In all macros: a = integral part, b = fractional part, 					*
 *                f = fixed point number									*
 *																		   	*
 ****************************************************************************/	

#define INT2FIXED(a)        ( (Int32) (((Int16) (a)) << kBitsPerFract) )
#define UINT2FIXED(a)       ( (UInt32) (((UInt16) (a)) << kBitsPerFract) )
#define FIXEDCONST(a,b) 	( INT2FIXED(a) + (UInt16) (b) )
#define UFIXEDCONST(a, b) 	( UINT2FIXED(a) + (UInt16) (b) )

#define FRACTPART(f) 		( (UInt16) ((UInt32) (f)) )
#define SIGNEDFRACTPART(f) 	( (Int16) ((Int32) (f)) )
#define INTPART(f) 			( ((Int32) (f)) >> kBitsPerFract )
#define UINTPART(f) 		( ((UInt32) (f)) >> kBitsPerFract )


/****************************************************************************
**
**	NQDRegion
**
**	Expand the region definition to include rgnData.
**
****************************************************************************/	
	
struct NQDRegion
	{
	Int16	rgnSize;	/* Size in bytes*/
	Rect	rgnBBox;	/* Enclosing rectangle*/
	Int16	rgnData[1];	/* Private region inversion points */
	};
typedef struct NQDRegion NQDRegion, *NQDRgnPtr, **NQDRgnHandle;

#define	kOffsetToRgnData		10
#define kRgnTerminator			0x7FFF		/* Marks the end of the region */
#define kRgnTerminatorMinus1	0x7FFE		/* Marker for coordinates that were 0x7fff to avoid confusion */


/****************************************************************************
**
**	TrimRect
**
**	=0	rectangular
**	<0	empty
**	>0	non-empty
**
****************************************************************************/	

enum
	{
	kTrimIsRect	   =  0,
	kTrimIsEmpty   = -1,
	kTrimIsNonRect =  1
	};
typedef UInt32 TrimRectResult;




/***************************************************************************
**
** NQDITInfoTable:
**
** Extension of the Inverse Table data structure.
** ITInfoTable is a private substructure located at the end of the ITab structure.
** It contains hidden entries for fine-tuned color matching.
**
***************************************************************************/

typedef struct NQDITInfoTable NQDITInfoTable, *NQDITInfoTablePtr;
struct NQDITInfoTable
	{
	Int16	iTabHidden;	/* number of hidden entries (0 for balanced table) */
	Int16	iTabFlags;	/* flags for this table (<0 if gray scale) */
	Int16	iTabReserved;	/* space for a handle or ptr to other info */
	UInt8	iTabInfo[1];	/* luminance table if gray scale table, hidden colors if 
                                not grays scale */
	};


/***************************************************************************
**
**	NQDPixMap
**
**	Extended PixMap record.
**	Long fields should be aligned on long boundaries.
**	Some day rowBytes will be a long.
**	Add new rowBytes field which has the special bits stripped off and is a long.
**	Add screenFlag and topLeft which are used for and by ShieldCursor.
**
***************************************************************************/

struct NQDPixMap
	{
	Int8 *		baseAddr;	/* same PixMap fields */
	Int32		rowBytes;	/* new row bytes is a long! */
	Rect		bounds;
	Int32		pmVersion;
	Int32		pixelType;
	Int32		pixelSize;
	Int32		cmpCount;
	Int32		cmpSize;
	CTabHandle	pmTable;
	Int32		shift;		/* shift to avoid multiplies and divides */
	Boolean	pixMapFlag;	/* true if original is a pixmap */
	Boolean	screenFlag;	/* this is a screen so call ShieldCursor */
	Int16		unused1;
	Point		topLeft;	/* used by ShieldCursor */
	};
typedef struct NQDPixMap NQDPixMap, *NQDPixMapPtr, **NQDPixMapHdl;



/***************************************************************************
**
**	NQDDrawVars
**
**	Variables used to draw everything.
**
***************************************************************************/

typedef struct NQDDrawVars NQDDrawVars;

typedef void	 (*NQDBlitProcPtr)( NQDDrawVars * );
typedef UInt32  (*NQDXferProcPtr)(  NQDDrawVars *, UInt32, UInt32 );
typedef UInt32 (*NQDCvtProcPtr)(   NQDDrawVars *, UInt32, UInt32 * );

#define	kCurDrawVarsVersion	0x0100

struct NQDDrawVars
	{
	/* version numbers */
	Int16	curVersion;	/* current version of this draw record */
	Int16	minVersion;	/* oldest compatible version number */
	
	/* general drawing variables */
	CGrafPtr	port;		/* destination port */
	Int32		isColorPort;	/* dest port is color */
	Int32		mode;		/* drawing mode */
	Int32		unmappedMode;	/* unmapped drawing mode */
	Point		penSize;	/* port's pen size */
	Int32		colorizeFlag;	/* need to colorize */
	UInt32		foreColor;	/* foreground color long */
	UInt32		backColor;	/* background color long */
	RGBColor	foreRGB;	/* foreground rgb color */
	RGBColor	backRGB;	/* background rgb color */
	NQDPixMap	srcPixMap;	/* src pix map */
	NQDPixMap	dstPixMap;	/* dst pix map */
	NQDPixMap	mskPixMap;	/* msk pix map */
	Rect		srcRect;	/* src bounds */
	Rect		dstRect;	/* dst bounds */
	Rect		mskRect;	/* msk bounds */
	Rect		origSrcRect;	/* unclipped src bounds */
	Rect		origDstRect;	/* unclipped dst bounds */
	Rect		origMskRect;	/* unclipped msk bounds */
	Rect		minRect;	/* drawing bounds */
	Int32		alignLeft;	/* min left aligned to long boundary */
	UInt32		firstMask;	/* left or first mask */
	UInt32		lastMask;	/* right or last mask */
	UInt32		dstPixCnt;	/* num pixels in one row of dstRect */
	UInt32		dstLngCnt;	/* num longs  in one row of dstRect */
	UInt32 *	srcRow;	/* topLeft of srcRect in src */
	UInt32 *	dstRow;	/* topLeft of dstRect in dst */
	UInt32 *	dstPtr;	/* start of slab */
	UInt32 *	maskPtr;	/* start position in rgnBuf */
	Int32		hasMask;	/* flag to use a mask in the transfer */
	Int32		fastFlag;	/* can do fast drawing */
	Int32		trimResult;	/* trim result of regions */
	Int32		bMustScale;	/* must create a scale table */
	Int32		forceBlackWhiteTable; /* create a special table for black and 
						   white */
	long		reserved1;	
	Int32		rgbTransfer;	/* force transfers to use RGB longs rather than 
 					   		pixels */
	long		reserved2;
	Int32		xferSrcShift;	/* the shift of the src data for the xfer proc */
	Int32		xferDstShift;	/* the shift of the dst data for the xfer proc */
	
	/* line drawing variables */
	Point		pt1;		/* endpoints of line being drawn */
	Point		pt2;		/* safe to assume pt1.v <= pt2.v */
	
	/* blitting variables */
	Int32		align;		/* bit offset to align src and dst longs */
	Int32		hBump;		/* horizontal transfer direction */
	UInt32		dstHeight;	/* num rows in dstRect */
	Int32		firstLine;	/* first vertical line to blit */
	Int32		lastLine;	/* last vertical line to blit */
	Int32		vertDir;	/* vertical offset to add to loop counter */
	
	/* drawing procedures */
	Int32		blitType;	/* type of blit loop assumed by xferProc */
	NQDBlitProcPtr blitProc;	/* major drawing procedure */
	NQDXferProcPtr xferProc;	/* transfer pixel procedure */
	UInt32		color1;	/* used by transfer routines */
	UInt32		color2;	/* used by transfer routines */
	UInt32		loBits;	/* used by transfer routines */
	UInt32		hiBits;	/* used by transfer routines */
	UInt32		invertMask;	/* inverts the entire source by pixel/long/rgb */
	Int32		rgb1Size;	/* size of one rgb component */
	UInt32		rgbMask;	/* masks one rgb component */
	UInt32		srcCache;	/* cache source data */
	UInt32		dstCache;	/* cache destination data */
	UInt32		newCache;	/* cache new data */
	
	/* clip region vars */
	RgnHandle	rgnA;		/* drawing mask region */
	RgnHandle	rgnB;		/* drawing mask region */
	RgnHandle	rgnC;		/* drawing mask region */
	Int32		runClip;	/* flag whether run or mask clipping */
	Int32		combineMask;	/* flag to combine the mask into the clip */
	long 		reservedArea[50];
	
	/* pattern expansion variables */
	Int32		patType;	/* type of pattern */
	Int32		patSolid;	/* true if pattern is solid black */
	Int32		patNewFlag;	/* true if pattern is new */
	PixPatHandle	patHdl;	/* handle to pattern */
	UInt32 *	patData;	/* pattern data */
	UInt32 *	patExData;	/* expanded data */
	Int32		patRowBytes;	/* row bytes */
	UInt32		patHMask;	/* rolls index into pattern */
	UInt32		patVMask;	/* rolls index into pattern */
	Int32		patHPos;	/* starting horz offset into pattern */
	Int32		patVPos;	/* starting vert offset into pattern */
	long		privatePattern[64]; 
	
	/* arithmetic mode and colorized variables */
	UInt32		rWt;		/* weight for blend or pins */
	UInt32		gWt;		/* weight for blend or pins */
	UInt32		bWt;		/* weight for blend or pins */
	UInt32		notRWt;	/* inverse weight for blend or pins */
	UInt32		notGWt;	/* inverse weight for blend or pins */
	UInt32		notBWt;	/* inverse weight for blend or pins */
	ColorSpec *	colorTable;	/* color table */
	ITabPtr	invTable;	/* inverse color table */
	NQDITInfoTable *invInfo;	/* inverse color table info */
	Int32		invSize;	/* inv table size (3-5) */
	UInt32		hilitColor;	/* hilite color */
	UInt32		pixMask;	/* masks one pixel of source */
	
	/* slant line vars */
	Fixed		horz;		/* fixed point coord */
	Fixed		leftEdge;	/* of slab */
	Fixed		rightEdge;	/* of slab */
	Fixed		slope;		/* of slanted line dh/dv */
	long		reserved3;
	Int32		leftBytes;	/* byte offset from dstPtr to begin of slab */
	
	/* arc drawing vars */
	Int32		bFrame;	/* true if arc framed */
	Int32		ovalWidth;	/* arc oval width */
	Int32		ovalHeight;	/* arc oval height */
	Int32		startAngle;	/* start arc angle */
	Int32		stopAngle;	/* stop  arc angle */
	long		reservedArea2[24];
	
	/* ScaleBlt vars */
	UInt32 *	scaleTable;	/* maps source index to destination pixel */
	UInt32		scaleTableIsRGBTable; /* flag if scale table is an RGB table or 
						    an index table */
	Int32		srcLeftExtra;	/* extra alignment pixels left of src left  */
	Int32		dstLeftExtra;	/* extra alignment pixels left of dst left */
	Int32		srcLeftAlign;	/* extra alignment bits left of src left  */
	Int32		dstLeftAlign;	/* extra alignment bits left of dst left */
	Int32		dstRightExtra; /* extra alignment pixels right of dst right */
	Int32		dstRightAlign; /* extra alignment bits right of dst right */
	
	long		reservedArea3[28];
	
	/* accelerator vars */
	Int32		refCon;	/* for use by the GetXXXProc and the blitProc */
	/* will not be used by system software */
	
	/* ============================ */
	/* end of version 0x0100 fields */
	};




/********************************************************************
**
**	Constants for the blit proc table.
**
********************************************************************/

enum
	{
	kBitBlitIndex = 0,
	kRgnBlitIndex,
	kPatBlitIndex,
	kPatRgnBlitIndex,
	kScaleBlitIndex,
	kLineBlitIndex,
	kSlabBlitIndex,
	kNumBlitProcs
	};

enum
	{
	kGetBlitProcVers = 0,
	kGetBlitProcPtr,
	kSetBlitProcPtr,
	kRegisterShieldRectangle,
	kUnRegisterShieldRectangle,
	kSetShieldRectangle,
	kAddBlitProcPtr,
	kRemoveBlitProcPtr,
	kAcceleratorActivate,
	kRegisterHdwrMemory,
	kUnregisterHdwrMemory,
	kRelocateAcceleratorMemory,
	kReleaseAcceleratorMemory
	};
	

	
typedef long		(*NQDGetBlitProcPtr)( NQDDrawVars * );
typedef long		(*NQDFinishBlitProcPtr)( long );

/****************************************************************************
**
**	NQDBlitServiceParamBlock
**
**	Used by NQDMisc.
**
****************************************************************************/	
		
struct NQDBlitServiceParamBlock
	{
	long			version;	/* NQDAccel API version number */
	NQDBlitProcPtr	pointer;	/* pointer to blit proc */
	unsigned long		index;		/* blit routine index */
	};
	
typedef struct NQDBlitServiceParamBlock NQDBlitServiceParamBlock, *NQDBlitServiceParamBlockPtr;

/****************************************************************************
**
**	NQDGetBlitProcParamBlock
**
**	Used by NQDAccel.
**
****************************************************************************/	
		
struct NQDGetBlitProcParamBlock
	{
	NQDGetBlitProcPtr		getBlitProc;	/* pointer to blit proc */
	NQDFinishBlitProcPtr	finishProc;		/* pointer to "finish" routine */
	unsigned long			index;			/* blit routine index */
	};

typedef struct NQDGetBlitProcParamBlock NQDGetBlitProcParamBlock, *NQDGetBlitProcParamBlockPtr;




/****************************************************************************
**
**	AcceleratorMemoryParamBlock
**
**	Used by NQDMisc.
**
****************************************************************************/	
		
#define kAcceleratorMemVersion				0x0100

typedef pascal OSErr	(*AllocAcceleratorMemProcPtr)( UInt32 type, UInt32 size, Ptr *Buffer, UInt32 *rowBytes );
typedef pascal OSErr	(*ReleaseAcceleratorMemProcPtr)( Ptr buffer );

struct AcceleratorMemoryParamBlock {
	UInt32								version;			/* returns version of NQDHdwrMem API */
	SInt32								gdRefNum;			/* refNum of attached GDevice */						
	AllocAcceleratorMemProcPtr			getMemProc;			/* ptr to vendor mem allocate routine */
	ReleaseAcceleratorMemProcPtr		releaseMemProc;		/* ptr to vendor mem release routine */
	unsigned long						refCon;				
	unsigned long						reserved;			/* reserved for future use */
};
	
typedef struct AcceleratorMemoryParamBlock AcceleratorMemoryParamBlock, *AcceleratorMemoryParamBlockPtr;


struct RelocateAcceleratorMemoryPB {
	Ptr									oldAddress;
	Ptr									newAddress;
};

typedef struct RelocateAcceleratorMemoryPB RelocateAcceleratorMemoryPB, *RelocateAcceleratorMemoryPBPtr;


extern pascal long NQDMisc(long selector, long *paramBlock);  	

// rcf Adding this prototype. InterfaceLib in UI 3.3 doesn't have NQDMisc, so using 
// this hacked version.
extern pascal long HackNQDMisc(long selector, long *paramBlock);  	

#endif //__NQDAcceleration__

