#
# generate table of pllCtrl setting as a function of freq
#
use lib './packages';

use sgram;
use h3defs;

$outfile = "memtable.h";

$fmin = 30;
$fmax = 120;

$i = 0;
$default8 = 0;
$default16 = 0;

open (OUT,">$outfile") ||
    die "*** $progName ERROR: unable to open $outfile: $!\n";
print OUT "//\n// generated by $0 at ",scalar localtime,"\n//\n";
print OUT "\n";
print OUT "#define MEM_TYPE_SGRAM  0\n";
print OUT "#define MEM_TYPE_SDRAM  1\n";
print OUT "\n";
print OUT "\nstruct memTable_st {\n";
print OUT "    char vendor[8];\n";
print OUT "    FxU32 type;\n";
print OUT "    FxU32 size;\n";
print OUT "    FxU32 dramInit0;\n";
print OUT "    FxU32 sgramMode;\n";
print OUT "} memTable[] = {\n";
print OUT "\n";
print OUT "//      vendor   \ttype \t\tsize \tdramInit0 \tsgramMode\n";
print OUT "//    ---------- \t---- \t\t---- \t--------- \t---------\n";
foreach $name (sort keys %sgramTiming) {
    $vendor = $name;
    next if ( $vendor =~ /DEFAULT/ );
    if ( $vendor =~ /_SGRAM/ ) { $sdram = 0; $typeStr = "MEM_TYPE_SGRAM"; $vendor =~ s/_SGRAM//g; }
    if ( $vendor =~ /_SDRAM/ ) { $sdram = 1; $typeStr = "MEM_TYPE_SDRAM"; $vendor =~ s/_SDRAM//g; }
    if ( $vendor =~ /_16$/ ) { $size = 16; $vendor =~ s/_16$//g; }
    if ( $vendor =~ /_8$/ ) { $size = 8; $vendor =~ s/_8$//g; }
    if ( $vendor eq $defaultVendor ) {
	if ( $typeStr eq "MEM_TYPE_SGRAM" ) {
	    $defaultSgram8 = $i if ($size==8);
	    $defaultSgram16 = $i if ($size==16);
	} else {
	    $defaultSdram16 = $i if ($size==16);
	}
    }
    
    ($dramInit0,$sgramMode,$sgramMask,$sgramColor) = &calcSgramParams($vendor,$sdram,$size);
    $vendorStr = "\"$vendor\"";
    printf OUT "    { %10s, \t%s, %2d, \t0x%08x, \t0x%08x} , \n",$vendorStr,$typeStr,$size,$dramInit0,$sgramMode;
    $i++;
}
print OUT "};\n";
print OUT "#define MEM_TABLE_SIZE (sizeof(memTable)/sizeof(struct memTable_st))\n";

print OUT "\n";
$i = 0;
foreach $name (sort keys %sgramTiming) {
    next if ( $name =~ /DEFAULT/ );
    printf OUT "#define MEM_%s\t%d\n",$name,$i++;
}
 
print OUT "\n";
print OUT "#define MEM_DEFAULT_SGRAM_8  MEM_GENERIC_SGRAM_8\n";
print OUT "#define MEM_DEFAULT_SGRAM_16 MEM_GENERIC_SGRAM_16\n";
print OUT "#define MEM_DEFAULT_SDRAM_16 MEM_GENERIC_SDRAM_16\n";

print OUT "\n";

close OUT;

