#
# Copyright (c) 1998, 3Dfx Interactive, Inc.
# All Rights Reserved.
# 
# This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
# the contents of this file may not be disclosed to third parties, copied or
# duplicated in any form, in whole or in part, without the prior written
# permission of 3Dfx Interactive, Inc.
# 
# RESTRICTED RIGHTS LEGEND:
# Use, duplication or disclosure by the Government is subject to restrictions
# as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
# and Computer Software clause at DFARS 252.227-7013, and/or in similar or
# successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished  -
# rights reserved under the Copyright Laws of the United States.
# 
# 
# $Revision: 5 $ 
# $Date: 5/20/99 8:53a $ 
# 
#

#
# gen_plltable.pl
#
# Generate table of pllCtrl setting as a function of freq for avenger or
# avenger oem. The minimum frequency for both chips is 30 MHz. But the
# maximum for avenger oem is 141 MHz, and this is enforced by logic on the
# chip which wedges the m factor to 24. The max frequency for avenger is
# set to 220 MHz here. To accomodate both devices with one table, the table
# will have only m values of 24 up to 141 MHz, and any m value above 141 MHz.
# The init code makes sure that if the chip is avenger oem, the frequency
# cannot be set above 141 MHz, so only the lower part of the table will be
# used for avenger oem.
#

use lib './packages';

use pll;
use h3defs;

$outfile = "plltable.h";

$fmin = 30;
$fmax_h4_oem = 141;
$fmax_h4 = 220;

open (OUT,">$outfile") ||
    die "*** $progName ERROR: unable to open $outfile: $!\n";
print OUT "//\n// generated by $0 at ",scalar localtime,"\n//\n";

print OUT "\nFxU32 _pllTable[] = {\n";
print OUT "//\t   pllCtrl        n   m   k actual(MHz)\n";
print OUT "//\t   -------      --- --- --- -----------\n";
for ($f=$fmin; $f<=$fmax_h4; $f++) {
    if ($f<=$fmax_h4_oem) {
	# calcOEMPllParams also wedges k to 1.
	$w_wedge = 11;
	($n,$m,$k) = calcOEMPllParams($f,$w_wedge);
    }
    else {
	# calcPllParams actually ignores $w_wedge.
	$w_wedge = -1;
	($n,$m,$k) = calcPllParams($f,$w_wedge);
    }
    $actual = 14.31818 * ($n + 2) / ($m + 2) / 2**$k;
    $pllCtrl  = ($k<<$h3def{SST_PLL_K_SHIFT}) & $h3def{SST_PLL_K};
    $pllCtrl |= ($m<<$h3def{SST_PLL_M_SHIFT}) & $h3def{SST_PLL_M};
    $pllCtrl |= ($n<<$h3def{SST_PLL_N_SHIFT}) & $h3def{SST_PLL_N};
    printf OUT "\t0x%08X,  // %3d %3d %3d %f\n",
    $pllCtrl,$n,$m,$k,$actual;
}
print OUT "};\n";
printf OUT "#define MIN_PLL_FREQ %d\n",$fmin;
# The MAX_PLL_FREQ is an old define possibly used by other files. So I
# left it in.
printf OUT "#define MAX_PLL_FREQ 120\n";
printf OUT "#define MAX_H4_OEM_PLL_FREQ %d\n",$fmax_h4_oem;
printf OUT "#define MAX_H4_PLL_FREQ %d\n",$fmax_h4;
print OUT "FxU32 *pllTable = (FxU32*) (_pllTable - MIN_PLL_FREQ);\n";
