/* -*-c++-*- */
/* */
/* Copyright (c) 1997-1998, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished  -
** rights reserved under the Copyright Laws of the United States.

Module Name:

    h3data.c

Abstract:

    This module contains all the global data used by the H3 driver.

Environment:

    Kernel mode

Revision History:


--*/

#ifdef INCSTBCUST
#ifndef MS_VIEW
#include "..\..\..\..\build\stbcust.inc"
#else
#include "stbcust.inc"
#endif
#endif

#include "dderror.h"
#include "devioctl.h"
#include "miniport.h"

#include "ntddvdeo.h"
#include "video.h"
#include "h3.h"
#include "localpci.h"
#include "cmdcnst.h"

/*****************************************************************************
 *
 * START OF PAGED DATA
 *
 * All of the data listed below is pageable.  Therefore the system can
 * swap the data out to disk when it needs to free some physical memory.
 *
 * Any data accessed while paging is unavailable should be placed above.
 *
 ****************************************************************************/

#if defined(ALLOC_PRAGMA)
#pragma data_seg("PAGE")
#endif

ULONG DebugLevel;

//
//              RangeStart        RangeLength
//              |                 |      RangeInIoSpace
//              |                 |      |  RangeVisible
//        +-----+-----+           |      |  |  RangeShareable
//        v           v           v      v  v  v

VIDEO_ACCESS_RANGE H3AccessRanges[] = {
    {0xfa000000, 0x00000000, 0x02000000, 0, 1, 1},  //   0 membase zero
    {0xf8000000, 0x00000000, 0x02000000, VIDEO_MEMORY_SPACE_P6CACHE, 1, 1},  //   1 membase one
    {0x00000000, 0x00000000, 0x00000100, 1, 1, 1}   //   2 I/O space
};

VIDEO_ACCESS_RANGE H3SmallAccessRanges[] = {
    {0xfa000000, 0x00000000, 0x00600000, 0, 1, 1},  //   0 membase zero
    {0xf8000000, 0x00000000, 0x00900000, 0, 1, 1},  //   1 membase one
    {0x00000000, 0x00000000, 0x00000100, 1, 1, 1},  //   2 I/O space
    {0x000c0000, 0x00000000, 0x00008000, 0, 1, 1},  //   3 ROM
    {0x00000000, 0x00000000, 0x00000000, 0, 0, 0}   //   4 linear range
};

VIDEO_ACCESS_RANGE H3LinearAccessRanges[] = {
    {0x00000000, 0x00000000, 0x01000000, 0, 0, 0}  // 1 3D Linear frame buffer
};

//****************************************************************************
// Memory Timings, and Size Table
//****************************************************************************

//
// These are the init values for the SGRAMs
//
ULONG H3SgramTimings[] =
{
    0x00169d25,     // 8Mbit chips, 1 chipset (4 chips)
    0x00169d25,     // 8Mbit chips, 2 chipsets (8 chips)
    0x00169d25,     // 16Mbit chips, 1 chipset (4 chips)
    0x00169d25      // 16Mbit chips, 2 chipsets (8 chips)
};

//
// Table for computing the display's amount of memory.
//

ULONG gacjMemorySize[] =
{
    0x00400000,  // 0 = 4mb
    0x00800000,  // 1 = 8mb
    0x01000000   // 2 = 16mb
};

//
// XXX XXX XXX
// If the default gamma for glide is changed, this table must be updated!!!
//
// This is the table correspnding to a gamma value of 1.3. We use this
// in cases where the the miniport comes up, and does not find a gamma
// table for Glide, because a gamma of 1.0 (the basic linear ramp) is
// too dark. So we precalculated the 1.3 table, and here it is...
//
//
ULONG DefaultGlideGammaTable[] = {
	0x00000000, 0x00070707, 0x000b0b0b, 0x000e0e0e,
	0x00111111, 0x00141414, 0x00161616, 0x00181818,
	0x001b1b1b, 0x001d1d1d, 0x001f1f1f, 0x00212121,
	0x00232323, 0x00242424, 0x00262626, 0x00282828,
	0x002a2a2a, 0x002b2b2b, 0x002d2d2d, 0x002f2f2f,
	0x00303030, 0x00323232, 0x00333333, 0x00353535,
	0x00363636, 0x00383838, 0x00393939, 0x003b3b3b,
	0x003c3c3c, 0x003e3e3e, 0x003f3f3f, 0x00404040,
	0x00424242, 0x00434343, 0x00444444, 0x00464646,
	0x00474747, 0x00484848, 0x00494949, 0x004b4b4b,
	0x004c4c4c, 0x004d4d4d, 0x004e4e4e, 0x00505050,
	0x00515151, 0x00525252, 0x00535353, 0x00545454,
	0x00565656, 0x00575757, 0x00585858, 0x00595959,
	0x005a5a5a, 0x005b5b5b, 0x005c5c5c, 0x005e5e5e,
	0x005f5f5f, 0x00606060, 0x00616161, 0x00626262,
	0x00636363, 0x00646464, 0x00656565, 0x00666666,
	0x00676767, 0x00686868, 0x00696969, 0x006a6a6a,
	0x006b6b6b, 0x006d6d6d, 0x006e6e6e, 0x006f6f6f,
	0x00707070, 0x00717171, 0x00727272, 0x00737373,
	0x00747474, 0x00757575, 0x00767676, 0x00777777,
	0x00787878, 0x00797979, 0x00797979, 0x007a7a7a,
	0x007b7b7b, 0x007c7c7c, 0x007d7d7d, 0x007e7e7e,
	0x007f7f7f, 0x00808080, 0x00818181, 0x00828282,
	0x00838383, 0x00848484, 0x00858585, 0x00868686,
	0x00878787, 0x00888888, 0x00888888, 0x00898989,
	0x008a8a8a, 0x008b8b8b, 0x008c8c8c, 0x008d8d8d,
	0x008e8e8e, 0x008f8f8f, 0x00909090, 0x00919191,
	0x00919191, 0x00929292, 0x00939393, 0x00949494,
	0x00959595, 0x00969696, 0x00979797, 0x00989898,
	0x00989898, 0x00999999, 0x009a9a9a, 0x009b9b9b,
	0x009c9c9c, 0x009d9d9d, 0x009d9d9d, 0x009e9e9e,
	0x009f9f9f, 0x00a0a0a0, 0x00a1a1a1, 0x00a2a2a2,
	0x00a3a3a3, 0x00a3a3a3, 0x00a4a4a4, 0x00a5a5a5,
	0x00a6a6a6, 0x00a7a7a7, 0x00a7a7a7, 0x00a8a8a8,
	0x00a9a9a9, 0x00aaaaaa, 0x00ababab, 0x00acacac,
	0x00acacac, 0x00adadad, 0x00aeaeae, 0x00afafaf,
	0x00b0b0b0, 0x00b0b0b0, 0x00b1b1b1, 0x00b2b2b2,
	0x00b3b3b3, 0x00b3b3b3, 0x00b4b4b4, 0x00b5b5b5,
	0x00b6b6b6, 0x00b7b7b7, 0x00b7b7b7, 0x00b8b8b8,
	0x00b9b9b9, 0x00bababa, 0x00bababa, 0x00bbbbbb,
	0x00bcbcbc, 0x00bdbdbd, 0x00bebebe, 0x00bebebe,
	0x00bfbfbf, 0x00c0c0c0, 0x00c1c1c1, 0x00c1c1c1,
	0x00c2c2c2, 0x00c3c3c3, 0x00c4c4c4, 0x00c4c4c4,
	0x00c5c5c5, 0x00c6c6c6, 0x00c7c7c7, 0x00c7c7c7,
	0x00c8c8c8, 0x00c9c9c9, 0x00cacaca, 0x00cacaca,
	0x00cbcbcb, 0x00cccccc, 0x00cdcdcd, 0x00cdcdcd,
	0x00cecece, 0x00cfcfcf, 0x00cfcfcf, 0x00d0d0d0,
	0x00d1d1d1, 0x00d2d2d2, 0x00d2d2d2, 0x00d3d3d3,
	0x00d4d4d4, 0x00d5d5d5, 0x00d5d5d5, 0x00d6d6d6,
	0x00d7d7d7, 0x00d7d7d7, 0x00d8d8d8, 0x00d9d9d9,
	0x00dadada, 0x00dadada, 0x00dbdbdb, 0x00dcdcdc,
	0x00dcdcdc, 0x00dddddd, 0x00dedede, 0x00dfdfdf,
	0x00dfdfdf, 0x00e0e0e0, 0x00e1e1e1, 0x00e1e1e1,
	0x00e2e2e2, 0x00e3e3e3, 0x00e3e3e3, 0x00e4e4e4,
	0x00e5e5e5, 0x00e5e5e5, 0x00e6e6e6, 0x00e7e7e7,
	0x00e8e8e8, 0x00e8e8e8, 0x00e9e9e9, 0x00eaeaea,
	0x00eaeaea, 0x00ebebeb, 0x00ececec, 0x00ececec,
	0x00ededed, 0x00eeeeee, 0x00eeeeee, 0x00efefef,
	0x00f0f0f0, 0x00f0f0f0, 0x00f1f1f1, 0x00f2f2f2,
	0x00f2f2f2, 0x00f3f3f3, 0x00f4f4f4, 0x00f4f4f4,
	0x00f5f5f5, 0x00f6f6f6, 0x00f6f6f6, 0x00f7f7f7,
	0x00f8f8f8, 0x00f8f8f8, 0x00f9f9f9, 0x00fafafa,
	0x00fafafa, 0x00fbfbfb, 0x00fcfcfc, 0x00fcfcfc,
	0x00fdfdfd, 0x00fefefe, 0x00fefefe, 0x00ffffff 
};

//
// Mode tables for architectures where int10 may fail
//

/////////////////////////////////////////////////////////////////////////////////
// This is the section containing the timing sequences for the banshee reference
// board. If these timings do not work for your board, you'll need to concoct a
// set of your own timings for all the supported mode/refresh rate combinations
// desired.
//
// We currently support 60, 70, 72, 75 and 85 Hz for all the workstation desktop
// modes. Sometime, it would be nice to extend the timing to include 65 Hz, and
// also some really high refresh rates.
//
// For the game modes, all bets are off.
//
// A timing table consisting entirely of zeroes is considered to be invalid
//

//
// Generic unavailable setting for mode/timings
//
UCHAR H3_Frequency_Not_Available[] = {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

// GAME MODE
// 320 x 200 refresh rates
//
//UCHAR H3_Reference_Freq_320x200x70[] = {
//    0x2f, 0x27, 0x27, 0x93, 0x2a, 0x8e, 0xbb,
//	0x1f, 0x40, 0x91, 0x24, 0x8f, 0x8f, 0xbc,
//	0x80, 0x00, 0x4f, 0x21, 0xee, 0xdb, 0x00
//};

UCHAR H3_STB_Reference_Freq_320x200x70[] = {	// STB
    0x2d, 0x27, 0x27, 0x91, 0x28, 0x8e, 0xbf, 0x1f, 0xc0, 0x9c,
	0x2e, 0x8f, 0x97, 0xb8, 0x80, 0x00, 0x4f, 0x21, 0x73, 0xd1, 0x00
};


UCHAR H3_Reference_Freq_320x200x85[] = {
    0x2f, 0x27, 0x27, 0x93, 0x29, 0x8d, 0xbb, 0x1f, 0xc0, 0x90,
	0x23, 0x8f, 0x8f, 0xbc, 0x80, 0x00, 0x4f, 0x21, 0xd5, 0x77, 0x00
};

// GAME MODE
// 320 x 240 refresh rates
//
UCHAR H3_Reference_Freq_320x240x60[] = {
    0x2d, 0x27, 0x27, 0x91, 0x28, 0x8e, 0x0b, 0x3e, 0xc0, 0xe9,
	0x2b, 0xdf, 0xdf, 0x0c, 0x80, 0x00, 0xcf, 0x21, 0x73, 0xd1, 0x00
};

UCHAR H3_Reference_Freq_320x240x72[] = {
    0x2f, 0x27, 0x27, 0x93, 0x29, 0x8b, 0x06, 0x3e, 0xc0, 0xe8,
	0x2b, 0xdf, 0xdf, 0x07, 0x80, 0x00, 0xcf, 0x21, 0xd5, 0x77, 0x00
};

UCHAR H3_Reference_Freq_320x240x75[] = {
    0x30, 0x27, 0x27, 0x94, 0x28, 0x8c, 0xf2, 0x1f, 0xc0, 0xe0,
	0x23, 0xdf, 0xdf, 0xf3, 0x80, 0x00, 0xcf, 0x21, 0xca, 0xe5, 0x00
};

UCHAR H3_Reference_Freq_320x240x85[] = {
    0x2f, 0x27, 0x27, 0x93, 0x2a, 0x8e, 0xfb, 0x1f, 0xc0, 0xe0,
	0x23, 0xdf, 0xdf, 0xfc, 0x80, 0x00, 0xcf, 0x21, 0x86, 0xae, 0x00
};

// GAME MODE
// 400 x 300 refresh rates
//
UCHAR H3_Reference_Freq_400x300x60[] = {
    0x3d, 0x31, 0x31, 0x81, 0x34, 0x1c, 0x72, 0xf0, 0xe0, 0x58,
	0x2c, 0x57, 0x57, 0x73, 0xa0, 0x00, 0x0f, 0x21, 0x3f, 0xbc, 0x00
};

UCHAR H3_Reference_Freq_400x300x72[] = {
    0x3c, 0x31, 0x31, 0x80, 0x35, 0x1c, 0x98, 0xf0, 0xe0, 0x7c,
	0x22, 0x57, 0x57, 0x99, 0xa0, 0x00, 0x0f, 0x21, 0xfd, 0xe1, 0x00
};

UCHAR H3_Reference_Freq_400x300x75[] = {
    0x3d, 0x31, 0x31, 0x81, 0x32, 0x17, 0x6f, 0xf0, 0xe0, 0x58,
	0x2b, 0x57, 0x57, 0x70, 0xa0, 0x00, 0x0f, 0x21, 0x8a, 0xf7, 0x00
};

UCHAR H3_Reference_Freq_400x300x85[] = {
    0x3d, 0x31, 0x31, 0x81, 0x33, 0x17, 0x75, 0xf0, 0xe0, 0x58,
	0x2b, 0x57, 0x57, 0x76, 0xa0, 0x00, 0x0f, 0x21, 0x2f, 0xcc, 0x00
};


// GAME MODE
// 512 x 384 refresh rates
//
UCHAR H3_Reference_Freq_512x384x60[] = {
    0x4f, 0x3f, 0x3f, 0x93, 0x41, 0x09, 0x24, 0xf5, 0xe0, 0x02,
	0x28, 0xff, 0xff, 0x25, 0x20, 0x00, 0xcf, 0x21, 0xc1, 0xe1, 0x00
};

//UCHAR H3_Reference_Freq_512x384x70[] = {
//    0x4e, 0x3f, 0x3f, 0x92, 0x41, 0x09, 0x24, 0xf5, 0xe0, 0x02,
//	0x28, 0xff, 0xff, 0x25, 0x20, 0x00, 0xcf, 0x21, 0x2b, 0xf9, 0x00
//};

UCHAR H3_STB_Reference_Freq_512x384x70[] = { // STB
	0x4e, 0x3f, 0x3f, 0x92, 0x42, 0x0a, 0x24, 0xf5, 0xe0, 0x02,
	0x28, 0xff, 0xff, 0x25, 0x20, 0x00, 0xcf, 0x21, 0x2b, 0xf9, 0x00};

//UCHAR H3_Reference_Freq_512x384x72[] = {
//    0x4e, 0x3f, 0x3f, 0x92, 0x41, 0x0a, 0x24,
//	0xf5, 0x60, 0x03, 0x29, 0xff, 0xff, 0x25,
//	0x20, 0x00, 0xcf, 0x21, 0xbc, 0x82, 0x00
//};

UCHAR H3_Reference_Freq_512x384x75[] = {
    0x4d, 0x3f, 0x3f, 0x91, 0x40, 0x06, 0x1e, 0xf5, 0xe0, 0x00,
	0x23, 0xff, 0xff, 0x1f, 0x20, 0x00, 0xcf, 0x21, 0x26, 0x77, 0x00
};

UCHAR H3_Reference_Freq_512x384x85[] = {
    0x51, 0x3f, 0x3f, 0x95, 0x42, 0x08, 0x26, 0xf5, 0xe0, 0x00,
	0x23, 0xff, 0xff, 0x27, 0x20, 0x00, 0xcf, 0x21, 0x21, 0x40, 0x00
};

// GAME MODE
// 640 x 400 refresh rates
//
//UCHAR H3_Reference_Freq_640x400x60[] = {
//    0x63, 0x4f, 0x4f, 0x87, 0x54, 0xdc, 0xbb,
//    0x1f, 0x40, 0x91, 0x94, 0x8f, 0x8f, 0xbc,
//    0x20, 0x00, 0x0f, 0x21, 0x17, 0x56, 0x00
//};

//UCHAR H3_Reference_Freq_640x400x70[] = {
//    0x5f, 0x4f, 0x50, 0x82, 0x51, 0x9d, 0xbf, 0x1f, 0x40, 0x9c,
//	0x2e, 0x8f, 0x97, 0xb8, 0x20, 0x00, 0x4f, 0x21, 0x72, 0xd1, 0x00
//};

UCHAR H3_STB_Reference_Freq_640x400x70[] = {	// STB
    0x5f, 0x4f, 0x50, 0x82, 0x51, 0x9d, 0xbf, 0x1f, 0x40, 0x9c,
	0x2e, 0x8f, 0x97, 0xb8, 0x20, 0x00, 0x4f, 0x21, 0x72, 0xd1, 0x00};

//UCHAR H3_Reference_Freq_640x400x72[] = {
//    0x63, 0x4f, 0x4f, 0x87, 0x54, 0xdc, 0xbb,
//    0x1f, 0x40, 0x91, 0x94, 0x8f, 0x8f, 0xbc,
//    0x20, 0x00, 0x4f, 0x21, 0x1b, 0x77, 0x00
//};

//UCHAR H3_Reference_Freq_640x400x85[] = {
//    0x63, 0x4f, 0x4f, 0x87, 0x54, 0xdc, 0xbb,
//    0x1f, 0x40, 0x91, 0x94, 0x8f, 0x8f, 0xbc,
//    0x20, 0x00, 0x4f, 0x21, 0x36, 0x82, 0x00
//};

UCHAR H3_STB_Reference_Freq_640x400x85[] = { // STB
    0x63, 0x4f, 0x4f, 0x87, 0x53, 0x9b, 0xbb, 0x1f, 0x40, 0x90,
	0x23, 0x8f, 0x8f, 0xbc, 0x20, 0x00, 0x4f, 0x21, 0xd4, 0x77, 0x00};

//
// 640 x 480 refresh rates
//
//UCHAR H3_Reference_Freq_640x480x60[] = {
//    0x5f, 0x4f, 0x4f, 0x83, 0x52, 0x9e, 0x0b,
//    0x3e, 0x40, 0xea, 0xec, 0xdf, 0xdf, 0x0c,
//    0x20, 0x00, 0xcf, 0x21, 0x2b, 0xa7, 0x00
//};

UCHAR H3_STB_Reference_Freq_640x480x60[] = {	// STB
    0x5f, 0x4f, 0x50, 0x82, 0x51, 0x9d, 0x0b, 0x3e, 0x40, 0xe9,
	0x2b, 0xdf, 0xe7, 0x04, 0x20, 0x00, 0xcf, 0x21, 0x72, 0xd1, 0x00};


//UCHAR H3_Reference_Freq_640x480x70[] = {
//    // need values
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
//};

//UCHAR H3_Reference_Freq_640x480x72[] = {
//    0x63, 0x4f, 0x4f, 0x87, 0x56, 0x9b, 0x06,
//    0x3e, 0x40, 0xe9, 0xec, 0xdf, 0xdf, 0x07,
//    0x20, 0x00, 0xcf, 0x21, 0x36, 0x82, 0x00
//};

UCHAR H3_STB_Reference_Freq_640x480x72[] = {	// STB
    0x63, 0x4f, 0x50, 0x86, 0x52, 0x97, 0x06, 0x3e, 0x40, 0xe8,
	0x2b, 0xdf, 0xe7, 0xff, 0x20, 0x00, 0xcf, 0x21, 0xd4, 0x77, 0x00};


//UCHAR H3_Reference_Freq_640x480x75[] = {
//    0x64, 0x4f, 0x4f, 0x88, 0x51, 0x99, 0xf2,
//    0x1f, 0x40, 0xe1, 0xe4, 0xdf, 0xdf, 0xf3,
//    0x20, 0x00, 0xcf, 0x21, 0x36, 0x82, 0x00
//};

UCHAR H3_STB_Reference_Freq_640x480x75[] = { // STB
    0x64, 0x4f, 0x4f, 0x88, 0x51, 0x99, 0xf2, 0x1f, 0x40, 0xe0,
	0x23, 0xdf, 0xdf, 0xf3, 0x20, 0x00, 0xcf, 0x21, 0xd4, 0x77, 0x00};


//UCHAR H3_Reference_Freq_640x480x85[] = {
//    0x63, 0x4f, 0x4f, 0x87, 0x56, 0x9e, 0xfb,
//    0x1f, 0x40, 0xe1, 0xe4, 0xdf, 0xdf, 0xfc,
//    0x20, 0x00, 0xcf, 0x21, 0x3e, 0xa9, 0x00
//};

UCHAR H3_STB_Reference_Freq_640x480x85[] = { // STB
    0x63, 0x4f, 0x4f, 0x87, 0x56, 0x9d, 0xfb, 0x1f, 0x40, 0xe0,
	0x23, 0xdf, 0xdf, 0xfc, 0x20, 0x00, 0xcf, 0x21, 0x85, 0xae, 0x00};


//UCHAR H3_Reference_Freq_640x480x100[] = {
//    0x63, 0x4f, 0x4f, 0x87, 0x56, 0x9e, 0xfb,
//    0x1f, 0x40, 0xe1, 0xe4, 0xdf, 0xdf, 0xfc,
//    0x20, 0x00, 0xcf, 0x21, 0x58, 0x45, 0x00
//};

UCHAR H3_STB_Reference_Freq_640x480x100[] = { // STB
    0x65, 0x4f, 0x4f, 0x89, 0x54, 0x9c, 0xfb, 0x1f, 0x40, 0xe0,
	0x23, 0xdf, 0xdf, 0xfc, 0x20, 0x00, 0x4f, 0x21, 0x1f, 0xd7, 0x00};


//UCHAR H3_Reference_Freq_640x480x120[] = {
//    0x63, 0x4f, 0x4f, 0x87, 0x56, 0x9e, 0xfb,
//    0x1f, 0x40, 0xe1, 0xe4, 0xdf, 0xdf, 0xfc,
//    0x20, 0x00, 0xcf, 0x21, 0x48, 0x45, 0x00
//};

UCHAR H3_STB_Reference_Freq_640x480x120[] = { // STB
    0x65, 0x4f, 0x4f, 0x89, 0x54, 0x9c, 0x01, 0x3e, 0x40, 0xe0,
	0x23, 0xdf, 0xdf, 0x02, 0x20, 0x00, 0x4f, 0x21, 0x17, 0xcb, 0x00};

//UCHAR H3_Reference_Freq_640x480x140[] = { 
//	0x63, 0x4f, 0x4f, 0x87, 0x57, 0x9e, 0xfb, 0x1f, 0x40, 0xe0,
//	0x23, 0xdf, 0xdf, 0xfc, 0x20, 0x00, 0xcf, 0x21, 0xdc, 0xea, 0x00
//};

UCHAR H3_STB_Reference_Freq_640x480x140[] = { // STB
    0x67, 0x4f, 0x4f, 0x8b, 0x54, 0x9d, 0x07, 0x3e, 0x40, 0xe0,
	0x23, 0xdf, 0xdf, 0x08, 0x20, 0x00, 0x4f, 0x21, 0x22, 0xae, 0x00};


//UCHAR H3_Reference_Freq_640x480x150[] = {
//    0x65, 0x4f, 0x4f, 0x89, 0x52, 0x82, 0xff,
//    0x1f, 0x40, 0xe1, 0xe4, 0xdf, 0xdf, 0x00,
//    0xa0, 0x00, 0xcf, 0x21, 0x1d, 0x50, 0x00
//};

//UCHAR H3_Reference_Freq_640x480x160[] = {
//	0x63, 0x4f, 0x4f, 0x87, 0x57, 0x9e, 0xfb, 0x1f, 0x40, 0xe0,
//	0x23, 0xdf, 0xdf, 0xfc, 0x20, 0x00, 0xcf, 0x21, 0x34, 0x45, 0x00
//};

UCHAR H3_STB_Reference_Freq_640x480x160[] = { // STB
    0x67, 0x4f, 0x4f, 0x8b, 0x54, 0x9d, 0x0d, 0x3e, 0x40, 0xe0,
	0x23, 0xdf, 0xdf, 0x0e, 0x20, 0x00, 0x4f, 0x21, 0x24, 0x36, 0x00};


#if VIDEO_720_MODES_SUPPORTED
// GAME MODE
// 720 x 400 refresh rates
//
UCHAR H3_Reference_Freq_720x400x60[] = {
    // need values
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

UCHAR H3_Reference_Freq_720x400x72[] = {
    0x70, 0x59, 0x59, 0x94, 0x5e, 0x87, 0xbc,
    0x1f, 0x40, 0x91, 0x24, 0x8f, 0x8f, 0xbd,
    0xa0, 0x00, 0x4f, 0x21, 0x20, 0x13, 0x00
};

UCHAR H3_Reference_Freq_720x400x85[] = {
    0x63, 0x4f, 0x4f, 0x87, 0x54, 0xdc, 0xbc,
    0x1f, 0x40, 0x91, 0x94, 0x8f, 0x8f, 0xbd,
    0x20, 0x00, 0x4f, 0x20, 0x1e, 0x56, 0x00
};
#endif

//
// 720 x 480 -- Mode for Video
//
//UCHAR H3_Reference_Freq_720x480x60[] = {
//    0x70, 0x59, 0x59, 0x94, 0x5e, 0x87, 0x15,
//    0x3e, 0x40, 0xe1, 0x24, 0xdf, 0xdf, 0x16,
//    0xa0, 0x00, 0x4f, 0x21, 0xc4, 0x69, 0x00
//};

UCHAR H3_STB_Reference_Freq_720x480x60[] = { // STB
    0x6b, 0x59, 0x59, 0x8f, 0x5b, 0x8a, 0x0b, 0x3e, 0x40, 0xe9,
	0x2b, 0xdf, 0xdf, 0x0c, 0xa0, 0x00, 0xcf, 0x21, 0xfc, 0x7e, 0x00};


//UCHAR H3_Reference_Freq_720x480x72[] = {
//    0x70, 0x59, 0x59, 0x94, 0x5e, 0x87, 0x15,
//    0x3e, 0x40, 0xe1, 0x24, 0xdf, 0xdf, 0x16,
//    0xa0, 0x00, 0x4f, 0x21, 0xd8, 0x8b, 0x00
//};

UCHAR H3_STB_Reference_Freq_720x480x72[] = { // STB
    0x6e, 0x59, 0x59, 0x92, 0x5b, 0x8c, 0x06, 0x3e, 0x40, 0xe8,
	0x2b, 0xdf, 0xdf, 0x07, 0xa0, 0x00, 0xcf, 0x21, 0xc9, 0xfb, 0x00};


UCHAR H3_STB_Reference_Freq_720x480x85[] = { // STB
    0x70, 0x59, 0x59, 0x94, 0x61, 0x89, 0xfb, 0x1f, 0x40, 0xe0,
	0x23, 0xdf, 0xdf, 0xfc, 0xa0, 0x00, 0xcf, 0x21, 0x1b, 0xb3, 0x00};


//
// 720 x 576 -- Mode for Video
//
//UCHAR H3_Reference_Freq_720x576x72[] = {
//    0x70, 0x59, 0x59, 0x94, 0x5e, 0x87, 0x80,
//    0xf0, 0x60, 0x41, 0x24, 0x3f, 0x3f, 0x81,
//    0xa0, 0x00, 0x4f, 0x21, 0xb0, 0x89, 0x00
//};

UCHAR H3_STB_Reference_Freq_720x576x72[] = { // STB
    0x70, 0x59, 0x59, 0x94, 0x5d, 0x89, 0x80, 0xf0, 0x60, 0x41,
	0x25, 0x3f, 0x3f, 0x81, 0xa0, 0x00, 0x0f, 0x21, 0x2a, 0x8f, 0x00};

//UCHAR H3_Reference_Freq_720x576x100[] = {
//    0x70, 0x59, 0x59, 0x94, 0x5e, 0x87, 0x80,
//    0xf0, 0x60, 0x41, 0x24, 0x3f, 0x3f, 0x81,
//    0xa0, 0x00, 0x4f, 0x21, 0xec, 0xfe, 0x00
//};

UCHAR H3_STB_Reference_Freq_720x576x100[] = { // STB
    0x70, 0x59, 0x59, 0x94, 0x5d, 0x89, 0x80, 0xf0, 0x60, 0x41,
	0x25, 0x3f, 0x3f, 0x81, 0xa0, 0x00, 0x0f, 0x21, 0x17, 0xe9, 0x00};

//
// 800 x 600 refresh rates
//
UCHAR H3_STB_Reference_Freq_800x600x56[] = { // STB
    0x7b, 0x63, 0x63, 0x9f, 0x66, 0x8f, 0x6f, 0xf0, 0x60, 0x58,
	0x2a, 0x57, 0x57, 0x70, 0xa0, 0x00, 0x0f, 0x21, 0x85, 0xae, 0x00};

//UCHAR H3_Reference_Freq_800x600x60[] = {
//    0x7b, 0x63, 0x63, 0x9f, 0x6a, 0x94, 0x6c,
//    0xf0, 0x60, 0x59, 0x5c, 0x57, 0x57, 0x6d,
//    0xa0, 0x00, 0x0f, 0x21, 0x06, 0x1e, 0x00
//};

UCHAR H3_STB_Reference_Freq_800x600x60[] = { // STB
    0x7f, 0x63, 0x63, 0x83, 0x68, 0x18, 0x72, 0xf0, 0x60, 0x58,
	0x2c, 0x57, 0x57, 0x73, 0x80, 0x00, 0x0f, 0x21, 0x26, 0x79, 0x00};


//UCHAR H3_Reference_Freq_800x600x70[] = {
//	// need values
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
//};

//UCHAR H3_Reference_Freq_800x600x72[] = {
//    0x7d, 0x63, 0x63, 0x81, 0x68, 0x17, 0x98,
//    0xf0, 0x60, 0x7d, 0x83, 0x57, 0x57, 0x99,
//    0x80, 0x00, 0x0f, 0x21, 0x06, 0x28, 0x00
//};

UCHAR H3_STB_Reference_Freq_800x600x72[] = { // STB
    0x7d, 0x63, 0x63, 0x81, 0x6a, 0x19, 0x98, 0xf0, 0x60, 0x7c,
	0x22, 0x57, 0x57, 0x99, 0x80, 0x00, 0x0f, 0x21, 0x1f, 0xf9, 0x00};

//UCHAR H3_Reference_Freq_800x600x75[] = {
//    0x7f, 0x63, 0x63, 0x83, 0x65, 0x0f, 0x6f,
//    0xf0, 0x60, 0x59, 0x5c, 0x57, 0x57, 0x70,
//    0x80, 0x00, 0x0f, 0x21, 0x55, 0x9d, 0x00
//};

UCHAR H3_STB_Reference_Freq_800x600x75[] = { // STB
    0x7f, 0x63, 0x63, 0x83, 0x65, 0x0f, 0x6f, 0xf0, 0x60, 0x58,
	0x2b, 0x57, 0x57, 0x70, 0x80, 0x00, 0x0f, 0x21, 0x1f, 0xf7, 0x00};


//UCHAR H3_Reference_Freq_800x600x85[] = {
//    0x7e, 0x63, 0x63, 0x82, 0x67, 0x0f, 0x75,
//    0xf0, 0x60, 0x59, 0x2c, 0x57, 0x57, 0x76,
//    0x80, 0x00, 0x0f, 0x21, 0x31, 0x6c, 0x00
//};

UCHAR H3_STB_Reference_Freq_800x600x85[] = { // STB
    0x7e, 0x63, 0x63, 0x82, 0x67, 0x0f, 0x75, 0xf0, 0x60, 0x58,
	0x2b, 0x57, 0x57, 0x76, 0x80, 0x00, 0x0f, 0x21, 0x17, 0xda, 0x00};

//UCHAR H3_Reference_Freq_800x600x100[] = {
//    0x7e, 0x63, 0x63, 0x82, 0x66, 0x0e, 0x75,
//    0xf0, 0x60, 0x59, 0x2c, 0x57, 0x57, 0x76,
//    0x80, 0x00, 0x0f, 0x21, 0x8c, 0xa9, 0x00
//};

UCHAR H3_STB_Reference_Freq_800x600x100[] = { // STB
    0x81, 0x63, 0x63, 0x85, 0x69, 0x14, 0x7a, 0xf0, 0x60, 0x58,
	0x2b, 0x57, 0x57, 0x7b, 0x80, 0x00, 0x4f, 0x21, 0x2a, 0xe3, 0x00};

//UCHAR H3_Reference_Freq_800x600x120[] = {
//    0x7e, 0x63, 0x63, 0x82, 0x66, 0x0e, 0x75,
//    0xf0, 0x60, 0x59, 0x2c, 0x57, 0x57, 0x76,
//    0x80, 0x00, 0x0f, 0x21, 0x24, 0x3b, 0x00
//};

UCHAR H3_STB_Reference_Freq_800x600x120[] = { // STB
    0x83, 0x63, 0x63, 0x87, 0x6a, 0x15, 0x81, 0xf0, 0x60, 0x58,
	0x2b, 0x57, 0x57, 0x82, 0x80, 0x00, 0x4f, 0x21, 0x1e, 0xd1, 0x00};

//UCHAR H3_Reference_Freq_800x600x140[] = {
//	0x7e, 0x63, 0x63, 0x82, 0x68, 0x10, 0x75, 0xf0, 0x60, 0x58,
//	0x2b, 0x57, 0x57, 0x76, 0x80, 0x00, 0x0f, 0x21, 0x3c, 0x6c, 0x00
//};

UCHAR H3_STB_Reference_Freq_800x600x140[] = { // STB
    0x83, 0x63, 0x63, 0x87, 0x6a, 0x15, 0x89, 0xf0, 0x60, 0x58,
	0x2b, 0x57, 0x57, 0x8a, 0x80, 0x00, 0x4f, 0x21, 0x15, 0x5f, 0x00};


//UCHAR H3_Reference_Freq_800x600x150[] = {
//    0x7f, 0x63, 0x63, 0x83, 0x66, 0x1a, 0x6f,
//    0xf0, 0x60, 0x59, 0x5c, 0x57, 0x57, 0x70,
//    0x80, 0x00, 0x0f, 0x21, 0x11, 0x51, 0x00
//};

//UCHAR H3_Reference_Freq_800x600x160[] = {
//	0x7e, 0x63, 0x63, 0x82, 0x68, 0x10, 0x75, 0xf0, 0x60, 0x58,
//	0x2b, 0x57, 0x57, 0x76, 0x80, 0x00, 0x0f, 0x21, 0x7c, 0xf2, 0x00
//};

UCHAR H3_STB_Reference_Freq_800x600x160[] = { // STB
    0x85, 0x63, 0x63, 0x89, 0x6b, 0x16, 0x91, 0xf0, 0x60, 0x58,
	0x2b, 0x57, 0x57, 0x92, 0x80, 0x00, 0x4f, 0x21, 0x29, 0xc1, 0x00};

//
// 960 x 720 refresh rates
//
UCHAR H3_STB_Reference_Freq_960x720x60[] = { // STB
    0x97, 0x77, 0x77, 0x9b, 0x7d, 0x09, 0xe8, 0xf0, 0x60, 0xd0,
	0x23, 0xcf, 0xcf, 0xe9, 0x00, 0x00, 0x4f, 0x21, 0x22, 0x9a, 0x00};

UCHAR H3_STB_Reference_Freq_960x720x75[] = { // STB
    0x9b, 0x77, 0x77, 0x9f, 0x7e, 0x0b, 0xee, 0xf0, 0x60, 0xd0,
	0x23, 0xcf, 0xcf, 0xef, 0x00, 0x00, 0x4f, 0x21, 0x29, 0x77, 0x00};

UCHAR H3_STB_Reference_Freq_960x720x85[] = { // STB
    0x9b, 0x77, 0x77, 0x9f, 0x7e, 0x0b, 0xf2, 0xf0, 0x60, 0xd0,
	0x23, 0xcf, 0xcf, 0xf3, 0x00, 0x00, 0x4f, 0x21, 0x21, 0x71, 0x00};

//
// 1024 x 768 refresh rates
//
//UCHAR H3_Reference_Freq_1024x768x60[] = {
//    0xa3, 0x7f, 0x7f, 0x87, 0x83, 0x94, 0x24,
//	0xf5, 0x60, 0x04, 0x2a, 0xff, 0xff, 0x25,
//	0x00, 0x00, 0xcf, 0x21, 0x5d, 0xe1, 0x00
//};

UCHAR H3_STB_Reference_Freq_1024x768x60[] = { // STB
    0xa3, 0x7f, 0x7f, 0x87, 0x82, 0x93, 0x24, 0xf5, 0x60, 0x02,
	0x28, 0xff, 0xff, 0x25, 0x00, 0x00, 0xcf, 0x21, 0x29, 0x6b, 0x00};

//UCHAR H3_Reference_Freq_1024x768x70[] = {
//    0xa1, 0x7f, 0x7f, 0x85, 0x83, 0x94, 0x24,
//	0xf5, 0x60, 0x03, 0x29, 0xff, 0xff, 0x25,
//	0x00, 0x00, 0xcf, 0x21, 0x4d, 0xda, 0x00
//};

UCHAR H3_STB_Reference_Freq_1024x768x70[] = { // STB
    0xa1, 0x7f, 0x7f, 0x85, 0x82, 0x93, 0x24, 0xf5, 0x60, 0x02,
	0x28, 0xff, 0xff, 0x25, 0x00, 0x00, 0xcf, 0x21, 0x2a, 0xf9, 0x00};

//UCHAR H3_Reference_Freq_1024x768x72[] = {
//    0xa1, 0x7f, 0x7f, 0x85, 0x83, 0x94, 0x24,
//	0xf5, 0x60, 0x03, 0x29, 0xff, 0xff, 0x25,
//	0x00, 0x00, 0xcf, 0x21, 0x74, 0xa5, 0x00
//};

//UCHAR H3_Reference_Freq_1024x768x75[] = {
//    0x9f, 0x7f, 0x7f, 0x83, 0x82, 0x8e, 0x1e,
//	0xf5, 0x60, 0x01, 0x24, 0xff, 0xff, 0x1f,
//	0x00, 0x00, 0x0f, 0x21, 0x29, 0x82, 0x00
//};

UCHAR H3_STB_Reference_Freq_1024x768x75[] = { // STB
    0x9f, 0x7f, 0x7f, 0x83, 0x81, 0x8d, 0x1e, 0xf5, 0x60, 0x00,
	0x23, 0xff, 0xff, 0x1f, 0x00, 0x00, 0x0f, 0x21, 0x25, 0x77, 0x00};

//UCHAR H3_Reference_Freq_1024x768x85[] = {
//    0xa7, 0x7f, 0x7f, 0x8b, 0x86, 0x92, 0x26,
//	0xf5, 0x60, 0x01, 0x24, 0xff, 0xff, 0x27,
//	0x00, 0x00, 0x0f, 0x21, 0x35, 0xc4, 0x00
//};

UCHAR H3_STB_Reference_Freq_1024x768x85[] = { // STB
    0xa7, 0x7f, 0x7f, 0x8b, 0x85, 0x91, 0x26, 0xf5, 0x60, 0x00,
	0x23, 0xff, 0xff, 0x27, 0x00, 0x00, 0x0f, 0x21, 0x20, 0x40, 0x00};

//UCHAR H3_Reference_Freq_1024x768x100[] = {
//    0xa7, 0x7f, 0x7f, 0x8b, 0x86, 0x92, 0x26,
//	0xf5, 0x60, 0x01, 0x24, 0xff, 0xff, 0x27,
//	0x00, 0x00, 0x0f, 0x21, 0x3c, 0x82, 0x00
//};

UCHAR H3_STB_Reference_Freq_1024x768x100[] = { // STB
    0xa9, 0x7f, 0x7f, 0x8d, 0x88, 0x96, 0x2c, 0xf5, 0x60, 0x00,
	0x23, 0xff, 0xff, 0x2d, 0x00, 0x00, 0x4f, 0x21, 0x28, 0x5d, 0x00};

//UCHAR H3_Reference_Freq_1024x768x120[] = {
//    0xa7, 0x7f, 0x7f, 0x8b, 0x86, 0x92, 0x26,
//	0xf5, 0x60, 0x01, 0x24, 0xff, 0xff, 0x27,
//	0x00, 0x00, 0x0f, 0x21, 0x50, 0xcb, 0x00
//};

UCHAR H3_STB_Reference_Freq_1024x768x120[] = { // STB
    0xab, 0x7f, 0x7f, 0x8f, 0x89, 0x97, 0x35, 0xf5, 0x60, 0x00,
	0x23, 0xff, 0xff, 0x36, 0x00, 0x00, 0x4f, 0x21, 0x15, 0x86, 0x00};

//UCHAR H3_Reference_Freq_1024x768x150[] = {
//    0x52, 0x3f, 0x3f, 0x96, 0x43, 0x4b, 0x1d,
//    0xf5, 0x60, 0x01, 0x04, 0xff, 0xff, 0x1e,
//    0x20, 0x00, 0x0f, 0x21, 0x28, 0x82, 0x01
//};

//
// 1152 x 864 refresh rates
//
//UCHAR H3_Reference_Freq_1152x864x60[] = {
//    0xb4, 0x8f, 0x8f, 0x98, 0x94, 0x80, 0x8e,
//	0xff, 0x60, 0x61, 0x24, 0x5f, 0x5f, 0x8f,
//	0x80, 0x00, 0x0f, 0x21, 0x84, 0xc4, 0x00
//};

UCHAR H3_STB_Reference_Freq_1152x864x60[] = { // STB
    0xb3, 0x8f, 0x8f, 0x97, 0x93, 0x9f, 0x87, 0xff, 0x60, 0x60,
	0x23, 0x5f, 0x5f, 0x88, 0x00, 0x00, 0x0f, 0x21, 0x25, 0x79, 0x00};

//UCHAR H3_Reference_Freq_1152x864x70[] = {
//    0xb4, 0x8f, 0x8f, 0x98, 0x94, 0x80, 0x8e,
//    0xff, 0x60, 0x61, 0x64, 0x5f, 0x5f, 0x8f,
//    0x80, 0x00, 0x0f, 0x21, 0x35, 0xc4, 0x00
//};

UCHAR H3_STB_Reference_Freq_1152x864x70[] = { // STB
    0xb4, 0x8f, 0x8f, 0x98, 0x93, 0x9f, 0x8e, 0xff, 0x60, 0x60,
	0x23, 0x5f, 0x5f, 0x8f, 0x00, 0x00, 0x0f, 0x21, 0x20, 0x40, 0x00};


//UCHAR H3_Reference_Freq_1152x864x72[] = {
//	// need values
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
//};

//UCHAR H3_Reference_Freq_1152x864x75[] = {
//    0xc3, 0x8f, 0x8f, 0x87, 0x98, 0x08, 0x82,
//	0xff, 0x60, 0x61, 0x24, 0x5f, 0x5f, 0x83,
//	0xa0, 0x00, 0x0f, 0x21, 0x50, 0xa4, 0x00
//};

UCHAR H3_STB_Reference_Freq_1152x864x75[] = { // STB
    0xc3, 0x8f, 0x8f, 0x87, 0x97, 0x07, 0x82, 0xff, 0x60, 0x60,
	0x23, 0x5f, 0x5f, 0x83, 0xa0, 0x00, 0x0f, 0x21, 0x29, 0xb3, 0x00};

//UCHAR H3_Reference_Freq_1152x864x85[] = {
//    0xc0, 0x8f, 0x8f, 0x84, 0x98, 0x08, 0x89,
//	0xff, 0x60, 0x61, 0x24, 0x5f, 0x5f, 0x8a,
//	0xa0, 0x00, 0x0f, 0x21, 0x6c, 0xf4, 0x00
//};

UCHAR H3_STB_Reference_Freq_1152x864x85[] = { // STB
    0xbf, 0x8f, 0x8f, 0x83, 0x97, 0x07, 0x8d, 0xff, 0x60, 0x60,
	0x23, 0x5f, 0x5f, 0x8e, 0xa0, 0x00, 0x0f, 0x21, 0x28, 0x64, 0x00};

//UCHAR H3_Reference_Freq_1152x864x100[] = {
//    0x5f, 0x47, 0x47, 0x83, 0x4c, 0x93, 0x82,
//	0xff, 0x60, 0x61, 0x24, 0x5f, 0x5f, 0x83,
//	0x20, 0x00, 0x0f, 0x21, 0x40, 0xb3, 0x01
//};

UCHAR H3_STB_Reference_Freq_1152x864x100[] = { // STB
    0xbf, 0x8f, 0x8f, 0x83, 0x99, 0x09, 0x91, 0xff, 0x60, 0x60,
	0x23, 0x5f, 0x5f, 0x92, 0xa0, 0x00, 0x4f, 0x21, 0x24, 0x6c, 0x00};

//UCHAR H3_Reference_Freq_1152x864x120[] = {
//	0xbf, 0x8f, 0x8f, 0x83, 0x98, 0x08, 0x8d, 0xff, 0x60, 0x60,
//	0x23, 0x5f, 0x5f, 0x8e, 0xa0, 0x00, 0x0f, 0x21, 0x04, 0x22, 0x00
//};

UCHAR H3_STB_Reference_Freq_1152x864x120[] = { // STB
    0xc1, 0x8f, 0x8f, 0x85, 0x9a, 0x0a, 0x9c, 0xff, 0x60, 0x60,
	0x23, 0x5f, 0x5f, 0x9d, 0xa0, 0x00, 0x4f, 0x21, 0x21, 0xf4, 0x00};

//
// 1280 x 960 refresh rates
//
//UCHAR H3_Reference_Freq_1280x960x60[] = {
//    0xdc, 0x9f, 0x9f, 0x80, 0xac, 0x9a, 0xe6,
//	0xff, 0x60, 0xc1, 0x24, 0xbf, 0xbf, 0xe7,
//	0xa0, 0x00, 0x0f, 0x21, 0x50, 0xa4, 0x00
//};

UCHAR H3_STB_Reference_Freq_1280x960x60[] = {
    0xdc, 0x9f, 0x9f, 0x80, 0xab, 0x99, 0xe6, 0xff, 0x60, 0xc0,
	0x23, 0xbf, 0xbf, 0xe7, 0xa0, 0x00, 0x0f, 0x21, 0x29, 0xb3, 0x00};

//UCHAR H3_Reference_Freq_1280x960x75[] = {
//    0xcd, 0x9f, 0x9f, 0x91, 0xa4, 0x16, 0xe6,
//	0xff, 0x60, 0xc1, 0x24, 0xbf, 0xbf, 0xe7,
//	0xa0, 0x00, 0x0f, 0x21, 0x0c, 0x2a, 0x00
//};

UCHAR H3_STB_Reference_Freq_1280x960x75[] = {
    0xd3, 0x9f, 0x9f, 0x97, 0xaa, 0x1b, 0xe8, 0xff, 0x60, 0xc0,
	0x23, 0xbf, 0xbf, 0xe9, 0xa0, 0x00, 0x4f, 0x21, 0x15, 0x7d, 0x00};

//UCHAR H3_Reference_Freq_1280x960x85[] = {
//    0x67, 0x4f, 0x4f, 0x8b, 0x54, 0x9e, 0xf1,
//	0xff, 0x60, 0xc1, 0x24, 0xbf, 0xbf, 0xf2,
//	0x20, 0x00, 0x0f, 0x21, 0x38, 0xa4, 0x01
//};

UCHAR H3_STB_Reference_Freq_1280x960x85[] = {
    0xd3, 0x9f, 0x9f, 0x97, 0xa7, 0x1b, 0xf1, 0xff, 0x60, 0xc0,
	0x23, 0xbf, 0xbf, 0xf2, 0xa0, 0x00, 0x0f, 0x21, 0x18, 0x51, 0x00};

//
// 1280 x 1024 refresh rates
//
//UCHAR H3_Reference_Freq_1280x1024x60[] = {
//	0xce, 0x9f, 0x9f, 0x92, 0xa6, 0x14, 0x28,
//	0x5a, 0x60, 0x01, 0x24, 0xff, 0xff, 0x29,
//	0xa0, 0x41, 0x0f, 0x21, 0x50, 0xa4, 0x00
//};

UCHAR H3_STB_Reference_Freq_1280x1024x60[] = { // STB
    0xce, 0x9f, 0x9f, 0x92, 0xa5, 0x13, 0x28, 0x5a, 0x60, 0x00,
	0x23, 0xff, 0xff, 0x29, 0xa0, 0x41, 0x0f, 0x21, 0x29, 0xb3, 0x00};


//UCHAR H3_Reference_Freq_1280x1024x70[] = {
//	// need values
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
//};

//UCHAR H3_Reference_Freq_1280x1024x72[] = {
//	// need values
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
//};

//UCHAR H3_Reference_Freq_1280x1024x75[] = {
//    0xce, 0x9f, 0x9f, 0x92, 0xa2, 0x14, 0x28,
//	0x5a, 0x60, 0x01, 0x24, 0xff, 0xff, 0x29,
//	0xa0, 0x41, 0x0f, 0x21, 0x30, 0x82, 0x00
//};

UCHAR H3_STB_Reference_Freq_1280x1024x75[] = { // STB
    0xce, 0x9f, 0x9f, 0x92, 0xa1, 0x13, 0x28, 0x5a, 0x60, 0x00,
	0x23, 0xff, 0xff, 0x29, 0xa0, 0x41, 0x0f, 0x21, 0x14, 0x40, 0x00};

//UCHAR H3_Reference_Freq_1280x1024x85[] = {
//    0x67, 0x4f, 0x4f, 0x8b, 0x54, 0x9e, 0x2e,
//	0x5a, 0x60, 0x01, 0x24, 0xff, 0xff, 0x2f,
//	0x20, 0x41, 0x0f, 0x21, 0x28, 0x82, 0x01
//};

UCHAR H3_STB_Reference_Freq_1280x1024x85[] = { // STB
    0xd3, 0x9f, 0x9f, 0x97, 0xa7, 0x1b, 0x2e, 0x5a, 0x60, 0x00,
	0x23, 0xff, 0xff, 0x2f, 0xa0, 0x41, 0x0f, 0x21, 0x24, 0x77, 0x00};

//UCHAR H3_Reference_Freq_1280x1024x100[] = {
//    0x67, 0x4f, 0x4f, 0x8b, 0x54, 0x9e, 0x2e,
//	0x5a, 0x60, 0x01, 0x24, 0xff, 0xff, 0x2f,
//	0x20, 0x41, 0x0f, 0x21, 0x3c, 0xda, 0x01
//};

UCHAR H3_STB_Reference_Freq_1280x1024x100[] = { // STB
    0x69, 0x4f, 0x4f, 0x8d, 0x55, 0x9e, 0x3b, 0x5a, 0x60, 0x00,
	0x23, 0xff, 0xff, 0x3c, 0x20, 0x41, 0x4f, 0x21, 0x28, 0x9e, 0x01};


//
// 1600 x 1024 refresh rates, derived from STB values
//
//UCHAR H3_Reference_Freq_1600x1024x60[] = {
//	0x01, 0xc7, 0xc7, 0x85, 0xcc, 0x00, 0x2c,
//	0x5a, 0x60, 0x02, 0x25, 0xff, 0xff, 0x2d,
//	0x81, 0x41, 0xcf, 0x21, 0x0c, 0x2d, 0x00
//};

UCHAR H3_STB_Reference_Freq_1600x1024x60[] = { // STB
    0x07, 0xc7, 0xc7, 0x8b, 0xd4, 0x09, 0x22, 0x5a, 0x60, 0x00,
	0x23, 0xff, 0xff, 0x23, 0x81, 0x41, 0x4f, 0x21, 0x29, 0xe3, 0x00};

//UCHAR H3_Reference_Freq_1600x1024x76[] = {
//	0x01, 0xc7, 0xc7, 0x85, 0xcc, 0x00, 0x2c,
//	0x5a, 0x60, 0x02, 0x25, 0xff, 0xff, 0x2d,
//	0x81, 0x41, 0xcf, 0x21, 0x4c, 0xf8, 0x00
//};

UCHAR H3_STB_Reference_Freq_1600x1024x76[] = { // STB
    0x7e, 0x63, 0x63, 0x82, 0x65, 0x0f, 0x2c, 0x5a, 0x60, 0x02,
	0x25, 0xff, 0xff, 0x2d, 0x80, 0x41, 0xcf, 0x21, 0x21, 0xec, 0x01};

//UCHAR H3_Reference_Freq_1600x1024x85[] = {
//	0x07, 0xc7, 0xc7, 0x8b, 0xd5, 0x0a, 0x31,
//	0x5a, 0x60, 0x00, 0x23, 0xff, 0xff, 0x32,
//	0x81, 0x41, 0xcf, 0x21, 0x38, 0xd9, 0x00
//};

UCHAR H3_STB_Reference_Freq_1600x1024x85[] = { // STB
    0x83, 0x63, 0x63, 0x87, 0x6a, 0x15, 0x31, 0x5a, 0x60, 0x00,
	0x23, 0xff, 0xff, 0x32, 0x80, 0x41, 0x4f, 0x21, 0x1d, 0xf8, 0x01};

//
// 1600 x 1200 refresh rates
//
//UCHAR H3_Reference_Freq_1600x1200x60[] = {
//    0x82, 0x63, 0x63, 0x86, 0x68, 0x14, 0xe0,
//	0x10, 0x40, 0xb1, 0x24, 0xaf, 0xaf, 0xe1,
//	0x80, 0x55, 0x0f, 0x21, 0x44, 0xd5, 0x01
//};

UCHAR H3_STB_Reference_Freq_1600x1200x60[] = { // STB
    0x82, 0x63, 0x63, 0x86, 0x67, 0x13, 0xe0, 0x10, 0x40, 0xb0,
	0x23, 0xaf, 0xaf, 0xe1, 0x80, 0x55, 0x0f, 0x21, 0x19, 0xb3, 0x01};

//UCHAR H3_Reference_Freq_1600x1200x65[] = {
//    0x82, 0x63, 0x63, 0x86, 0x68, 0x14, 0xe0,
//	0x10, 0x40, 0xb1, 0x24, 0xaf, 0xaf, 0xe1,
//	0x80, 0x55, 0x0f, 0x21, 0x44, 0xe7, 0x01
//};

UCHAR H3_STB_Reference_Freq_1600x1200x65[] = { // STB
    0x82, 0x63, 0x63, 0x86, 0x67, 0x13, 0xe0, 0x10, 0x40, 0xb0,
	0x23, 0xaf, 0xaf, 0xe1, 0x80, 0x55, 0x0f, 0x21, 0x28, 0x91, 0x01};

//UCHAR H3_Reference_Freq_1600x1200x70[] = {
//    0x82, 0x63, 0x63, 0x86, 0x68, 0x14, 0xe0,
//	0x10, 0x40, 0xb1, 0x24, 0xaf, 0xaf, 0xe1,
//	0x80, 0x55, 0x0f, 0x21, 0x34, 0xc4, 0x01
//};

UCHAR H3_STB_Reference_Freq_1600x1200x70[] = { // STB
    0x82, 0x63, 0x63, 0x86, 0x67, 0x13, 0xe0, 0x10, 0x40, 0xb0,
	0x23, 0xaf, 0xaf, 0xe1, 0x80, 0x55, 0x0f, 0x21, 0x20, 0x82, 0x01};

//UCHAR H3_Reference_Freq_1600x1200x72[] = {
//	// need values
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
//    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
//};

//UCHAR H3_Reference_Freq_1600x1200x75[] = {
//    0x82, 0x63, 0x63, 0x86, 0x68, 0x14, 0xe0,
//	0x10, 0x40, 0xb1, 0x24, 0xaf, 0xaf, 0xe1,
//	0x80, 0x55, 0x0f, 0x21, 0x30, 0xc4, 0x01
//};

UCHAR H3_STB_Reference_Freq_1600x1200x75[] = { // STB
    0x82, 0x63, 0x63, 0x86, 0x67, 0x13, 0xe0, 0x10, 0x40, 0xb0,
	0x23, 0xaf, 0xaf, 0xe1, 0x80, 0x55, 0x0f, 0x21, 0x14, 0x61, 0x01};

//UCHAR H3_Reference_Freq_1600x1200x80[] = {
//    0x82, 0x63, 0x63, 0x86, 0x68, 0x14, 0xe0,
//	0x10, 0x40, 0xb1, 0x24, 0xaf, 0xaf, 0xe1,
//	0x80, 0x55, 0x0f, 0x21, 0x24, 0xa4, 0x01
//};

UCHAR H3_STB_Reference_Freq_1600x1200x80[] = { // STB
    0x82, 0x63, 0x63, 0x86, 0x67, 0x13, 0xe0, 0x10, 0x40, 0xb0,
	0x23, 0xaf, 0xaf, 0xe1, 0x80, 0x55, 0x0f, 0x21, 0x11, 0xb3, 0x01};

//UCHAR H3_Reference_Freq_1600x1200x85[] = {
//    0x82, 0x63, 0x63, 0x86, 0x68, 0x14, 0xe0,
//	0x10, 0x40, 0xb1, 0x24, 0xaf, 0xaf, 0xe1,
//	0x80, 0x55, 0x0f, 0x21, 0x0c, 0x4e, 0x01
//};

UCHAR H3_STB_Reference_Freq_1600x1200x85[] = { // STB
    0x82, 0x63, 0x63, 0x86, 0x67, 0x13, 0xe0, 0x10, 0x40, 0xb0,
	0x23, 0xaf, 0xaf, 0xe1, 0x80, 0x55, 0x0f, 0x21, 0x0c, 0x4e, 0x01};

UCHAR H3_STB_Reference_Freq_1600x1200x100[] = { // STB
    0x85, 0x63, 0x63, 0x89, 0x6b, 0x16, 0xf5, 0x10, 0x40, 0xb0,
	0x23, 0xaf, 0xaf, 0xf6, 0x80, 0x55, 0x4f, 0x21, 0x0c, 0x60, 0x01};


//
// 1792 x 1344 refresh rates
//
//UCHAR H3_Reference_Freq_1792x1344x60[] = {
//    0x8e, 0x6f, 0x6f, 0x92, 0x71, 0x1f, 0x74,
//	0x1f, 0x40, 0x43, 0x29, 0x3f, 0x3f, 0x75,
//	0x80, 0x55, 0x0f, 0x21, 0x30, 0xbf, 0x01
//};

UCHAR H3_STB_Reference_Freq_1792x1344x60[] = { // STB
    0x94, 0x6f, 0x6f, 0x98, 0x77, 0x04, 0x70, 0x1f, 0x40, 0x40,
	0x23, 0x3f, 0x3f, 0x71, 0x00, 0x55, 0x4f, 0x21, 0x0d, 0x8d, 0x01};

//UCHAR H3_Reference_Freq_1792x1344x70[] = {
//    0x8e, 0x6f, 0x6f, 0x92, 0x73, 0x1f, 0x74,
//    0x1f, 0x40, 0x43, 0xC9, 0x3f, 0x3f, 0x75,
//    0x80, 0x55, 0x0f, 0x21, 0x24, 0xaf, 0x01
//};

//UCHAR H3_Reference_Freq_1792x1344x72[] = {
//    0x8e, 0x6f, 0x6f, 0x92, 0x73, 0x1f, 0x74,
//	0x1f, 0x40, 0x43, 0x29, 0x3f, 0x3f, 0x75,
//	0x80, 0x55, 0x0f, 0x21, 0x24, 0xb4, 0x01
//};

//UCHAR H3_Reference_Freq_1792x1344x75[] = {
//    0x94, 0x6f, 0x6f, 0x98, 0x74, 0x01, 0x87,
//    0x1f, 0x40, 0x4a, 0x4d, 0x3f, 0x3f, 0x88,
//    0x00, 0x55, 0x0f, 0x21, 0x0c, 0x59, 0x01
//};

UCHAR H3_STB_Reference_Freq_1792x1344x75[] = { // STB
    0x95, 0x6f, 0x6f, 0x99, 0x75, 0x03, 0x87, 0x1f, 0x40, 0x40,
	0x23, 0x3f, 0x3f, 0x88, 0x00, 0x55, 0x4f, 0x21, 0x15, 0xfd, 0x01};

//
// 1800 x 1440 refresh rates
//
//UCHAR H3_Reference_Freq_1800x1440x60[] = {
//    // need values
//    0x8e, 0x6f, 0x6f, 0x92, 0x71, 0x1f, 0xd4,
//    0x1f, 0x40, 0xa3, 0xa9, 0x9f, 0x9f, 0xd5,
//    0x80, 0x55, 0x0f, 0x21, 0x30, 0xcd, 0x01
//};

//UCHAR H3_Reference_Freq_1800x1440x65[] = {
//    0x8e, 0x6f, 0x6f, 0x92, 0x71, 0x1f, 0xd4,
//    0x1f, 0x40, 0xa3, 0xa9, 0x9f, 0x9f, 0xd5,
//    0x80, 0x55, 0x0f, 0x21, 0x0c, 0x4e, 0x01
//};

//UCHAR H3_Reference_Freq_1800x1440x70[] = {
//    0x8f, 0x6f, 0x6f, 0x92, 0x71, 0x59, 0xd4,
//    0x1f, 0x40, 0xa1, 0xa4, 0x9f, 0x9f, 0xd5,
//    0x80, 0x55, 0x0f, 0x21, 0x24, 0xbe, 0x01
//};

//
// 1856 x 1392 refresh rates
//
//UCHAR H3_Reference_Freq_1856x1392x60[] = {
//	0x92, 0x73, 0x73, 0x96, 0x7c, 0x03, 0xa4,
//	0x1f, 0x40, 0x73, 0x29, 0x6f, 0x6f, 0xa5,
//	0x00, 0x55, 0x0f, 0x21, 0x3c, 0xf7, 0x01
//};

UCHAR H3_STB_Reference_Freq_1856x1392x60[] = { // STB
    0x99, 0x73, 0x73, 0x9d, 0x79, 0x07, 0x9d, 0x1f, 0x40, 0x70,
	0x23, 0x6f, 0x6f, 0x9e, 0x00, 0x55, 0x4f, 0x21, 0x11, 0xb5, 0x01};

//UCHAR H3_Reference_Freq_1856x1392x75[] = {
//	0x3b, 0xe7, 0xe7, 0x9f, 0xf8, 0x94, 0xda, 0x1f, 0x40, 0x70,
//	0x23, 0x6f, 0x6f, 0xdb, 0x01, 0x55, 0x0f, 0x21, 0x1c, 0xb3, 0x00
//};

UCHAR H3_STB_Reference_Freq_1856x1392x75[] = { // STB
    0x9b, 0x73, 0x73, 0x9f, 0x7b, 0x09, 0xda, 0x1f, 0x40, 0x70,
	0x23, 0x6f, 0x6f, 0xdb, 0x00, 0x55, 0x4f, 0x21, 0x09, 0x9f, 0x01};


//
// 1900 x 1200 refresh rates
//
UCHAR H3_Reference_Freq_1900x1200x60[] = {
    0x98, 0x76, 0x76, 0x9c, 0x7b, 0x08, 0xe0,
    0x10, 0x40, 0xba, 0xbd, 0xaf, 0xaf, 0xdd,
    0x00, 0x55, 0x0f, 0x21, 0x34, 0xc4, 0x01
};

UCHAR H3_Reference_Freq_1900x1200x70[] = {
    0x98, 0x76, 0x76, 0x9c, 0x79, 0x06, 0xe0,
    0x10, 0x40, 0xba, 0xbd, 0xaf, 0xaf, 0xdd,
    0x00, 0x55, 0x0f, 0x21, 0x0c, 0x4b, 0x01
};

UCHAR H3_Reference_Freq_1900x1200x72[] = {
    0x9d, 0x76, 0x76, 0x81, 0x7b, 0x88, 0xe0,
    0x10, 0x40, 0xba, 0xbd, 0xaf, 0xaf, 0xdd,
    0x00, 0x55, 0x0f, 0x21, 0x04, 0x2f, 0x01
};

UCHAR H3_Reference_Freq_1900x1200x75[] = {
    0x9a, 0x76, 0x76, 0x9e, 0x7a, 0x07, 0xe0,
    0x10, 0x40, 0xba, 0xbd, 0xaf, 0xaf, 0xdd,
    0x00, 0x55, 0x0f, 0x21, 0x04, 0x30, 0x01
};
//
// 1920 x 1080 refresh rates, derived from STB values
//
//UCHAR H3_Reference_Freq_1920x1080x60[] = {
//    0x23, 0xef, 0xef, 0x87, 0xf4, 0x86, 0x63,
//	0x10, 0x40, 0x3a, 0x2d, 0x37, 0x37, 0x64,
//	0x01, 0x55, 0xcf, 0x21, 0x10, 0x41, 0x00
//};

UCHAR H3_STB_Reference_Freq_1920x1080x60[] = { // STB
    0x23, 0xef, 0xef, 0x87, 0xf3, 0x85, 0x63, 0x10, 0x40, 0x3a,
	0x2d, 0x37, 0x37, 0x64, 0x01, 0x55, 0xcf, 0x21, 0x28, 0x84, 0x00};

//UCHAR H3_Reference_Freq_1920x1080x72[] = {
//    0x3b, 0xef, 0xef, 0x9f, 0xf6, 0x91, 0x92,
//	0x10, 0x40, 0x3a, 0x2d, 0x37, 0x37, 0x93,
//	0x01, 0x55, 0xcf, 0x21, 0x24, 0xa4, 0x00
//};

UCHAR H3_STB_Reference_Freq_1920x1080x72[] = { // STB
    0x9b, 0x77, 0x77, 0x9f, 0x7a, 0x08, 0x92, 0x10, 0x40, 0x3a,
	0x2d, 0x37, 0x37, 0x93, 0x00, 0x55, 0xcf, 0x21, 0x11, 0xb3, 0x01};

//UCHAR H3_Reference_Freq_1920x1080x75[] = {
//    0x3b, 0xef, 0xef, 0x9f, 0xfe, 0x98, 0x66,
//	0x10, 0x40, 0x38, 0x2b, 0x37, 0x37, 0x67,
//	0x01, 0x55, 0xcf, 0x21, 0x18, 0x77, 0x00
//};

UCHAR H3_STB_Reference_Freq_1920x1080x75[] = { // STB
    0x9e, 0x77, 0x77, 0x82, 0x7f, 0x8c, 0x66, 0x10, 0x40, 0x38,
	0x2b, 0x37, 0x37, 0x67, 0x00, 0x55, 0x4f, 0x21, 0x11, 0xb7, 0x01};

//UCHAR H3_Reference_Freq_1920x1080x85[] = {
//    0x3d, 0xef, 0xef, 0x81, 0xff, 0x19, 0x6c,
//	0x10, 0x40, 0x38, 0x2b, 0x37, 0x37, 0x6d,
//	0x21, 0x55, 0xcf, 0x21, 0x04, 0x32, 0x00
//};

UCHAR H3_STB_Reference_Freq_1920x1080x85[] = { // STB
    0x9f, 0x77, 0x77, 0x83, 0x80, 0x8d, 0x6c, 0x10, 0x40, 0x38,
	0x2b, 0x37, 0x37, 0x6d, 0x00, 0x55, 0x4f, 0x21, 0x11, 0xd2, 0x01};


//
// 1920 x 1200 refresh rates, derived from STB values
//
//UCHAR H3_Reference_Freq_1920x1200x60[] = {
//    0x3e, 0xef, 0xef, 0x82, 0xf8, 0x18, 0xe0,
//	0x10, 0x40, 0xb2, 0x25, 0xaf, 0xaf, 0xe1,
//	0x21, 0x55, 0xcf, 0x21, 0x2c, 0xae, 0x00
//};

UCHAR H3_STB_Reference_Freq_1920x1200x60[] = { // STB
    0x9d, 0x77, 0x77, 0x81, 0x7f, 0x8c, 0xd8, 0x10, 0x40, 0xb0,
	0x23, 0xaf, 0xaf, 0xd9, 0x00, 0x55, 0x4f, 0x21, 0x15, 0xbb, 0x01};

//UCHAR H3_Reference_Freq_1920x1200x76[] = {
//    0x3e, 0xef, 0xef, 0x82, 0xf8, 0x18, 0xe0,
//	0x10, 0x40, 0xb2, 0x25, 0xaf, 0xaf, 0xe1,
//	0x21, 0x55, 0xcf, 0x21, 0x14, 0x76, 0x00};

UCHAR H3_STB_Reference_Freq_1920x1200x76[] = { // STB
    0x9c, 0x77, 0x77, 0x80, 0x7b, 0x8b, 0xe0, 0x10, 0x40, 0xb2,
	0x25, 0xaf, 0xaf, 0xe1, 0x00, 0x55, 0xcf, 0x21, 0x14, 0x76, 0x01};

UCHAR H3_STB_Reference_Freq_1920x1200x85[] = { // STB
    0xa0, 0x77, 0x77, 0x84, 0x80, 0x8d, 0xea, 0x10, 0x40, 0xb0,
	0x23, 0xaf, 0xaf, 0xeb, 0x00, 0x55, 0x4f, 0x21, 0x08, 0x4d, 0x01};


//
// 1920 x 1440 refresh rates
//
//UCHAR H3_Reference_Freq_1920x1440x60[] = {
//    0x9d, 0x77, 0x77, 0x81, 0x7a, 0x87, 0xda,
//	0x1f, 0x40, 0xa1, 0x24, 0x9f, 0x9f, 0xdb,
//	0x00, 0x55, 0x0f, 0x21, 0x30, 0xe3, 0x01
//};

UCHAR H3_STB_Reference_Freq_1920x1440x60[] = { // STB
    0x9d, 0x77, 0x77, 0x81, 0x7f, 0x8c, 0xda, 0x1f, 0x40, 0xa0,
	0x23, 0x9f, 0x9f, 0xdb, 0x00, 0x55, 0x4f, 0x21, 0x10, 0x60, 0x01};

//UCHAR H3_Reference_Freq_1920x1440x75[] = {
//	0x45, 0xef, 0xef, 0x89, 0x02, 0x1e, 0xda, 0x1f, 0x40, 0xa0,
//	0x23, 0x9f, 0x9f, 0xdb, 0x61, 0x55, 0x0f, 0x21, 0x08, 0x51, 0x00
//};

UCHAR H3_STB_Reference_Freq_1920x1440x75[] = { // STB
    0xa0, 0x77, 0x77, 0x84, 0x80, 0x8e, 0xda, 0x1f, 0x40, 0xa0,
	0x23, 0x9f, 0x9f, 0xdb, 0x00, 0x55, 0x4f, 0x21, 0x08, 0x51, 0x01};

//
// 2048 x 1536 refresh rates, derived from STB values
//
//UCHAR H3_Reference_Freq_2048x1536x60[] = {
//	0x34, 0xff, 0xff, 0x98, 0x06, 0x81, 0x3b,
//	0xba, 0x40, 0x00, 0x23, 0xff, 0xff, 0x3c,
//	0xc1, 0x55, 0x0f, 0x21, 0x08, 0x41, 0x00
//};

UCHAR H3_STB_Reference_Freq_2048x1536x60[] = { // STB
    0x97, 0x7f, 0x7f, 0x9b, 0x82, 0x0f, 0x3b, 0xba, 0x40, 0x00,
	0x23, 0xff, 0xff, 0x3c, 0x00, 0x55, 0x0f, 0x21, 0x11, 0xc7, 0x01};

//UCHAR H3_Reference_Freq_2048x1536x75[] = {
//	0x47, 0xff, 0xff, 0x8b, 0x08, 0x03, 0x44,
//	0xba, 0x40, 0x00, 0x23, 0xff, 0xff, 0x45,
//	0xe1, 0x55, 0x0f, 0x21, 0x04, 0x41, 0x00
//};

UCHAR H3_STB_Reference_Freq_2048x1536x75[] = { // STB
    0xa1, 0x7f, 0x7f, 0x85, 0x83, 0x90, 0x44, 0xba, 0x40, 0x00,
	0x23, 0xff, 0xff, 0x45, 0x00, 0x55, 0x0f, 0x21, 0x10, 0x84, 0x01};

//UCHAR H3_Reference_Freq_2048x1536x85[] = {
//	0x3c, 0xff, 0xff, 0x80, 0x00, 0x17, 0x30,
//	0x3e, 0x40, 0xff, 0x22, 0xff, 0xff, 0x31,
//	0x61, 0x55, 0x0f, 0x21, 0x10, 0x8f, 0x00
//};

UCHAR H3_STB_Reference_Freq_2048x1536x85[] = { // STB
    0x9c, 0x7f, 0x7f, 0x80, 0x7f, 0x8b, 0x30, 0x3e, 0x40, 0xff,
	0x22, 0xff, 0xff, 0x31, 0x00, 0x55, 0x0f, 0x21, 0x10, 0x8f, 0x01};


/*****************************************************************************
 * Generic H3 mode set bits table
 *
 *  Uses the hardware refresh setting for all the listed modes.
 *
 *  Note that any new modes should be added here and to the H3_VIDEO_MODES
 *  table.
 *
 ****************************************************************************/
H3_VIDEO_FREQUENCIES H4GenericFrequencyTable[] = {
    // 640x480x8x60 is the default
    { 8,   640,  480, 60,  H3_STB_Reference_Freq_640x480x60, TRUE },
    { 16,  640,  480, 60,  H3_STB_Reference_Freq_640x480x60, TRUE },
//not supported    { 24,  640,  480, 60,  H3_STB_Reference_Freq_640x480x60, TRUE },
    { 32,  640,  480, 60,  H3_STB_Reference_Freq_640x480x60, TRUE },

    { 8,   640,  480, 72,  H3_STB_Reference_Freq_640x480x72, FALSE },
    { 16,  640,  480, 72,  H3_STB_Reference_Freq_640x480x72, FALSE },
//not supported    { 24,  640,  480, 72,  H3_STB_Reference_Freq_640x480x72, FALSE },
    { 32,  640,  480, 72,  H3_STB_Reference_Freq_640x480x72, FALSE },

    { 8,   640,  480, 75,  H3_STB_Reference_Freq_640x480x75, FALSE },
    { 16,  640,  480, 75,  H3_STB_Reference_Freq_640x480x75, FALSE },
//not supported    { 24,  640,  480, 75,  H3_STB_Reference_Freq_640x480x75, FALSE },
    { 32,  640,  480, 75,  H3_STB_Reference_Freq_640x480x75, FALSE },

    { 8,   640,  480, 85,  H3_STB_Reference_Freq_640x480x85, FALSE },
    { 16,  640,  480, 85,  H3_STB_Reference_Freq_640x480x85, FALSE },
//not supported    { 24,  640,  480, 85,  H3_STB_Reference_Freq_640x480x85, FALSE },
    { 32,  640,  480, 85,  H3_STB_Reference_Freq_640x480x85, FALSE },

    { 8,   640,  480, 100, H3_STB_Reference_Freq_640x480x100, FALSE },
    { 16,  640,  480, 100, H3_STB_Reference_Freq_640x480x100, FALSE },
//not supported    { 24,  640,  480, 100, H3_STB_Reference_Freq_640x480x100, FALSE },
    { 32,  640,  480, 100, H3_STB_Reference_Freq_640x480x100, FALSE },

    { 8,   640,  480, 120, H3_STB_Reference_Freq_640x480x120, FALSE },
    { 16,  640,  480, 120, H3_STB_Reference_Freq_640x480x120, FALSE },
//not supported    { 24,  640,  480, 120, H3_STB_Reference_Freq_640x480x120, FALSE },
    { 32,  640,  480, 120, H3_STB_Reference_Freq_640x480x120, FALSE },

    { 8,   640,  480, 140, H3_STB_Reference_Freq_640x480x140, FALSE },
    { 16,  640,  480, 140, H3_STB_Reference_Freq_640x480x140, FALSE },
//not supported    { 24,  640,  480, 140, H3_STB_Reference_Freq_640x480x140, FALSE },
    { 32,  640,  480, 140, H3_STB_Reference_Freq_640x480x140, FALSE },

    { 8,   640,  480, 160, H3_STB_Reference_Freq_640x480x160, FALSE },
    { 16,  640,  480, 160, H3_STB_Reference_Freq_640x480x160, FALSE },
//not supported    { 24,  640,  480, 160, H3_STB_Reference_Freq_640x480x160, FALSE },
    { 32,  640,  480, 160, H3_STB_Reference_Freq_640x480x160, FALSE },

    { 8,  720,  480, 60, H3_STB_Reference_Freq_720x480x60, TRUE },
    { 16, 720,  480, 60, H3_STB_Reference_Freq_720x480x60, TRUE },
//not supported    { 24, 720,  480, 60, H3_STB_Reference_Freq_720x480x60, TRUE },
    { 32, 720,  480, 60, H3_STB_Reference_Freq_720x480x60, TRUE },

    { 8,  720,  480, 72, H3_STB_Reference_Freq_720x480x72, FALSE },
    { 16, 720,  480, 72, H3_STB_Reference_Freq_720x480x72, FALSE },
//not supported    { 24, 720,  480, 72, H3_STB_Reference_Freq_720x480x72, FALSE },
    { 32, 720,  480, 72, H3_STB_Reference_Freq_720x480x72, FALSE },

    { 8,  720,  480, 85, H3_STB_Reference_Freq_720x480x85, FALSE },
    { 16, 720,  480, 85, H3_STB_Reference_Freq_720x480x85, FALSE },
//not supported    { 24, 720,  480, 85, H3_STB_Reference_Freq_720x480x85, FALSE },
    { 32, 720,  480, 85, H3_STB_Reference_Freq_720x480x85, FALSE },

    { 8,  720,  576, 72, H3_STB_Reference_Freq_720x576x72, FALSE },
    { 16, 720,  576, 72, H3_STB_Reference_Freq_720x576x72, FALSE },
//not supported    { 24, 720,  576, 72, H3_STB_Reference_Freq_720x576x72, FALSE },
    { 32, 720,  576, 72, H3_STB_Reference_Freq_720x576x72, FALSE },

    { 8,  720,  576, 100, H3_STB_Reference_Freq_720x576x100, FALSE },
    { 16, 720,  576, 100, H3_STB_Reference_Freq_720x576x100, FALSE },
//not supported    { 24, 720,  576, 100, H3_STB_Reference_Freq_720x576x100, FALSE },
    { 32, 720,  576, 100, H3_STB_Reference_Freq_720x576x100, FALSE },

    { 8,   800,  600, 56,  H3_STB_Reference_Freq_800x600x56, FALSE },
    { 16,  800,  600, 56,  H3_STB_Reference_Freq_800x600x56, FALSE },
//not supported    { 24,  800,  600, 56,  H3_STB_Reference_Freq_800x600x56, FALSE },
    { 32,  800,  600, 56,  H3_STB_Reference_Freq_800x600x56, FALSE },

    { 8,   800,  600, 60,  H3_STB_Reference_Freq_800x600x60, TRUE },
    { 16,  800,  600, 60,  H3_STB_Reference_Freq_800x600x60, TRUE },
//not supported    { 24,  800,  600, 60,  H3_STB_Reference_Freq_800x600x60, TRUE },
    { 32,  800,  600, 60,  H3_STB_Reference_Freq_800x600x60, TRUE },

    { 8,   800,  600, 72,  H3_STB_Reference_Freq_800x600x72, FALSE },
    { 16,  800,  600, 72,  H3_STB_Reference_Freq_800x600x72, FALSE },
//not supported    { 24,  800,  600, 72,  H3_STB_Reference_Freq_800x600x72, FALSE },
    { 32,  800,  600, 72,  H3_STB_Reference_Freq_800x600x72, FALSE },

    { 8,   800,  600, 75,  H3_STB_Reference_Freq_800x600x75, FALSE },
    { 16,  800,  600, 75,  H3_STB_Reference_Freq_800x600x75, FALSE },
//not supported    { 24,  800,  600, 75,  H3_STB_Reference_Freq_800x600x75, FALSE },
    { 32,  800,  600, 75,  H3_STB_Reference_Freq_800x600x75, FALSE },

    { 8,   800,  600, 85,  H3_STB_Reference_Freq_800x600x85, FALSE },
    { 16,  800,  600, 85,  H3_STB_Reference_Freq_800x600x85, FALSE },
//not supported    { 24,  800,  600, 85,  H3_STB_Reference_Freq_800x600x85, FALSE },
    { 32,  800,  600, 85,  H3_STB_Reference_Freq_800x600x85, FALSE },

    { 8,   800,  600, 100, H3_STB_Reference_Freq_800x600x100, FALSE },
    { 16,  800,  600, 100, H3_STB_Reference_Freq_800x600x100, FALSE },
//not supported    { 24,  800,  600, 100, H3_STB_Reference_Freq_800x600x100, FALSE },
    { 32,  800,  600, 100, H3_STB_Reference_Freq_800x600x100, FALSE },

    { 8,   800,  600, 120, H3_STB_Reference_Freq_800x600x120, FALSE },
    { 16,  800,  600, 120, H3_STB_Reference_Freq_800x600x120, FALSE },
//not supported    { 24,  800,  600, 120, H3_STB_Reference_Freq_800x600x120, FALSE },
    { 32,  800,  600, 120, H3_STB_Reference_Freq_800x600x120, FALSE },

    { 8,   800,  600, 140, H3_STB_Reference_Freq_800x600x140, FALSE },
    { 16,  800,  600, 140, H3_STB_Reference_Freq_800x600x140, FALSE },
//not supported    { 24,  800,  600, 140, H3_STB_Reference_Freq_800x600x140, FALSE },
    { 32,  800,  600, 140, H3_STB_Reference_Freq_800x600x140, FALSE },

    { 8,   800,  600, 160, H3_STB_Reference_Freq_800x600x160, FALSE },
    { 16,  800,  600, 160, H3_STB_Reference_Freq_800x600x160, FALSE },
//not supported    { 24,  800,  600, 160, H3_STB_Reference_Freq_800x600x160, FALSE },
    { 32,  800,  600, 160, H3_STB_Reference_Freq_800x600x160, FALSE },

    { 8,  960, 720, 60,  H3_STB_Reference_Freq_960x720x60, FALSE },
    { 16, 960, 720, 60,  H3_STB_Reference_Freq_960x720x60, FALSE },
//not supported    { 24, 960, 720, 60,  H3_STB_Reference_Freq_960x720x60, FALSE },
    { 32, 960, 720, 60,  H3_STB_Reference_Freq_960x720x60, FALSE },

    { 8,  960, 720, 75,  H3_STB_Reference_Freq_960x720x75, FALSE },
    { 16, 960, 720, 75,  H3_STB_Reference_Freq_960x720x75, FALSE },
//not supported    { 24, 960, 720, 75,  H3_STB_Reference_Freq_960x720x75, FALSE },
    { 32, 960, 720, 75,  H3_STB_Reference_Freq_960x720x75, FALSE },

    { 8,  960, 720, 85,  H3_STB_Reference_Freq_960x720x85, FALSE },
    { 16, 960, 720, 85,  H3_STB_Reference_Freq_960x720x85, FALSE },
//not supported    { 24, 960, 720, 85,  H3_STB_Reference_Freq_960x720x85, FALSE },
    { 32, 960, 720, 85,  H3_STB_Reference_Freq_960x720x85, FALSE },

    { 8,  1024,  768, 60,  H3_STB_Reference_Freq_1024x768x60, FALSE },
    { 16, 1024,  768, 60,  H3_STB_Reference_Freq_1024x768x60, FALSE },
//not supported    { 24, 1024,  768, 60,  H3_STB_Reference_Freq_1024x768x60, FALSE },
    { 32, 1024,  768, 60,  H3_STB_Reference_Freq_1024x768x60,  FALSE },

    { 8,  1024,  768, 70,  H3_STB_Reference_Freq_1024x768x70, FALSE },
    { 16, 1024,  768, 70,  H3_STB_Reference_Freq_1024x768x70, FALSE },
//not supported    { 24, 1024,  768, 70,  H3_STB_Reference_Freq_1024x768x70, FALSE },
    { 32, 1024,  768, 70,  H3_STB_Reference_Freq_1024x768x70, FALSE },

    { 8,  1024,  768, 75,  H3_STB_Reference_Freq_1024x768x75, FALSE },
    { 16, 1024,  768, 75,  H3_STB_Reference_Freq_1024x768x75, FALSE },
//not supported    { 24, 1024,  768, 75,  H3_STB_Reference_Freq_1024x768x75, FALSE },
    { 32, 1024,  768, 75,  H3_STB_Reference_Freq_1024x768x75, FALSE },

    { 8,  1024,  768, 85,  H3_STB_Reference_Freq_1024x768x85, FALSE },
    { 16, 1024,  768, 85,  H3_STB_Reference_Freq_1024x768x85, FALSE },
//not supported    { 24, 1024,  768, 85,  H3_STB_Reference_Freq_1024x768x85, FALSE },
    { 32, 1024,  768, 85,  H3_STB_Reference_Freq_1024x768x85, FALSE },

    { 8,  1024,  768, 100, H3_STB_Reference_Freq_1024x768x100, FALSE },
    { 16, 1024,  768, 100, H3_STB_Reference_Freq_1024x768x100, FALSE },
//not supported    { 24, 1024,  768, 100, H3_STB_Reference_Freq_1024x768x100, FALSE },
    { 32, 1024,  768, 100, H3_STB_Reference_Freq_1024x768x100, FALSE },

    { 8,  1024,  768, 120, H3_STB_Reference_Freq_1024x768x120, FALSE },
    { 16, 1024,  768, 120, H3_STB_Reference_Freq_1024x768x120, FALSE },
//not supported    { 24, 1024,  768, 120, H3_STB_Reference_Freq_1024x768x120, FALSE },
    { 32, 1024,  768, 120, H3_STB_Reference_Freq_1024x768x120, FALSE },

    { 8,  1152,  864, 60,  H3_STB_Reference_Freq_1152x864x60, FALSE },
    { 16, 1152,  864, 60,  H3_STB_Reference_Freq_1152x864x60, FALSE },
//not supported    { 24, 1152,  864, 60,  H3_STB_Reference_Freq_1152x864x60, FALSE },
    { 32, 1152,  864, 60,  H3_STB_Reference_Freq_1152x864x60, FALSE },

    { 8,  1152,  864, 70,  H3_STB_Reference_Freq_1152x864x70, FALSE },
    { 16, 1152,  864, 70,  H3_STB_Reference_Freq_1152x864x70, FALSE },
//not supported    { 24, 1152,  864, 70,  H3_STB_Reference_Freq_1152x864x70, FALSE },
    { 32, 1152,  864, 70,  H3_STB_Reference_Freq_1152x864x70, FALSE },

    { 8,  1152,  864, 75,  H3_STB_Reference_Freq_1152x864x75, FALSE },
    { 16, 1152,  864, 75,  H3_STB_Reference_Freq_1152x864x75, FALSE },
//not supported    { 24, 1152,  864, 75,  H3_STB_Reference_Freq_1152x864x75, FALSE },
    { 32, 1152,  864, 75,  H3_STB_Reference_Freq_1152x864x75, FALSE },

    { 8,  1152,  864, 85,  H3_STB_Reference_Freq_1152x864x85, FALSE },
    { 16, 1152,  864, 85,  H3_STB_Reference_Freq_1152x864x85, FALSE },
//not supported    { 24, 1152,  864, 85,  H3_STB_Reference_Freq_1152x864x85, FALSE },
    { 32, 1152,  864, 85,  H3_STB_Reference_Freq_1152x864x85, FALSE },

    { 8,  1152,  864, 100, H3_STB_Reference_Freq_1152x864x100, FALSE },
    { 16, 1152,  864, 100, H3_STB_Reference_Freq_1152x864x100, FALSE },
//not supported    { 24, 1152,  864, 100, H3_STB_Reference_Freq_1152x864x100, FALSE },
    { 32, 1152,  864, 100, H3_STB_Reference_Freq_1152x864x100, FALSE },

    { 8,  1152,  864, 120, H3_STB_Reference_Freq_1152x864x120, FALSE },
    { 16, 1152,  864, 120, H3_STB_Reference_Freq_1152x864x120, FALSE },
//not supported    { 24, 1152,  864, 120, H3_STB_Reference_Freq_1152x864x120, FALSE },
    { 32, 1152,  864, 120, H3_STB_Reference_Freq_1152x864x120, FALSE },

    { 8,  1280, 960, 60,  H3_STB_Reference_Freq_1280x960x60, FALSE },
    { 16, 1280, 960, 60,  H3_STB_Reference_Freq_1280x960x60, FALSE },
//not supported    { 24, 1280, 960, 60,  H3_STB_Reference_Freq_1280x960x60, FALSE },
    { 32, 1280, 960, 60,  H3_STB_Reference_Freq_1280x960x60, FALSE },

    { 8,  1280, 960, 75,  H3_STB_Reference_Freq_1280x960x75, FALSE },
    { 16, 1280, 960, 75,  H3_STB_Reference_Freq_1280x960x75, FALSE },
//not supported    { 24, 1280, 960, 75,  H3_STB_Reference_Freq_1280x960x75, FALSE },
    { 32, 1280, 960, 75,  H3_STB_Reference_Freq_1280x960x75, FALSE },

    { 8,  1280, 960, 85,  H3_STB_Reference_Freq_1280x960x85, FALSE },
    { 16, 1280, 960, 85,  H3_STB_Reference_Freq_1280x960x85, FALSE },
//not supported    { 24, 1280, 960, 85,  H3_STB_Reference_Freq_1280x960x85, FALSE },
    { 32, 1280, 960, 85,  H3_STB_Reference_Freq_1280x960x85, FALSE },

    { 8,  1280, 1024, 60,  H3_STB_Reference_Freq_1280x1024x60, FALSE },
    { 16, 1280, 1024, 60,  H3_STB_Reference_Freq_1280x1024x60, FALSE },
//not supported    { 24, 1280, 1024, 60,  H3_STB_Reference_Freq_1280x1024x60, FALSE },
    { 32, 1280, 1024, 60,  H3_STB_Reference_Freq_1280x1024x60, FALSE },

    { 8,  1280, 1024, 75,  H3_STB_Reference_Freq_1280x1024x75, FALSE },
    { 16, 1280, 1024, 75,  H3_STB_Reference_Freq_1280x1024x75, FALSE },
//not supported    { 24, 1280, 1024, 75,  H3_STB_Reference_Freq_1280x1024x75, FALSE },
    { 32, 1280, 1024, 75,  H3_STB_Reference_Freq_1280x1024x75, FALSE },

    { 8,  1280, 1024, 85,  H3_STB_Reference_Freq_1280x1024x85, FALSE },
    { 16, 1280, 1024, 85,  H3_STB_Reference_Freq_1280x1024x85, FALSE },
//not supported    { 24, 1280, 1024, 85,  H3_STB_Reference_Freq_1280x1024x85, FALSE },
    { 32, 1280, 1024, 85,  H3_STB_Reference_Freq_1280x1024x85, FALSE },

    { 8,  1280, 1024, 100, H3_STB_Reference_Freq_1280x1024x100, FALSE },
    { 16, 1280, 1024, 100, H3_STB_Reference_Freq_1280x1024x100, FALSE },
//not supported    { 24, 1280, 1024, 100, H3_STB_Reference_Freq_1280x1024x100, FALSE },
    { 32, 1280, 1024, 100, H3_STB_Reference_Freq_1280x1024x100, FALSE },

    { 8,  1600, 1024, 60,  H3_STB_Reference_Freq_1600x1024x60, FALSE },
    { 16, 1600, 1024, 60,  H3_STB_Reference_Freq_1600x1024x60, FALSE },
//not supported    { 24, 1600, 1024, 60,  H3_STB_Reference_Freq_1600x1024x60, FALSE },
    { 32, 1600, 1024, 60,  H3_STB_Reference_Freq_1600x1024x60, FALSE },

    { 8,  1600, 1024, 76,  H3_STB_Reference_Freq_1600x1024x76, FALSE },
    { 16, 1600, 1024, 76,  H3_STB_Reference_Freq_1600x1024x76, FALSE },
//not supported    { 24, 1600, 1024, 76,  H3_STB_Reference_Freq_1600x1024x76, FALSE },
    { 32, 1600, 1024, 76,  H3_STB_Reference_Freq_1600x1024x76, FALSE },

    { 8,  1600, 1024, 85,  H3_STB_Reference_Freq_1600x1024x85, FALSE },
    { 16, 1600, 1024, 85,  H3_STB_Reference_Freq_1600x1024x85, FALSE },
//not supported    { 24, 1600, 1024, 85,  H3_STB_Reference_Freq_1600x1024x85, FALSE },
    { 32, 1600, 1024, 85,  H3_STB_Reference_Freq_1600x1024x85, FALSE },

    { 8,  1600, 1200, 60,  H3_STB_Reference_Freq_1600x1200x60, FALSE },
    { 16, 1600, 1200, 60,  H3_STB_Reference_Freq_1600x1200x60, FALSE },
//not supported    { 24, 1600, 1200, 60,  H3_STB_Reference_Freq_1600x1200x60, FALSE },
    { 32, 1600, 1200, 60,  H3_STB_Reference_Freq_1600x1200x60, FALSE },

    { 8,  1600, 1200, 65,  H3_STB_Reference_Freq_1600x1200x65, FALSE },
    { 16, 1600, 1200, 65,  H3_STB_Reference_Freq_1600x1200x65, FALSE },
//not supported    { 24, 1600, 1200, 65,  H3_STB_Reference_Freq_1600x1200x65, FALSE },
    { 32, 1600, 1200, 65,  H3_STB_Reference_Freq_1600x1200x65, FALSE },

    { 8,  1600, 1200, 70,  H3_STB_Reference_Freq_1600x1200x70, FALSE },
    { 16, 1600, 1200, 70,  H3_STB_Reference_Freq_1600x1200x70, FALSE },
//not supported    { 24, 1600, 1200, 70,  H3_STB_Reference_Freq_1600x1200x70, FALSE },
    { 32, 1600, 1200, 70,  H3_STB_Reference_Freq_1600x1200x70, FALSE },

    { 8,  1600, 1200, 75,  H3_STB_Reference_Freq_1600x1200x75, FALSE },
    { 16, 1600, 1200, 75,  H3_STB_Reference_Freq_1600x1200x75, FALSE },
//not supported    { 24, 1600, 1200, 75,  H3_STB_Reference_Freq_1600x1200x75, FALSE },
    { 32, 1600, 1200, 75,  H3_STB_Reference_Freq_1600x1200x75, FALSE },

    { 8,  1600, 1200, 80,  H3_STB_Reference_Freq_1600x1200x80, FALSE },
    { 16, 1600, 1200, 80,  H3_STB_Reference_Freq_1600x1200x80, FALSE },
//not supported    { 24, 1600, 1200, 80,  H3_STB_Reference_Freq_1600x1200x80, FALSE },
    { 32, 1600, 1200, 80,  H3_STB_Reference_Freq_1600x1200x80, FALSE },

    { 8,  1600, 1200, 85,  H3_STB_Reference_Freq_1600x1200x85, FALSE },
    { 16, 1600, 1200, 85,  H3_STB_Reference_Freq_1600x1200x85, FALSE },
//not supported    { 24, 1600, 1200, 85,  H3_STB_Reference_Freq_1600x1200x85, FALSE },
    { 32, 1600, 1200, 85,  H3_STB_Reference_Freq_1600x1200x85, FALSE },

    { 8,  1600, 1200, 100,  H3_STB_Reference_Freq_1600x1200x100, FALSE },
    { 16, 1600, 1200, 100,  H3_STB_Reference_Freq_1600x1200x100, FALSE },
//not supported    { 24, 1600, 1200, 100,  H3_STB_Reference_Freq_1600x1200x100, FALSE },
    { 32, 1600, 1200, 100,  H3_STB_Reference_Freq_1600x1200x100, FALSE },

    { 8,  1792, 1344, 60,  H3_STB_Reference_Freq_1792x1344x60, FALSE },
    { 16, 1792, 1344, 60,  H3_STB_Reference_Freq_1792x1344x60, FALSE },
//not supported    { 24, 1792, 1344, 60,  H3_STB_Reference_Freq_1792x1344x60, FALSE },
    { 32, 1792, 1344, 60,  H3_STB_Reference_Freq_1792x1344x60, FALSE },

    { 8,  1792, 1344, 75,  H3_STB_Reference_Freq_1792x1344x75, FALSE },
    { 16, 1792, 1344, 75,  H3_STB_Reference_Freq_1792x1344x75, FALSE },
//not supported    { 24, 1792, 1344, 75,  H3_STB_Reference_Freq_1792x1344x75, FALSE },
    { 32, 1792, 1344, 75,  H3_STB_Reference_Freq_1792x1344x75, FALSE },

    { 8,  1856, 1392, 60,  H3_STB_Reference_Freq_1856x1392x60, FALSE },
    { 16, 1856, 1392, 60,  H3_STB_Reference_Freq_1856x1392x60, FALSE },
//not supported    { 24, 1856, 1392, 60,  H3_STB_Reference_Freq_1856x1392x60, FALSE },
    { 32, 1856, 1392, 60,  H3_STB_Reference_Freq_1856x1392x60, FALSE },

    { 8,  1856, 1392, 75,  H3_STB_Reference_Freq_1856x1392x75, FALSE },
    { 16, 1856, 1392, 75,  H3_STB_Reference_Freq_1856x1392x75, FALSE },
//not supported    { 24, 1856, 1392, 75,  H3_STB_Reference_Freq_1856x1392x75, FALSE },
    { 32, 1856, 1392, 75,  H3_STB_Reference_Freq_1856x1392x75, FALSE },

    { 8,  1920, 1080, 60,  H3_STB_Reference_Freq_1920x1080x60, FALSE },
    { 16, 1920, 1080, 60,  H3_STB_Reference_Freq_1920x1080x60, FALSE },
//not supported    { 24, 1920, 1080, 60,  H3_STB_Reference_Freq_1920x1080x60, FALSE },
    { 32, 1920, 1080, 60,  H3_STB_Reference_Freq_1920x1080x60, FALSE },

    { 8,  1920, 1080, 72,  H3_STB_Reference_Freq_1920x1080x72, FALSE },
    { 16, 1920, 1080, 72,  H3_STB_Reference_Freq_1920x1080x72, FALSE },
//not supported    { 24, 1920, 1080, 72,  H3_STB_Reference_Freq_1920x1080x72, FALSE },
    { 32, 1920, 1080, 72,  H3_STB_Reference_Freq_1920x1080x72, FALSE },

    { 8,  1920, 1080, 75,  H3_STB_Reference_Freq_1920x1080x75, FALSE },
    { 16, 1920, 1080, 75,  H3_STB_Reference_Freq_1920x1080x75, FALSE },
//not supported    { 24, 1920, 1080, 75,  H3_STB_Reference_Freq_1920x1080x75, FALSE },
    { 32, 1920, 1080, 75,  H3_STB_Reference_Freq_1920x1080x75, FALSE },

    { 8,  1920, 1080, 85,  H3_STB_Reference_Freq_1920x1080x85, FALSE },
    { 16, 1920, 1080, 85,  H3_STB_Reference_Freq_1920x1080x85, FALSE },
//not supported    { 24, 1920, 1080, 85,  H3_STB_Reference_Freq_1920x1080x85, FALSE },
    { 32, 1920, 1080, 85,  H3_STB_Reference_Freq_1920x1080x85, FALSE },

    { 8,  1920, 1200, 60,  H3_STB_Reference_Freq_1920x1200x60, FALSE },
    { 16, 1920, 1200, 60,  H3_STB_Reference_Freq_1920x1200x60, FALSE },
//not supported    { 24, 1920, 1200, 60,  H3_STB_Reference_Freq_1920x1200x60, FALSE },
    { 32, 1920, 1200, 60,  H3_STB_Reference_Freq_1920x1200x60, FALSE },

    { 8,  1920, 1200, 76,  H3_STB_Reference_Freq_1920x1200x76, FALSE },
    { 16, 1920, 1200, 76,  H3_STB_Reference_Freq_1920x1200x76, FALSE },
//not supported    { 24, 1920, 1200, 76,  H3_STB_Reference_Freq_1920x1200x76, FALSE },
    { 32, 1920, 1200, 76,  H3_STB_Reference_Freq_1920x1200x76, FALSE },

    { 8,  1920, 1200, 85,  H3_STB_Reference_Freq_1920x1200x85, FALSE },
    { 16, 1920, 1200, 85,  H3_STB_Reference_Freq_1920x1200x85, FALSE },
//not supported    { 24, 1920, 1200, 85,  H3_STB_Reference_Freq_1920x1200x85, FALSE },
    { 32, 1920, 1200, 85,  H3_STB_Reference_Freq_1920x1200x85, FALSE },

    { 8,  1920, 1440, 60,  H3_STB_Reference_Freq_1920x1440x60, FALSE },
    { 16, 1920, 1440, 60,  H3_STB_Reference_Freq_1920x1440x60, FALSE },
//not supported    { 24, 1920, 1440, 60,  H3_STB_Reference_Freq_1920x1440x60, FALSE },
    { 32, 1920, 1440, 60,  H3_STB_Reference_Freq_1920x1440x60, FALSE },

    { 8,  1920, 1440, 75,  H3_STB_Reference_Freq_1920x1440x75, FALSE },
    { 16, 1920, 1440, 75,  H3_STB_Reference_Freq_1920x1440x75, FALSE },
//not supported    { 24, 1920, 1440, 75,  H3_STB_Reference_Freq_1920x1440x75, FALSE },
    { 32, 1920, 1440, 75,  H3_STB_Reference_Freq_1920x1440x75, FALSE },

    // Special case 2046 mode.
    // External to the miniport Windows believes this mode is 2046 pixels, internally it actually is
    //       2048 wide with 2 pixels left black.  This is a workaround for a Voodoo3 hardware limitation.
    { 8,  2048, 1536, 60,  H3_STB_Reference_Freq_2048x1536x60, FALSE },
    { 16, 2048, 1536, 60,  H3_STB_Reference_Freq_2048x1536x60, FALSE },
//not supported    { 24, 2048, 1536, 60,  H3_STB_Reference_Freq_2048x1536x60, FALSE },
    { 32, 2048, 1536, 60,  H3_STB_Reference_Freq_2048x1536x60, FALSE },

    { 8,  2048, 1536, 75,  H3_STB_Reference_Freq_2048x1536x75, FALSE },
    { 16, 2048, 1536, 75,  H3_STB_Reference_Freq_2048x1536x75, FALSE },
//not supported    { 24, 2048, 1536, 75,  H3_STB_Reference_Freq_2048x1536x75, FALSE },
    { 32, 2048, 1536, 75,  H3_STB_Reference_Freq_2048x1536x75, FALSE },

//not supported    { 8,  2048, 1536, 85,  H3_STB_Reference_Freq_2048x1536x85, FALSE },
//not supported    { 16, 2048, 1536, 85,  H3_STB_Reference_Freq_2048x1536x85, FALSE },
//not supported    { 24, 2048, 1536, 85,  H3_STB_Reference_Freq_2048x1536x85, FALSE },
//not supported    { 32, 2048, 1536, 85,  H3_STB_Reference_Freq_2048x1536x85, FALSE },
    //
    // From here on, we're either in video or game modes, not
    // desktop modes...
    //
    // GAME MODES:
    //
    // We don't support 8bpp modes at this time, but we will soon.
    //
    {  8, 320,  200, 70,  H3_STB_Reference_Freq_320x200x70, FALSE },
    { 16, 320,  200, 70,  H3_STB_Reference_Freq_320x200x70, FALSE },
//not supported    { 24, 320,  200, 70,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 320,  200, 70,  H3_STB_Reference_Freq_320x200x70, FALSE },

    {  8, 320,  200, 85,  H3_Reference_Freq_320x200x85, FALSE },
    { 16, 320,  200, 85,  H3_Reference_Freq_320x200x85, FALSE },
//not supported    { 24, 320,  200, 85,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 320,  200, 85,  H3_Reference_Freq_320x200x85, FALSE },

    {  8, 320,  240, 60,  H3_Reference_Freq_320x240x60, FALSE },
    { 16, 320,  240, 60,  H3_Reference_Freq_320x240x60, FALSE },
//not supported    { 24, 320,  240, 60,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 320,  240, 60,  H3_Reference_Freq_320x240x60, FALSE },

    {  8, 320,  240, 72,  H3_Reference_Freq_320x240x72, FALSE },
    { 16, 320,  240, 72,  H3_Reference_Freq_320x240x72, FALSE },
//not supported    { 24, 320,  240, 72,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 320,  240, 72,  H3_Reference_Freq_320x240x72, FALSE },

    {  8, 320,  240, 75,  H3_Reference_Freq_320x240x75, FALSE },
    { 16, 320,  240, 75,  H3_Reference_Freq_320x240x75, FALSE },
//not supported    { 24, 320,  240, 75,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 320,  240, 75,  H3_Reference_Freq_320x240x75, FALSE },

    {  8, 320,  240, 85,  H3_Reference_Freq_320x240x85, FALSE },
    { 16, 320,  240, 85,  H3_Reference_Freq_320x240x85, FALSE },
//not supported    { 24, 320,  240, 85,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 320,  240, 85,  H3_Reference_Freq_320x240x85, FALSE },

    {  8, 400,  300, 60,  H3_Reference_Freq_400x300x60, FALSE },
    { 16, 400,  300, 60,  H3_Reference_Freq_400x300x60, FALSE },
//not supported    { 24, 400,  300, 60,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 400,  300, 60,  H3_Reference_Freq_400x300x60, FALSE },

    {  8, 400,  300, 72,  H3_Reference_Freq_400x300x72, FALSE },
    { 16, 400,  300, 72,  H3_Reference_Freq_400x300x72, FALSE },
//not supported    { 24, 400,  300, 72,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 400,  300, 72,  H3_Reference_Freq_400x300x72, FALSE },

    {  8, 400,  300, 75,  H3_Reference_Freq_400x300x75, FALSE },
    { 16, 400,  300, 75,  H3_Reference_Freq_400x300x75, FALSE },
//not supported    { 24, 400,  300, 75,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 400,  300, 75,  H3_Reference_Freq_400x300x75, FALSE },

    {  8, 400,  300, 85,  H3_Reference_Freq_400x300x85, FALSE },
    { 16, 400,  300, 85,  H3_Reference_Freq_400x300x85, FALSE },
//not supported    { 24, 400,  300, 85,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 400,  300, 85,  H3_Reference_Freq_400x300x85, FALSE },

    {  8, 512,  384, 60,  H3_Reference_Freq_512x384x60, FALSE },
    { 16, 512,  384, 60,  H3_Reference_Freq_512x384x60, FALSE },
//not supported    { 24, 512,  384, 60,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 512,  384, 60,  H3_Reference_Freq_512x384x60, FALSE },

    {  8, 512,  384, 70,  H3_STB_Reference_Freq_512x384x70, FALSE },
    { 16, 512,  384, 70,  H3_STB_Reference_Freq_512x384x70, FALSE },
//not supported    { 24, 512,  384, 70,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 512,  384, 70,  H3_STB_Reference_Freq_512x384x70, FALSE },

// add 512x384@72 to meet Gateway Spec.
//not supported    {  8, 512,  384, 72,  H3_Reference_Freq_512x384x72, FALSE },
//not supported    { 16, 512,  384, 72,  H3_Reference_Freq_512x384x72, FALSE },

    {  8, 512,  384, 75,  H3_Reference_Freq_512x384x75, FALSE },
    { 16, 512,  384, 75,  H3_Reference_Freq_512x384x75, FALSE },
//not supported    { 24, 512,  384, 75,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 512,  384, 75,  H3_Reference_Freq_512x384x75, FALSE },

    {  8, 512,  384, 85,  H3_Reference_Freq_512x384x85, FALSE },
    { 16, 512,  384, 85,  H3_Reference_Freq_512x384x85, FALSE },
//not supported    { 24, 512,  384, 85,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 512,  384, 85,  H3_Reference_Freq_512x384x85, FALSE },

    {  8, 640,  400, 70,  H3_STB_Reference_Freq_640x400x70, FALSE },
    { 16, 640,  400, 70,  H3_STB_Reference_Freq_640x400x70, FALSE },
//not supported    { 24, 640,  400, 70,  H3_STB_Reference_Freq_640x400x70, FALSE },	// no 24 bit mode
    { 32, 640,  400, 70,  H3_STB_Reference_Freq_640x400x70, FALSE },

    {  8, 640,  400, 85,  H3_STB_Reference_Freq_640x400x85, FALSE },
    { 16, 640,  400, 85,  H3_STB_Reference_Freq_640x400x85, FALSE },
//not supported    { 24, 640,  400, 85,  H3_STB_Reference_Freq_640x400x85, FALSE },	// no 24 bit mode
    { 32, 640,  400, 85,  H3_STB_Reference_Freq_640x400x85, FALSE },

	//
	// These are the "funny" modes required for stuff like
	// video
	//
#if VIDEO_720_MODES_SUPPORTED
//not supported    { 8,  720,  400, 60, H3_Reference_Freq_720x400x60, FALSE },
//not supported    { 16, 720,  400, 60, H3_Reference_Freq_720x400x60, FALSE },
//not supported    { 24, 720,  400, 60, H3_Reference_Freq_720x400x60, FALSE },
//not supported    { 32, 720,  400, 60, H3_Reference_Freq_720x400x60, FALSE },

    { 8,  720,  400, 72, H3_Reference_Freq_720x400x72, FALSE },
    { 16, 720,  400, 72, H3_Reference_Freq_720x400x72, FALSE },
//not supported    { 24, 720,  400, 72, H3_Reference_Freq_720x400x72, FALSE },
    { 32, 720,  400, 72, H3_Reference_Freq_720x400x72, FALSE },
#endif

#if 0
// This is the graveyard for frequency/mode pairs that are not
// supported now or in the forseeable future...
    // 1280x960
    //
    { 8,  1280, 960, 60,  H3_Reference_Freq_1280x960x60, FALSE },
    { 16, 1280, 960, 60,  H3_Reference_Freq_1280x960x60, FALSE },
    { 24, 1280, 960, 60,  H3_Reference_Freq_1280x960x60, FALSE },
    { 32, 1280, 960, 60,  H3_Reference_Freq_1280x960x60, FALSE },

    { 8,  1280, 960, 72,  H3_Reference_Freq_1280x960x72, FALSE },
    { 16, 1280, 960, 72,  H3_Reference_Freq_1280x960x72, FALSE },
    { 24, 1280, 960, 72,  H3_Reference_Freq_1280x960x72, FALSE },
    { 32, 1280, 960, 72,  H3_Reference_Freq_1280x960x72, FALSE },

    { 8,  1280, 960, 75,  H3_Reference_Freq_1280x960x75, FALSE },
    { 16, 1280, 960, 75,  H3_Reference_Freq_1280x960x75, FALSE },
    { 24, 1280, 960, 75,  H3_Reference_Freq_1280x960x75, FALSE },
    { 32, 1280, 960, 75,  H3_Reference_Freq_1280x960x75, FALSE },

    { 8,  1280, 960, 85,  H3_Reference_Freq_1280x960x85, FALSE },
    { 16, 1280, 960, 85,  H3_Reference_Freq_1280x960x85, FALSE },
    { 24, 1280, 960, 85,  H3_Reference_Freq_1280x960x85, FALSE },
    { 32, 1280, 960, 85,  H3_Reference_Freq_1280x960x85, FALSE },

    { 8,  1920, 1440, 70,  H3_Frequency_Not_Available, FALSE },
    { 16, 1920, 1440, 70,  H3_Frequency_Not_Available, FALSE },
    { 24, 1920, 1440, 70,  H3_Frequency_Not_Available, FALSE },
    { 32, 1920, 1440, 70,  H3_Frequency_Not_Available, FALSE },

    { 8,  1920, 1440, 72,  H3_Frequency_Not_Available, FALSE },
    { 16, 1920, 1440, 72,  H3_Frequency_Not_Available, FALSE },
    { 24, 1920, 1440, 72,  H3_Frequency_Not_Available, FALSE },
    { 32, 1920, 1440, 72,  H3_Frequency_Not_Available, FALSE },

    { 8,  1920, 1440, 75,  H3_Frequency_Not_Available, FALSE },
    { 16, 1920, 1440, 75,  H3_Frequency_Not_Available, FALSE },
    { 24, 1920, 1440, 75,  H3_Frequency_Not_Available, FALSE },
    { 32, 1920, 1440, 75,  H3_Frequency_Not_Available, FALSE },

    { 8,  1920, 1440, 80,  H3_Frequency_Not_Available, FALSE },
    { 16, 1920, 1440, 80,  H3_Frequency_Not_Available, FALSE },
    { 24, 1920, 1440, 80,  H3_Frequency_Not_Available, FALSE },
    { 32, 1920, 1440, 80,  H3_Frequency_Not_Available, FALSE },

    { 8,  1920, 1440, 85,  H3_Frequency_Not_Available, FALSE },
    { 16, 1920, 1440, 85,  H3_Frequency_Not_Available, FALSE },
    { 24, 1920, 1440, 85,  H3_Frequency_Not_Available, FALSE },
    { 32, 1920, 1440, 85,  H3_Frequency_Not_Available, FALSE },
#endif
    { 0 }   // Mark the end

 };

H3_VIDEO_FREQUENCIES H4GenericInt10FrequencyTable[] = {
    { 8,   640,  480, 1, H3_STB_Reference_Freq_640x480x60, FALSE },
    { 8,   800,  600, 1, H3_STB_Reference_Freq_800x600x60, FALSE },
    { 8,  1024,  768, 1, H3_STB_Reference_Freq_1024x768x60, FALSE },
#if ALLOW_ENHANCED_MODES
#ifndef	CUST_PRUNE_MODES
    { 8,  1152,  864, 1, H3_STB_Reference_Freq_1152x864x60, FALSE },
#endif // CUST_PRUNE_MODES
#endif // ALLOW_ENHANCED_MODES
    { 8,  1280, 1024, 1, H3_STB_Reference_Freq_1280x1024x60, FALSE },
    { 8,  1600, 1200, 1, H3_STB_Reference_Freq_1600x1200x60, FALSE },

    { 16,  640,  480, 1, H3_STB_Reference_Freq_640x480x60, FALSE },
    { 16,  800,  600, 1, H3_STB_Reference_Freq_800x600x60, FALSE },
    { 16, 1024,  768, 1, H3_STB_Reference_Freq_1024x768x60, FALSE },
#if ALLOW_ENHANCED_MODES
#ifndef	CUST_PRUNE_MODES
    { 16, 1152,  864, 1, H3_STB_Reference_Freq_1152x864x60, FALSE },
#endif // CUST_PRUNE_MODES
#endif // ALLOW_ENHANCED_MODES
    { 16, 1280, 1024, 1, H3_STB_Reference_Freq_1280x1024x60, FALSE },
    { 16, 1600, 1200, 1, H3_STB_Reference_Freq_1600x1200x60, FALSE },

//not supported    { 24,  640,  480, 1, H3_STB_Reference_Freq_640x480x60, FALSE },
//not supported    { 24,  800,  600, 1, H3_STB_Reference_Freq_800x600x60, FALSE },
//not supported    { 24, 1024,  768, 1, H3_STB_Reference_Freq_1024x768x60, FALSE },
#if ALLOW_ENHANCED_MODES
#ifndef	CUST_PRUNE_MODES
//not supported    { 24, 1152,  864, 1, H3_Reference_Freq_1152x864x60, FALSE },
#endif // CUST_PRUNE_MODES
#endif // ALLOW_ENHANCED_MODES
//not supported    { 24, 1280, 1024, 1, H3_STB_Reference_Freq_1280x1024x60, FALSE },
//not supported    { 24, 1600, 1200, 1, H3_STB_Reference_Freq_1600x1200x60, FALSE },

    { 32,  640,  480, 1, H3_STB_Reference_Freq_640x480x60, FALSE },
    { 32,  800,  600, 1, H3_STB_Reference_Freq_800x600x60, FALSE },
    { 32, 1024,  768, 1, H3_STB_Reference_Freq_1024x768x60, FALSE },
#if ALLOW_ENHANCED_MODES
#ifndef	CUST_PRUNE_MODES
    { 32, 1152,  864, 1, H3_STB_Reference_Freq_1152x864x60, FALSE },
#endif // CUST_PRUNE_MODES
#endif // ALLOW_ENHANCED_MODES
    { 32, 1280, 1024, 1, H3_STB_Reference_Freq_1280x1024x60, FALSE },
    { 32, 1600, 1200, 1, H3_STB_Reference_Freq_1600x1200x60, FALSE },

    { 0 }   // Mark the end

};

H3_VIDEO_FREQUENCIES H4VariableInt10FrequencyTable[] = {
    // 640x480x8x60 is the default
    { 8,   640,  480, 60,  H3_STB_Reference_Freq_640x480x60, FALSE },
    { 16,  640,  480, 60,  H3_STB_Reference_Freq_640x480x60, FALSE },
//not supported    { 24,  640,  480, 60,  H3_STB_Reference_Freq_640x480x60, FALSE },
    { 32,  640,  480, 60,  H3_STB_Reference_Freq_640x480x60, FALSE },

    { 8,   640,  480, 72,  H3_STB_Reference_Freq_640x480x72, FALSE },
    { 16,  640,  480, 72,  H3_STB_Reference_Freq_640x480x72, FALSE },
//not supported    { 24,  640,  480, 72,  H3_STB_Reference_Freq_640x480x72, FALSE },
    { 32,  640,  480, 72,  H3_STB_Reference_Freq_640x480x72, FALSE },

    { 8,   640,  480, 75,  H3_STB_Reference_Freq_640x480x75, FALSE },
    { 16,  640,  480, 75,  H3_STB_Reference_Freq_640x480x75, FALSE },
//not supported    { 24,  640,  480, 75,  H3_STB_Reference_Freq_640x480x75, FALSE },
    { 32,  640,  480, 75,  H3_STB_Reference_Freq_640x480x75, FALSE },

    { 8,   640,  480, 85,  H3_STB_Reference_Freq_640x480x85, FALSE },
    { 16,  640,  480, 85,  H3_STB_Reference_Freq_640x480x85, FALSE },
//not supported    { 24,  640,  480, 85,  H3_STB_Reference_Freq_640x480x85, FALSE },
    { 32,  640,  480, 85,  H3_STB_Reference_Freq_640x480x85, FALSE },

    { 8,   640,  480, 100, H3_STB_Reference_Freq_640x480x100, FALSE },
    { 16,  640,  480, 100, H3_STB_Reference_Freq_640x480x100, FALSE },
//not supported    { 24,  640,  480, 100, H3_STB_Reference_Freq_640x480x100, FALSE },
    { 32,  640,  480, 100, H3_STB_Reference_Freq_640x480x100, FALSE },

    { 8,   640,  480, 120, H3_STB_Reference_Freq_640x480x120, FALSE },
    { 16,  640,  480, 120, H3_STB_Reference_Freq_640x480x120, FALSE },
//not supported    { 24,  640,  480, 120, H3_STB_Reference_Freq_640x480x120, FALSE },
    { 32,  640,  480, 120, H3_STB_Reference_Freq_640x480x120, FALSE },

    { 8,   640,  480, 140, H3_STB_Reference_Freq_640x480x140, FALSE },
    { 16,  640,  480, 140, H3_STB_Reference_Freq_640x480x140, FALSE },
//not supported    { 24,  640,  480, 140, H3_STB_Reference_Freq_640x480x140, FALSE },
    { 32,  640,  480, 140, H3_STB_Reference_Freq_640x480x140, FALSE },

    { 8,   640,  480, 160, H3_STB_Reference_Freq_640x480x160, FALSE },
    { 16,  640,  480, 160, H3_STB_Reference_Freq_640x480x160, FALSE },
//not supported    { 24,  640,  480, 160, H3_STB_Reference_Freq_640x480x160, FALSE },
    { 32,  640,  480, 160, H3_STB_Reference_Freq_640x480x160, FALSE },

    { 8,  720,  480, 60, H3_STB_Reference_Freq_720x480x60, FALSE },
    { 16, 720,  480, 60, H3_STB_Reference_Freq_720x480x60, FALSE },
//not supported    { 24, 720,  480, 60, H3_STB_Reference_Freq_720x480x60, FALSE },
    { 32, 720,  480, 60, H3_STB_Reference_Freq_720x480x60, FALSE },

    { 8,  720,  480, 72, H3_STB_Reference_Freq_720x480x72, FALSE },
    { 16, 720,  480, 72, H3_STB_Reference_Freq_720x480x72, FALSE },
//not supported    { 24, 720,  480, 72, H3_STB_Reference_Freq_720x480x72, FALSE },
    { 32, 720,  480, 72, H3_STB_Reference_Freq_720x480x72, FALSE },

    { 8,  720,  480, 85, H3_STB_Reference_Freq_720x480x85, FALSE },
    { 16, 720,  480, 85, H3_STB_Reference_Freq_720x480x85, FALSE },
//not supported    { 24, 720,  480, 85, H3_STB_Reference_Freq_720x480x85, FALSE },
    { 32, 720,  480, 85, H3_STB_Reference_Freq_720x480x85, FALSE },

    { 8,  720,  576, 72, H3_STB_Reference_Freq_720x576x72, FALSE },
    { 16, 720,  576, 72, H3_STB_Reference_Freq_720x576x72, FALSE },
//not supported    { 24, 720,  576, 72, H3_STB_Reference_Freq_720x576x72, FALSE },
    { 32, 720,  576, 72, H3_STB_Reference_Freq_720x576x72, FALSE },

    { 8,  720,  576, 100, H3_STB_Reference_Freq_720x576x100, FALSE },
    { 16, 720,  576, 100, H3_STB_Reference_Freq_720x576x100, FALSE },
//not supported    { 24, 720,  576, 100, H3_STB_Reference_Freq_720x576x100, FALSE },
    { 32, 720,  576, 100, H3_STB_Reference_Freq_720x576x100, FALSE },

    { 8,   800,  600, 56,  H3_STB_Reference_Freq_800x600x56, FALSE },
    { 16,  800,  600, 56,  H3_STB_Reference_Freq_800x600x56, FALSE },
//not supported    { 24,  800,  600, 56,  H3_STB_Reference_Freq_800x600x56, FALSE },
    { 32,  800,  600, 56,  H3_STB_Reference_Freq_800x600x56, FALSE },

    { 8,   800,  600, 60,  H3_STB_Reference_Freq_800x600x60, FALSE },
    { 16,  800,  600, 60,  H3_STB_Reference_Freq_800x600x60, FALSE },
//not supported    { 24,  800,  600, 60,  H3_STB_Reference_Freq_800x600x60, FALSE },
    { 32,  800,  600, 60,  H3_STB_Reference_Freq_800x600x60, FALSE },

    { 8,   800,  600, 72,  H3_STB_Reference_Freq_800x600x72, FALSE },
    { 16,  800,  600, 72,  H3_STB_Reference_Freq_800x600x72, FALSE },
//not supported    { 24,  800,  600, 72,  H3_STB_Reference_Freq_800x600x72, FALSE },
    { 32,  800,  600, 72,  H3_STB_Reference_Freq_800x600x72, FALSE },

    { 8,   800,  600, 75,  H3_STB_Reference_Freq_800x600x75, FALSE },
    { 16,  800,  600, 75,  H3_STB_Reference_Freq_800x600x75, FALSE },
//not supported    { 24,  800,  600, 75,  H3_STB_Reference_Freq_800x600x75, FALSE },
    { 32,  800,  600, 75,  H3_STB_Reference_Freq_800x600x75, FALSE },

    { 8,   800,  600, 85,  H3_STB_Reference_Freq_800x600x85, FALSE },
    { 16,  800,  600, 85,  H3_STB_Reference_Freq_800x600x85, FALSE },
//not supported    { 24,  800,  600, 85,  H3_STB_Reference_Freq_800x600x85, FALSE },
    { 32,  800,  600, 85,  H3_STB_Reference_Freq_800x600x85, FALSE },

    { 8,   800,  600, 100, H3_STB_Reference_Freq_800x600x100, FALSE },
    { 16,  800,  600, 100, H3_STB_Reference_Freq_800x600x100, FALSE },
//not supported    { 24,  800,  600, 100, H3_STB_Reference_Freq_800x600x100, FALSE },
    { 32,  800,  600, 100, H3_STB_Reference_Freq_800x600x100, FALSE },

    { 8,   800,  600, 120, H3_STB_Reference_Freq_800x600x120, FALSE },
    { 16,  800,  600, 120, H3_STB_Reference_Freq_800x600x120, FALSE },
//not supported    { 24,  800,  600, 120, H3_STB_Reference_Freq_800x600x120, FALSE },
    { 32,  800,  600, 120, H3_STB_Reference_Freq_800x600x120, FALSE },

    { 8,   800,  600, 140, H3_STB_Reference_Freq_800x600x140, FALSE },
    { 16,  800,  600, 140, H3_STB_Reference_Freq_800x600x140, FALSE },
//not supported    { 24,  800,  600, 140, H3_STB_Reference_Freq_800x600x140, FALSE },
    { 32,  800,  600, 140, H3_STB_Reference_Freq_800x600x140, FALSE },

    { 8,   800,  600, 160, H3_STB_Reference_Freq_800x600x160, FALSE },
    { 16,  800,  600, 160, H3_STB_Reference_Freq_800x600x160, FALSE },
//not supported    { 24,  800,  600, 160, H3_STB_Reference_Freq_800x600x160, FALSE },
    { 32,  800,  600, 160, H3_STB_Reference_Freq_800x600x160, FALSE },

    { 8,  960, 720, 60,  H3_STB_Reference_Freq_960x720x60, FALSE },
    { 16, 960, 720, 60,  H3_STB_Reference_Freq_960x720x60, FALSE },
//not supported    { 24, 960, 720, 60,  H3_STB_Reference_Freq_960x720x60, FALSE },
    { 32, 960, 720, 60,  H3_STB_Reference_Freq_960x720x60, FALSE },

    { 8,  960, 720, 75,  H3_STB_Reference_Freq_960x720x75, FALSE },
    { 16, 960, 720, 75,  H3_STB_Reference_Freq_960x720x75, FALSE },
//not supported    { 24, 960, 720, 75,  H3_STB_Reference_Freq_960x720x75, FALSE },
    { 32, 960, 720, 75,  H3_STB_Reference_Freq_960x720x75, FALSE },

    { 8,  960, 720, 85,  H3_STB_Reference_Freq_960x720x85, FALSE },
    { 16, 960, 720, 85,  H3_STB_Reference_Freq_960x720x85, FALSE },
//not supported    { 24, 960, 720, 85,  H3_STB_Reference_Freq_960x720x85, FALSE },
    { 32, 960, 720, 85,  H3_STB_Reference_Freq_960x720x85, FALSE },

    { 8,  1024,  768, 60,  H3_STB_Reference_Freq_1024x768x60, FALSE },
    { 16, 1024,  768, 60,  H3_STB_Reference_Freq_1024x768x60, FALSE },
//not supported    { 24, 1024,  768, 60,  H3_STB_Reference_Freq_1024x768x60, FALSE },
    { 32, 1024,  768, 60,  H3_STB_Reference_Freq_1024x768x60, FALSE },

    { 8,  1024,  768, 70,  H3_STB_Reference_Freq_1024x768x70, FALSE },
    { 16, 1024,  768, 70,  H3_STB_Reference_Freq_1024x768x70, FALSE },
//not supported    { 24, 1024,  768, 70,  H3_STB_Reference_Freq_1024x768x70, FALSE },
    { 32, 1024,  768, 70,  H3_STB_Reference_Freq_1024x768x70, FALSE },

    { 8,  1024,  768, 75,  H3_STB_Reference_Freq_1024x768x75, FALSE },
    { 16, 1024,  768, 75,  H3_STB_Reference_Freq_1024x768x75, FALSE },
//not supported    { 24, 1024,  768, 75,  H3_STB_Reference_Freq_1024x768x75, FALSE },
    { 32, 1024,  768, 75,  H3_STB_Reference_Freq_1024x768x75, FALSE },

    { 8,  1024,  768, 85,  H3_STB_Reference_Freq_1024x768x85, FALSE },
    { 16, 1024,  768, 85,  H3_STB_Reference_Freq_1024x768x85, FALSE },
//not supported    { 24, 1024,  768, 85,  H3_STB_Reference_Freq_1024x768x85, FALSE },
    { 32, 1024,  768, 85,  H3_STB_Reference_Freq_1024x768x85, FALSE },

    { 8,  1024,  768, 100, H3_STB_Reference_Freq_1024x768x100, FALSE },
    { 16, 1024,  768, 100, H3_STB_Reference_Freq_1024x768x100, FALSE },
//not supported    { 24, 1024,  768, 100, H3_STB_Reference_Freq_1024x768x100, FALSE },
    { 32, 1024,  768, 100, H3_STB_Reference_Freq_1024x768x100, FALSE },

    { 8,  1024,  768, 120, H3_STB_Reference_Freq_1024x768x120, FALSE },
    { 16, 1024,  768, 120, H3_STB_Reference_Freq_1024x768x120, FALSE },
//not supported    { 24, 1024,  768, 120, H3_STB_Reference_Freq_1024x768x120, FALSE },
    { 32, 1024,  768, 120, H3_STB_Reference_Freq_1024x768x120, FALSE },

    { 8,  1152,  864, 60,  H3_STB_Reference_Freq_1152x864x60, FALSE },
    { 16, 1152,  864, 60,  H3_STB_Reference_Freq_1152x864x60, FALSE },
//not supported    { 24, 1152,  864, 60,  H3_STB_Reference_Freq_1152x864x60, FALSE },
    { 32, 1152,  864, 60,  H3_STB_Reference_Freq_1152x864x60, FALSE },

    { 8,  1152,  864, 70,  H3_STB_Reference_Freq_1152x864x70, FALSE },
    { 16, 1152,  864, 70,  H3_STB_Reference_Freq_1152x864x70, FALSE },
//not supported    { 24, 1152,  864, 70,  H3_STB_Reference_Freq_1152x864x70, FALSE },
    { 32, 1152,  864, 70,  H3_STB_Reference_Freq_1152x864x70, FALSE },

    { 8,  1152,  864, 75,  H3_STB_Reference_Freq_1152x864x75, FALSE },
    { 16, 1152,  864, 75,  H3_STB_Reference_Freq_1152x864x75, FALSE },
//not supported    { 24, 1152,  864, 75,  H3_STB_Reference_Freq_1152x864x75, FALSE },
    { 32, 1152,  864, 75,  H3_STB_Reference_Freq_1152x864x75, FALSE },

    { 8,  1152,  864, 85,  H3_STB_Reference_Freq_1152x864x85, FALSE },
    { 16, 1152,  864, 85,  H3_STB_Reference_Freq_1152x864x85, FALSE },
//not supported    { 24, 1152,  864, 85,  H3_STB_Reference_Freq_1152x864x85, FALSE },
    { 32, 1152,  864, 85,  H3_STB_Reference_Freq_1152x864x85, FALSE },

    { 8,  1152,  864, 100, H3_STB_Reference_Freq_1152x864x100, FALSE },
    { 16, 1152,  864, 100, H3_STB_Reference_Freq_1152x864x100, FALSE },
//not supported    { 24, 1152,  864, 100, H3_STB_Reference_Freq_1152x864x100, FALSE },
    { 32, 1152,  864, 100, H3_STB_Reference_Freq_1152x864x100, FALSE },

    { 8,  1152,  864, 120, H3_STB_Reference_Freq_1152x864x120, FALSE },
    { 16, 1152,  864, 120, H3_STB_Reference_Freq_1152x864x120, FALSE },
//not supported    { 24, 1152,  864, 120, H3_STB_Reference_Freq_1152x864x120, FALSE },
    { 32, 1152,  864, 120, H3_STB_Reference_Freq_1152x864x120, FALSE },

    { 8,  1280, 960, 60,  H3_STB_Reference_Freq_1280x960x60, FALSE },
    { 16, 1280, 960, 60,  H3_STB_Reference_Freq_1280x960x60, FALSE },
//not supported    { 24, 1280, 960, 60,  H3_STB_Reference_Freq_1280x960x60, FALSE },
    { 32, 1280, 960, 60,  H3_STB_Reference_Freq_1280x960x60, FALSE },

    { 8,  1280, 960, 75,  H3_STB_Reference_Freq_1280x960x75, FALSE },
    { 16, 1280, 960, 75,  H3_STB_Reference_Freq_1280x960x75, FALSE },
//not supported    { 24, 1280, 960, 75,  H3_STB_Reference_Freq_1280x960x75, FALSE },
    { 32, 1280, 960, 75,  H3_STB_Reference_Freq_1280x960x75, FALSE },

    { 8,  1280, 960, 85,  H3_STB_Reference_Freq_1280x960x85, FALSE },
    { 16, 1280, 960, 85,  H3_STB_Reference_Freq_1280x960x85, FALSE },
//not supported    { 24, 1280, 960, 85,  H3_STB_Reference_Freq_1280x960x85, FALSE },
    { 32, 1280, 960, 85,  H3_STB_Reference_Freq_1280x960x85, FALSE },

    { 8,  1280, 1024, 60,  H3_STB_Reference_Freq_1280x1024x60, FALSE },
    { 16, 1280, 1024, 60,  H3_STB_Reference_Freq_1280x1024x60, FALSE },
//not supported    { 24, 1280, 1024, 60,  H3_STB_Reference_Freq_1280x1024x60, FALSE },
    { 32, 1280, 1024, 60,  H3_STB_Reference_Freq_1280x1024x60, FALSE },

    { 8,  1280, 1024, 75,  H3_STB_Reference_Freq_1280x1024x75, FALSE },
    { 16, 1280, 1024, 75,  H3_STB_Reference_Freq_1280x1024x75, FALSE },
//not supported    { 24, 1280, 1024, 75,  H3_STB_Reference_Freq_1280x1024x75, FALSE },
    { 32, 1280, 1024, 75,  H3_STB_Reference_Freq_1280x1024x75, FALSE },

    { 8,  1280, 1024, 85,  H3_STB_Reference_Freq_1280x1024x85, FALSE },
    { 16, 1280, 1024, 85,  H3_STB_Reference_Freq_1280x1024x85, FALSE },
//not supported    { 24, 1280, 1024, 85,  H3_STB_Reference_Freq_1280x1024x85, FALSE },
    { 32, 1280, 1024, 85,  H3_STB_Reference_Freq_1280x1024x85, FALSE },

    { 8,  1280, 1024, 100, H3_STB_Reference_Freq_1280x1024x100, FALSE },
    { 16, 1280, 1024, 100, H3_STB_Reference_Freq_1280x1024x100, FALSE },
//not supported    { 24, 1280, 1024, 100, H3_STB_Reference_Freq_1280x1024x100, FALSE },
    { 32, 1280, 1024, 100, H3_STB_Reference_Freq_1280x1024x100, FALSE },

    { 8,  1600, 1024, 60,  H3_STB_Reference_Freq_1600x1024x60, FALSE },
    { 16, 1600, 1024, 60,  H3_STB_Reference_Freq_1600x1024x60, FALSE },
//not supported    { 24, 1600, 1024, 60,  H3_STB_Reference_Freq_1600x1024x60, FALSE },
    { 32, 1600, 1024, 60,  H3_STB_Reference_Freq_1600x1024x60, FALSE },

    { 8,  1600, 1024, 76,  H3_STB_Reference_Freq_1600x1024x76, FALSE },
    { 16, 1600, 1024, 76,  H3_STB_Reference_Freq_1600x1024x76, FALSE },
//not supported    { 24, 1600, 1024, 76,  H3_STB_Reference_Freq_1600x1024x76, FALSE },
    { 32, 1600, 1024, 76,  H3_STB_Reference_Freq_1600x1024x76, FALSE },

    { 8,  1600, 1024, 85,  H3_STB_Reference_Freq_1600x1024x85, FALSE },
    { 16, 1600, 1024, 85,  H3_STB_Reference_Freq_1600x1024x85, FALSE },
//not supported    { 24, 1600, 1024, 85,  H3_STB_Reference_Freq_1600x1024x85, FALSE },
    { 32, 1600, 1024, 85,  H3_STB_Reference_Freq_1600x1024x85, FALSE },

    { 8,  1600, 1200, 60,  H3_STB_Reference_Freq_1600x1200x60, FALSE },
    { 16, 1600, 1200, 60,  H3_STB_Reference_Freq_1600x1200x60, FALSE },
//not supported    { 24, 1600, 1200, 60,  H3_STB_Reference_Freq_1600x1200x60, FALSE },
    { 32, 1600, 1200, 60,  H3_STB_Reference_Freq_1600x1200x60, FALSE },

    { 8,  1600, 1200, 65,  H3_STB_Reference_Freq_1600x1200x65, FALSE },
    { 16, 1600, 1200, 65,  H3_STB_Reference_Freq_1600x1200x65, FALSE },
//not supported    { 24, 1600, 1200, 65,  H3_STB_Reference_Freq_1600x1200x65, FALSE },
    { 32, 1600, 1200, 65,  H3_STB_Reference_Freq_1600x1200x65, FALSE },

    { 8,  1600, 1200, 70,  H3_STB_Reference_Freq_1600x1200x70, FALSE },
    { 16, 1600, 1200, 70,  H3_STB_Reference_Freq_1600x1200x70, FALSE },
//not supported    { 24, 1600, 1200, 70,  H3_STB_Reference_Freq_1600x1200x70, FALSE },
    { 32, 1600, 1200, 70,  H3_STB_Reference_Freq_1600x1200x70, FALSE },

    { 8,  1600, 1200, 75,  H3_STB_Reference_Freq_1600x1200x75, FALSE },
    { 16, 1600, 1200, 75,  H3_STB_Reference_Freq_1600x1200x75, FALSE },
//not supported    { 24, 1600, 1200, 75,  H3_STB_Reference_Freq_1600x1200x75, FALSE },
    { 32, 1600, 1200, 75,  H3_STB_Reference_Freq_1600x1200x75, FALSE },

    { 8,  1600, 1200, 80,  H3_STB_Reference_Freq_1600x1200x80, FALSE },
    { 16, 1600, 1200, 80,  H3_STB_Reference_Freq_1600x1200x80, FALSE },
//not supported    { 24, 1600, 1200, 80,  H3_STB_Reference_Freq_1600x1200x80, FALSE },
    { 32, 1600, 1200, 80,  H3_STB_Reference_Freq_1600x1200x80, FALSE },

    { 8,  1600, 1200, 85,  H3_STB_Reference_Freq_1600x1200x85, FALSE },
    { 16, 1600, 1200, 85,  H3_STB_Reference_Freq_1600x1200x85, FALSE },
//not supported    { 24, 1600, 1200, 85,  H3_STB_Reference_Freq_1600x1200x85, FALSE },
    { 32, 1600, 1200, 85,  H3_STB_Reference_Freq_1600x1200x85, FALSE },

    { 8,  1600, 1200, 100,  H3_STB_Reference_Freq_1600x1200x100, FALSE },
    { 16, 1600, 1200, 100,  H3_STB_Reference_Freq_1600x1200x100, FALSE },
//not supported    { 24, 1600, 1200, 100,  H3_STB_Reference_Freq_1600x1200x100, FALSE },
    { 32, 1600, 1200, 100,  H3_STB_Reference_Freq_1600x1200x100, FALSE },

    { 8,  1792, 1344, 60,  H3_STB_Reference_Freq_1792x1344x60, FALSE },
    { 16, 1792, 1344, 60,  H3_STB_Reference_Freq_1792x1344x60, FALSE },
//not supported    { 24, 1792, 1344, 60,  H3_STB_Reference_Freq_1792x1344x60, FALSE },
    { 32, 1792, 1344, 60,  H3_STB_Reference_Freq_1792x1344x60, FALSE },

    { 8,  1792, 1344, 75,  H3_STB_Reference_Freq_1792x1344x75, FALSE },
    { 16, 1792, 1344, 75,  H3_STB_Reference_Freq_1792x1344x75, FALSE },
//not supported    { 24, 1792, 1344, 75,  H3_STB_Reference_Freq_1792x1344x75, FALSE },
    { 32, 1792, 1344, 75,  H3_STB_Reference_Freq_1792x1344x75, FALSE },

    { 8,  1856, 1392, 60,  H3_STB_Reference_Freq_1856x1392x60, FALSE },
    { 16, 1856, 1392, 60,  H3_STB_Reference_Freq_1856x1392x60, FALSE },
//not supported    { 24, 1856, 1392, 60,  H3_STB_Reference_Freq_1856x1392x60, FALSE },
    { 32, 1856, 1392, 60,  H3_STB_Reference_Freq_1856x1392x60, FALSE },

    { 8,  1856, 1392, 75,  H3_STB_Reference_Freq_1856x1392x75, FALSE },
    { 16, 1856, 1392, 75,  H3_STB_Reference_Freq_1856x1392x75, FALSE },
//not supported    { 24, 1856, 1392, 75,  H3_STB_Reference_Freq_1856x1392x75, FALSE },
    { 32, 1856, 1392, 75,  H3_STB_Reference_Freq_1856x1392x75, FALSE },

    { 8,  1920, 1080, 60,  H3_STB_Reference_Freq_1920x1080x60, FALSE },
    { 16, 1920, 1080, 60,  H3_STB_Reference_Freq_1920x1080x60, FALSE },
//not supported    { 24, 1920, 1080, 60,  H3_STB_Reference_Freq_1920x1080x60, FALSE },
    { 32, 1920, 1080, 60,  H3_STB_Reference_Freq_1920x1080x60, FALSE },

    { 8,  1920, 1080, 72,  H3_STB_Reference_Freq_1920x1080x72, FALSE },
    { 16, 1920, 1080, 72,  H3_STB_Reference_Freq_1920x1080x72, FALSE },
//not supported    { 24, 1920, 1080, 72,  H3_STB_Reference_Freq_1920x1080x72, FALSE },
    { 32, 1920, 1080, 72,  H3_STB_Reference_Freq_1920x1080x72, FALSE },

    { 8,  1920, 1080, 75,  H3_STB_Reference_Freq_1920x1080x75, FALSE },
    { 16, 1920, 1080, 75,  H3_STB_Reference_Freq_1920x1080x75, FALSE },
//not supported    { 24, 1920, 1080, 75,  H3_STB_Reference_Freq_1920x1080x75, FALSE },
    { 32, 1920, 1080, 75,  H3_STB_Reference_Freq_1920x1080x75, FALSE },

    { 8,  1920, 1080, 85,  H3_STB_Reference_Freq_1920x1080x85, FALSE },
    { 16, 1920, 1080, 85,  H3_STB_Reference_Freq_1920x1080x85, FALSE },
//not supported    { 24, 1920, 1080, 85,  H3_STB_Reference_Freq_1920x1080x85, FALSE },
    { 32, 1920, 1080, 85,  H3_STB_Reference_Freq_1920x1080x85, FALSE },

    { 8,  1920, 1200, 60,  H3_STB_Reference_Freq_1920x1200x60, FALSE },
    { 16, 1920, 1200, 60,  H3_STB_Reference_Freq_1920x1200x60, FALSE },
//not supported    { 24, 1920, 1200, 60,  H3_STB_Reference_Freq_1920x1200x60, FALSE },
    { 32, 1920, 1200, 60,  H3_STB_Reference_Freq_1920x1200x60, FALSE },

    { 8,  1920, 1200, 76,  H3_STB_Reference_Freq_1920x1200x76, FALSE },
    { 16, 1920, 1200, 76,  H3_STB_Reference_Freq_1920x1200x76, FALSE },
//not supported    { 24, 1920, 1200, 76,  H3_STB_Reference_Freq_1920x1200x76, FALSE },
    { 32, 1920, 1200, 76,  H3_STB_Reference_Freq_1920x1200x76, FALSE },

    { 8,  1920, 1200, 85,  H3_STB_Reference_Freq_1920x1200x85, FALSE },
    { 16, 1920, 1200, 85,  H3_STB_Reference_Freq_1920x1200x85, FALSE },
//not supported    { 24, 1920, 1200, 85,  H3_STB_Reference_Freq_1920x1200x85, FALSE },
    { 32, 1920, 1200, 85,  H3_STB_Reference_Freq_1920x1200x85, FALSE },

    { 8,  1920, 1440, 60,  H3_STB_Reference_Freq_1920x1440x60, FALSE },
    { 16, 1920, 1440, 60,  H3_STB_Reference_Freq_1920x1440x60, FALSE },
//not supported    { 24, 1920, 1440, 60,  H3_STB_Reference_Freq_1920x1440x60, FALSE },
    { 32, 1920, 1440, 60,  H3_STB_Reference_Freq_1920x1440x60, FALSE },

    { 8,  1920, 1440, 75,  H3_STB_Reference_Freq_1920x1440x75, FALSE },
    { 16, 1920, 1440, 75,  H3_STB_Reference_Freq_1920x1440x75, FALSE },
//not supported    { 24, 1920, 1440, 75,  H3_STB_Reference_Freq_1920x1440x75, FALSE },
    { 32, 1920, 1440, 75,  H3_STB_Reference_Freq_1920x1440x75, FALSE },

    // Special case 2046 mode.
    // External to the miniport Windows believes this mode is 2046 pixels, internally it actually is
    //       2048 wide with 2 pixels left black.  This is a workaround for a Voodoo3 hardware limitation.
    { 8,  2048, 1536, 60,  H3_STB_Reference_Freq_2048x1536x60, FALSE },
    { 16, 2048, 1536, 60,  H3_STB_Reference_Freq_2048x1536x60, FALSE },
//not supported    { 24, 2048, 1536, 60,  H3_STB_Reference_Freq_2048x1536x60, FALSE },
//not supported    { 32, 2048, 1536, 60,  H3_STB_Reference_Freq_2048x1536x60, FALSE },

    { 8,  2048, 1536, 75,  H3_STB_Reference_Freq_2048x1536x75, FALSE },
    { 16, 2048, 1536, 75,  H3_STB_Reference_Freq_2048x1536x75, FALSE },
//not supported    { 24, 2048, 1536, 75,  H3_STB_Reference_Freq_2048x1536x75, FALSE },
//not supported    { 32, 2048, 1536, 75,  H3_STB_Reference_Freq_2048x1536x75, FALSE },

//not supported    { 8,  2048, 1536, 85,  H3_STB_Reference_Freq_2048x1536x85, FALSE },
//not supported    { 16, 2048, 1536, 85,  H3_STB_Reference_Freq_2048x1536x85, FALSE },
//not supported    { 24, 2048, 1536, 85,  H3_STB_Reference_Freq_2048x1536x85, FALSE },
//not supported    { 32, 2048, 1536, 85,  H3_STB_Reference_Freq_2048x1536x85, FALSE },

    //
    // From here on, we're either in video or game modes, not
    // desktop modes...
    //
    // GAME MODES:
    //
    // We don't support 8bpp modes at this time, but we will soon.
    //
    {  8, 320,  200, 70,  H3_STB_Reference_Freq_320x200x70, FALSE },
    { 16, 320,  200, 70,  H3_STB_Reference_Freq_320x200x70, FALSE },
//not supported    { 24, 320,  200, 70,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 320,  200, 70,  H3_STB_Reference_Freq_320x200x70, FALSE },

    {  8, 320,  200, 85,  H3_Reference_Freq_320x200x85, FALSE },
    { 16, 320,  200, 85,  H3_Reference_Freq_320x200x85, FALSE },
//not supported    { 24, 320,  200, 85,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 320,  200, 85,  H3_Reference_Freq_320x200x85, FALSE },

    {  8, 320,  240, 60,  H3_Reference_Freq_320x240x60, FALSE },
    { 16, 320,  240, 60,  H3_Reference_Freq_320x240x60, FALSE },
//not supported    { 24, 320,  240, 60,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 320,  240, 60,  H3_Reference_Freq_320x240x60, FALSE },

    {  8, 320,  240, 72,  H3_Reference_Freq_320x240x72, FALSE },
    { 16, 320,  240, 72,  H3_Reference_Freq_320x240x72, FALSE },
//not supported    { 24, 320,  240, 72,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 320,  240, 72,  H3_Reference_Freq_320x240x72, FALSE },

    {  8, 320,  240, 75,  H3_Reference_Freq_320x240x75, FALSE },
    { 16, 320,  240, 75,  H3_Reference_Freq_320x240x75, FALSE },
//not supported    { 24, 320,  240, 75,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 320,  240, 75,  H3_Reference_Freq_320x240x75, FALSE },

    {  8, 320,  240, 85,  H3_Reference_Freq_320x240x85, FALSE },
    { 16, 320,  240, 85,  H3_Reference_Freq_320x240x85, FALSE },
//not supported    { 24, 320,  240, 85,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 320,  240, 85,  H3_Reference_Freq_320x240x85, FALSE },

    {  8, 400,  300, 60,  H3_Reference_Freq_400x300x60, FALSE },
    { 16, 400,  300, 60,  H3_Reference_Freq_400x300x60, FALSE },
//not supported    { 24, 400,  300, 60,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 400,  300, 60,  H3_Reference_Freq_400x300x60, FALSE },

    {  8, 400,  300, 72,  H3_Reference_Freq_400x300x72, FALSE },
    { 16, 400,  300, 72,  H3_Reference_Freq_400x300x72, FALSE },
//not supported    { 24, 400,  300, 72,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 400,  300, 72,  H3_Reference_Freq_400x300x72, FALSE },

    {  8, 400,  300, 75,  H3_Reference_Freq_400x300x75, FALSE },
    { 16, 400,  300, 75,  H3_Reference_Freq_400x300x75, FALSE },
//not supported    { 24, 400,  300, 75,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 400,  300, 75,  H3_Reference_Freq_400x300x75, FALSE },

    {  8, 400,  300, 85,  H3_Reference_Freq_400x300x85, FALSE },
    { 16, 400,  300, 85,  H3_Reference_Freq_400x300x85, FALSE },
//not supported    { 24, 400,  300, 85,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 400,  300, 85,  H3_Reference_Freq_400x300x85, FALSE },

    {  8, 512,  384, 60,  H3_Reference_Freq_512x384x60, FALSE },
    { 16, 512,  384, 60,  H3_Reference_Freq_512x384x60, FALSE },
//not supported    { 24, 512,  384, 60,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 512,  384, 60,  H3_Reference_Freq_512x384x60, FALSE },

    {  8, 512,  384, 70,  H3_STB_Reference_Freq_512x384x70, FALSE },
    { 16, 512,  384, 70,  H3_STB_Reference_Freq_512x384x70, FALSE },
//not supported    { 24, 512,  384, 70,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 512,  384, 70,  H3_STB_Reference_Freq_512x384x70, FALSE },

// add 512x384@72 to meet Gateway Spec.
//not supported    {  8, 512,  384, 72,  H3_Reference_Freq_512x384x72, FALSE },
//not supported    { 16, 512,  384, 72,  H3_Reference_Freq_512x384x72, FALSE },

    {  8, 512,  384, 75,  H3_Reference_Freq_512x384x75, FALSE },
    { 16, 512,  384, 75,  H3_Reference_Freq_512x384x75, FALSE },
//not supported    { 24, 512,  384, 75,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 512,  384, 75,  H3_Reference_Freq_512x384x75, FALSE },

    {  8, 512,  384, 85,  H3_Reference_Freq_512x384x85, FALSE },
    { 16, 512,  384, 85,  H3_Reference_Freq_512x384x85, FALSE },
//not supported    { 24, 512,  384, 85,  H3_Frequency_Not_Available, FALSE },	// no 24 bit mode
    { 32, 512,  384, 85,  H3_Reference_Freq_512x384x85, FALSE },

    {  8, 640,  400, 70,  H3_STB_Reference_Freq_640x400x70, FALSE },
    { 16, 640,  400, 70,  H3_STB_Reference_Freq_640x400x70, FALSE },
//not supported    { 24, 640,  400, 70,  H3_STB_Reference_Freq_640x400x70, FALSE },	// no 24 bit mode
    { 32, 640,  400, 70,  H3_STB_Reference_Freq_640x400x70, FALSE },

    {  8, 640,  400, 85,  H3_STB_Reference_Freq_640x400x85, FALSE },
    { 16, 640,  400, 85,  H3_STB_Reference_Freq_640x400x85, FALSE },
//not supported    { 24, 640,  400, 85,  H3_STB_Reference_Freq_640x400x85, FALSE },	// no 24 bit mode
    { 32, 640,  400, 85,  H3_STB_Reference_Freq_640x400x85, FALSE },

	//
	// These are the "funny" modes required for stuff like
	// video
	//
#if VIDEO_720_MODES_SUPPORTED
//    { 8,  720,  400, 60, H3_Reference_Freq_720x400x60, FALSE },
//    { 16, 720,  400, 60, H3_Reference_Freq_720x400x60, FALSE },
//    { 24, 720,  400, 60, H3_Reference_Freq_720x400x60, FALSE },
//    { 32, 720,  400, 60, H3_Reference_Freq_720x400x60, FALSE },

    { 8,  720,  400, 72, H3_Reference_Freq_720x400x72, FALSE },
    { 16, 720,  400, 72, H3_Reference_Freq_720x400x72, FALSE },
    { 24, 720,  400, 72, H3_Reference_Freq_720x400x72, FALSE },
    { 32, 720,  400, 72, H3_Reference_Freq_720x400x72, FALSE },
#endif
    { 0 }   // Mark the end
};

#include "h4plltable.h"

#if defined(ALLOC_PRAGMA)
#pragma data_seg()
#endif
