/******************************Module*Header*******************************\
*
* Module Name: dx95type.h
*
* Portability module for NT kernel-mode DirectX components created from
* Win95 sources
*
* Copyright (c) 1996-1997 Microsoft Corp.
* Copyright (c) 1998 3Dfx Interactive, Inc.
*
\**************************************************************************/

#ifndef __DX95TYPE_H__
#define __DX95TYPE_H__

// Basic includes
#ifndef NODXINCLUDES
#ifdef WINNT
#include <stddef.h>
// Includes ddrawint.h and d3dnthal.h
#include <winddi.h>
#else // WINNT
#include <ddrawi.h>
#include <d3dhal.h>
#endif // WINNT
#endif // NODXINCLUDES

// Simple memory allocation of N bytes
#ifndef NODXMALLOC
#ifdef WINNT
#define DXMALLOC(n) EngAllocMem(0, n, '3HxD')
#define DXMALLOCZ(n) EngAllocMem(FL_ZERO_MEMORY, n, '3HxD')
#define DXFREE(p) EngFreeMem(p)
#else // WINNT
#define DXMALLOC(n)   HeapAlloc(GLOBAL_MEM_HEAP_PTR, 0, n)
#define DXMALLOCZ(n)  HeapAlloc(GLOBAL_MEM_HEAP_PTR, HEAP_ZERO_MEMORY, n)
#define DXFREE(ptr)   HeapFree(GLOBAL_MEM_HEAP_PTR, 0, (void*) ptr)
#endif // WINNT
#endif // NODXMALLOC

// Surface and palette access macros
#ifndef NODXLCL
#ifdef WINNT
#define DDS_LCL(pdds) (pdds)
#define DDP_LCL(pddp) (pddp)
#else // WINNT
#define DDS_LCL(pdds) ((DDRAWI_DDRAWSURFACE_INT *)(pdds)->lpLcl)
#define DDP_LCL(pddp) ((DDRAWI_DDRAWPALETTE_INT *)(pddp)->lpLcl)
#endif // WINNT
#define DDS_LCL_OR_NULL(pdds) ((pdds) == NULL ? NULL : DDS_LCL(pdds))
#define DDP_LCL_OR_NULL(pddp) ((pddp) == NULL ? NULL : DDP_LCL(pddp))
#endif // NODXLCL

// Typedefs and defines to create a Win95-like compilation environment

#ifndef NODX95TYPES
#ifdef WINNT

#if _WIN32_WINNT >= 0x0500
typedef D3DNTHALDEVICEDESC_V1           D3DDEVICEDESC_V1;
#endif
typedef PDD_SURFACE_LOCAL               LPDDRAWI_DDRAWSURFACE_LCL;
#if _WIN32_WINNT >= 0x0500
typedef PDD_SURFACE_MORE                LPDDRAWI_DDRAWSURFACE_MORE;
#endif
typedef PDD_SURFACE_GLOBAL              LPDDRAWI_DDRAWSURFACE_GBL;
typedef PDD_DIRECTDRAW_GLOBAL           LPDDRAWI_DIRECTDRAW_GBL;
#if _WIN32_WINNT >= 0x0500
typedef PDD_ATTACHLIST                  LPATTACHLIST;
#endif

#if _WIN32_WINNT >= 0x0500
typedef DDNTCORECAPS                    DDCORECAPS;
typedef PDDNTCORECAPS                   LPDDCORECAPS;
#endif
typedef DD_HALINFO                      DDHALINFO;
typedef PDD_HALINFO                     LPDDHALINFO;
#if _WIN32_WINNT >= 0x0500
typedef DD_MORECAPS                     DDMORECAPS;
typedef PDD_MORECAPS                    LPDDMORECAPS;
#endif

// DirectDraw HAL function data structures

typedef PDD_CREATEPALETTEDATA           LPDDHAL_CREATEPALETTEDATA;
typedef PDD_CREATESURFACEDATA           LPDDHAL_CREATESURFACEDATA;
typedef PDD_CANCREATESURFACEDATA        LPDDHAL_CANCREATESURFACEDATA;
typedef PDD_WAITFORVERTICALBLANKDATA    LPDDHAL_WAITFORVERTICALBLANKDATA;
typedef PDD_DESTROYDRIVERDATA           LPDDHAL_DESTROYDRIVERDATA;
typedef PDD_SETMODEDATA                 LPDDHAL_SETMODEDATA;
typedef PDD_DRVSETCOLORKEYDATA          LPDDHAL_DRVSETCOLORKEYDATA;
typedef PDD_GETSCANLINEDATA             LPDDHAL_GETSCANLINEDATA;
typedef PDD_MAPMEMORYDATA               LPDDHAL_MAPMEMORYDATA;
typedef DD_CREATEPALETTEDATA            DDHAL_CREATEPALETTEDATA;
typedef DD_CREATESURFACEDATA            DDHAL_CREATESURFACEDATA;
typedef DD_CANCREATESURFACEDATA         DDHAL_CANCREATESURFACEDATA;
typedef DD_WAITFORVERTICALBLANKDATA     DDHAL_WAITFORVERTICALBLANKDATA;
typedef DD_DRVSETCOLORKEYDATA           DDHAL_DRVSETCOLORKEYDATA;
typedef DD_GETSCANLINEDATA              DDHAL_GETSCANLINEDATA;
typedef DD_MAPMEMORYDATA                DDHAL_MAPMEMORYDATA;

typedef PDD_DESTROYPALETTEDATA          LPDDHAL_DESTROYPALETTEDATA;
typedef PDD_SETENTRIESDATA              LPDDHAL_SETENTRIESDATA;
typedef DD_DESTROYPALETTEDATA           DDHAL_DESTROYPALETTEDATA;
typedef DD_SETENTRIESDATA               DDHAL_SETENTRIESDATA;

typedef PDD_BLTDATA                     LPDDHAL_BLTDATA;
typedef PDD_LOCKDATA                    LPDDHAL_LOCKDATA;
typedef PDD_UNLOCKDATA                  LPDDHAL_UNLOCKDATA;
typedef PDD_UPDATEOVERLAYDATA           LPDDHAL_UPDATEOVERLAYDATA;
typedef PDD_SETOVERLAYPOSITIONDATA      LPDDHAL_SETOVERLAYPOSITIONDATA;
typedef PDD_SETPALETTEDATA              LPDDHAL_SETPALETTEDATA;
typedef PDD_FLIPDATA                    LPDDHAL_FLIPDATA;
typedef PDD_DESTROYSURFACEDATA          LPDDHAL_DESTROYSURFACEDATA;
typedef PDD_SETCLIPLISTDATA             LPDDHAL_SETCLIPLISTDATA;
typedef PDD_ADDATTACHEDSURFACEDATA      LPDDHAL_ADDATTACHEDSURFACEDATA;
typedef PDD_SETCOLORKEYDATA             LPDDHAL_SETCOLORKEYDATA;
typedef PDD_GETBLTSTATUSDATA            LPDDHAL_GETBLTSTATUSDATA;
typedef PDD_GETFLIPSTATUSDATA           LPDDHAL_GETFLIPSTATUSDATA;
#if _WIN32_WINNT >= 0x0500
typedef PDD_GETAVAILDRIVERMEMORYDATA    LPDDHAL_GETAVAILDRIVERMEMORYDATA;
typedef PDD_GETDRIVERINFODATA           LPDDHAL_GETDRIVERINFODATA;
#endif
typedef DD_BLTDATA                      DDHAL_BLTDATA;
typedef DD_LOCKDATA                     DDHAL_LOCKDATA;
typedef DD_UNLOCKDATA                   DDHAL_UNLOCKDATA;
typedef DD_UPDATEOVERLAYDATA            DDHAL_UPDATEOVERLAYDATA;
typedef DD_SETOVERLAYPOSITIONDATA       DDHAL_SETOVERLAYPOSITIONDATA;
typedef DD_SETPALETTEDATA               DDHAL_SETPALETTEDATA;
typedef DD_FLIPDATA                     DDHAL_FLIPDATA;
typedef DD_DESTROYSURFACEDATA           DDHAL_DESTROYSURFACEDATA;
typedef DD_SETCLIPLISTDATA              DDHAL_SETCLIPLISTDATA;
typedef DD_ADDATTACHEDSURFACEDATA       DDHAL_ADDATTACHEDSURFACEDATA;
typedef DD_SETCOLORKEYDATA              DDHAL_SETCOLORKEYDATA;
typedef DD_GETBLTSTATUSDATA             DDHAL_GETBLTSTATUSDATA;
typedef DD_GETFLIPSTATUSDATA            DDHAL_GETFLIPSTATUSDATA;
#if _WIN32_WINNT >= 0x0500
typedef DD_GETAVAILDRIVERMEMORYDATA     DDHAL_GETAVAILDRIVERMEMORYDATA;
typedef DD_GETDRIVERINFODATA            DDHAL_GETDRIVERINFODATA;

//check this -> typedef PDD_ALPHABLTDATA                LPDDHAL_ALPHABLTDATA;
//check this -> typedef PDD_SETSPRITEDISPLAYLISTDATA    LPDDHAL_SETSPRITEDISPLAYLISTDATA;
typedef PDD_RESIZEDATA                  LPDDHAL_RESIZEDATA;
typedef PDD_CAPTURECOMPOSITIONBUFFERDATA LPDDHAL_CAPTURECOMPOSITIONBUFFERDATA;
//check this -> typedef DD_ALPHABLTDATA                 DDHAL_ALPHABLTDATA;
typedef DD_SETSPRITEDISPLAYLISTDATA     DDHAL_SETSPRITEDISPLAYLISTDATA;
typedef DD_RESIZEDATA                   DDHAL_RESIZEDATA;
typedef DD_CAPTURECOMPOSITIONBUFFERDATA DDHAL_CAPTURECOMPOSITIONBUFFERDATA;


typedef PDD_CANCREATEVPORTDATA          LPDDHAL_CANCREATEVPORTDATA;
typedef PDD_CREATEVPORTDATA             LPDDHAL_CREATEVPORTDATA;
typedef PDD_FLIPVPORTDATA               LPDDHAL_FLIPVPORTDATA;
typedef PDD_GETVPORTCONNECTDATA         LPDDHAL_GETVPORTCONNECTDATA;
typedef PDD_GETVPORTBANDWIDTHDATA       LPDDHAL_GETVPORTBANDWIDTHDATA;
typedef PDD_GETVPORTINPUTFORMATDATA     LPDDHAL_GETVPORTINPUTFORMATDATA;
typedef PDD_GETVPORTOUTPUTFORMATDATA    LPDDHAL_GETVPORTOUTPUTFORMATDATA;
typedef PDD_GETVPORTFIELDDATA           LPDDHAL_GETVPORTFIELDDATA;
typedef PDD_GETVPORTLINEDATA            LPDDHAL_GETVPORTLINEDATA;
typedef PDD_DESTROYVPORTDATA            LPDDHAL_DESTROYVPORTDATA;
typedef PDD_GETVPORTFLIPSTATUSDATA      LPDDHAL_GETVPORTFLIPSTATUSDATA;
typedef PDD_UPDATEVPORTDATA             LPDDHAL_UPDATEVPORTDATA;
typedef PDD_WAITFORVPORTSYNCDATA        LPDDHAL_WAITFORVPORTSYNCDATA;
typedef PDD_GETVPORTSIGNALDATA          LPDDHAL_GETVPORTSIGNALDATA;
typedef PDD_VPORTCOLORDATA              LPDDHAL_VPORTCOLORDATA;
typedef DD_CANCREATEVPORTDATA           DDHAL_CANCREATEVPORTDATA;
typedef DD_CREATEVPORTDATA              DDHAL_CREATEVPORTDATA;
typedef DD_FLIPVPORTDATA                DDHAL_FLIPVPORTDATA;
typedef DD_GETVPORTCONNECTDATA          DDHAL_GETVPORTCONNECTDATA;
typedef DD_GETVPORTBANDWIDTHDATA        DDHAL_GETVPORTBANDWIDTHDATA;
typedef DD_GETVPORTINPUTFORMATDATA      DDHAL_GETVPORTINPUTFORMATDATA;
typedef DD_GETVPORTOUTPUTFORMATDATA     DDHAL_GETVPORTOUTPUTFORMATDATA;
typedef DD_GETVPORTFIELDDATA            DDHAL_GETVPORTFIELDDATA;
typedef DD_GETVPORTLINEDATA             DDHAL_GETVPORTLINEDATA;
typedef DD_DESTROYVPORTDATA             DDHAL_DESTROYVPORTDATA;
typedef DD_GETVPORTFLIPSTATUSDATA       DDHAL_GETVPORTFLIPSTATUSDATA;
typedef DD_UPDATEVPORTDATA              DDHAL_UPDATEVPORTDATA;
typedef DD_WAITFORVPORTSYNCDATA         DDHAL_WAITFORVPORTSYNCDATA;
typedef DD_GETVPORTSIGNALDATA           DDHAL_GETVPORTSIGNALDATA;
typedef DD_VPORTCOLORDATA               DDHAL_VPORTCOLORDATA;

typedef PDD_COLORCONTROLDATA            LPDDHAL_COLORCONTROLDATA;
typedef DD_COLORCONTROLDATA             DDHAL_COLORCONTROLDATA;

typedef PDD_SYNCSURFACEDATA             LPDDHAL_SYNCSURFACEDATA;
typedef PDD_SYNCVIDEOPORTDATA           LPDDHAL_SYNCVIDEOPORTDATA;
typedef DD_SYNCSURFACEDATA              DDHAL_SYNCSURFACEDATA;
typedef DD_SYNCVIDEOPORTDATA            DDHAL_SYNCVIDEOPORTDATA;

//check this -> typedef PDD_GETVIDEOGUIDSDATA           LPDDHAL_GETVIDEOGUIDSDATA;
//check this -> typedef PDD_GETVIDEOFORMATSDATA         LPDDHAL_GETVIDEOFORMATSDATA;
//check this -> typedef PDD_CREATEVIDEODATA             LPDDHAL_CREATEVIDEODATA;
//check this -> typedef PDD_BEGINVIDEOFRAMEDATA         LPDDHAL_BEGINVIDEOFRAMEDATA;
//check this -> typedef PDD_ENDVIDEOFRAMEDATA           LPDDHAL_ENDVIDEOFRAMEDATA;
//check this -> typedef PDD_EXECUTEVIDEODATA            LPDDHAL_EXECUTEVIDEODATA;
//check this -> typedef PDD_QUERYVIDEOSTATUSDATA        LPDDHAL_QUERYVIDEOSTATUSDATA;
//check this -> typedef PDD_DESTROYVIDEODATA            LPDDHAL_DESTROYVIDEODATA;
//check this -> typedef DD_GETVIDEOGUIDSDATA            DDHAL_GETVIDEOGUIDSDATA;
//check this -> typedef DD_GETVIDEOFORMATSDATA          DDHAL_GETVIDEOFORMATSDATA;
//check this -> typedef DD_CREATEVIDEODATA              DDHAL_CREATEVIDEODATA;
//check this -> typedef DD_BEGINVIDEOFRAMEDATA          DDHAL_BEGINVIDEOFRAMEDATA;
//check this -> typedef DD_ENDVIDEOFRAMEDATA            DDHAL_ENDVIDEOFRAMEDATA;
//check this -> typedef DD_EXECUTEVIDEODATA             DDHAL_EXECUTEVIDEODATA;
//check this -> typedef DD_QUERYVIDEOSTATUSDATA         DDHAL_QUERYVIDEOSTATUSDATA;
//check this -> typedef DD_DESTROYVIDEODATA             DDHAL_DESTROYVIDEODATA;
#endif // _WIN32_WINNT >= 0x0500

// DirectDraw HAL call backs

typedef PDD_SETCOLORKEY                 LPDDHAL_SETCOLORKEY;
typedef PDD_CANCREATESURFACE            LPDDHAL_CANCREATESURFACE;
typedef PDD_WAITFORVERTICALBLANK        LPDDHAL_WAITFORVERTICALBLANK;
typedef PDD_CREATESURFACE               LPDDHAL_CREATESURFACE;
typedef PDD_DESTROYDRIVER               LPDDHAL_DESTROYDRIVER;
typedef PDD_SETMODE                     LPDDHAL_SETMODE;
typedef PDD_CREATEPALETTE               LPDDHAL_CREATEPALETTE;
typedef PDD_GETSCANLINE                 LPDDHAL_GETSCANLINE;
typedef PDD_MAPMEMORY                   LPDDHAL_MAPMEMORY;
#if _WIN32_WINNT >= 0x0500
typedef PDD_GETAVAILDRIVERMEMORY        LPDDHAL_GETAVAILDRIVERMEMORY;
typedef PDD_GETDRIVERINFODATA           LPDDHAL_GETDRIVERINFODATA;
typedef PDD_SYNCSURFACEDATA             LPDDHAL_SYNCSURFACEDATA;
typedef PDD_SYNCVIDEOPORTDATA           LPDDHAL_SYNCVIDEOPORTDATA;
#endif

typedef PDD_PALCB_DESTROYPALETTE        LPDDHALPALCB_DESTROYPALETTE;
typedef PDD_PALCB_SETENTRIES            LPDDHALPALCB_SETENTRIES;

typedef PDD_SURFCB_LOCK                 LPDDHALSURFCB_LOCK;
typedef PDD_SURFCB_UNLOCK               LPDDHALSURFCB_UNLOCK;
typedef PDD_SURFCB_BLT                  LPDDHALSURFCB_BLT;
typedef PDD_SURFCB_UPDATEOVERLAY        LPDDHALSURFCB_UPDATEOVERLAY;
typedef PDD_SURFCB_SETOVERLAYPOSITION   LPDDHALSURFCB_SETOVERLAYPOSITION;
typedef PDD_SURFCB_SETPALETTE           LPDDHALSURFCB_SETPALETTE;
typedef PDD_SURFCB_FLIP                 LPDDHALSURFCB_FLIP;
typedef PDD_SURFCB_DESTROYSURFACE       LPDDHALSURFCB_DESTROYSURFACE;
typedef PDD_SURFCB_SETCLIPLIST          LPDDHALSURFCB_SETCLIPLIST;
typedef PDD_SURFCB_ADDATTACHEDSURFACE   LPDDHALSURFCB_ADDATTACHEDSURFACE;
typedef PDD_SURFCB_SETCOLORKEY          LPDDHALSURFCB_SETCOLORKEY;
typedef PDD_SURFCB_GETBLTSTATUS         LPDDHALSURFCB_GETBLTSTATUS;
typedef PDD_SURFCB_GETFLIPSTATUS        LPDDHALSURFCB_GETFLIPSTATUS;

#if _WIN32_WINNT >= 0x0500
typedef PDD_VPORTCB_CANCREATEVIDEOPORT  LPDDHALVPORTCB_CANCREATEVIDEOPORT;
typedef PDD_VPORTCB_CREATEVIDEOPORT     LPDDHALVPORTCB_CREATEVIDEOPORT;
typedef PDD_VPORTCB_FLIP                LPDDHALVPORTCB_FLIP;
typedef PDD_VPORTCB_GETBANDWIDTH        LPDDHALVPORTCB_GETBANDWIDTH;
typedef PDD_VPORTCB_GETINPUTFORMATS     LPDDHALVPORTCB_GETINPUTFORMATS;
typedef PDD_VPORTCB_GETOUTPUTFORMATS    LPDDHALVPORTCB_GETOUTPUTFORMATS;
typedef PDD_VPORTCB_GETFIELD            LPDDHALVPORTCB_GETFIELD;
typedef PDD_VPORTCB_GETLINE             LPDDHALVPORTCB_GETLINE;
typedef PDD_VPORTCB_GETVPORTCONNECT     LPDDHALVPORTCB_GETVPORTCONNECT;
typedef PDD_VPORTCB_DESTROYVPORT        LPDDHALVPORTCB_DESTROYVPORT;
typedef PDD_VPORTCB_GETFLIPSTATUS       LPDDHALVPORTCB_GETFLIPSTATUS;
typedef PDD_VPORTCB_UPDATE              LPDDHALVPORTCB_UPDATE;
typedef PDD_VPORTCB_WAITFORSYNC         LPDDHALVPORTCB_WAITFORSYNC;
typedef PDD_VPORTCB_GETSIGNALSTATUS     LPDDHALVPORTCB_GETSIGNALSTATUS;
typedef PDD_VPORTCB_COLORCONTROL        LPDDHALVPORTCB_COLORCONTROL;

//check this -> typedef PDD_VIDEOCB_GETGUIDS            LPDDHALVIDEOCB_GETGUIDS;
//check this -> typedef PDD_VIDEOCB_GETFORMATS          LPDDHALVIDEOCB_GETFORMATS;
//check this -> typedef PDD_VIDEOCB_CREATE              LPDDHALVIDEOCB_CREATE;
//check this -> typedef PDD_VIDEOCB_BEGINFRAME          LPDDHALVIDEOCB_BEGINFRAME;
//check this -> typedef PDD_VIDEOCB_ENDFRAME            LPDDHALVIDEOCB_ENDFRAME;
//check this -> typedef PDD_VIDEOCB_EXECUTE             LPDDHALVIDEOCB_EXECUTE;
//check this -> typedef PDD_VIDEOCB_QUERYSTATUS         LPDDHALVIDEOCB_QUERYSTATUS;
//check this -> typedef PDD_VIDEOCB_DESTROY             LPDDHALVIDEOCB_DESTROY;
#endif // _WIN32_WINNT >= 0x0500

typedef DD_CALLBACKS                    DDHAL_DDCALLBACKS;
typedef DD_PALETTECALLBACKS             DDHAL_DDPALETTECALLBACKS;
typedef DD_SURFACECALLBACKS             DDHAL_DDSURFACECALLBACKS;
#if _WIN32_WINNT >= 0x0500
typedef DD_VIDEOPORTCALLBACKS           DDHAL_DDVIDEOPORTCALLBACKS;
typedef DD_COLORCONTROLCALLBACKS        DDHAL_DDCOLORCONTROLCALLBACKS;
typedef DD_KERNELCALLBACKS              DDHAL_DDKERNELCALLBACKS;
//check this -> typedef DD_VIDEOCALLBACKS               DDHAL_DDVIDEOCALLBACKS;
typedef DD_MISCELLANEOUS2CALLBACKS      DDHAL_DDMISCELLANEOUS2CALLBACKS;
#endif

typedef PDD_CALLBACKS                   LPDDHAL_DDCALLBACKS;
typedef PDD_PALETTECALLBACKS            LPDDHAL_DDPALETTECALLBACKS;
typedef PDD_SURFACECALLBACKS            LPDDHAL_DDSURFACECALLBACKS;
#if _WIN32_WINNT >= 0x0500
typedef PDD_VIDEOPORTCALLBACKS          LPDDHAL_DDVIDEOPORTCALLBACKS;
typedef PDD_COLORCONTROLCALLBACKS       LPDDHAL_DDCOLORCONTROLCALLBACKS;
typedef PDD_KERNELCALLBACKS             LPDDHAL_DDKERNELCALLBACKS;
//check this -> typedef PDD_VIDEOCALLBACKS              LPDDHAL_DDVIDEOCALLBACKS;
typedef PDD_MISCELLANEOUS2CALLBACKS     LPDDHAL_DDMISCELLANEOUS2CALLBACKS;


// NOTE - Not supported by NT in the kernel

typedef struct _DDRAWI_DDRAWPALETTE_GBL
{
    void *lpColorTable;
} DDRAWI_DDRAWPALETTE_GBL;
typedef DDRAWI_DDRAWPALETTE_GBL *LPDDRAWI_DDRAWPALETTE_GBL;

typedef struct _DDRAWI_DDRAWPALETTE_LCL
{
    LPDDRAWI_DDRAWPALETTE_GBL lpGbl;
} DDRAWI_DDRAWPALETTE_LCL;
typedef DDRAWI_DDRAWPALETTE_LCL *LPDDRAWI_DDRAWPALETTE_LCL;

// NOTE - Not supported by NT in the kernel

typedef struct _DDHAL_DDEXEBUFCALLBACKS
{
    DWORD dwSize;
    DWORD dwFlags;
    void *CanCreateExecuteBuffer;
    void *CreateExecuteBuffer;
    void *DestroyExecuteBuffer;
    void *LockExecuteBuffer;
    void *UnlockExecuteBuffer;
} DDHAL_DDEXEBUFCALLBACKS;

typedef DDHAL_DDEXEBUFCALLBACKS        *LPDDHAL_DDEXEBUFCALLBACKS;

typedef D3DNTHAL_CALLBACKS              D3DHAL_CALLBACKS;
typedef D3DNTHAL_GLOBALDRIVERDATA       D3DHAL_GLOBALDRIVERDATA;
typedef D3DNTHAL_CONTEXTCREATEDATA      D3DHAL_CONTEXTCREATEDATA;
typedef D3DNTHAL_CONTEXTDESTROYDATA     D3DHAL_CONTEXTDESTROYDATA;
typedef D3DNTHAL_CONTEXTDESTROYALLDATA  D3DHAL_CONTEXTDESTROYALLDATA;
//check this -> typedef D3DNTHAL_RENDERSTATEDATA        D3DHAL_RENDERSTATEDATA;
//check this -> typedef D3DNTHAL_RENDERPRIMITIVEDATA    D3DHAL_RENDERPRIMITIVEDATA;
//check this -> typedef D3DNTHAL_GETSTATEDATA           D3DHAL_GETSTATEDATA;
typedef D3DNTHAL_SCENECAPTUREDATA       D3DHAL_SCENECAPTUREDATA;
//check this -> typedef D3DNTHAL_EXECUTEDATA            D3DHAL_EXECUTEDATA;
typedef D3DNTHAL_TEXTURECREATEDATA      D3DHAL_TEXTURECREATEDATA;
typedef D3DNTHAL_TEXTUREDESTROYDATA     D3DHAL_TEXTUREDESTROYDATA;
typedef D3DNTHAL_TEXTURESWAPDATA        D3DHAL_TEXTURESWAPDATA;
typedef D3DNTHAL_TEXTUREGETSURFDATA     D3DHAL_TEXTUREGETSURFDATA;
typedef D3DNTHAL_SETRENDERTARGETDATA    D3DHAL_SETRENDERTARGETDATA;
//check this -> typedef D3DNTHAL_CLEARDATA              D3DHAL_CLEARDATA;
typedef D3DNTHAL_CLEAR2DATA             D3DHAL_CLEAR2DATA;
//check this -> typedef D3DNTHAL_SETVIEWPORT2DATA       D3DHAL_SETVIEWPORT2DATA;
//check this -> typedef D3DNTHAL_DRAWONEPRIMITIVEDATA   D3DHAL_DRAWONEPRIMITIVEDATA;
//check this -> typedef D3DNTHAL_DRAWONEINDEXEDPRIMITIVEDATA D3DHAL_DRAWONEINDEXEDPRIMITIVEDATA;
//check this -> typedef D3DNTHAL_DRAWPRIMCOUNTS         D3DHAL_DRAWPRIMCOUNTS;
//check this -> typedef D3DNTHAL_DRAWPRIMITIVESDATA     D3DHAL_DRAWPRIMITIVESDATA;
typedef D3DNTHAL_CALLBACKS2             D3DHAL_CALLBACKS2;
typedef LPD3DNTHAL_CALLBACKS            LPD3DHAL_CALLBACKS;
typedef LPD3DNTHAL_GLOBALDRIVERDATA     LPD3DHAL_GLOBALDRIVERDATA;
typedef LPD3DNTHAL_CONTEXTCREATEDATA    LPD3DHAL_CONTEXTCREATEDATA;
typedef LPD3DNTHAL_CONTEXTDESTROYDATA   LPD3DHAL_CONTEXTDESTROYDATA;
typedef LPD3DNTHAL_CONTEXTDESTROYALLDATA LPD3DHAL_CONTEXTDESTROYALLDATA;
//check this -> typedef LPD3DNTHAL_RENDERSTATEDATA      LPD3DHAL_RENDERSTATEDATA;
//check this -> typedef LPD3DNTHAL_RENDERPRIMITIVEDATA  LPD3DHAL_RENDERPRIMITIVEDATA;
//check this -> typedef LPD3DNTHAL_GETSTATEDATA         LPD3DHAL_GETSTATEDATA;
typedef LPD3DNTHAL_SCENECAPTUREDATA     LPD3DHAL_SCENECAPTUREDATA;
//check this -> typedef LPD3DNTHAL_EXECUTEDATA          LPD3DHAL_EXECUTEDATA;
typedef LPD3DNTHAL_TEXTURECREATEDATA    LPD3DHAL_TEXTURECREATEDATA;
typedef LPD3DNTHAL_TEXTUREDESTROYDATA   LPD3DHAL_TEXTUREDESTROYDATA;
typedef LPD3DNTHAL_TEXTURESWAPDATA      LPD3DHAL_TEXTURESWAPDATA;
typedef LPD3DNTHAL_TEXTUREGETSURFDATA   LPD3DHAL_TEXTUREGETSURFDATA;
typedef LPD3DNTHAL_SETRENDERTARGETDATA  LPD3DHAL_SETRENDERTARGETDATA;
//check this -> typedef LPD3DNTHAL_CLEARDATA            LPD3DHAL_CLEARDATA;
typedef LPD3DNTHAL_CLEAR2DATA           LPD3DHAL_CLEAR2DATA;
//check this -> typedef LPD3DNTHAL_SETVIEWPORT2DATA     LPD3DHAL_SETVIEWPORT2DATA;
//check this -> typedef LPD3DNTHAL_DRAWONEPRIMITIVEDATA LPD3DHAL_DRAWONEPRIMITIVEDATA;
//check this -> typedef LPD3DNTHAL_DRAWONEINDEXEDPRIMITIVEDATA LPD3DHAL_DRAWONEINDEXEDPRIMITIVEDATA;
//check this -> typedef LPD3DNTHAL_DRAWPRIMCOUNTS       LPD3DHAL_DRAWPRIMCOUNTS;
//check this -> typedef LPD3DNTHAL_DRAWPRIMITIVESDATA   LPD3DHAL_DRAWPRIMITIVESDATA;
typedef LPD3DNTHAL_CALLBACKS2           LPD3DHAL_CALLBACKS2;

#define D3DHAL2_CB32_SETRENDERTARGET          D3DNTHAL2_CB32_SETRENDERTARGET
#define D3DHAL2_CB32_CLEAR                    D3DNTHAL2_CB32_CLEAR
#define D3DHAL2_CB32_DRAWONEPRIMITIVE         D3DNTHAL2_CB32_DRAWONEPRIMITIVE
#define D3DHAL2_CB32_DRAWONEINDEXEDPRIMITIVE  D3DNTHAL2_CB32_DRAWONEINDEXEDPRIMITIVE
#define D3DHAL2_CB32_DRAWPRIMITIVES           D3DNTHAL2_CB32_DRAWPRIMITIVES
#define D3DHAL2_CB32_CLEAR2                   D3DNTHAL2_CB32_CLEAR2
#define D3DHAL2_CB32_SETVIEWPORT2             D3DNTHAL2_CB32_SETVIEWPORT2

#define D3DHAL_CONTEXT_BAD              D3DNTHAL_CONTEXT_BAD
#define D3DHAL_OUTOFCONTEXTS            D3DNTHAL_OUTOFCONTEXTS
#define D3DHALSTATE_GET_RENDER          D3DNTHALSTATE_GET_RENDER
#define D3DHAL_SCENE_CAPTURE_START      D3DNTHAL_SCENE_CAPTURE_START
#define D3DHAL_SCENE_CAPTURE_END        D3DNTHAL_SCENE_CAPTURE_END
#endif // _WIN32_WINNT >= 0x0500

#endif // WINNT
#endif // NODX95TYPES

#endif // __DX95TYPE_H__
