/*
** Copyright (c) 1995, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished  -
** rights reserved under the Copyright Laws of the United States.
**
** ALPHA PRODUCT!
*/

/*
** 3DS Loader Macros
**
** This file contains the 3DS Loader macros.
*/

#ifndef TDSMACRO_H__
#define TDSMACRO_H__

#ifdef __cplusplus
extern "C" {
#endif

/*-----------------------------  Includes -----------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include "3ds.h"

/*------------------------------  Macros ------------------------------*/

/*---------------------------------------------------------------------
  Macro: DUMP
  Date: 4/3/96
  Library: 3DStudio Loader
  Description: Dumps debugging information out to the bug.out file.
  Arguments: none
  Return: none
---------------------------------------------------------------------*/
#define DUMP\
  { \
    long offset, len; \
    FILE *fp; \
    unsigned char buf[1000000]; \
    offset = ftell( file->fp ); \
    fseek( file->fp, 0, SEEK_END ); \
    len = ftell( file->fp ) - offset; \
    fseek( file->fp, offset, SEEK_SET ); \
    fread( buf, len, 1, file->fp ); \
    fp = fopen( "bug.out", "w" ); \
    fwrite( buf, len, 1, fp ); \
    fclose( fp ); \
    exit( 0 ); \
  }

/*---------------------------------------------------------------------
  Macro: ASSERT_CHUNK
  Date: 4/3/96
  Library: 3DStudio Loader
  Description: check to make sure that the next chunk in the 3DS file is
               the next chunk expected by the program
  Arguments: 
    type - Contains the hex value defining the next 3DS chunk
    name - Name of the next expected chunk in the 3DS file
  Return: none
---------------------------------------------------------------------*/
#define ASSERT_CHUNK(TYPE,NAME) \
  do { \
  TDSReadChunkHeader( file, &header ); \
  if( header.id != TYPE ) \
    { \
        atuError( FXTRUE, "got 0x%04x\n", ( int )header.id ); \
    } \
  } while( 0 )

#ifdef __cplusplus
};
#endif

#endif /* TDSMACRO_H__ */
