/*
** Copyright (c) 1999, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished  -
** rights reserved under the Copyright Laws of the United States.
*/

/* 
// Trap.h: interface for the Trap functions
//
// Hector Yee
// yhy1@cornell.edu
// yee@3dfx.com
// 6/8/98
//
// This set of functions records glide calls for playback
// Based on glide 3.0
// 
*/

#if !defined(AFX_TRAP_H__991E6543_FF8E_11D1_B7A3_00A0C9C84654__INCLUDED_)
#define AFX_TRAP_H__991E6543_FF8E_11D1_B7A3_00A0C9C84654__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include <glide.h>
#include <stdio.h>
#ifdef __WIN32__
#include <windows.h>
#define GETCH getch
#else
#define GETCH getchar
#endif
#ifdef __cplusplus
extern "C" {
#endif

#include "glideres.h"

#define MAX(A,B) ((A)>(B) ? (A) : (B))

/* tokens file prefix*/
#define TOKENF "tokens.trp"
/* textures file prefix */
#define TEXF "t"
/* LFB file prefix */
#define LFBF "f"
/* Invalid data */
#define INVALID "invalid!"
/* texture/lfb file names */
#define TEXNAMELEN 15    
/* some arbitrary number to store temporary fog values that are discarded */
/* actual tables use grGet */
#define BIG_NUM 300
/* initial size of dynamic array */
#define DARRAYSIZE 50

#define MAX_PATH_LEN 128

extern GrProc MYEXT_grChromaRangeModeExt;
extern GrProc MYEXT_grChromaRangeExt;
extern GrProc MYEXT_grTexChromaModeExt;
extern GrProc MYEXT_grTexChromaRangeExt;
extern GrProc MYEXT_grDrawTextureLineExt;

#include "filebuff.h"
#include "state.h"


    /* linked list structure */
    typedef struct _list {
        unsigned long token;
        void *data;
        struct _list *next;
    } List;
    

    /* user defined */
    /* Local functions... inlined in trap.c */
    
    void my_grVertexLayout(FxU32 param, FxI32 offset, FxU32 mode);
    __inline void my_outputvertex(void *p);
    __inline void * my_inputvertex(FILEBUFF *infile);
    __inline void my_inputvnalloc(void *vtx, FILEBUFF *infile);
    int getvertexsize();

    void Init();
    void CTrapInit(char * trapdir);
    FILEBUFF* PlayInit();
    FILEBUFF* FilterInit(int *startframe, int *endframe);
    FILEBUFF* CommandLinePlayInit(int argc, char *argv[],
                              int *framestart, int *frameend);
    void PlayDone(FILEBUFF* infile);
    void CTrapDone();
    void SetScreenEcho(FxBool echo);
    void SetFastComp(FxBool fast);
    void SetNoDraw(FxBool nd);
    FxU32 GetMipMapSize(GrAspectRatio_t aspectRatio, GrLOD_t thisLod,GrTextureFormat_t format);
    void load_texture(char *filename, void **data, GrTextureFormat_t format);

    /* extensions */
    void FX_CALL trap_grChromaRangeModeExt(GrChromakeyMode_t mode);
    void FX_CALL trap_grChromaRangeExt(GrColor_t color, GrColor_t range, GrChromaRangeMode_t match_mode);
    void FX_CALL trap_grTexChromaModeExt(GrChipID_t tmu, GrChromakeyMode_t mode);
    void FX_CALL trap_grTexChromaRangeExt(GrChipID_t tmu, GrColor_t min, GrColor_t max, GrTexChromakeyMode_t mode);
    void FX_CALL trap_grDrawTextureLineExt(const void *a, const void *b);

    void play_grChromaRangeModeExt(FILEBUFF *infile);
    void play_grChromaRangeExt(FILEBUFF *infile);
    void play_grTexChromaModeExt(FILEBUFF *infile);
    void play_grTexChromaRangeExt(FILEBUFF *infile);
    void play_grDrawTextureLineExt(FILEBUFF *infile);

    void filter_grChromaRangeModeExt(int token, FILEBUFF *infile);
    void filter_grChromaRangeExt(int token, FILEBUFF *infile);
    void filter_grTexChromaModeExt(int token, FILEBUFF *infile);
    void filter_grTexChromaRangeExt(int token, FILEBUFF *infile);
    void filter_grDrawTextureLineExt(int token, FILEBUFF *infile);

    /* generated by perl script */

    void trap_grAADrawTriangle(const void *a,
                            const void *b,
                            const void *c,
                            FxBool ab_antialias,
                            FxBool bc_antialias,
                            FxBool ca_antialias);
    void trap_grAlphaBlendFunction(GrAlphaBlendFnc_t rgb_sf,
                                GrAlphaBlendFnc_t rgb_df,
                                GrAlphaBlendFnc_t alpha_sf,
                                GrAlphaBlendFnc_t alpha_df);
    void trap_grAlphaCombine(GrCombineFunction_t function,
                          GrCombineFactor_t factor,
                          GrCombineLocal_t local,
                          GrCombineOther_t other,
                          FxBool invert);
    void trap_grAlphaControlsITRGBLighting(FxBool enable);
    void trap_grAlphaTestFunction(GrCmpFnc_t function);
    void trap_grAlphaTestReferenceValue(GrAlpha_t value);
    void trap_grBufferClear(GrColor_t color,
                         GrAlpha_t alpha,
                         FxU32 depth);
    void trap_grBufferSwap(int swap_interval);
    void trap_grChromakeyMode(GrChromakeyMode_t mode);
    void trap_grChromakeyValue(GrColor_t value);
    void trap_grClipWindow(FxU32 minx,
                        FxU32 miny,
                        FxU32 maxx,
                        FxU32 maxy);
    void trap_grColorCombine(GrCombineFunction_t function,
                          GrCombineFactor_t factor,
                          GrCombineLocal_t local,
                          GrCombineOther_t other,
                          FxBool invert);
    void trap_grColorMask(FxBool rgb,
                       FxBool a);
    void trap_grConstantColorValue(GrColor_t value);
    void trap_grCoordinateSpace(GrCoordinateSpaceMode_t mode);
    void trap_grCullMode(GrCullMode_t mode);
    void trap_grDepthBiasLevel(FxI32 level);
    void trap_grDepthBufferFunction(GrCmpFnc_t function);
    void trap_grDepthBufferMode(GrDepthBufferMode_t mode);
    void trap_grDepthMask(FxBool mask);
    void trap_grDepthRange(FxFloat n,
                        FxFloat f);
    void trap_grDisable(GrEnableMode_t mode);
    void trap_grDisableAllEffects( void);
    void trap_grDitherMode(GrDitherMode_t mode);
    void trap_grDrawLine(const void *v1,
                      const void *v2);
    void trap_grDrawPoint(const void *pt);
    void trap_grDrawTriangle(const void *a,
                          const void *b,
                          const void *c);
    void trap_grDrawVertexArray(FxU32 mode,
                             FxU32 Count,
                             void *pointers);
    void trap_grDrawVertexArrayContiguous(FxU32 mode,
                                       FxU32 Count,
                                       void *pointers,
                                       FxU32 stride);
    void trap_grEnable(GrEnableMode_t mode);
    void trap_grErrorSetCallback(GrErrorCallbackFnc_t fnc);
    void trap_grFinish( void);
    void trap_grFlush( void);
    void trap_grFogColorValue(GrColor_t fogcolor);
    void trap_grFogMode(GrFogMode_t mode);
    void trap_grFogTable(const GrFog_t ft[],FxU32 tsize);
    void trap_grGet(FxU32 pname,
                 FxU32 plength,
                 FxI32 *params);
    GrProc trap_grGetProcAddress(char *procName, GrProc ext);
    void trap_grGetString(FxU32 pname);
    void trap_grGlideGetState(void *state, FxU32 length);
    void trap_grGlideGetVertexLayout(void *layout, FxU32 length);
    void trap_grGlideInit( void);
    void trap_grGlideSetState(const void *state, FxU32 length);
    void trap_grGlideSetVertexLayout(const void *layout, FxU32 length);
    void trap_grGlideShutdown( void);
    void trap_grLfbConstantAlpha(GrAlpha_t alpha);
    void trap_grLfbConstantDepth(FxU32 depth);
    void trap_grLfbLock(GrLock_t type,
                     GrBuffer_t buffer,
                     GrLfbWriteMode_t writeMode,
                     GrOriginLocation_t origin,
                     FxBool pixelPipeline,
                     FxBool result);
    void trap_grLfbReadRegion(GrBuffer_t src_buffer,
                           FxU32 src_x,
                           FxU32 src_y,
                           FxU32 src_width,
                           FxU32 src_height,
                           FxU32 dst_stride,
                           void *dst_data);
    void trap_grLfbUnlock(GrLock_t type,
                       GrBuffer_t buffer);
    void trap_grLfbWriteColorFormat(GrColorFormat_t colorFormat);
    void trap_grLfbWriteColorSwizzle(FxBool swizzleBytes,
                                  FxBool swapWords);
    void trap_grLfbWriteRegion(GrBuffer_t dst_buffer,
                            FxU32 dst_x,
                            FxU32 dst_y,
                            GrLfbSrcFmt_t src_format,
                            FxU32 src_width,
                            FxU32 src_height,
                            FxBool pixelPipeline,
                            FxI32 src_stride,
                            void *src_data);
    void trap_grLoadGammaTable(FxU32 nentries,
                            FxU32 *red,
                            FxU32 *green,
                            FxU32 *blue);
    void trap_grQueryResolutions(const GlideResolution *resTemplate,
                              GlideResolution *output,FxI32 result);
    void trap_grRenderBuffer(GrBuffer_t buffer);
    void trap_grReset(FxU32 what);
    void trap_grSelectContext(GrContext_t context);
    void trap_grSplash(float x,
                    float y,
                    float width,
                    float height,
                    FxU32 frame);
    void trap_grSstOrigin(GrOriginLocation_t origin);
    void trap_grSstSelect(int which_sst);
    void trap_grSstWinClose(GrContext_t context);
    void trap_grSstWinOpen(FxU32 hWnd,
                        GrScreenResolution_t screen_resolution,
                        GrScreenRefresh_t refresh_rate,
                        GrColorFormat_t color_format,
                        GrOriginLocation_t origin_location,
                        int nColBuffers,
                        int nAuxBuffers);
    void trap_grTexCalcMemRequired(GrLOD_t lodmin,
                                GrLOD_t lodmax,
                                GrAspectRatio_t aspect,
                                GrTextureFormat_t fmt);
    void trap_grTexClampMode(GrChipID_t tmu,
                          GrTextureClampMode_t s_clampmode,
                          GrTextureClampMode_t t_clampmode);
    void trap_grTexCombine(GrChipID_t tmu,
                        GrCombineFunction_t rgb_function,
                        GrCombineFactor_t rgb_factor,
                        GrCombineFunction_t alpha_function,
                        GrCombineFactor_t alpha_factor,
                        FxBool rgb_invert,
                        FxBool alpha_invert);
    void trap_grTexDetailControl(GrChipID_t tmu,
                              int lod_bias,
                              FxU8 detail_scale,
                              float detail_max);
    void trap_grTexDownloadMipMap(GrChipID_t tmu,
                               FxU32 startAddress,
                               FxU32 evenOdd,
                               GrTexInfo *info,FxU32 size);
    void trap_grTexDownloadMipMapLevel(GrChipID_t tmu,
                                    FxU32 startAddress,
                                    GrLOD_t thisLod,
                                    GrLOD_t largeLod,
                                    GrAspectRatio_t aspectRatio,
                                    GrTextureFormat_t format,
                                    FxU32 evenOdd,
                                    void *data,
                                    FxU32 size);
    void trap_grTexDownloadMipMapLevelPartial(GrChipID_t tmu,
                                           FxU32 startAddress,
                                           GrLOD_t thisLod,
                                           GrLOD_t largeLod,
                                           GrAspectRatio_t aspectRatio,
                                           GrTextureFormat_t format,
                                           FxU32 evenOdd,
                                           void *data,
                                           int start,
                                           int end,
                                           FxU32 size);
    void trap_grTexDownloadTable(GrTexTable_t type,
                              void *data);
    void trap_grTexDownloadTablePartial(GrTexTable_t type,
                                     void *data,
                                     int start,
                                     int end);
    void trap_grTexFilterMode(GrChipID_t tmu,
                           GrTextureFilterMode_t minfilter_mode,
                           GrTextureFilterMode_t magfilter_mode);
    void trap_grTexLodBiasValue(GrChipID_t tmu,
                             float bias);
    void trap_grTexMaxAddress(GrChipID_t tmu);
    void trap_grTexMinAddress(GrChipID_t tmu);
    void trap_grTexMipMapMode(GrChipID_t tmu,
                           GrMipMapMode_t mode,
                           FxBool lodBlend);
    void trap_grTexMultibase(GrChipID_t tmu,
                          FxBool enable);
    void trap_grTexMultibaseAddress(GrChipID_t tmu,
                                 GrTexBaseRange_t range,
                                 FxU32 startAddress,
                                 FxU32 evenOdd,
                                 GrTexInfo *info);
    void trap_grTexNCCTable(GrNCCTable_t table);
    void trap_grTexSource(GrChipID_t tmu,
                       FxU32 startAddress,
                       FxU32 evenOdd,
                       GrTexInfo *info);
    void trap_grTexTextureMemRequired(FxU32 evenOdd,
                                   GrTexInfo *info);
    void trap_grVertexLayout(FxU32 param,
                          FxI32 offset,
                          FxU32 mode);
    void trap_grViewport(FxI32 x,
                      FxI32 y,
                      FxI32 width,
                      FxI32 height);
    void trap_gu3dfGetInfo(const char *filename,
                        Gu3dfInfo *info);
    void trap_gu3dfLoad(const char *filename,
                     Gu3dfInfo *data);
    void trap_guFogGenerateExp(GrFog_t *fogtable,
                            float density);
    void trap_guFogGenerateExp2(GrFog_t *fogtable,
                             float density);
    void trap_guFogGenerateLinear(GrFog_t *fogtable,
                               float nearZ,
                               float farZ);
    void trap_guFogTableIndexToW(int i);
    void trap_guGammaCorrectionRGB(FxFloat red,
                                FxFloat green,
                                FxFloat blue);

    void play_grAADrawTriangle(FILEBUFF* infile);
    void play_grAlphaBlendFunction(FILEBUFF* infile);
    void play_grAlphaCombine(FILEBUFF* infile);
    void play_grAlphaControlsITRGBLighting(FILEBUFF* infile);
    void play_grAlphaTestFunction(FILEBUFF* infile);
    void play_grAlphaTestReferenceValue(FILEBUFF* infile);
    void play_grBufferClear(FILEBUFF* infile);
    void play_grBufferSwap(FILEBUFF* infile);
    void play_grChromakeyMode(FILEBUFF* infile);
    void play_grChromakeyValue(FILEBUFF* infile);
    void play_grClipWindow(FILEBUFF* infile);
    void play_grColorCombine(FILEBUFF* infile);
    void play_grColorMask(FILEBUFF* infile);
    void play_grConstantColorValue(FILEBUFF* infile);
    void play_grCoordinateSpace(FILEBUFF* infile);
    void play_grCullMode(FILEBUFF* infile);
    void play_grDepthBiasLevel(FILEBUFF* infile);
    void play_grDepthBufferFunction(FILEBUFF* infile);
    void play_grDepthBufferMode(FILEBUFF* infile);
    void play_grDepthMask(FILEBUFF* infile);
    void play_grDepthRange(FILEBUFF* infile);
    void play_grDisable(FILEBUFF* infile);
    void play_grDisableAllEffects(FILEBUFF* infile);
    void play_grDitherMode(FILEBUFF* infile);
    void play_grDrawLine(FILEBUFF* infile);
    void play_grDrawPoint(FILEBUFF* infile);
    void play_grDrawTriangle(FILEBUFF* infile);
    void play_grDrawVertexArray(FILEBUFF* infile);
    void play_grDrawVertexArrayContiguous(FILEBUFF* infile);
    void play_grEnable(FILEBUFF* infile);
    void play_grErrorSetCallback(FILEBUFF* infile);
    void play_grFinish(FILEBUFF* infile);
    void play_grFlush(FILEBUFF* infile);
    void play_grFogColorValue(FILEBUFF* infile);
    void play_grFogMode(FILEBUFF* infile);
    void play_grFogTable(FILEBUFF* infile);
    void play_grGet(FILEBUFF* infile);
    void play_grGetProcAddress(FILEBUFF* infile);
    void play_grGetString(FILEBUFF* infile);
    void play_grGlideGetState(FILEBUFF* infile);
    void play_grGlideGetVertexLayout(FILEBUFF* infile);
    void play_grGlideInit(FILEBUFF* infile);
    void play_grGlideSetState(FILEBUFF* infile);
    void play_grGlideSetVertexLayout(FILEBUFF* infile);
    void play_grGlideShutdown(FILEBUFF* infile);
    void play_grLfbConstantAlpha(FILEBUFF* infile);
    void play_grLfbConstantDepth(FILEBUFF* infile);
    void play_grLfbLock(FILEBUFF* infile);
    void play_grLfbReadRegion(FILEBUFF* infile);
    void play_grLfbUnlock(FILEBUFF* infile);
    void play_grLfbWriteColorFormat(FILEBUFF* infile);
    void play_grLfbWriteColorSwizzle(FILEBUFF* infile);
    void play_grLfbWriteRegion(FILEBUFF* infile);
    void play_grLoadGammaTable(FILEBUFF* infile);
    void play_grQueryResolutions(FILEBUFF* infile);
    void play_grRenderBuffer(FILEBUFF* infile);
    void play_grReset(FILEBUFF* infile);
    void play_grSelectContext(FILEBUFF* infile);
    void play_grSplash(FILEBUFF* infile);
    void play_grSstOrigin(FILEBUFF* infile);
    void play_grSstSelect(FILEBUFF* infile);
    void play_grSstWinClose(FILEBUFF* infile);
    void play_grSstWinOpen(FILEBUFF* infile);
    void play_grTexCalcMemRequired(FILEBUFF* infile);
    void play_grTexClampMode(FILEBUFF* infile);
    void play_grTexCombine(FILEBUFF* infile);
    void play_grTexDetailControl(FILEBUFF* infile);
    void play_grTexDownloadMipMap(FILEBUFF* infile);
    void play_grTexDownloadMipMapLevel(FILEBUFF* infile);
    void play_grTexDownloadMipMapLevelPartial(FILEBUFF* infile);
    void play_grTexDownloadTable(FILEBUFF* infile);
    void play_grTexDownloadTablePartial(FILEBUFF* infile);
    void play_grTexFilterMode(FILEBUFF* infile);
    void play_grTexLodBiasValue(FILEBUFF* infile);
    void play_grTexMaxAddress(FILEBUFF* infile);
    void play_grTexMinAddress(FILEBUFF* infile);
    void play_grTexMipMapMode(FILEBUFF* infile);
    void play_grTexMultibase(FILEBUFF* infile);
    void play_grTexMultibaseAddress(FILEBUFF* infile);
    void play_grTexNCCTable(FILEBUFF* infile);
    void play_grTexSource(FILEBUFF* infile);
    void play_grTexTextureMemRequired(FILEBUFF* infile);
    void play_grVertexLayout(FILEBUFF* infile);
    void play_grViewport(FILEBUFF* infile);
    void play_gu3dfGetInfo(FILEBUFF* infile);
    void play_gu3dfLoad(FILEBUFF* infile);
    void play_guFogGenerateExp(FILEBUFF* infile);
    void play_guFogGenerateExp2(FILEBUFF* infile);
    void play_guFogGenerateLinear(FILEBUFF* infile);
    void play_guFogTableIndexToW(FILEBUFF* infile);
    void play_guGammaCorrectionRGB(FILEBUFF* infile);

    void filter_grAADrawTriangle(int token, FILEBUFF* infile);
    void filter_grAlphaBlendFunction(int token, FILEBUFF* infile);
    void filter_grAlphaCombine(int token, FILEBUFF* infile);
    void filter_grAlphaControlsITRGBLighting(int token, FILEBUFF* infile);
    void filter_grAlphaTestFunction(int token, FILEBUFF* infile);
    void filter_grAlphaTestReferenceValue(int token, FILEBUFF* infile);
    void filter_grBufferClear(int token, FILEBUFF* infile);
    void filter_grBufferSwap(int token, FILEBUFF* infile);
    void filter_grChromakeyMode(int token, FILEBUFF* infile);
    void filter_grChromakeyValue(int token, FILEBUFF* infile);
    void filter_grClipWindow(int token, FILEBUFF* infile);
    void filter_grColorCombine(int token, FILEBUFF* infile);
    void filter_grColorMask(int token, FILEBUFF* infile);
    void filter_grConstantColorValue(int token, FILEBUFF* infile);
    void filter_grCoordinateSpace(int token, FILEBUFF* infile);
    void filter_grCullMode(int token, FILEBUFF* infile);
    void filter_grDepthBiasLevel(int token, FILEBUFF* infile);
    void filter_grDepthBufferFunction(int token, FILEBUFF* infile);
    void filter_grDepthBufferMode(int token, FILEBUFF* infile);
    void filter_grDepthMask(int token, FILEBUFF* infile);
    void filter_grDepthRange(int token, FILEBUFF* infile);
    void filter_grDisable(int token, FILEBUFF* infile);
    void filter_grDisableAllEffects(int token, FILEBUFF* infile);
    void filter_grDitherMode(int token, FILEBUFF* infile);
    void filter_grDrawLine(int token, FILEBUFF* infile);
    void filter_grDrawPoint(int token, FILEBUFF* infile);
    void filter_grDrawTriangle(int token, FILEBUFF* infile);
    void filter_grDrawVertexArray(int token, FILEBUFF* infile);
    void filter_grDrawVertexArrayContiguous(int token, FILEBUFF* infile);
    void filter_grEnable(int token, FILEBUFF* infile);
    void filter_grErrorSetCallback(int token, FILEBUFF* infile);
    void filter_grFinish(int token, FILEBUFF* infile);
    void filter_grFlush(int token, FILEBUFF* infile);
    void filter_grFogColorValue(int token, FILEBUFF* infile);
    void filter_grFogMode(int token, FILEBUFF* infile);
    void filter_grFogTable(int token, FILEBUFF* infile);
    void filter_grGet(int token, FILEBUFF* infile);
    void filter_grGetProcAddress(int token, FILEBUFF* infile);
    void filter_grGetString(int token, FILEBUFF* infile);
    void filter_grGlideInit(int token, FILEBUFF* infile);
    void filter_grGlideShutdown(int token, FILEBUFF* infile);
    void filter_grLfbConstantAlpha(int token, FILEBUFF* infile);
    void filter_grLfbConstantDepth(int token, FILEBUFF* infile);
    void filter_grLfbLock(int token, FILEBUFF* infile);
    void filter_grLfbReadRegion(int token, FILEBUFF* infile);
    void filter_grLfbUnlock(int token, FILEBUFF* infile);
    void filter_grLfbWriteColorFormat(int token, FILEBUFF* infile);
    void filter_grLfbWriteColorSwizzle(int token, FILEBUFF* infile);
    void filter_grLfbWriteRegion(int token, FILEBUFF* infile);
    void filter_grLoadGammaTable(int token, FILEBUFF* infile);
    void filter_grQueryResolutions(int token, FILEBUFF* infile);
    void filter_grRenderBuffer(int token, FILEBUFF* infile);
    void filter_grReset(int token, FILEBUFF* infile);
    void filter_grSelectContext(int token, FILEBUFF* infile);
    void filter_grSplash(int token, FILEBUFF* infile);
    void filter_grSstOrigin(int token, FILEBUFF* infile);
    void filter_grTexCalcMemRequired(int token, FILEBUFF* infile);
    void filter_grTexClampMode(int token, FILEBUFF* infile);
    void filter_grTexCombine(int token, FILEBUFF* infile);
    void filter_grTexDetailControl(int token, FILEBUFF* infile);
    void filter_grTexDownloadMipMap(int token, FILEBUFF* infile);
    void filter_grTexDownloadMipMapLevel(int token, FILEBUFF* infile);
    void filter_grTexDownloadMipMapLevelPartial(int token, FILEBUFF* infile);
    void filter_grTexDownloadTable(int token, FILEBUFF* infile);
    void filter_grTexDownloadTablePartial(int token, FILEBUFF* infile);
    void filter_grTexFilterMode(int token, FILEBUFF* infile);
    void filter_grTexLodBiasValue(int token, FILEBUFF* infile);
    void filter_grTexMaxAddress(int token, FILEBUFF* infile);
    void filter_grTexMinAddress(int token, FILEBUFF* infile);
    void filter_grTexMipMapMode(int token, FILEBUFF* infile);
    void filter_grTexMultibase(int token, FILEBUFF* infile);
    void filter_grTexMultibaseAddress(int token, FILEBUFF* infile);
    void filter_grTexNCCTable(int token, FILEBUFF* infile);
    void filter_grTexSource(int token, FILEBUFF* infile);
    void filter_grTexTextureMemRequired(int token, FILEBUFF* infile);
    void filter_grVertexLayout(int token, FILEBUFF* infile);
    void filter_grViewport(int token, FILEBUFF* infile);
    void filter_gu3dfGetInfo(int token, FILEBUFF* infile);
    void filter_gu3dfLoad(int token, FILEBUFF* infile);
    void filter_guFogGenerateExp(int token, FILEBUFF* infile);
    void filter_guFogGenerateExp2(int token, FILEBUFF* infile);
    void filter_guFogGenerateLinear(int token, FILEBUFF* infile);
    void filter_guFogTableIndexToW(int token, FILEBUFF* infile);
    void filter_guGammaCorrectionRGB(int token, FILEBUFF* infile);

    enum
    {
      // leave "0" as an error condition
      token_grAADrawTriangle=1,
      token_grAlphaBlendFunction,
      token_grAlphaCombine,
      token_grAlphaControlsITRGBLighting,
      token_grAlphaTestFunction,
      token_grAlphaTestReferenceValue,
      token_grBufferClear,
      token_grBufferSwap,
      token_grChromakeyMode,
      token_grChromakeyValue,
      token_grClipWindow,
      token_grColorCombine,
      token_grColorMask,
      token_grConstantColorValue,
      token_grCoordinateSpace,
      token_grCullMode,
      token_grDepthBiasLevel,
      token_grDepthBufferFunction,
      token_grDepthBufferMode,
      token_grDepthMask,
      token_grDepthRange,
      token_grDisable,
      token_grDisableAllEffects,
      token_grDitherMode,
      token_grDrawLine,
      token_grDrawPoint,
      token_grDrawTriangle,
      token_grDrawVertexArray,
      token_grDrawVertexArrayContiguous,
      token_grEnable,
      token_grErrorSetCallback,
      token_grFinish,
      token_grFlush,
      token_grFogColorValue,
      token_grFogMode,
      token_grFogTable,
      token_grGet,
      token_grGetProcAddress,
      token_grGetString,
      token_grGlideGetState,
      token_grGlideGetVertexLayout,
      token_grGlideInit,
      token_grGlideSetState,
      token_grGlideSetVertexLayout,
      token_grGlideShutdown,
      token_grLfbConstantAlpha,
      token_grLfbConstantDepth,
      token_grLfbLock,
      token_grLfbReadRegion,
      token_grLfbUnlock,
      token_grLfbWriteColorFormat,
      token_grLfbWriteColorSwizzle,
      token_grLfbWriteRegion,
      token_grLoadGammaTable,
      token_grQueryResolutions,
      token_grRenderBuffer,
      token_grReset,
      token_grSelectContext,
      token_grSplash,
      token_grSstOrigin,
      token_grSstSelect,
      token_grSstWinClose,
      token_grSstWinOpen,
      token_grTexCalcMemRequired,
      token_grTexClampMode,
      token_grTexCombine,
      token_grTexDetailControl,
      token_grTexDownloadMipMap,
      token_grTexDownloadMipMapLevel,
      token_grTexDownloadMipMapLevelPartial,
      token_grTexDownloadTable,
      token_grTexDownloadTablePartial,
      token_grTexFilterMode,
      token_grTexLodBiasValue,
      token_grTexMaxAddress,
      token_grTexMinAddress,
      token_grTexMipMapMode,
      token_grTexMultibase,
      token_grTexMultibaseAddress,
      token_grTexNCCTable,
      token_grTexSource,
      token_grTexTextureMemRequired,
      token_grVertexLayout,
      token_grViewport,
      token_gu3dfGetInfo,
      token_gu3dfLoad,
      token_guFogGenerateExp,
      token_guFogGenerateExp2,
      token_guFogGenerateLinear,
      token_guFogTableIndexToW,
      token_guGammaCorrectionRGB,
      token_grChromaRangeModeExt,
      token_grChromaRangeExt,
      token_grTexChromaModeExt,
      token_grTexChromaRangeExt,
      token_grDrawTextureLineExt
    };

#ifdef __cplusplus
}
#endif
#endif // !defined(AFX_TRAP_H__991E6543_FF8E_11D1_B7A3_00A0C9C84654__INCLUDED_)
