/*
** Copyright 1991-1997, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Revision: 2$
** $Date: 10/11/00 8:06:51 PM$
*/
#include <stdio.h>
#include <string.h>

#define BUF_LEN 256
char buf[BUF_LEN];

FILE *glh;
FILE *glcltpt;

char pre[350][50];
char funcs[350][100];
char post[350][300];
int count;
char order[350][50];
int ocount;

int lookUp(int ix)
{
    int i;

    /* lookup function in order table */
    for (i=0; i < count; i++) {
	if (strcmp(funcs[i], order[ix]) == 0) {
	    /* found it */
	    return i;
	}
    }

    return -1;
}
    

main(int argc, char **argv)
{
    int state;

    if (argc != 3) {
	fprintf(stderr, "Usage: %s gl.h glcltpt.c\n", argv[0]);
	exit(1);
    }

    glh = fopen(argv[1], "r");
    glcltpt = fopen(argv[2], "r");
    if ((glh == NULL) ||
	(glcltpt == NULL)) {
	fprintf(stderr, "Usage: Can't open input file\n");
	exit(1);
    }

    /* first read our prototypes */
    state = 0;
    count = 0;
    while(fgets(buf, BUF_LEN, glh)) {
	switch(state) {
	  case 0:
	      /* trying to find defs */
	      if (strstr(buf, "WINGDIAPI") == NULL) {
		  continue;
	      }
	      state = 1;
	      /* fall thru */
	  case 1:
	      if (strstr(buf, "/*")) {
		  /* we're done */
		  state = 2;
		  break;
	      }
	      if (strstr(buf, "WINGDIAPI") == NULL) {
		  continue;
	      }
	      {
		  char *sp;
		  char *eot;
		  char *gl;
		  char *name;

		  sp = strstr(buf, " ");
		  eot = strstr(buf, "APIENTRY");
		  memcpy(pre[count], sp, eot-sp);
		  gl = strstr(eot, "gl");
		  gl += 2;
		  name = strstr(gl, " ");
		  memcpy(funcs[count], gl, name-gl);
		  strcpy(post[count], name);
		  count++;
	      }
	      break;
	  default:
	      ;
	}
    }

    /* read order file from glcltpt.c */
    state = 0;
    ocount = 0;
    while(fgets(buf, BUF_LEN, glcltpt)) {
	switch(state) {
	  case 0:
	      /* look for entry comment */
	      if (strstr(buf, "gl Entry points")) {
		  state = 1;
	      }
	      break;
	  case 1:
	      if (strstr(buf, "gl") == NULL) {
		  /* nothing here */
		  continue;
	      }
	      {
		  char *gl;
		  char *sp;

		  gl = strstr(buf, "glclt");
		  gl += 5;
		  sp = strstr(gl, " ");
		  memcpy(order[ocount], gl, sp-gl);
		  ocount++;
	      }
	      break;
	  default:
	      ;
	}
    }
    
	      
    {
	int	i;

	printf("/* DO NOT EDIT - THIS FILE IS AUTOMATICALLY GENERATED */\n\n");

	printf("#define SZPTR (%d)\n", sizeof(void *));
	printf("#define SZINT (%d)\n\n", sizeof(int));

	for (i=0; i < count; i++) {
	    printf("#define __GL_TAB_%s %d\n", order[i], i);
	}
	printf("#define __GL_TAB_SIZE (%d*SZPTR)\n", count);
	printf("#define __GL_TAB_ENTRIES %d\n\n", count);

	printf("\n#define __GL_DISPATCHSTATE_SIZE ");
	printf("((__GL_TAB_ENTRIES+__GL_EXTTAB_ENTRIES)*SZPTR+SZINT)\n");
	printf("\n#if defined(_LANGUAGE_C) || defined(__cplusplus)\n");
	printf("struct __GLdispatchTableRec {\n");

	for (i=0; i < count; i++) {
	    int ix = lookUp(i);
	    if (ix == -1) {
		printf("    void\t(APIENTRY *Foo%d)(void)", i);
		printf("\t /* ALERT - NOT IN DISPATCH TABLE ?? */\n");
	    } else {
		printf("    %s\t(APIENTRY *%s)%s", 
		       pre[ix], funcs[ix], post[ix]);
	    }
	}
	printf("};\n\n");


	printf("#endif /* _LANGUAGE_C || __cplusplus */\n\n");
    }

    return 0;
}
