#
# Copyright (c) 1997, 3Dfx Interactive, Inc.
# All Rights Reserved.
#
# This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
# the contents of this file may not be disclosed to third parties, copied or
# duplicated in any form, in whole or in part, without the prior written
# permission of 3Dfx Interactive, Inc.
#
# RESTRICTED RIGHTS LEGEND:
# Use, duplication or disclosure by the Government is subject to restrictions
# as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
# and Computer Software clause at DFARS 252.227-7013, and/or in similar or
# successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished  -
# rights reserved under the Copyright Laws of the United States.
#
# $Revision: $
# $Date: $
#

#----------------------------------------------------------------------
#                    Avenger constants
#----------------------------------------------------------------------

package h3defs;
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(%h3def %h3field %h3fieldv BIT SST_MASK PARSE);
@EXPORT_OK = qw();

################################################################################
################################################################################
##
## from $/devel/h3/incsrc/h3defs.h
##
################################################################################
################################################################################

$h3def{MAX_NUM_TMUS} = 2;

# From fxhal.h

$h3def{SST_DEVICE_ID_H4_OEM} = 4;
$h3def{SST_DEVICE_ID_H4} = 5;
$h3def{SST_DEVICE_ID_H5_OEM} = 6;
$h3def{SST_DEVICE_ID_H5} = 7;

# SST commands
$h3def{SST_NOPCMD} =       1;
$h3def{SST_TRIANGLECMD} =  2;
$h3def{SST_FASTFILLCMD} =  3;
$h3def{SST_SWAPBUFCMD} =   4;
$h3def{SST_SBEGINTRICMD} =  5;
$h3def{SST_SDRAWTRICMD} = 	6;
$h3def{SST_USERINTRCMD} =  7;

# this crazy macro fixes the binary point within a floating point
# number so that it has 'fracbits' to the right of the binary point
# it also adds a '1' bit to the MSB so that slightly negative numbers
# end up looking like twos complement numbers (they carry out from the '1')
sub FLOAT_FIX	{ my($f,$fracbits) = @_; return (($f)+(3<<(22-($fracbits)))); }
sub SST_FLOAT	{ my($f,$scale,$shift) = @_; return (($f)*(($scale)*(1<<$shift))); }
sub SST_TLOD_MINMAX { my($lodmin,$lodmax) = @_; 
                      my $t = ((($lodmin) << ($h3def{SST_LODMIN_SHIFT})) |
			       (($lodmax) << ($h3def{SST_LODMAX_SHIFT})));
                      return $t;
		}
sub SST_TLOD_MINMAX_INT { my($lodmin,$lodmax) = @_; 
                      my $t = ((($lodmin) << ($h3def{SST_LOD_FRACBITS}+$h3def{SST_LODMIN_SHIFT})) |
			       (($lodmax) << ($h3def{SST_LOD_FRACBITS}+$h3def{SST_LODMAX_SHIFT})));
                      return $t;
		}

# this crazy macro tests the sign bit of a float by loading it into
# an integer register and then testing the sign bit of the integer
sub FLOAT_ISNEG	{ my($f) = @_; return (int($f) < 0); }

# these crazy macros returns the sign of a number (1 if >= 0; -1 if < 0)
sub ISIGN	{ my($x) = @_; return ((($x) | 0x40000000) >> 30); }
sub FSIGN	{ my($f) = @_; return ISIGN($f); }

sub BIT 	{ my($n) = @_; return 1 << @_[0]; }
sub SST_MASK 	{ my($n) = @_; my($s)=0;
		  $s = $s + BIT($n) while ( $n-- > 0 );
		  return $s; }

#----------------- SST binary point locations ---------------
$h3def{SST_LOD_SIZE} =      6;
$h3def{SST_LOD_FRACBITS} =  2;
$h3def{SST_XY_SIZE} =      17;
$h3def{SST_XY_INTBITS} =   13;
$h3def{SST_XY_FRACBITS} =   4;
$h3def{SST_RGBA_SIZE} =    24;
$h3def{SST_RGBA_INTBITS} =  8;
$h3def{SST_RGBA_FRACBITS} =  12;
# NOTE: below are ONLY for the fixed point registers
$h3def{SST_Z_32BPP_SIZE}  =   32;
$h3def{SST_Z_32BPP_INTBITS} = 24;
$h3def{SST_Z_32BPP_FRACBITS}=  4;
$h3def{SST_Z_16BPP_SIZE}    = 32;
$h3def{SST_Z_16BPP_INTBITS} = 16;
$h3def{SST_Z_16BPP_FRACBITS}= 12;
$h3def{SST_ST_SIZE} =      32;
$h3def{SST_ST_INTBITS} =   14;
$h3def{SST_ST_FRACBITS} =  18;
$h3def{SST_W_SIZE} =       32;
$h3def{SST_W_INTBITS} =     2;
$h3def{SST_W_FRACBITS} =   30;

#----------------- SST status bits ---------------------------
$STATUS{SST_FIFOLEVEL} =            0x3F;
$STATUS{SST_PCIFIFO_FREE} =         0x1F;
$STATUS{SST_PCIFIFO_BUSY} =         BIT(5);
$STATUS{SST_VRETRACE} =             BIT(6);
$STATUS{SST_FBI_BUSY} =             BIT(7);
$STATUS{SST_TMU_BUSY} =             BIT(8);
$STATUS{SST_TREX_BUSY} =            $STATUS{SST_TMU_BUSY};
$STATUS{SST_BUSY} =                 BIT(9);
$STATUS{SST_GUI_BUSY} =             BIT(10);
$STATUS{SST_CMD0_BUSY} =            BIT(11);
$STATUS{SST_CMD1_BUSY} =            BIT(12);
$STATUS{SST_TMU0_BUSY} =            BIT(13);
$STATUS{SST_TMU1_BUSY} =            BIT(14);
$STATUS{SST_SWAPBUFPENDING_SHIFT} =  28;
$STATUS{SST_SWAPBUFPENDING} =       (0x7<<$STATUS{SST_SWAPBUFPENDING_SHIFT});
$STATUS{SST_PCIINTERRUPTED} =       BIT(31);

#----------------- SST renderMode bits --------------------------
$RENDERMODE{SST_RM_3D_SHIFT}	=	0;
$RENDERMODE{SST_RM_3D_MODE}	=	(0x3<<$RENDERMODE{SST_RM_3D_SHIFT});
    $SST_RM_3D_MODE{SST_RM_16BPP}=	(0x0<<$RENDERMODE{SST_RM_3D_SHIFT});
    $SST_RM_3D_MODE{SST_RM_15BPP}=	(0x1<<$RENDERMODE{SST_RM_3D_SHIFT});
    $SST_RM_3D_MODE{SST_RM_32BPP}=	(0x2<<$RENDERMODE{SST_RM_3D_SHIFT});
$RENDERMODE{SST_RM_YORIGIN_SELECT} =	BIT(2);
$RENDERMODE{SST_RM_YORIGIN_TOP_SHIFT} = 3;
$RENDERMODE{SST_RM_YORIGIN_TOP} =	(0xFFF<<$RENDERMODE{SST_RM_YORIGIN_TOP_SHIFT});
$RENDERMODE{SST_RM_ALPHAMODE_SHIFT} =	15;
$RENDERMODE{SST_RM_ALPHAMODE} =	(0x3<<$RENDERMODE{SST_RM_ALPHAMODE_SHIFT});
    $SST_RM_ALPHAMODE{SST_RM_ALPHA_ZERO} = (0x0<<$RENDERMODE{SST_RM_ALPHAMODE_SHIFT});
    $SST_RM_ALPHAMODE{SST_RM_ALPHA_ONE}  = (0x1<<$RENDERMODE{SST_RM_ALPHAMODE_SHIFT});
    $SST_RM_ALPHAMODE{SST_RM_ALPHA_MSB}  = (0x2<<$RENDERMODE{SST_RM_ALPHAMODE_SHIFT});
$RENDERMODE{SST_RM_RED_WMASK} =	BIT(17);
$RENDERMODE{SST_RM_GREEN_WMASK}	= BIT(18);
$RENDERMODE{SST_RM_BLUE_WMASK} =  BIT(19);
$RENDERMODE{SST_RM_ALPHA_WMASK}	= BIT(20);
$RENDERMODE{SST_RM_ENGUARDBAND} =  BIT(21);
$RENDERMODE{SST_RM_TWO_PIXELS_PER_CLOCK_BAND_SELECTION_SHIFT} =   22;
$RENDERMODE{SST_RM_TWO_PIXELS_PER_CLOCK_BAND_SELECTION} =  (0x7 << $RENDERMODE{SST_RM_TWO_PIXELS_PER_CLOCK_BAND_SELECTION_SHIFT});
$RENDERMODE{SST_RM_DITHER_ROTATION} =	BIT(25);

# Some useful additional constants.
$RENDERMODE{SST_RM_RGB_WMASK} =     ($RENDERMODE{SST_RM_RED_WMASK} | 
				     $RENDERMODE{SST_RM_GREEN_WMASK} | 
				     $RENDERMODE{$SST_RM_BLUE_WMASK});
$RENDERMODE{SST_RM_RGBA_WMASK}=     ($RENDERMODE{SST_RM_RED_WMASK} | 
				     $RENDERMODE{SST_RM_GREEN_WMASK} | 
				     $RENDERMODE{SST_RM_BLUE_WMASK} | 
				     $RENDERMODE{SST_RM_ALPHA_WMASK});

$RENDERMODE{SST_RM_EN15BPP_RGBA} =              $RENDERMODE{SST_RM_15BPP} | $RENDERMODE{SST_RM_RGBA_WMASK};
$RENDERMODE{SST_RM_EN16BPP_RGBA} =              $RENDERMODE{SST_RM_16BPP} | $RENDERMODE{SST_RM_RGBA_WMASK};
$RENDERMODE{SST_RM_EN32BPP_RGBA} =              $RENDERMODE{SST_RM_32BPP} | $RENDERMODE{SST_RM_RGBA_WMASK};

$RENDERMODE{SST_RM_RESETMODEANDCHANNELS}=       ~($RENDERMODE{SST_RM_3D_MODE} | $RENDERMODE{SST_RM_RGBA_WMASK});


#----------------- SST stencilMode bits --------------------------
$STENCILMODE{SST_STENCIL_REF_SHIFT}=	0;
$STENCILMODE{SST_STENCIL_REF}=		(0xFF<<$STENCILMODE{SST_STENCIL_REF_SHIFT});
$STENCILMODE{SST_STENCIL_MASK_SHIFT}=	8;
$STENCILMODE{SST_STENCIL_MASK}=	        (0xFF<<$STENCILMODE{SST_STENCIL_MASK_SHIFT});
$STENCILMODE{SST_STENCIL_WMASK_SHIFT}=	16;
$STENCILMODE{SST_STENCIL_WMASK}=	(0xFF<<$STENCILMODE{SST_STENCIL_WMASK_SHIFT});
$STENCILMODE{SST_STENCIL_ENABLE}=	BIT(24);
$STENCILMODE{SST_STENCIL_FUNC_SHIFT}=	25;
$STENCILMODE{SST_STENCIL_FUNC}=	        (0x7<<$STENCILMODE{SST_STENCIL_FUNC_SHIFT});
    $SST_STENCIL_FUNC{SST_SFUNC_LT} =     BIT(25);
    $SST_STENCIL_FUNC{SST_SFUNC_EQ} =     BIT(26);
    $SST_STENCIL_FUNC{SST_SFUNC_GT} =     BIT(27);


#----------------- SST stencilOp bits --------------------------
$STENCILOP{SST_STENCIL_SFAIL_OP_SHIFT}=	0;
$STENCILOP{SST_STENCIL_SFAIL_OP}=	(0xF<<$STENCILOP{SST_STENCIL_SFAIL_OP_SHIFT});
$STENCILOP{SST_STENCIL_ZFAIL_OP_SHIFT}=	4;
$STENCILOP{SST_STENCIL_ZFAIL_OP}=	(0xF<<$STENCILOP{SST_STENCIL_ZFAIL_OP_SHIFT});
$STENCILOP{SST_STENCIL_ZPASS_OP_SHIFT}=	8;
$STENCILOP{SST_STENCIL_ZPASS_OP}=	(0xF<<$STENCILOP{SST_STENCIL_ZPASS_OP_SHIFT});
    $SST_STENCIL_OP{SST_SOP_KEEP}=	0;
    $SST_STENCIL_OP{SST_SOP_ZERO}=	1;
    $SST_STENCIL_OP{SST_SOP_REPLACE}=	2;
    $SST_STENCIL_OP{SST_SOP_INCSAT}=	3;
    $SST_STENCIL_OP{SST_SOP_DECSAT}=	4;
    $SST_STENCIL_OP{SST_SOP_NEG}=	5;
    $SST_STENCIL_OP{SST_SOP_INC}=	6;
    $SST_STENCIL_OP{SST_SOP_DEC}=	7;

#----------------- SST lfbMode bits --------------------------
$LFBMODE{SST_LFB_FORMAT_SHIFT} =     0;
$LFBMODE{SST_LFB_FORMAT} =           (0xF<<$LFBMODE{SST_LFB_FORMAT_SHIFT});
       $SST_LFB_FORMAT{SST_LFB_565} =              (0<<$LFBMODE{SST_LFB_FORMAT_SHIFT});
       $SST_LFB_FORMAT{SST_LFB_555} =              (1<<$LFBMODE{SST_LFB_FORMAT_SHIFT});
       $SST_LFB_FORMAT{SST_LFB_1555} =             (2<<$LFBMODE{SST_LFB_FORMAT_SHIFT});
       $SST_LFB_FORMAT{SST_LFB_u1} =               (3<<$LFBMODE{SST_LFB_FORMAT_SHIFT});
       $SST_LFB_FORMAT{SST_LFB_888} =              (4<<$LFBMODE{SST_LFB_FORMAT_SHIFT});
       $SST_LFB_FORMAT{SST_LFB_8888} =             (5<<$LFBMODE{SST_LFB_FORMAT_SHIFT});
       $SST_LFB_FORMAT{SST_LFB_u2} =               (6<<$LFBMODE{SST_LFB_FORMAT_SHIFT});
       $SST_LFB_FORMAT{SST_LFB_u3} =               (7<<$LFBMODE{SST_LFB_FORMAT_SHIFT});
       $SST_LFB_FORMAT{SST_LFB_Z32} =              (8<<$LFBMODE{SST_LFB_FORMAT_SHIFT});
       $SST_LFB_FORMAT{SST_LFB_Z565} =             (12<<$LFBMODE{SST_LFB_FORMAT_SHIFT});
       $SST_LFB_FORMAT{SST_LFB_Z555} =             (13<<$LFBMODE{SST_LFB_FORMAT_SHIFT});
       $SST_LFB_FORMAT{SST_LFB_Z1555} =            (14<<$LFBMODE{SST_LFB_FORMAT_SHIFT});
       $SST_LFB_FORMAT{SST_LFB_ZZ} =               (15<<$LFBMODE{SST_LFB_FORMAT_SHIFT});
$LFBMODE{SST_LFB_READBUFSELECT_SHIFT} =      6;
$LFBMODE{SST_LFB_READBUFSELECT} =    (0x3<<$LFBMODE{SST_LFB_READBUFSELECT_SHIFT});
       $SST_LFB_READBUFSELECT{SST_LFB_READCOLORBUFFER} =  (1<<$LFBMODE{SST_LFB_READBUFSELECT_SHIFT});
       $SST_LFB_READBUFSELECT{SST_LFB_READDEPTHABUFFER} = (2<<$LFBMODE{SST_LFB_READBUFSELECT_SHIFT});
       $SST_LFB_READBUFSELECT{SST_LFB_READAUXBUFFER} =    $LFBMODE{SST_LFB_READDEPTHABUFFER};
$LFBMODE{SST_LFB_ENPIXPIPE} =        BIT(8);
$LFBMODE{SST_LFB_RGBALANES_SHIFT} =          9;
$LFBMODE{SST_LFB_RGBALANES} =        (0x3<<$LFBMODE{SST_LFB_RGBALANES_SHIFT});
       $SST_LFB_RGBALANES{SST_LFB_RGBALANES_ARGB} = (0<<$LFBMODE{SST_LFB_RGBALANES_SHIFT});
       $SST_LFB_RGBALANES{SST_LFB_RGBALANES_ABGR} = (1<<$LFBMODE{SST_LFB_RGBALANES_SHIFT});
       $SST_LFB_RGBALANES{SST_LFB_RGBALANES_RGBA} = (2<<$LFBMODE{SST_LFB_RGBALANES_SHIFT});
       $SST_LFB_RGBALANES{SST_LFB_RGBALANES_BGRA} = (3<<$LFBMODE{SST_LFB_RGBALANES_SHIFT});
$LFBMODE{SST_LFB_WRITE_SWAP16} =     BIT(11);
$LFBMODE{SST_LFB_WRITE_BYTESWAP} =   BIT(12);
$LFBMODE{SST_LFB_YORIGIN} =          BIT(13);
$LFBMODE{SST_LFB_WSELECT} =          BIT(14);
$LFBMODE{SST_LFB_READ_SWAP16} =      BIT(15);
$LFBMODE{SST_LFB_READ_BYTESWAP} =    BIT(16);

$LFBMODE{SST_LFB_ADDR_STRIDE} =      2048;
$LFBMODE{SST_LFB_ADDR_X_SHIFT} =     0;
$LFBMODE{SST_LFB_ADDR_X} =           (0x7FF<<$LFBMODE{SST_LFB_ADDR_X_SHIFT});
$LFBMODE{SST_LFB_ADDR_Y_SHIFT} =     11;
$LFBMODE{SST_LFB_ADDR_Y} =           (0x7FF<<$LFBMODE{SST_LFB_ADDR_Y_SHIFT});

#----------------- SST videoFilterRgbThreshold
$h3def{SST_VIDEOFILTER_BLUE_SHIFT} =       0;
$h3def{SST_VIDEOFILTER_BLUE} =     (0xff << $h3def{SST_VIDEOFILTER_BLUE_SHIFT});;
$h3def{SST_VIDEOFILTER_GREEN_SHIFT} =      8;
$h3def{SST_VIDEOFILTER_GREEN} =    (0xff << $h3def{SST_VIDEOFILTER_GREEN_SHIFT});;
$h3def{SST_VIDEOFILTER_RED_SHIFT} =  16            ;
$h3def{SST_VIDEOFILTER_RED} =      (0xff << $h3def{SST_VIDEOFILTER_RED_SHIFT});;

#----------------- SST textureMode bits -----------------------
$TEXTUREMODE{SST_TPERSP_ST} =            BIT(0);
$TEXTUREMODE{SST_TMINFILTER} =           BIT(1);
$TEXTUREMODE{SST_TMAGFILTER} =           BIT(2);
$TEXTUREMODE{SST_TCLAMPW} =              BIT(3);
$TEXTUREMODE{SST_TLODDITHER} =           BIT(4);
$TEXTUREMODE{SST_TNCCSELECT} =           BIT(5);  # selects which table
$TEXTUREMODE{SST_TCLAMPS} =              BIT(6);
$TEXTUREMODE{SST_TCLAMPT} =              BIT(7);
sub SST_T4BIT_COMPRESSED {my($mode) = @_; return(((($mode & SST_TFORMAT) == SST_FXT1) ||
						  (($mode & SST_TFORMAT) == SST_DXT1)) &&
						 (($mode & SST_COMPRESSED_TEXTURES)));}
sub SST_T8BIT_COMPRESSED {my($mode) = @_; return(((($mode & SST_TFORMAT) == SST_DXT2) ||
						  (($mode & SST_TFORMAT) == SST_DXT4)) &&
						 ($mode & SST_COMPRESSED_TEXTURES));}
sub SST_T8BIT	{ my($mode) = @_; return (($mode&$TEXTUREMODE{SST_TFORMAT})<$TEXTUREMODE{SST_ARGB8332}); }
sub SST_T16BIT  { my($mode) = @_; return ((($mode & SST_TFORMAT)>=SST_ARGB8332) && (($mode & SST_TFORMAT)<=SST_AP88) && 
					  (($mode & SST_COMPRESSED_TEXTURES) == 0));}
sub SST_T32BIT  { my($mode) = @_; return ((($mode & SST_TFORMAT) == SST_ARGB8888) &&
					  (($mode & SST_COMPRESSED_TEXTURES) == 0));}

$TEXTUREMODE{SST_TFORMAT_SHIFT} =        8;
$TEXTUREMODE{SST_TFORMAT} =              (0xF<<$TEXTUREMODE{SST_TFORMAT_SHIFT});

#Here are the non-compressed formats  (when textureMode::SST_COMPRESSED_TEXTURES=0)
       $SST_TFORMAT{SST_RGB332} =               (0<<$TEXTUREMODE{SST_TFORMAT_SHIFT});
       $SST_TFORMAT{SST_YIQ422} =               (1<<$TEXTUREMODE{SST_TFORMAT_SHIFT});
       $SST_TFORMAT{SST_A8} =                   (2<<$TEXTUREMODE{SST_TFORMAT_SHIFT});
       $SST_TFORMAT{SST_I8} =                   (3<<$TEXTUREMODE{SST_TFORMAT_SHIFT});
       $SST_TFORMAT{SST_AI44} =                 (4<<$TEXTUREMODE{SST_TFORMAT_SHIFT});
       $SST_TFORMAT{SST_P8} =                   (5<<$TEXTUREMODE{SST_TFORMAT_SHIFT});
       $SST_TFORMAT{SST_P8_ARGB6666} =          (6<<$TEXTUREMODE{SST_TFORMAT_SHIFT});
       $SST_TFORMAT{SST_ARGB8332} =             (8<<$TEXTUREMODE{SST_TFORMAT_SHIFT});
       $SST_TFORMAT{SST_AYIQ8422} =             (9<<$TEXTUREMODE{SST_TFORMAT_SHIFT});
       $SST_TFORMAT{SST_RGB565} =               (10<<$TEXTUREMODE{SST_TFORMAT_SHIFT});
       $SST_TFORMAT{SST_ARGB1555} =             (11<<$TEXTUREMODE{SST_TFORMAT_SHIFT});
       $SST_TFORMAT{SST_ARGB4444} =             (12<<$TEXTUREMODE{SST_TFORMAT_SHIFT});
       $SST_TFORMAT{SST_AI88} =                 (13<<$TEXTUREMODE{SST_TFORMAT_SHIFT});
       $SST_TFORMAT{SST_AP88} =                 (14<<$TEXTUREMODE{SST_TFORMAT_SHIFT});
       $SST_TFORMAT{SST_ARGB8888} =             (15<<$TEXTUREMODE{SST_TFORMAT_SHIFT});

#Here are the compressed formats      (when textureMode::SST_COMPRESSED_TEXTURES=1)
$SST_TFORMAT{SST_3DFX_COMPRESSED}=     (0<<SST_TFORMAT_SHIFT);
$SST_TFORMAT{SST_FXT1}=                (0<<SST_TFORMAT_SHIFT);
$SST_TFORMAT{SST_DXT1}=                (1<<SST_TFORMAT_SHIFT);   #This is 4 bits
$SST_TFORMAT{SST_DXT2}=                (2<<SST_TFORMAT_SHIFT);   #This is 8 bits  (pre mult alpha)
$SST_TFORMAT{SST_DXT3}=                (2<<SST_TFORMAT_SHIFT);   #This is 8 bits
$SST_TFORMAT{SST_DXT4}=                (3<<SST_TFORMAT_SHIFT);   #This is 8 bits  (pre mult alpha)
$SST_TFORMAT{SST_DXT5}=                (3<<SST_TFORMAT_SHIFT);   #This is 8 bits

#4 bit per texel compressed textures have 8x4 microtiles (DXT1 has 2 side by side 4x4 microtiles)
$h3def{SST_MICROTILE_LOG2_WIDTH4}=           3;
$h3def{SST_MICROTILE_LOG2_HEIGHT4}=          2;
$h3def{SST_COMPRESSED_HORIZONTAL_TEXELS_PER_TILE4}=  64;
$h3def{SST_COMPRESSED_VERTICAL_TEXELS_PER_TILE4}=    128;

#8 bit per texel compressed textures have 4x4 microtiles
$h3def{SST_MICROTILE_LOG2_WIDTH8}=           2;
$h3def{SST_MICROTILE_LOG2_HEIGHT8}=          2;
$h3def{SST_COMPRESSED_HORIZONTAL_TEXELS_PER_TILE8}=  32;
$h3def{SST_COMPRESSED_VERTICAL_TEXELS_PER_TILE8}=    128;


$TEXTUREMODE{SST_TC_ZERO_OTHER} =        BIT(12);
$TEXTUREMODE{SST_TC_SUB_CLOCAL} =        BIT(13);
$TEXTUREMODE{SST_TC_MSELECT_SHIFT} =     14;
$TEXTUREMODE{SST_TC_MSELECT} =           (0x7<<$TEXTUREMODE{SST_TC_MSELECT_SHIFT});
       $SST_TC_MSELECT{SST_TC_MONE} =              (0<<$TEXTUREMODE{SST_TC_MSELECT_SHIFT});
       $SST_TC_MSELECT{SST_TC_MCLOCAL} =           (1<<$TEXTUREMODE{SST_TC_MSELECT_SHIFT});
       $SST_TC_MSELECT{SST_TC_MAOTHER} =           (2<<$TEXTUREMODE{SST_TC_MSELECT_SHIFT});
       $SST_TC_MSELECT{SST_TC_MALOCAL} =           (3<<$TEXTUREMODE{SST_TC_MSELECT_SHIFT});
       $SST_TC_MSELECT{SST_TC_MLOD} =              (4<<$TEXTUREMODE{SST_TC_MSELECT_SHIFT});
       $SST_TC_MSELECT{SST_TC_MLODFRAC} =          (5<<$TEXTUREMODE{SST_TC_MSELECT_SHIFT});
       $SST_TC_MSELECT{SST_TC_MONE6} =             (6<<$TEXTUREMODE{SST_TC_MSELECT_SHIFT});
       $SST_TC_MSELECT{SST_TC_MCMSELECT7} =        (7<<$TEXTUREMODE{SST_TC_MSELECT_SHIFT});
$TEXTUREMODE{SST_TC_REVERSE_BLEND} =     BIT(17);
$TEXTUREMODE{SST_TC_ADD_CLOCAL} =        BIT(18);
$TEXTUREMODE{SST_TC_ADD_ALOCAL} =        BIT(19);
$TEXTUREMODE{SST_TC_INVERT_OUTPUT} =     BIT(20);
$TEXTUREMODE{SST_TCA_ZERO_OTHER} =       BIT(21);
$TEXTUREMODE{SST_TCA_SUB_CLOCAL} =       BIT(22);
$TEXTUREMODE{SST_TCA_MSELECT_SHIFT} =    23;
$TEXTUREMODE{SST_TCA_MSELECT} =          (0x7<<$TEXTUREMODE{SST_TCA_MSELECT_SHIFT});
       $SST_TCA_MSELECT{SST_TCA_MONE} =             (0<<$TEXTUREMODE{SST_TCA_MSELECT_SHIFT});
       $SST_TCA_MSELECT{SST_TCA_MCLOCAL} =          (1<<$TEXTUREMODE{SST_TCA_MSELECT_SHIFT});
       $SST_TCA_MSELECT{SST_TCA_MAOTHER} =          (2<<$TEXTUREMODE{SST_TCA_MSELECT_SHIFT});
       $SST_TCA_MSELECT{SST_TCA_MALOCAL} =          (3<<$TEXTUREMODE{SST_TCA_MSELECT_SHIFT});
       $SST_TCA_MSELECT{SST_TCA_MLOD} =             (4<<$TEXTUREMODE{SST_TCA_MSELECT_SHIFT});
       $SST_TCA_MSELECT{SST_TCA_MLODFRAC} =         (5<<$TEXTUREMODE{SST_TCA_MSELECT_SHIFT});
       $SST_TCA_MSELECT{SST_TCA_MAITER} =           (6<<$TEXTUREMODE{SST_TCA_MSELECT_SHIFT});
       $SST_TCA_MSELECT{SST_TCA_MCR} =              (7<<$TEXTUREMODE{SST_TCA_MSELECT_SHIFT});
$TEXTUREMODE{SST_TCA_REVERSE_BLEND} =    BIT(26);
$TEXTUREMODE{SST_TCA_ADD_CLOCAL} =       BIT(27);
$TEXTUREMODE{SST_TCA_ADD_ALOCAL} =       BIT(28);
$TEXTUREMODE{SST_TCA_INVERT_OUTPUT} =    BIT(29);
$TEXTUREMODE{SST_TRILINEAR} =            BIT(30);
$TEXTUREMODE{SST_COMPRESSED_TEXTURES} =  BIT(31);

# here are some abstract constants that most people will be using
# TC stands for Texture Combine (RGB channels)
# TCA stands for Texture Combine Alpha (just the Alpha channel)
# BLEND_ATT blends on the Alpha channel of the color passed from TMU to TMU
# BLEND_ALOCAL blends on the Alpha channel of the local texture color
# NOTE: and of the BLEND modes can be modified using SST_TC_REVERSE_BLEND
#       and/or SST_TCA_REVERSE_BLEND, these modifiers reverse the "polarity"
#       of the blend

addToDefs(\%TEXTUREMODE);      # load defn's into %h3def for use below
addToDefs(\%SST_TC_MSELECT);
addToDefs(\%SST_TCA_MSELECT);

$h3def{SST_TC_MZERO} =   ($h3def{SST_TC_MONE} | $h3def{SST_TC_REVERSE_BLEND});;
$h3def{SST_TCA_MZERO} =  ($h3def{SST_TCA_MONE} | $h3def{SST_TCA_REVERSE_BLEND});;
		  
$h3def{SST_TCOMBINE_SHIFT} =       12;       # RGB COMBINE MODES
$h3def{SST_TCOMBINE} =             (0x1FF<<$h3def{SST_TCOMBINE_SHIFT});
       $h3def{SST_TC_REPLACE} =           ($h3def{SST_TC_ZERO_OTHER} | $h3def{SST_TC_ADD_CLOCAL});
       $h3def{SST_TC_PASS} =              ($h3def{SST_TC_MONE});
       $h3def{SST_TC_ADD} =               ($h3def{SST_TC_MONE}  | $h3def{SST_TC_ADD_CLOCAL});
       $h3def{SST_TC_SUB} =               ($h3def{SST_TC_SUB_CLOCAL} | $h3def{SST_TC_MONE});
       $h3def{SST_TC_MULT} =              ($h3def{SST_TC_MCLOCAL} | $h3def{SST_TC_REVERSE_BLEND});
       $h3def{SST_TC_ZERO} =              ($h3def{SST_TC_MZERO});
       $h3def{SST_TC_ONE} =               ($h3def{SST_TC_MZERO} | $h3def{SST_TC_INVERT_OUTPUT});
       $h3def{SST_TC_BLEND} =      	  ($h3def{SST_TC_SUB_CLOCAL} | $h3def{SST_TC_ADD_CLOCAL});
       $h3def{SST_TC_BLEND_LOD} =         ($h3def{SST_TC_BLEND} | $h3def{SST_TC_MLOD});
       $h3def{SST_TC_BLEND_LODFRAC} =     ($h3def{SST_TC_BLEND} | $h3def{SST_TC_MLODFRAC});
       $h3def{SST_TC_BLEND_ATT} =         ($h3def{SST_TC_BLEND} | $h3def{SST_TC_MAOTHER});
       $h3def{SST_TC_BLEND_ALOCAL} =      ($h3def{SST_TC_BLEND} | $h3def{SST_TC_MALOCAL});
$h3def{SST_TACOMBINE_SHIFT} =      21;      # ALPHA COMBINE MODES
$h3def{SST_TACOMBINE} =            (0x1FF<<$h3def{SST_TACOMBINE_SHIFT});
       $h3def{SST_TCA_REPLACE} =          ($h3def{SST_TCA_ZERO_OTHER} | $h3def{SST_TCA_ADD_CLOCAL});
       $h3def{SST_TCA_PASS} =             ($h3def{SST_TCA_MONE});
       $h3def{SST_TCA_ADD} =              ($h3def{SST_TCA_MONE}  | $h3def{SST_TCA_ADD_CLOCAL});
       $h3def{SST_TCA_SUB} =              ($h3def{SST_TCA_SUB_CLOCAL} | $h3def{SST_TCA_MONE});
       $h3def{SST_TCA_MULT} =             ($h3def{SST_TCA_MCLOCAL} | $h3def{SST_TCA_REVERSE_BLEND});
       $h3def{SST_TCA_ONE} =              ($h3def{SST_TCA_MZERO} | $h3def{SST_TCA_INVERT_OUTPUT});
       $h3def{SST_TCA_ZERO} =             ($h3def{SST_TCA_MZERO});
       $h3def{SST_TCA_BLEND} =            ($h3def{SST_TCA_SUB_CLOCAL} | $h3def{SST_TCA_ADD_CLOCAL});
       $h3def{SST_TCA_BLEND_LOD} =        ($h3def{SST_TCA_BLEND} | $h3def{SST_TCA_MLOD});
       $h3def{SST_TCA_BLEND_LODFRAC} =    ($h3def{SST_TCA_BLEND} | $h3def{SST_TCA_MLODFRAC});
       $h3def{SST_TCA_BLEND_ATT} =        ($h3def{SST_TCA_BLEND} | $h3def{SST_TCA_MAOTHER});
       $h3def{SST_TCA_BLEND_ALOCAL} =     ($h3def{SST_TCA_BLEND} | $h3def{SST_TCA_MALOCAL});
sub SST_TMU_ACTIVE	{ my($texMode) = @_; 
			  return (($texMode & ($h3def{SST_TCOMBINE}|$h3def{SST_TACOMBINE})) 
				   != ($h3def{SST_TC_PASS}|$h3def{SST_TCA_PASS}));
			}
sub SST_TREX_ACTIVE	{ my($texMode) = @_; return SST_TMU_ACTIVE($texMode); }

#----------------- SST combineMode bits ---------------------
#FBI combineMode interpretation
$COMBINEMODE{SST_CM_CC_OTHERSELECT_SHIFT}= 0;
$COMBINEMODE{SST_CM_CC_OTHERSELECT}=          (7<<$COMBINEMODE{SST_CM_CC_OTHERSELECT_SHIFT});
  $SST_CM_CC_OTHERSELECT{SST_CM_CC_OTHERSELECT_IRGB}=     (0<<$COMBINEMODE{SST_CM_CC_OTHERSELECT_SHIFT});   #Iterated RGB
  $SST_CM_CC_OTHERSELECT{SST_CM_CC_OTHERSELECT_TRGB}=     (1<<$COMBINEMODE{SST_CM_CC_OTHERSELECT_SHIFT});   #Texture RGB
  $SST_CM_CC_OTHERSELECT{SST_CM_CC_OTHERSELECT_C1_RGB}=   (2<<$COMBINEMODE{SST_CM_CC_OTHERSELECT_SHIFT});   #Color 1 RGB
  $SST_CM_CC_OTHERSELECT{SST_CM_CC_OTHERSELECT_LFB_RGB}=  (3<<$COMBINEMODE{SST_CM_CC_OTHERSELECT_SHIFT});   #LFB RGB
  $SST_CM_CC_OTHERSELECT{SST_CM_CC_OTHERSELECT_IA}=       (4<<$COMBINEMODE{SST_CM_CC_OTHERSELECT_SHIFT});   #Iterated Alpha
  $SST_CM_CC_OTHERSELECT{SST_CM_CC_OTHERSELECT_TA}=       (5<<$COMBINEMODE{SST_CM_CC_OTHERSELECT_SHIFT});   #Texture Alpha
  $SST_CM_CC_OTHERSELECT{SST_CM_CC_OTHERSELECT_C1_A}=     (6<<$COMBINEMODE{SST_CM_CC_OTHERSELECT_SHIFT});   #Color 1 Alpha
  $SST_CM_CC_OTHERSELECT{SST_CM_CC_OTHERSELECT_ZERO}=     (7<<$COMBINEMODE{SST_CM_CC_OTHERSELECT_SHIFT});
$COMBINEMODE{SST_CM_CC_LOCALSELECT_SHIFT}= 3;
$COMBINEMODE{SST_CM_CC_LOCALSELECT} =         (7<<$COMBINEMODE{SST_CM_CC_LOCALSELECT_SHIFT});
  $SST_CM_CC_LOCALSELECT{SST_CM_CC_LOCALSELECT_IRGB}=     (0<<$COMBINEMODE{SST_CM_CC_LOCALSELECT_SHIFT});   #Iterated RGB
  $SST_CM_CC_LOCALSELECT{SST_CM_CC_LOCALSELECT_C0_RGB}=   (1<<$COMBINEMODE{SST_CM_CC_LOCALSELECT_SHIFT});   #Color 0 RGB
  $SST_CM_CC_LOCALSELECT{SST_CM_CC_LOCALSELECT_TRGB}=     (2<<$COMBINEMODE{SST_CM_CC_LOCALSELECT_SHIFT});   #Texture RGB   
  $SST_CM_CC_LOCALSELECT{SST_CM_CC_LOCALSELECT_IA}=       (3<<$COMBINEMODE{SST_CM_CC_LOCALSELECT_SHIFT});   #Iterated Alpha
  $SST_CM_CC_LOCALSELECT{SST_CM_CC_LOCALSELECT_C0_A}=     (4<<$COMBINEMODE{SST_CM_CC_LOCALSELECT_SHIFT});   #Color 0 Alpha
  $SST_CM_CC_LOCALSELECT{SST_CM_CC_LOCALSELECT_TA}=       (5<<$COMBINEMODE{SST_CM_CC_LOCALSELECT_SHIFT});   #Texture Alpha
  $SST_CM_CC_LOCALSELECT{SST_CM_CC_LOCALSELECT_ZERO}=     (6<<$COMBINEMODE{SST_CM_CC_LOCALSELECT_SHIFT});   
  $SST_CM_CC_LOCALSELECT{SST_CM_CC_LOCALSELECT_ZERO7}=    (7<<$COMBINEMODE{SST_CM_CC_LOCALSELECT_SHIFT});   
$COMBINEMODE{SST_CM_CC_MSELECT_7_SHIFT}= 6;
$COMBINEMODE{SST_CM_CC_MSELECT_7}=            (3<<$COMBINEMODE{SST_CM_CC_MSELECT_7_SHIFT});
  $SST_CM_CC_MSELECT_7{SST_CM_CC_MSELECT_7_IRGB}=       (0<<$COMBINEMODE{SST_CM_CC_MSELECT_7_SHIFT});   #Iterated RGB
  $SST_CM_CC_MSELECT_7{SST_CM_CC_MSELECT_7_C1_RGB}=     (1<<$COMBINEMODE{SST_CM_CC_MSELECT_7_SHIFT});   #Color 1 RGB
  $SST_CM_CC_MSELECT_7{SST_CM_CC_MSELECT_7_IA}=         (2<<$COMBINEMODE{SST_CM_CC_MSELECT_7_SHIFT});   #Iterated Alpha
  $SST_CM_CC_MSELECT_7{SST_CM_CC_MSELECT_7_C1_A}=       (3<<$COMBINEMODE{SST_CM_CC_MSELECT_7_SHIFT});   #Color 1 Alpha
$COMBINEMODE{SST_CM_CC_INVERT_OTHER_SHIFT}= 9;
$COMBINEMODE{SST_CM_CC_INVERT_OTHER}=                 (3<<$COMBINEMODE{SST_CM_CC_INVERT_OTHER_SHIFT});
  $SST_CM_CC_INVERT_OTHER{SST_CM_CC_INVERT_OTHER_X}=               (0<<$COMBINEMODE{SST_CM_CC_INVERT_OTHER_SHIFT});
  $SST_CM_CC_INVERT_OTHER{SST_CM_CC_INVERT_OTHER_ZERO_MINUS_X}=    (1<<$COMBINEMODE{SST_CM_CC_INVERT_OTHER_SHIFT});
  $SST_CM_CC_INVERT_OTHER{SST_CM_CC_INVERT_OTHER_ONE_MINUS_X}=     (2<<$COMBINEMODE{SST_CM_CC_INVERT_OTHER_SHIFT});
  $SST_CM_CC_INVERT_OTHER{SST_CM_CC_INVERT_OTHER_X_MINUS_HALF}=    (3<<$COMBINEMODE{SST_CM_CC_INVERT_OTHER_SHIFT});
$COMBINEMODE{SST_CM_CC_INVERT_LOCAL_SHIFT}= 11;
$COMBINEMODE{SST_CM_CC_INVERT_LOCAL}=                 (3<<$COMBINEMODE{SST_CM_CC_INVERT_LOCAL_SHIFT});
  $SST_CM_CC_INVERT_LOCAL{SST_CM_CC_INVERT_LOCAL_ZERO_MINUS_X}=    (0<<$COMBINEMODE{SST_CM_CC_INVERT_LOCAL_SHIFT});
  $SST_CM_CC_INVERT_LOCAL{SST_CM_CC_INVERT_LOCAL_X}=               (1<<$COMBINEMODE{SST_CM_CC_INVERT_LOCAL_SHIFT});
  $SST_CM_CC_INVERT_LOCAL{SST_CM_CC_INVERT_LOCAL_ONE_MINUS_X}=     (2<<$COMBINEMODE{SST_CM_CC_INVERT_LOCAL_SHIFT});
  $SST_CM_CC_INVERT_LOCAL{SST_CM_CC_INVERT_LOCAL_X_MINUS_HALF}=    (3<<$COMBINEMODE{SST_CM_CC_INVERT_LOCAL_SHIFT});
$COMBINEMODE{SST_CM_CC_INVERT_ADD_LOCAL}=           BIT(13);
$COMBINEMODE{SST_CM_CC_OUTSHIFT_SHIFT}= 14;
$COMBINEMODE{SST_CM_CC_OUTSHIFT}=            (3<<$COMBINEMODE{SST_CM_CC_OUTSHIFT_SHIFT});
  $SST_CM_CC_OUTSHIFT{SST_CM_CC_OUTSHIFT_1X}=         (0<<$COMBINEMODE{SST_CM_CC_OUTSHIFT_SHIFT});
  $SST_CM_CC_OUTSHIFT{SST_CM_CC_OUTSHIFT_2X}=         (1<<$COMBINEMODE{SST_CM_CC_OUTSHIFT_SHIFT});
  $SST_CM_CC_OUTSHIFT{SST_CM_CC_OUTSHIFT_4X}=         (2<<$COMBINEMODE{SST_CM_CC_OUTSHIFT_SHIFT});
$COMBINEMODE{SST_CM_CCA_OTHERSELECT_SHIFT}= 16;
$COMBINEMODE{SST_CM_CCA_OTHERSELECT}=        (3<<$COMBINEMODE{SST_CM_CCA_OTHERSELECT_SHIFT});
  $SST_CM_CCA_OTHERSELECT{SST_CM_CCA_OTHERSELECT_IA}=     (0<<$COMBINEMODE{SST_CM_CCA_OTHERSELECT_SHIFT}); #Iterated alpha
  $SST_CM_CCA_OTHERSELECT{SST_CM_CCA_OTHERSELECT_TA}=     (1<<$COMBINEMODE{SST_CM_CCA_OTHERSELECT_SHIFT}); #Texture alpha
  $SST_CM_CCA_OTHERSELECT{SST_CM_CCA_OTHERSELECT_C1_A}=   (2<<$COMBINEMODE{SST_CM_CCA_OTHERSELECT_SHIFT}); #Color 1 Alpha
  $SST_CM_CCA_OTHERSELECT{SST_CM_CCA_OTHERSELECT_LFB_A}=  (3<<$COMBINEMODE{SST_CM_CCA_OTHERSELECT_SHIFT}); #LFB Alpha
$COMBINEMODE{SST_CM_CCA_LOCALSELECT_SHIFT}= 18;
$COMBINEMODE{SST_CM_CCA_LOCALSELECT}=        (7<<$COMBINEMODE{SST_CM_CCA_LOCALSELECT_SHIFT}); 
  $SST_CM_CCA_LOCALSELECT{SST_CM_CCA_LOCALSELECT_IA}=     (0<<$COMBINEMODE{SST_CM_CCA_LOCALSELECT_SHIFT}); #Iterated Alpha
  $SST_CM_CCA_LOCALSELECT{SST_CM_CCA_LOCALSELECT_C0_A}=   (1<<$COMBINEMODE{SST_CM_CCA_LOCALSELECT_SHIFT}); #Color 0 Alpha
  $SST_CM_CCA_LOCALSELECT{SST_CM_CCA_LOCALSELECT_IZ}=     (2<<$COMBINEMODE{SST_CM_CCA_LOCALSELECT_SHIFT}); #Iterated Z
  $SST_CM_CCA_LOCALSELECT{SST_CM_CCA_LOCALSELECT_IW}=     (3<<$COMBINEMODE{SST_CM_CCA_LOCALSELECT_SHIFT}); #Iterated W
  $SST_CM_CCA_LOCALSELECT{SST_CM_CCA_LOCALSELECT_TA}=     (4<<$COMBINEMODE{SST_CM_CCA_LOCALSELECT_SHIFT}); #Texture Alpha
  $SST_CM_CCA_LOCALSELECT{SST_CM_CCA_LOCALSELECT_ZERO}=   (5<<$COMBINEMODE{SST_CM_CCA_LOCALSELECT_SHIFT}); 
$COMBINEMODE{SST_CM_CCA_INVERT_OTHER_SHIFT}= 21;
$COMBINEMODE{SST_CM_CCA_INVERT_OTHER}=               (3<<$COMBINEMODE{SST_CM_CCA_INVERT_OTHER_SHIFT});
  $SST_CM_CCA_INVERT_OTHER{SST_CM_CCA_INVERT_OTHER_X}=             (0<<$COMBINEMODE{SST_CM_CCA_INVERT_OTHER_SHIFT});
  $SST_CM_CCA_INVERT_OTHER{SST_CM_CCA_INVERT_OTHER_ZERO_MINUS_X}=  (1<<$COMBINEMODE{SST_CM_CCA_INVERT_OTHER_SHIFT});
  $SST_CM_CCA_INVERT_OTHER{SST_CM_CCA_INVERT_OTHER_ONE_MINUS_X}=   (2<<$COMBINEMODE{SST_CM_CCA_INVERT_OTHER_SHIFT});
  $SST_CM_CCA_INVERT_OTHER{SST_CM_CCA_INVERT_OTHER_X_MINUS_HALF}=  (3<<$COMBINEMODE{SST_CM_CCA_INVERT_OTHER_SHIFT});
$COMBINEMODE{SST_CM_CCA_INVERT_LOCAL_SHIFT}= 23;
$COMBINEMODE{SST_CM_CCA_INVERT_LOCAL}=               (3<<$COMBINEMODE{SST_CM_CCA_INVERT_LOCAL_SHIFT});
  $SST_CM_CCA_INVERT_LOCAL{SST_CM_CCA_INVERT_LOCAL_ZERO_MINUS_X}=  (0<<$COMBINEMODE{SST_CM_CCA_INVERT_LOCAL_SHIFT});
  $SST_CM_CCA_INVERT_LOCAL{SST_CM_CCA_INVERT_LOCAL_X}=             (1<<$COMBINEMODE{SST_CM_CCA_INVERT_LOCAL_SHIFT});
  $SST_CM_CCA_INVERT_LOCAL{SST_CM_CCA_INVERT_LOCAL_ONE_MINUS_X}=   (2<<$COMBINEMODE{SST_CM_CCA_INVERT_LOCAL_SHIFT});
  $SST_CM_CCA_INVERT_LOCAL{SST_CM_CCA_INVERT_LOCAL_X_MINUS_HALF}=  (3<<$COMBINEMODE{SST_CM_CCA_INVERT_LOCAL_SHIFT});
$COMBINEMODE{SST_CM_CCA_INVERT_ADD_LOCAL}=          BIT(25);
$COMBINEMODE{SST_CM_CCA_OUTSHIFT_SHIFT}= 26;
$COMBINEMODE{SST_CM_CCA_OUTSHIFT}=           (3<<$COMBINEMODE{SST_CM_CCA_OUTSHIFT_SHIFT});
  $SST_CM_CCA_OUTSHIFT{SST_CM_CCA_OUTSHIFT_1X}=        (0<<$COMBINEMODE{SST_CM_CCA_OUTSHIFT_SHIFT});
  $SST_CM_CCA_OUTSHIFT{SST_CM_CCA_OUTSHIFT_2X}=        (1<<$COMBINEMODE{SST_CM_CCA_OUTSHIFT_SHIFT});
  $SST_CM_CCA_OUTSHIFT{SST_CM_CCA_OUTSHIFT_4X}=        (2<<$COMBINEMODE{SST_CM_CCA_OUTSHIFT_SHIFT});
$COMBINEMODE{SST_CM_ENABLE_TWO_PIXELS_PER_CLOCK}=   BIT(29);
$COMBINEMODE{SST_CM_DISABLE_CHROMA_SUBSTITUTION}=   BIT(30);
$COMBINEMODE{SST_CM_USE_COMBINE_MODE}=              BIT(31);

#TMU combineMode interpretation
$COMBINEMODE{SST_CM_TC_OTHERSELECT_SHIFT}= 0;
$COMBINEMODE{SST_CM_TC_OTHERSELECT}=             (7<<$COMBINEMODE{SST_CM_TC_OTHERSELECT_SHIFT});
  $SST_CM_TC_OTHERSELECT{SST_CM_TC_OTHERSELECT_OTHER_TRGB}=  (0<<$COMBINEMODE{SST_CM_TC_OTHERSELECT_SHIFT});  #Other texture RGB
  $SST_CM_TC_OTHERSELECT{SST_CM_TC_OTHERSELECT_OTHER_TA}=    (1<<$COMBINEMODE{SST_CM_TC_OTHERSELECT_SHIFT});  #Other texture Alpha
  $SST_CM_TC_OTHERSELECT{SST_CM_TC_OTHERSELECT_LOCAL_TRGB}=  (2<<$COMBINEMODE{SST_CM_TC_OTHERSELECT_SHIFT});  #Local texture RGB
  $SST_CM_TC_OTHERSELECT{SST_CM_TC_OTHERSELECT_LOCAL_TA}=    (3<<$COMBINEMODE{SST_CM_TC_OTHERSELECT_SHIFT});  #Local texture Alpha
  $SST_CM_TC_OTHERSELECT{SST_CM_TC_OTHERSELECT_IRGB}=        (4<<$COMBINEMODE{SST_CM_TC_OTHERSELECT_SHIFT});  #Iterated RGB
  $SST_CM_TC_OTHERSELECT{SST_CM_TC_OTHERSELECT_IA}=          (5<<$COMBINEMODE{SST_CM_TC_OTHERSELECT_SHIFT});  #Iterated Alpha
  $SST_CM_TC_OTHERSELECT{SST_CM_TC_OTHERSELECT_CR_RGB}=      (6<<$COMBINEMODE{SST_CM_TC_OTHERSELECT_SHIFT});  #ChromaRange RGB
  $SST_CM_TC_OTHERSELECT{SST_CM_TC_OTHERSELECT_CR_A}=        (7<<$COMBINEMODE{SST_CM_TC_OTHERSELECT_SHIFT});  #ChromaRange Alpha
$COMBINEMODE{SST_CM_TC_LOCALSELECT_SHIFT}= 3;
$COMBINEMODE{SST_CM_TC_LOCALSELECT}=             (7<<$COMBINEMODE{SST_CM_TC_LOCALSELECT_SHIFT});
  $SST_CM_TC_LOCALSELECT{SST_CM_TC_LOCALSELECT_LOCAL_TRGB}=  (0<<$COMBINEMODE{SST_CM_TC_LOCALSELECT_SHIFT});  #Local texture RGB
  $SST_CM_TC_LOCALSELECT{SST_CM_TC_LOCALSELECT_LOCAL_TA}=    (1<<$COMBINEMODE{SST_CM_TC_LOCALSELECT_SHIFT});  #Local texture Alpha
  $SST_CM_TC_LOCALSELECT{SST_CM_TC_LOCALSELECT_OTHER_TRGB}=  (2<<$COMBINEMODE{SST_CM_TC_LOCALSELECT_SHIFT});  #Other texture RGB
  $SST_CM_TC_LOCALSELECT{SST_CM_TC_LOCALSELECT_OTHER_TA}=    (3<<$COMBINEMODE{SST_CM_TC_LOCALSELECT_SHIFT});  #Other texture Alpha
  $SST_CM_TC_LOCALSELECT{SST_CM_TC_LOCALSELECT_IRGB}=        (4<<$COMBINEMODE{SST_CM_TC_LOCALSELECT_SHIFT});  #Iterated RGB
  $SST_CM_TC_LOCALSELECT{SST_CM_TC_LOCALSELECT_IA}=          (5<<$COMBINEMODE{SST_CM_TC_LOCALSELECT_SHIFT});  #Iterated Alpha
  $SST_CM_TC_LOCALSELECT{SST_CM_TC_LOCALSELECT_CK_RGB}=      (6<<$COMBINEMODE{SST_CM_TC_LOCALSELECT_SHIFT});  #Chroma Key RGB
  $SST_CM_TC_LOCALSELECT{SST_CM_TC_LOCALSELECT_CK_A}=        (7<<$COMBINEMODE{SST_CM_TC_LOCALSELECT_SHIFT});  #Chroma Key Alpha
$COMBINEMODE{SST_CM_TC_MSELECT_7_SHIFT}= 6;
$COMBINEMODE{SST_CM_TC_MSELECT_7}=                (7<<$COMBINEMODE{SST_CM_TC_MSELECT_7_SHIFT});
  $SST_CM_TC_MSELECT_7{SST_CM_TC_MSELECT_7_LOCAL_TRGB}=     (0<<$COMBINEMODE{SST_CM_TC_MSELECT_7_SHIFT});  #Local texture RGB
  $SST_CM_TC_MSELECT_7{SST_CM_TC_MSELECT_7_ZERO}=           (1<<$COMBINEMODE{SST_CM_TC_MSELECT_7_SHIFT}); 
  $SST_CM_TC_MSELECT_7{SST_CM_TC_MSELECT_7_OTHER_TRGB}=     (2<<$COMBINEMODE{SST_CM_TC_MSELECT_7_SHIFT});  #Other texture RGB
  $SST_CM_TC_MSELECT_7{SST_CM_TC_MSELECT_7_ZERO3}=          (3<<$COMBINEMODE{SST_CM_TC_MSELECT_7_SHIFT});
  $SST_CM_TC_MSELECT_7{SST_CM_TC_MSELECT_7_IRGB}=           (4<<$COMBINEMODE{SST_CM_TC_MSELECT_7_SHIFT});  #Iterated RGB
  $SST_CM_TC_MSELECT_7{SST_CM_TC_MSELECT_7_IA}=             (5<<$COMBINEMODE{SST_CM_TC_MSELECT_7_SHIFT});  #Iterated Alpha
  $SST_CM_TC_MSELECT_7{SST_CM_TC_MSELECT_7_CR_RGB}=         (6<<$COMBINEMODE{SST_CM_TC_MSELECT_7_SHIFT});  #Chroma Range RGB
  $SST_CM_TC_MSELECT_7{SST_CM_TC_MSELECT_7_CR_A}=           (7<<$COMBINEMODE{SST_CM_TC_MSELECT_7_SHIFT});  #Chroma Range Alpha
$COMBINEMODE{SST_CM_TC_INVERT_OTHER_SHIFT}= 9;
$COMBINEMODE{SST_CM_TC_INVERT_OTHER}=                (3<<$COMBINEMODE{SST_CM_TC_INVERT_OTHER_SHIFT});
  $SST_CM_TC_INVERT_OTHER{SST_CM_TC_INVERT_OTHER_X}=              (0<<$COMBINEMODE{SST_CM_TC_INVERT_OTHER_SHIFT});
  $SST_CM_TC_INVERT_OTHER{SST_CM_TC_INVERT_OTHER_ZERO_MINUS_X}=   (1<<$COMBINEMODE{SST_CM_TC_INVERT_OTHER_SHIFT});
  $SST_CM_TC_INVERT_OTHER{SST_CM_TC_INVERT_OTHER_ONE_MINUS_X}=    (2<<$COMBINEMODE{SST_CM_TC_INVERT_OTHER_SHIFT});
  $SST_CM_TC_INVERT_OTHER{SST_CM_TC_INVERT_OTHER_X_MINUS_HALF}=   (3<<$COMBINEMODE{SST_CM_TC_INVERT_OTHER_SHIFT});
$COMBINEMODE{SST_CM_TC_INVERT_LOCAL_SHIFT}= 11;
$COMBINEMODE{SST_CM_TC_INVERT_LOCAL}=                (3<<$COMBINEMODE{SST_CM_TC_INVERT_LOCAL_SHIFT});
  $SST_CM_TC_INVERT_LOCAL{SST_CM_TC_INVERT_LOCAL_ZERO_MINUS_X}=   (0<<$COMBINEMODE{SST_CM_TC_INVERT_LOCAL_SHIFT});
  $SST_CM_TC_INVERT_LOCAL{SST_CM_TC_INVERT_LOCAL_X}=              (1<<$COMBINEMODE{SST_CM_TC_INVERT_LOCAL_SHIFT});
  $SST_CM_TC_INVERT_LOCAL{SST_CM_TC_INVERT_LOCAL_ONE_MINUS_X}=    (2<<$COMBINEMODE{SST_CM_TC_INVERT_LOCAL_SHIFT});
  $SST_CM_TC_INVERT_LOCAL{SST_CM_TC_INVERT_LOCAL_X_MINUS_HALF}=   (3<<$COMBINEMODE{SST_CM_TC_INVERT_LOCAL_SHIFT});
$COMBINEMODE{SST_CM_TC_INVERT_ADD_LOCAL}=           BIT(13);
$COMBINEMODE{SST_CM_TC_OUTSHIFT_SHIFT}= 14;
$COMBINEMODE{SST_CM_TC_OUTSHIFT}=            (3<<$COMBINEMODE{SST_CM_TC_OUTSHIFT_SHIFT});
  $SST_CM_TC_OUTSHIFT{SST_CM_TC_OUTSHIFT_1X}=         (0<<$COMBINEMODE{SST_CM_TC_OUTSHIFT_SHIFT});
  $SST_CM_TC_OUTSHIFT{SST_CM_TC_OUTSHIFT_2X}=         (1<<$COMBINEMODE{SST_CM_TC_OUTSHIFT_SHIFT});
  $SST_CM_TC_OUTSHIFT{SST_CM_TC_OUTSHIFT_4X}=         (2<<$COMBINEMODE{SST_CM_TC_OUTSHIFT_SHIFT});
$COMBINEMODE{SST_CM_TCA_OTHERSELECT_SHIFT}= 16;
$COMBINEMODE{SST_CM_TCA_OTHERSELECT}=           (3<<$COMBINEMODE{SST_CM_TCA_OTHERSELECT_SHIFT});
  $SST_CM_TCA_OTHERSELECT{SST_CM_TCA_OTHERSELECT_OTHER_TA}=  (0<<$COMBINEMODE{SST_CM_TCA_OTHERSELECT_SHIFT});  #Other texture alpha
  $SST_CM_TCA_OTHERSELECT{SST_CM_TCA_OTHERSELECT_LOCAL_TA}=  (1<<$COMBINEMODE{SST_CM_TCA_OTHERSELECT_SHIFT});  #Local texture alpha
  $SST_CM_TCA_OTHERSELECT{SST_CM_TCA_OTHERSELECT_IA}=        (2<<$COMBINEMODE{SST_CM_TCA_OTHERSELECT_SHIFT});  #Iterated alpha
  $SST_CM_TCA_OTHERSELECT{SST_CM_TCA_OTHERSELECT_CR_A}=      (3<<$COMBINEMODE{SST_CM_TCA_OTHERSELECT_SHIFT});  #Chroma Range alpha
$COMBINEMODE{SST_CM_TCA_LOCALSELECT_SHIFT}= 18;
$COMBINEMODE{SST_CM_TCA_LOCALSELECT}=           (3<<$COMBINEMODE{SST_CM_TCA_LOCALSELECT_SHIFT});
  $SST_CM_TCA_LOCALSELECT{SST_CM_TCA_LOCALSELECT_LOCAL_TA}=  (0<<$COMBINEMODE{SST_CM_TCA_LOCALSELECT_SHIFT});  #Local texture alpha
  $SST_CM_TCA_LOCALSELECT{SST_CM_TCA_LOCALSELECT_OTHER_TA}=  (1<<$COMBINEMODE{SST_CM_TCA_LOCALSELECT_SHIFT});  #Other texture alpha
  $SST_CM_TCA_LOCALSELECT{SST_CM_TCA_LOCALSELECT_IA}=        (2<<$COMBINEMODE{SST_CM_TCA_LOCALSELECT_SHIFT});  #Iterated alpha
  $SST_CM_TCA_LOCALSELECT{SST_CM_TCA_LOCALSELECT_CK_A}=      (3<<$COMBINEMODE{SST_CM_TCA_LOCALSELECT_SHIFT});  #Chroma key alpha
$COMBINEMODE{SST_CM_TCA_INVERT_OTHER_SHIFT}= 21;
$COMBINEMODE{SST_CM_TCA_INVERT_OTHER}=                (3<<$COMBINEMODE{SST_CM_TCA_INVERT_OTHER_SHIFT});
  $SST_CM_TCA_INVERT_OTHER{SST_CM_TCA_INVERT_OTHER_X}=              (0<<$COMBINEMODE{SST_CM_TCA_INVERT_OTHER_SHIFT});
  $SST_CM_TCA_INVERT_OTHER{SST_CM_TCA_INVERT_OTHER_ZERO_MINUS_X}=   (1<<$COMBINEMODE{SST_CM_TCA_INVERT_OTHER_SHIFT});
  $SST_CM_TCA_INVERT_OTHER{SST_CM_TCA_INVERT_OTHER_ONE_MINUS_X}=    (2<<$COMBINEMODE{SST_CM_TCA_INVERT_OTHER_SHIFT});
  $SST_CM_TCA_INVERT_OTHER{SST_CM_TCA_INVERT_OTHER_X_MINUS_HALF}=   (3<<$COMBINEMODE{SST_CM_TCA_INVERT_OTHER_SHIFT});
$COMBINEMODE{SST_CM_TCA_INVERT_LOCAL_SHIFT}= 23;
$COMBINEMODE{SST_CM_TCA_INVERT_LOCAL}=                (3<<$COMBINEMODE{SST_CM_TCA_INVERT_LOCAL_SHIFT});
  $SST_CM_TCA_INVERT_LOCAL{SST_CM_TCA_INVERT_LOCAL_ZERO_MINUS_X}=   (0<<$COMBINEMODE{SST_CM_TCA_INVERT_LOCAL_SHIFT});
  $SST_CM_TCA_INVERT_LOCAL{SST_CM_TCA_INVERT_LOCAL_X}=              (1<<$COMBINEMODE{SST_CM_TCA_INVERT_LOCAL_SHIFT});
  $SST_CM_TCA_INVERT_LOCAL{SST_CM_TCA_INVERT_LOCAL_ONE_MINUS_X}=    (2<<$COMBINEMODE{SST_CM_TCA_INVERT_LOCAL_SHIFT});
  $SST_CM_TCA_INVERT_LOCAL{SST_CM_TCA_INVERT_LOCAL_X_MINUS_HALF}=   (3<<$COMBINEMODE{SST_CM_TCA_INVERT_LOCAL_SHIFT});
$COMBINEMODE{SST_CM_TCA_INVERT_ADD_LOCAL}=          BIT(25);
$COMBINEMODE{SST_CM_TCA_OUTSHIFT_SHIFT}= 26;
$COMBINEMODE{SST_CM_TCA_OUTSHIFT}=           (3<<$COMBINEMODE{SST_CM_TCA_OUTSHIFT_SHIFT});
  $SST_CM_TCA_OUTSHIFT{SST_CM_TCA_OUTSHIFT_1X}=        (0<<$COMBINEMODE{SST_CM_TCA_OUTSHIFT_SHIFT});
  $SST_CM_TCA_OUTSHIFT{SST_CM_TCA_OUTSHIFT_2X}=        (1<<$COMBINEMODE{SST_CM_TCA_OUTSHIFT_SHIFT});
  $SST_CM_TCA_OUTSHIFT{SST_CM_TCA_OUTSHIFT_4X}=        (2<<$COMBINEMODE{SST_CM_TCA_OUTSHIFT_SHIFT});

#-------------------- SST sliCtrl bits --------------------
$SLICTRL{SST_SLI_CONTROL_RENDER_MASK_SHIFT}=              0;
$SLICTRL{SST_SLI_CONTROL_RENDER_MASK}=                    (0xFF << $SLICTRL{SST_SLI_CONTROL_RENDER_MASK_SHIFT});
$SLICTRL{SST_SLI_CONTROL_COMPARE_MASK_SHIFT}=             8;
$SLICTRL{SST_SLI_CONTROL_COMPARE_MASK}=                   (0xFF << $SLICTRL{SST_SLI_CONTROL_COMPARE_MASK_SHIFT});
$SLICTRL{SST_SLI_CONTROL_SCAN_MASK_SHIFT}=                16;
$SLICTRL{SST_SLI_CONTROL_SCAN_MASK}=                      (0xFF << $SLICTRL{SST_SLI_CONTROL_SCAN_MASK_SHIFT});
$SLICTRL{SST_SLI_CONTROL_LOG2_CHIP_COUNT_SHIFT}=          24;
$SLICTRL{SST_SLI_CONTROL_LOG2_CHIP_COUNT}=                (0x3 << $SLICTRL{SST_SLI_CONTROL_LOG2_CHIP_COUNT_SHIFT});
$SLICTRL{SST_SLI_CONTROL_SLI_ENABLE}=                     BIT(26);


#-------------------- SST aaCtrl bits --------------------
# The default values came from Gary Tarolli.
# Jitter values are in 3.4 format.
#  2/16 = 0.125,  4/16 = 0.250,  6/16 = 0.375,  8/16 = 0.500
# 10/16 = 0.625, 12/16 = 0.750, 14/16 = 0.875, 16/16 = 1.000
$AACTRL{SST_AA_CONTROL_PRIMARY_X_OFFSET_SHIFT}=          0;
$AACTRL{SST_AA_CONTROL_PRIMARY_X_OFFSET}=                (0x7F << $AACTRL{SST_AA_CONTROL_PRIMARY_X_OFFSET_SHIFT});
$AACTRL{SST_AA_CONTROL_PRIMARY_Y_OFFSET_SHIFT}=          7;
$AACTRL{SST_AA_CONTROL_PRIMARY_Y_OFFSET}=                (0x7F << $AACTRL{SST_AA_CONTROL_PRIMARY_Y_OFFSET_SHIFT});
$AACTRL{SST_AA_CONTROL_SECONDARY_X_OFFSET_SHIFT}=        14;
$AACTRL{SST_AA_CONTROL_SECONDARY_X_OFFSET}=              (0x7F << $AACTRL{SST_AA_CONTROL_SECONDARY_X_OFFSET_SHIFT});
$AACTRL{SST_AA_CONTROL_SECONDARY_Y_OFFSET_SHIFT}=        21;
$AACTRL{SST_AA_CONTROL_SECONDARY_Y_OFFSET}=              (0x7F << $AACTRL{SST_AA_CONTROL_SECONDARY_Y_OFFSET_SHIFT});
$AACTRL{SST_AA_CONTROL_AA_ENABLE}=                       BIT(28);

#-------------------- SST chipMask bits --------------------
$CHIPMASK{SST_CHIP_MASK_CHIP_0_ENABLE}=                    BIT(0);
$CHIPMASK{SST_CHIP_MASK_CHIP_1_ENABLE}=                    BIT(1);
$CHIPMASK{SST_CHIP_MASK_CHIP_2_ENABLE}=                    BIT(2);
$CHIPMASK{SST_CHIP_MASK_CHIP_3_ENABLE}=                    BIT(3);
$CHIPMASK{SST_CHIP_MASK_CHIP_4_ENABLE}=                    BIT(4);
$CHIPMASK{SST_CHIP_MASK_CHIP_5_ENABLE}=                    BIT(5);
$CHIPMASK{SST_CHIP_MASK_CHIP_6_ENABLE}=                    BIT(6);
$CHIPMASK{SST_CHIP_MASK_CHIP_7_ENABLE}=                    BIT(7);
$CHIPMASK{SST_CHIP_MASK_CHIP_8_ENABLE}=                    BIT(8);
$CHIPMASK{SST_CHIP_MASK_CHIP_9_ENABLE}=                    BIT(9);
$CHIPMASK{SST_CHIP_MASK_CHIP_10_ENABLE}=                    BIT(10);
$CHIPMASK{SST_CHIP_MASK_CHIP_11_ENABLE}=                    BIT(11);
$CHIPMASK{SST_CHIP_MASK_CHIP_12_ENABLE}=                    BIT(12);
$CHIPMASK{SST_CHIP_MASK_CHIP_13_ENABLE}=                    BIT(13);
$CHIPMASK{SST_CHIP_MASK_CHIP_14_ENABLE}=                    BIT(14);
$CHIPMASK{SST_CHIP_MASK_CHIP_15_ENABLE}=                    BIT(15);
$CHIPMASK{SST_CHIP_MASK_CHIP_16_ENABLE}=                    BIT(16);
$CHIPMASK{SST_CHIP_MASK_CHIP_17_ENABLE}=                    BIT(17);
$CHIPMASK{SST_CHIP_MASK_CHIP_18_ENABLE}=                    BIT(18);
$CHIPMASK{SST_CHIP_MASK_CHIP_19_ENABLE}=                    BIT(19);
$CHIPMASK{SST_CHIP_MASK_CHIP_20_ENABLE}=                    BIT(20);
$CHIPMASK{SST_CHIP_MASK_CHIP_21_ENABLE}=                    BIT(21);
$CHIPMASK{SST_CHIP_MASK_CHIP_22_ENABLE}=                    BIT(22);
$CHIPMASK{SST_CHIP_MASK_CHIP_23_ENABLE}=                    BIT(23);
$CHIPMASK{SST_CHIP_MASK_CHIP_24_ENABLE}=                    BIT(24);
$CHIPMASK{SST_CHIP_MASK_CHIP_25_ENABLE}=                    BIT(25);
$CHIPMASK{SST_CHIP_MASK_CHIP_26_ENABLE}=                    BIT(26);
$CHIPMASK{SST_CHIP_MASK_CHIP_27_ENABLE}=                    BIT(27);
$CHIPMASK{SST_CHIP_MASK_CHIP_28_ENABLE}=                    BIT(28);
$CHIPMASK{SST_CHIP_MASK_CHIP_29_ENABLE}=                    BIT(29);
$CHIPMASK{SST_CHIP_MASK_CHIP_30_ENABLE}=                    BIT(30);
$CHIPMASK{SST_CHIP_MASK_CHIP_31_ENABLE}=                    BIT(31);
$CHIPMASK{SST_CHIP_MASK_ALL_CHIPS}=                         0xFFFFFFFF;


#----------------- SST tLOD bits -----------------------------
$TLOD{SST_LODMIN_SHIFT} =         0;       # 4.2 format
$TLOD{SST_LODMIN} =               (0x3F<<$TLOD{SST_LODMIN_SHIFT});
$TLOD{SST_LODMAX_SHIFT} =         6;
$TLOD{SST_LODMAX} =               (0x3F<<$TLOD{SST_LODMAX_SHIFT});
$TLOD{SST_LODBIAS_SHIFT} =        12;
$TLOD{SST_LODBIAS} =              (0x3F<<$TLOD{SST_LODBIAS_SHIFT});
$TLOD{SST_LOD_ODD} =              BIT(18);
$TLOD{SST_LOD_TSPLIT} =           BIT(19);
$TLOD{SST_LOD_S_IS_WIDER} =       BIT(20);
$TLOD{SST_LOD_ASPECT_SHIFT} =     21;
$TLOD{SST_LOD_ASPECT} =           (0x3<<$TLOD{SST_LOD_ASPECT_SHIFT});
$TLOD{SST_LOD_ZEROFRAC} =         BIT(23);
$TLOD{SST_TMULTIBASEADDR} =       BIT(24);
$TLOD{SST_TDATA_SWIZZLE} =        BIT(25);
$TLOD{SST_TDATA_SWAP} =           BIT(26);
$TLOD{SST_TDIRECT_WRITE} =        BIT(27);
$TLOD{SST_TMIRRORS} = 		BIT(28);
$TLOD{SST_TMIRRORT} = 		BIT(29);
$TLOD{SST_TBIG} = 		BIT(30);

#----------------- SST tDetail bits --------------------------
$TDETAIL{SST_DETAIL_MAX_SHIFT} =     0;       # 8.0 format
$TDETAIL{SST_DETAIL_MAX} =           (0xFF<<$TDETAIL{SST_DETAIL_MAX_SHIFT});
$TDETAIL{SST_DETAIL_BIAS_SHIFT} =    8;       # 6.0 format signed
$TDETAIL{SST_DETAIL_BIAS} =          (0x3F<<$TDETAIL{SST_DETAIL_BIAS_SHIFT});
$TDETAIL{SST_DETAIL_SCALE_SHIFT} =   14;      # 3.0 format
$TDETAIL{SST_DETAIL_SCALE} =         (0x7<<$TDETAIL{SST_DETAIL_SCALE_SHIFT});
$TDETAIL{SST_TMINFILTER_RGB} = 	BIT(17);
$TDETAIL{SST_TMAGFILTER_RGB} = 	BIT(18);
$TDETAIL{SST_TMINFILTER_A} = 	BIT(19);
$TDETAIL{SST_TMAGFILTER_A} = 	BIT(20);
$TDETAIL{SST_TFILTER_SEPARATE} = 	BIT(21);

#----------------- SST fbzMode bits --------------------------
$FBZMODE{SST_ENRECTCLIP} =           BIT(0);
$FBZMODE{SST_ENCHROMAKEY} =          BIT(1);
$FBZMODE{SST_ENSTIPPLE} =            BIT(2);
$FBZMODE{SST_WBUFFER} =              BIT(3);
$FBZMODE{SST_ENDEPTHBUFFER} =        BIT(4);
$FBZMODE{SST_ZFUNC_SHIFT} =  5;
$FBZMODE{SST_ZFUNC} =        ( 0x7 << $FBZMODE{SST_ZFUNC_SHIFT} );
       $SST_ZFUNC{SST_ZFUNC_LT} =     BIT(5);
       $SST_ZFUNC{SST_ZFUNC_EQ} =     BIT(6);
       $SST_ZFUNC{SST_ZFUNC_GT} =     BIT(7);
$FBZMODE{SST_ENDITHER} =             BIT(8);
$FBZMODE{SST_RGBWRMASK} =            BIT(9);
$FBZMODE{SST_ZAWRMASK} =             BIT(10);
$FBZMODE{SST_DITHER2x2} =            BIT(11);
$FBZMODE{SST_ENSTIPPLEPATTERN} =     BIT(12);
$FBZMODE{SST_ENALPHAMASK} =          BIT(13);
$FBZMODE{SST_ENZBIAS} =              BIT(16);
$FBZMODE{SST_YORIGIN} =              BIT(17);
$FBZMODE{SST_ENALPHABUFFER} =        BIT(18);
$FBZMODE{SST_ENDITHERSUBTRACT} =     BIT(19);
$FBZMODE{SST_ABLEND_DITHER_SUB_EN} =  $FBZMODE{SST_ENDITHERSUBTRACT};;
$FBZMODE{SST_ENDITHERSUBTRACT} =     BIT(19);
$FBZMODE{SST_ZCOMPARE_TO_ZACOLOR} =  BIT(20);
$FBZMODE{SST_DEPTH_FLOAT_SEL} =      BIT(21);

#----------------- SST fbzColorPath bits ---------------------------
$FBZCOLORPATH{SST_RGBSELECT_SHIFT} =      0;
$FBZCOLORPATH{SST_RGBSELECT} =            (0x3<<$FBZCOLORPATH{SST_RGBSELECT_SHIFT});
       $SST_RGBSELECT{SST_RGBSEL_RGBA} =          (0<<$FBZCOLORPATH{SST_RGBSELECT_SHIFT});
       $SST_RGBSELECT{SST_RGBSEL_TMUOUT} =        (1<<$FBZCOLORPATH{SST_RGBSELECT_SHIFT});
       $SST_RGBSELECT{SST_RGBSEL_TREXOUT} =       $FBZCOLORPATH{SST_RGBSEL_TMUOUT};
       $SST_RGBSELECT{SST_RGBSEL_C1} =            (2<<$FBZCOLORPATH{SST_RGBSELECT_SHIFT});
       $SST_RGBSELECT{SST_RGBSEL_LFB} =           (3<<$FBZCOLORPATH{SST_RGBSELECT_SHIFT});
$FBZCOLORPATH{SST_ASELECT_SHIFT} =        2;
$FBZCOLORPATH{SST_ASELECT} =              (0x3<<$FBZCOLORPATH{SST_ASELECT_SHIFT});
       $SST_ASELECT{SST_ASEL_RGBA} =            (0<<$FBZCOLORPATH{SST_ASELECT_SHIFT});
       $SST_ASELECT{SST_ASEL_TMUOUT} =          (1<<$FBZCOLORPATH{SST_ASELECT_SHIFT});
       $SST_ASELECT{SST_ASEL_TREXOUT} =         $FBZCOLORPATH{SST_ASEL_TMUOUT};
       $SST_ASELECT{SST_ASEL_C1} =              (2<<$FBZCOLORPATH{SST_ASELECT_SHIFT});
       $SST_ASELECT{SST_ASEL_LFB} =             (3<<$FBZCOLORPATH{SST_ASELECT_SHIFT});
$FBZCOLORPATH{SST_LOCALSELECT_SHIFT} =    4;
$FBZCOLORPATH{SST_LOCALSELECT} =          BIT(4);
$FBZCOLORPATH{SST_ALOCALSELECT_SHIFT} =   5;
$FBZCOLORPATH{SST_ALOCALSELECT} =         (0x3<<$FBZCOLORPATH{SST_ALOCALSELECT_SHIFT});
$FBZCOLORPATH{SST_ALOCAL_ITERATOR} =              (0<<$FBZCOLORPATH{SST_ALOCALSELECT_SHIFT});
$FBZCOLORPATH{SST_ALOCAL_C0} =                    (1<<$FBZCOLORPATH{SST_ALOCALSELECT_SHIFT});
$FBZCOLORPATH{SST_ALOCAL_Z} =                     (2<<$FBZCOLORPATH{SST_ALOCALSELECT_SHIFT});
$FBZCOLORPATH{SST_ALOCAL_W} =                     (3<<$FBZCOLORPATH{SST_ALOCALSELECT_SHIFT});
$FBZCOLORPATH{SST_LOCALSELECT_OVERRIDE_WITH_ATEX} =       BIT(7);
$FBZCOLORPATH{SST_CC_ZERO_OTHER} =        BIT(8);
$FBZCOLORPATH{SST_CC_SUB_CLOCAL} =        BIT(9);
$FBZCOLORPATH{SST_CC_MSELECT_SHIFT} =     10;
$FBZCOLORPATH{SST_CC_MSELECT} =           (0x7<<$FBZCOLORPATH{SST_CC_MSELECT_SHIFT});
       $SST_CC_MSELECT{SST_CC_MONE} =              (0<<$FBZCOLORPATH{SST_CC_MSELECT_SHIFT});
       $SST_CC_MSELECT{SST_CC_MCLOCAL} =           (1<<$FBZCOLORPATH{SST_CC_MSELECT_SHIFT});
       $SST_CC_MSELECT{SST_CC_MAOTHER} =           (2<<$FBZCOLORPATH{SST_CC_MSELECT_SHIFT});
       $SST_CC_MSELECT{SST_CC_MALOCAL} =           (3<<$FBZCOLORPATH{SST_CC_MSELECT_SHIFT});
       $SST_CC_MSELECT{SST_CC_MATMU} =             (4<<$FBZCOLORPATH{SST_CC_MSELECT_SHIFT});
       $SST_CC_MSELECT{SST_CC_MATREX} =            $FBZCOLORPATH{SST_CC_MATMU};
       $SST_CC_MSELECT{SST_CC_MRGBTMU} =           (5<<$FBZCOLORPATH{SST_CC_MSELECT_SHIFT});
       $SST_CC_MSELECT{SST_CC_MONE6} =	           (6<<$FBZCOLORPATH{SST_CC_MSELECT_SHIFT});
       $SST_CC_MSELECT{SST_CC_MCMSELECT7}=         (7<<$FBZCOLORPATH{SST_CC_MSELECT_SHIFT});
$FBZCOLORPATH{SST_CC_REVERSE_BLEND} =     BIT(13);
$FBZCOLORPATH{SST_CC_ADD_CLOCAL} =        BIT(14);
$FBZCOLORPATH{SST_CC_ADD_ALOCAL} =        BIT(15);
$FBZCOLORPATH{SST_CC_INVERT_OUTPUT} =     BIT(16);
$FBZCOLORPATH{SST_CCA_ZERO_OTHER} =       BIT(17);
$FBZCOLORPATH{SST_CCA_SUB_CLOCAL} =       BIT(18);
$FBZCOLORPATH{SST_CCA_MSELECT_SHIFT} =    19;
$FBZCOLORPATH{SST_CCA_MSELECT} =          (0x7<<$FBZCOLORPATH{SST_CCA_MSELECT_SHIFT});
       $SST_CCA_MSELECT{SST_CCA_MONE} =             (0<<$FBZCOLORPATH{SST_CCA_MSELECT_SHIFT});
       $SST_CCA_MSELECT{SST_CCA_MCLOCAL} =          (1<<$FBZCOLORPATH{SST_CCA_MSELECT_SHIFT});
       $SST_CCA_MSELECT{SST_CCA_MAOTHER} =          (2<<$FBZCOLORPATH{SST_CCA_MSELECT_SHIFT});
       $SST_CCA_MSELECT{SST_CCA_MALOCAL} =          (3<<$FBZCOLORPATH{SST_CCA_MSELECT_SHIFT});
       $SST_CCA_MSELECT{SST_CCA_MATMU} =            (4<<$FBZCOLORPATH{SST_CCA_MSELECT_SHIFT});
       $SST_CCA_MSELECT{SST_CCA_MATREX} =           $FBZCOLORPATH{SST_CCA_MATMU};
       $SST_CCA_MSELECT{SST_CCA_MAITER} =           (5<<$FBZCOLORPATH{SST_CCA_MSELECT_SHIFT});  #Iterated Alpha
       $SST_CCA_MSELECT{SST_CCA_MAC1}   =           (6<<$FBZCOLORPATH{SST_CCA_MSELECT_SHIFT});  #Color 1 Alpha
$FBZCOLORPATH{SST_CCA_REVERSE_BLEND} =    BIT(22);
$FBZCOLORPATH{SST_CCA_ADD_CLOCAL} =       BIT(23);
$FBZCOLORPATH{SST_CCA_ADD_ALOCAL} =       BIT(24);
$FBZCOLORPATH{SST_CCA_INVERT_OUTPUT} =    BIT(25);
$FBZCOLORPATH{SST_PARMADJUST} =           BIT(26);
$FBZCOLORPATH{SST_ENTEXTUREMAP} =         BIT(27);
$FBZCOLORPATH{SST_RGBAZ_CLAMP} =          BIT(28);
$FBZCOLORPATH{SST_ENANTIALIAS} =          BIT(29);
$FBZCOLORPATH{SST_TRIANGLE_ITERATOR_COLUMN_BAND_CONTROL_SHIFT}=   30;
$FBZCOLORPATH{SST_TRIANGLE_ITERATOR_COLUMN_BAND_CONTROL} = (0x3 << $FBZCOLORPATH{SST_TRIANGLE_ITERATOR_COLUMN_BAND_CONTROL_SHIFT});


# here are some abstract constants that most people will be using
# CC stands for Color Combine (RGB channels)
# CCA stands for Color Combine Alpha (just the Alpha channel)
# BLEND_ATT blends on the Alpha channel of the color passed from TMU to TMU
# BLEND_ALOCAL blends on the Alpha channel of the local texture color
# NOTE: and of the BLEND modes can be modified using SST_CC_REVERSE_BLEND
#       and/or SST_CCA_REVERSE_BLEND, these modifiers reverse the "polarity"
#       of the blend

addToDefs(\%SST_RGBSELECT);             # load defn's into %h3def for use below
addToDefs(\%SST_ASELECT);
addToDefs(\%SST_CCA_MSELECT);

$h3def{SST_CC_MZERO} =   ($h3def{SST_CC_MONE}  | $h3def{SST_CC_REVERSE_BLEND});;
$h3def{SST_CCA_MZERO} =  ($h3def{SST_CCA_MONE} | $h3def{SST_CCA_REVERSE_BLEND});;

$h3def{SST_CCOMBINE_SHIFT} =       8;       # RGB COMBINE MODES
$h3def{SST_CCOMBINE} =             (0x1FF<<$h3def{SST_CCOMBINE_SHIFT});
       $h3def{SST_CC_REPLACE} =    ($h3def{SST_CC_ZERO_OTHER} | $h3def{SST_CC_ADD_CLOCAL});
       $h3def{SST_CC_PASS} =       ($h3def{SST_CC_MONE});
       $h3def{SST_CC_ADD} =        ($h3def{SST_CC_MONE}  | $h3def{SST_CC_ADD_CLOCAL});
       $h3def{SST_CC_SUB} =        ($h3def{SST_CC_SUB_CLOCAL} | $h3def{SST_CC_MONE});
       $h3def{SST_CC_MULT} =       ($h3def{SST_CC_MCLOCAL} | $h3def{SST_CC_REVERSE_BLEND});
       $h3def{SST_CC_ZERO} =       ($h3def{SST_CC_MZERO});
       $h3def{SST_CC_ONE} =        ($h3def{SST_CC_MZERO} | $h3def{SST_CC_INVERT_OUTPUT});
       $h3def{SST_CC_BLEND} =      ($h3def{SST_CC_SUB_CLOCAL} | $h3def{SST_CC_ADD_CLOCAL});
       $h3def{SST_CC_BLEND_ATT} =         ($h3def{SST_CC_BLEND} | $h3def{SST_CC_MAOTHER});
       $h3def{SST_CC_BLEND_ATEX} =        ($h3def{SST_CC_BLEND} | $h3def{SST_CC_MATMU});
       $h3def{SST_CC_BLEND_ATEX_REV} =    ($h3def{SST_CC_BLEND} | SST_CC_MATMU | $h3def{SST_CC_REVERSE_BLEND});
       $h3def{SST_CC_BLEND_ATT_REV} =     ($h3def{SST_CC_BLEND} | SST_CC_MAOTHER | $h3def{SST_CC_REVERSE_BLEND});
       $h3def{SST_CC_BLEND_ALOCAL} =      ($h3def{SST_CC_BLEND} | $h3def{SST_CC_MALOCAL});
$h3def{SST_CACOMBINE_SHIFT} =      17;      # ALPHA COMBINE MODES
$h3def{SST_CACOMBINE} =            (0x1FF<<$h3def{SST_CACOMBINE_SHIFT});
       $h3def{SST_CCA_REPLACE} =   ($h3def{SST_CCA_ZERO_OTHER} | $h3def{SST_CCA_ADD_CLOCAL});
       $h3def{SST_CCA_PASS} =      ($h3def{SST_CCA_MONE});
       $h3def{SST_CCA_ADD} =       ($h3def{SST_CCA_MONE}  | $h3def{SST_CCA_ADD_CLOCAL});
       $h3def{SST_CCA_SUB} =       ($h3def{SST_CCA_SUB_CLOCAL} | $h3def{SST_CCA_MONE});
       $h3def{SST_CCA_MULT} =      ($h3def{SST_CCA_MCLOCAL} | $h3def{SST_CCA_REVERSE_BLEND});
       $h3def{SST_CCA_ONE} =       ($h3def{SST_CCA_MZERO} | $h3def{SST_CCA_INVERT_OUTPUT});
       $h3def{SST_CCA_ZERO} =      ($h3def{SST_CCA_MZERO});
       $h3def{SST_CCA_BLEND} =     ($h3def{SST_CCA_SUB_CLOCAL} | $h3def{SST_CCA_ADD_CLOCAL});
       $h3def{SST_CCA_BLEND_ATT} =        ($h3def{SST_CCA_BLEND} | $h3def{SST_CCA_MAOTHER});
       $h3def{SST_CCA_BLEND_ATEX} =       ($h3def{SST_CCA_BLEND} | $h3def{SST_CCA_MATMU});
       $h3def{SST_CCA_BLEND_ATEX_REV} =   ($h3def{SST_CCA_BLEND} | SST_CCA_MATMU | $h3def{SST_CCA_REVERSE_BLEN});
       $h3def{SST_CCA_BLEND_ATT_REV} =    ($h3def{SST_CCA_BLEND} | SST_CCA_MAOTHER | $h3def{SST_CCA_REVERSE_BLEND});
       $h3def{SST_CCA_BLEND_ALOCAL} =     ($h3def{SST_CCA_BLEND} | $h3def{SST_CCA_MALOCAL});

#----------------- SST fogMode bits ---------------------------
$FOGMODE{SST_ENFOGGING} =            BIT(0);  # Afog*Cfog + (1-Afog)*Color
$FOGMODE{SST_FOGADD} =               BIT(1);	# generates (1-Afog)*Color
$FOGMODE{SST_FOGMULT} =              BIT(2);  # generates Afog*Cfog
$FOGMODE{SST_FOG_ALPHA} =            BIT(3);  # use Alpha iterator to control fog
$FOGMODE{SST_FOG_Z} =                BIT(4);  # use Z iterator to control fog
$FOGMODE{SST_FOG_CONSTANT} =         BIT(5);  # simply add Cfog to color and clamp
$FOGMODE{SST_FOG_DITHER} =           BIT(6);  # dither the Afog
$FOGMODE{SST_FOG_ZONES} =            BIT(7);  # enable signed fog deltas

		# NOTE: these bits control alphablending!!!
$FOGMODE{SST_RGB_BLEND_REVERSE}=   BIT(8);  # reverse RGB alphablend operation
$FOGMODE{SST_RGB_BLEND_SUB}=       BIT(9);  # RGB alphablend subtract
$FOGMODE{SST_A_BLEND_REVERSE}=     BIT(10); # reverse RGB alphablend operation
$FOGMODE{SST_A_BLEND_SUB}=         BIT(11); # RGB alphablend subtract

$FOGMODE{SST_DITHER_ROTATE_SHIFT}=		     12;
$FOGMODE{SST_DITHER_ROTATE}=		         (0x3<<$FOGMODE{SST_DITHER_ROTATE_SHIFT});
$FOGMODE{SST_DITHER_ROTATE_BLEND_SHIFT}=	 14;
$FOGMODE{SST_DITHER_ROTATE_BLEND}=		     (0x3<<$FOGMODE{SST_DITHER_ROTATE_BLEND_SHIFT});
$FOGMODE{SST_DITHER_ROTATE_AA_SHIFT}=	     16;
$FOGMODE{SST_DITHER_ROTATE_AA}=		     (0x3<<$FOGMODE{SST_DITHER_ROTATE_AA_SHIFT});
$FOGMODE{SST_DITHER_ROTATE_BLEND_AA_SHIFT}= 18;
$FOGMODE{SST_DITHER_ROTATE_BLEND_AA}=	     (0x3<<$FOGMODE{SST_DITHER_ROTATE_BLEND_AA_SHIFT});

#
# Rotated dither matrix selection
#
# Default values are:
# 1 chip 2 sample antialiasing:
#   fogMode[19:18] = 2, fogMode[17:16] = 2, fogMode[15:14] = 0, fogMode[13:12] = 0
# 2 chip 4 sample antialiasing:
#   chip 0:
#    fogMode[19:18] = 1, fogMode[17:16] = 1, fogMode[15:14] = 0, fogMode[13:12] = 0
#   chip 1:
#    fogMode[19:18] = 3, fogMode[17:16] = 3, fogMode[15:14] = 2, fogMode[13:12] = 2
#
$FOGMODE{SST_ROTDITHMATSEL}=              (0x3 << $FOGMODE{SST_DITHER_ROTATE_SHIFT});
  $SST_ROTDITHMATSEL{DITHMATSEL_2SMPL_DEF}=             (0x0 << $FOGMODE{SST_DITHER_ROTATE_SHIFT});
  $SST_ROTDITHMATSEL{DITHMATSEL_4SMPL_CHP0_DEF}=        (0x0 << $FOGMODE{SST_DITHER_ROTATE_SHIFT});
  $SST_ROTDITHMATSEL{DITHMATSEL_4SMPL_CHP1_DEF}=        (0x2 << $FOGMODE{SST_DITHER_ROTATE_SHIFT});

$FOGMODE{SST_ROTDITHMATSELABLND}=         (0x3 << $FOGMODE{SST_DITHER_ROTATE_BLEND_SHIFT});
  $SST_ROTDITHMATSELABLND{DITHMATSELABLND_2SMPL_DEF}=        (0x0 << $FOGMODE{SST_DITHER_ROTATE_BLEND_SHIFT});
  $SST_ROTDITHMATSELABLND{DITHMATSELABLND_4SMPL_CHP0_DEF}=   (0x0 << $FOGMODE{SST_DITHER_ROTATE_BLEND_SHIFT});
  $SST_ROTDITHMATSELABLND{DITHMATSELABLND_4SMPL_CHP1_DEF}=   (0x2 << $FOGMODE{SST_DITHER_ROTATE_BLEND_SHIFT});

$FOGMODE{SST_ROTDITHMATSELAA}=            (0x3 << $FOGMODE{SST_DITHER_ROTATE_AA_SHIFT});
  $SST_ROTDITHMATSELAA{DITHMATSELAA_2SMPL_DEF}=           (0x2 << $FOGMODE{SST_DITHER_ROTATE_AA_SHIFT});
  $SST_ROTDITHMATSELAA{DITHMATSELAA_4SMPL_CHP0_DEF}=      (0x1 << $FOGMODE{SST_DITHER_ROTATE_AA_SHIFT});
  $SST_ROTDITHMATSELAA{DITHMATSELAA_4SMPL_CHP1_DEF}=      (0x3 << $FOGMODE{SST_DITHER_ROTATE_AA_SHIFT});

$FOGMODE{SST_ROTDITHMATSELABLNDAA}=       (0x3 << $FOGMODE{SST_DITHER_ROTATE_BLEND_AA_SHIFT});
  $SST_ROTDITHMATSELABLNDAA{DITHMATSELABLNDAA_2SMPL_DEF}=      (0x2 << $FOGMODE{SST_DITHER_ROTATE_BLEND_AA_SHIFT});
  $SST_ROTDITHMATSELABLNDAA{DITHMATSELABLNDAA_4SMPL_CHP0_DEF}= (0x1 << $FOGMODE{SST_DITHER_ROTATE_BLEND_AA_SHIFT});
  $SST_ROTDITHMATSELABLNDAA{DITHMATSELABLNDAA_4SMPL_CHP1_DEF}= (0x3 << $FOGMODE{SST_DITHER_ROTATE_BLEND_AA_SHIFT});

$FOGMODE{SST_ROTDITHMATSELRESET}=         ~(                    
                                          $FOGMODE{SST_ROTDITHMATSEL}        |
                                          $FOGMODE{SST_ROTDITHMATSELABLND}   |
                                          $FOGMODE{SST_ROTDITHMATSELAA}      |
                                          $FOGMODE{SST_ROTDITHMATSELABLNDAA}  
								);


#----------------- SST alphaMode bits ------------------------
$ALPHAMODE{SST_ENALPHAFUNC} =          BIT(0);
$ALPHAMODE{SST_ALPHAFNC_SHIFT} =       1;
$ALPHAMODE{SST_ALPHAFUNC_SHIFT} =      1;
$ALPHAMODE{SST_ALPHAFNC} =     ( 0x7 << $ALPHAMODE{SST_ALPHAFNC_SHIFT} );
$ALPHAMODE{SST_ALPHAFUNC} =    ( 0x7 << $ALPHAMODE{SST_ALPHAFUNC_SHIFT} );
       $SST_ALPHAFUNC{SST_ALPHAFUNC_LT} = BIT(1);
       $SST_ALPHAFUNC{SST_ALPHAFUNC_EQ} = BIT(2);
       $SST_ALPHAFUNC{SST_ALPHAFUNC_GT} = BIT(3);
$ALPHAMODE{SST_ENALPHABLEND} =         BIT(4);
	# unused BIT(5)
$ALPHAMODE{SST_RGBSRCFACT_SHIFT} =     8;       # RGB source factor
$ALPHAMODE{SST_RGBSRCFACT} =           (0xF<<$ALPHAMODE{SST_RGBSRCFACT_SHIFT});
       $h3def{SST_A_ZERO} =       0;
       $h3def{SST_A_SRCALPHA} =   1;
       $h3def{SST_A_COLOR} =      2;
       $h3def{SST_A_DSTALPHA} =   3;
       $h3def{SST_A_ONE} =        4;
       $h3def{SST_AOM_SRCALPHA} = 5;
       $h3def{SST_AOM_COLOR} =    6;
       $h3def{SST_AOM_DSTALPHA} = 7;
       $h3def{SST_A_SAMECOLOR} =    8;
       $h3def{SST_AOM_SAMECOLOR} = 9;
       $h3def{SST_A_SATURATE} =   0xF;     # for SRC factor only
       $h3def{SST_A_COLORBEFOREFOG} = 0xF; # for DST factor only
$ALPHAMODE{SST_RGBDSTFACT_SHIFT} =     12;      # RGB destination factor
$ALPHAMODE{SST_RGBDSTFACT} =           (0xF<<$ALPHAMODE{SST_RGBDSTFACT_SHIFT});
$ALPHAMODE{SST_ASRCFACT_SHIFT} =       16;      # A source factor
$ALPHAMODE{SST_ASRCFACT} =             (0xF<<$ALPHAMODE{SST_ASRCFACT_SHIFT});
$ALPHAMODE{SST_ADSTFACT_SHIFT} =       20;      # A destination factor
$ALPHAMODE{SST_ADSTFACT} =             (0xF<<$ALPHAMODE{SST_ADSTFACT_SHIFT});
$ALPHAMODE{SST_ALPHAREF_SHIFT} =       24;
$ALPHAMODE{SST_ALPHAREF} =             (0xFF<<$ALPHAMODE{SST_ALPHAREF_SHIFT});

#---------------- SST zaColor bits ----------------------
$ZACOLOR{SST_ZACOLOR_DEPTH_SHIFT} =    0;
$ZACOLOR{SST_ZACOLOR_DEPTH} =          ( 0xFFFFFF << $ZACOLOR{SST_ZACOLOR_DEPTH_SHIFT} );;
$ZACOLOR{SST_ZACOLOR_ALPHA_SHIFT} =    24;
$ZACOLOR{SST_ZACOLOR_ALPHA} =          ( 0xFF << $ZACOLOR{SST_ZACOLOR_ALPHA_SHIFT} );;

#----------------- SST chromaRange bits ------------------------
$CHROMARANGE{SST_CHROMARANGE_BLUE_EX} = 	 BIT(24);# Blue value in exclusive mode
$CHROMARANGE{SST_CHROMARANGE_GREEN_EX} =  BIT(25);# Green value in exclusive mode
$CHROMARANGE{SST_CHROMARANGE_RED_EX} =    BIT(26);# Red  value in exclusive mode
$CHROMARANGE{SST_CHROMARANGE_BLOCK_OR} =  BIT(27);# Union of all colors.
$CHROMARANGE{SST_ENCHROMARANGE} =         BIT(28);
$CHROMARANGE{SST_ENCHROMAKEY_TMU} =       BIT(29);	# only for TMU
$CHROMARANGE{SST_ENCOLORSUBSTITUTION} =   BIT(30);	# only for TMU

#----------------- SST clip bits ------------------------
#clipLeftRight
$CLIPLEFTRIGHT{SST_CLIPLEFT_SHIFT} =         16;
$CLIPLEFTRIGHT{SST_CLIPRIGHT_SHIFT} =        0;
#clipBottomTop
$CLIPBOTTOMTOP{SST_CLIPBOTTOM_SHIFT} =       16;
$CLIPBOTTOMTOP{SST_CLIPTOP_SHIFT} =          0;
# clipLeftRight1, clipBottomTop1
$CLIPLEFTRIGHT1{SST_CLIPLEFT_SHIFT} =         16;
$CLIPLEFTRIGHT1{SST_CLIPRIGHT_SHIFT} =        0;
$CLIPLEFTRIGHT1{SST_ENRECTCLIP1} = 		BIT(31);
$CLIPLEFTRIGHT1{SST_RECTCLIP1_EX} = 	BIT(31);

$CLIPBOTTOMTOP1{SST_CLIPBOTTOM_SHIFT} =       16;
$CLIPBOTTOMTOP1{SST_CLIPTOP_SHIFT} =          0;
$CLIPBOTTOMTOP1{SST_ENRECTCLIP1} = 		BIT(31);
$CLIPBOTTOMTOP1{SST_RECTCLIP1_EX} = 	BIT(31);

#----------------- SST sSetupMode bits ------------------------
$SSETUPMODE{SST_SETUP_RGB} = 		BIT(0);
$SSETUPMODE{SST_SETUP_A} = 		BIT(1);
$SSETUPMODE{SST_SETUP_Z} = 		BIT(2);
$SSETUPMODE{SST_SETUP_Wfbi} = 		BIT(3);
$SSETUPMODE{SST_SETUP_W0} = 		BIT(4);
$SSETUPMODE{SST_SETUP_ST0} = 		BIT(5);
$SSETUPMODE{SST_SETUP_W1} = 		BIT(6);
$SSETUPMODE{SST_SETUP_ST1} = 		BIT(7);
$SSETUPMODE{SST_SETUP_FAN} = 		BIT(16);
$SSETUPMODE{SST_SETUP_EN_CULLING} = 	BIT(17);
$SSETUPMODE{SST_SETUP_CULL_NEGATIVE} = 	BIT(18);
$SSETUPMODE{SST_SETUP_DIS_PINGPONG} = 	BIT(19);

#-------------- SST col/aux buffer addressing ---------------
# auxBufferAddr
$AUXBUFFERADDR{SST_BUFFER_BASE_SHIFT} =     (0);
$AUXBUFFERADDR{SST_BUFFER_BASE_ADDR} =      (0xFFFFFF<<$AUXBUFFERADDR{SST_BUFFER_BASE_SHIFT});
$AUXBUFFERADDR{SST_BUFFER_BASE_T_SHIFT} =   (12);
$AUXBUFFERADDR{SST_BUFFER_BASE_T} =         (0xFFF<<$AUXBUFFERADDR{SST_BUFFER_BASE_T_SHIFT});
$AUXBUFFERADDR{SST_BUFFER_BASE_Y_SHIFT} =   (7);
$AUXBUFFERADDR{SST_BUFFER_BASE_Y} =         (0x1F<<$AUXBUFFERADDR{SST_BUFFER_BASE_Y_SHIFT});
$AUXBUFFERADDR{SST_BUFFER_BASE_X_SHIFT} =   (0);
$AUXBUFFERADDR{SST_BUFFER_BASE_X} =         (0x7F<<$AUXBUFFERADDR{SST_BUFFER_BASE_X_SHIFT});

# auxBufferStride
$AUXBUFFERSTRIDE{SST_BUFFER_STRIDE_SHIFT} =   (0);
$AUXBUFFERSTRIDE{SST_BUFFER_LINEAR_STRIDE} =  (0x3FFF<<$AUXBUFFERSTRIDE{SST_BUFFER_STRIDE_SHIFT});
$AUXBUFFERSTRIDE{SST_BUFFER_TILE_STRIDE} =    (0x7F<<$AUXBUFFERSTRIDE{SST_BUFFER_STRIDE_SHIFT});
$AUXBUFFERSTRIDE{SST_BUFFER_MEMORY_TYPE} =    BIT(15);
$AUXBUFFERSTRIDE{SST_BUFFER_MEMORY_TILED} =   BIT(15);
$AUXBUFFERSTRIDE{SST_BUFFER_MEMORY_LINEAR} =  (0);

# colBufferAddr
$COLBUFFERADDR{SST_BUFFER_BASE_SHIFT} =     (0);
$COLBUFFERADDR{SST_BUFFER_BASE_ADDR} =      (0xFFFFFF<<$COLBUFFERADDR{SST_BUFFER_BASE_SHIFT});
$COLBUFFERADDR{SST_BUFFER_BASE_T_SHIFT} =   (12);
$COLBUFFERADDR{SST_BUFFER_BASE_T} =         (0xFFF<<$COLBUFFERADDR{SST_BUFFER_BASE_T_SHIFT});
$COLBUFFERADDR{SST_BUFFER_BASE_Y_SHIFT} =   (7);
$COLBUFFERADDR{SST_BUFFER_BASE_Y} =         (0x1F<<$COLBUFFERADDR{SST_BUFFER_BASE_Y_SHIFT});
$COLBUFFERADDR{SST_BUFFER_BASE_X_SHIFT} =   (0);
$COLBUFFERADDR{SST_BUFFER_BASE_X} =         (0x7F<<$COLBUFFERADDR{SST_BUFFER_BASE_X_SHIFT});

# colBufferStride
$COLBUFFERSTRIDE{SST_BUFFER_STRIDE_SHIFT} =   (0);
$COLBUFFERSTRIDE{SST_BUFFER_LINEAR_STRIDE} =  (0x3FFF<<$COLBUFFERSTRIDE{SST_BUFFER_STRIDE_SHIFT});
$COLBUFFERSTRIDE{SST_BUFFER_TILE_STRIDE} =    (0x7F<<$COLBUFFERSTRIDE{SST_BUFFER_STRIDE_SHIFT});
$COLBUFFERSTRIDE{SST_BUFFER_MEMORY_TYPE} =    BIT(15);
$COLBUFFERSTRIDE{SST_BUFFER_MEMORY_TILED} =   BIT(15);
$COLBUFFERSTRIDE{SST_BUFFER_MEMORY_LINEAR} =  (0);


# tile dimensions
$h3def{SST_TILE_WIDTH_BITS} =       (7);
$h3def{SST_TILE_WIDTH} =            (1<<($h3def{SST_TILE_WIDTH_BITS}));
$h3def{SST_TILE_WIDTH_MASK} =       SST_MASK($h3def{SST_TILE_WIDTH_BITS});
$h3def{SST_TILE_HEIGHT_BITS} =      (5);
$h3def{SST_TILE_HEIGHT} =           (1<<($h3def{SST_TILE_HEIGHT_BITS}));
$h3def{SST_TILE_HEIGHT_MASK} =      SST_MASK($h3def{SST_TILE_HEIGHT_BITS});
$h3def{SST_TILE_SIZE} =             (SST_TILE_WIDTH * $h3def{SST_TILE_HEIGHT});;

#------------------- SST Planar YUV addressing --------------------
$h3def{SST_YUV_STRIDE_SHIFT} =   (0);
$h3def{SST_YUV_LINEAR_STRIDE} =  (0x1FFF<<$h3def{SST_YUV_STRIDE_SHIFT});
$h3def{SST_YUV_TILE_STRIDE} =    (0x7F<<$h3def{SST_YUV_STRIDE_SHIFT});
$h3def{SST_YUV_MEMORY_TYPE} =    BIT(31);
$h3def{SST_YUV_MEMORY_TILED} =   BIT(31);
$h3def{SST_YUV_MEMORY_LINEAR} =  (0);

$h3def{SST_YUV_BASE_ADDR_SHIFT} =  0;
$h3def{SST_YUV_BASE_ADDR} =        (0x01FFFFFF<<$h3def{SST_YUV_BASE_ADDR_SHIFT});

$h3def{SST_YUV_ADDR} =             0xFFFFF;
$h3def{SST_YUV_ADDR_STRIDE} =      1024;
$h3def{SST_YUV_ADDR_X_SHIFT} =     0;
$h3def{SST_YUV_ADDR_X} =           (0x3FF<<$h3def{SST_YUV_ADDR_X_SHIFT});
$h3def{SST_YUV_ADDR_Y_SHIFT} =     10;
$h3def{SST_YUV_ADDR_Y} =           (0x3FF<<$h3def{SST_YUV_ADDR_Y_SHIFT});

#----------------- SST texBaseAddr bits ---------------------------
$TEXBASEADDR{SST_TEXTURE_ADDRESS} = 	(SST_MASK(21)<<4 | BIT(1));
$TEXBASEADDR{SST_TEXTURE_FULL_ADDRESS} = 	SST_MASK(26);
sub SST_TEXTURE_UNMUNGE_ADDRESS {my($addr) = @_; return((($addr & SST_TEXTURE_ADDRESS) & (~BIT(1))) |
							((($addr & SST_TEXTURE_ADDRESS) & BIT(1)) << 24));}
sub SST_TEXTURE_MUNGE_ADDRESS {my($addr) = @_; return((($addr & BIT(25)) >> 24) |
						      ($addr & (SST_MASK(21)<<4)));}

$TEXBASEADDR{SST_TEXTURE_TILESTRIDE_SHIFT} = 	25;
$TEXBASEADDR{SST_TEXTURE_TILESTRIDE} = 	(0x7F<<$TEXBASEADDR{SST_TEXTURE_TILESTRIDE_SHIFT});
$TEXBASEADDR{SST_TEXTURE_IS_TILED} = 	BIT(0);

#For 256x256 textures

#For compressed textures, the s & t index microtiles, not individual texels
$h3def{SST_TEXTURE_LOD4_COMPRESSED_SHIFT}=  15;
$h3def{SST_TEXTURE_LOD4_COMPRESSED}=        (0xF<<$h3def{SST_TEXTURE_LOD4_COMPRESSED_SHIFT});
$h3def{SST_TEXTURE_T4_COMPRESSED_SHIFT}=    9;
$h3def{SST_TEXTURE_T4_COMPRESSED}=          (0x3F<<$h3def{SST_TEXTURE_T4_COMPRESSED_SHIFT});
$h3def{SST_TEXTURE_S4_COMPRESSED_SHIFT}=    4;
$h3def{SST_TEXTURE_S4_COMPRESSED}=          (0x1F<<$h3def{SST_TEXTURE_S4_COMPRESSED_SHIFT});
$h3def{SST_TEXTURE_WORDSELECT4_COMPRESSED_SHIFT}=  2;
$h3def{SST_TEXTURE_WORDSELECT4_COMPRESSED}=        (0x3<<$h3def{SST_TEXTURE_WORDSELECT4_COMPRESSED_SHIFT});
  
$h3def{SST_TEXTURE_LOD8_COMPRESSED_SHIFT}=  16;
$h3def{SST_TEXTURE_LOD8_COMPRESSED}=        (0xF<<$h3def{SST_TEXTURE_LOD8_COMPRESSED_SHIFT});
$h3def{SST_TEXTURE_T8_COMPRESSED_SHIFT}=    10;
$h3def{SST_TEXTURE_T8_COMPRESSED}=          (0x3F<<$h3def{SST_TEXTURE_T8_COMPRESSED_SHIFT});
$h3def{SST_TEXTURE_S8_COMPRESSED_SHIFT}=      4;
$h3def{SST_TEXTURE_S8_COMPRESSED}=          (0x3F<<$h3def{SST_TEXTURE_S8_COMPRESSED_SHIFT});
$h3def{SST_TEXTURE_WORDSELECT8_COMPRESSED_SHIFT}=  2;
$h3def{SST_TEXTURE_WORDSELECT8_COMPRESSED}=        (0x3<<$h3def{SST_TEXTURE_WORDSELECT8_COMPRESSED_SHIFT});

$h3def{SST_TEXTURE_LOD8_SHIFT} =   16;
$h3def{SST_TEXTURE_LOD8} =         (0xF<<$h3def{SST_TEXTURE_LOD8_SHIFT});
$h3def{SST_TEXTURE_T8_SHIFT} =     8;
$h3def{SST_TEXTURE_T8} =           (0xFF<<$h3def{SST_TEXTURE_T8_SHIFT});
$h3def{SST_TEXTURE_S8_SHIFT} =     0;
$h3def{SST_TEXTURE_S8} =           (0xFF<<$h3def{SST_TEXTURE_S8_SHIFT});

$h3def{SST_TEXTURE_LOD16_SHIFT} =   17;
$h3def{SST_TEXTURE_LOD16} =         (0xF<<$h3def{SST_TEXTURE_LOD16_SHIFT});
$h3def{SST_TEXTURE_T16_SHIFT} =     9;
$h3def{SST_TEXTURE_T16} =           (0xFF<<$h3def{SST_TEXTURE_T16_SHIFT});
$h3def{SST_TEXTURE_S16_SHIFT} =     1;
$h3def{SST_TEXTURE_S16} =           (0xFF<<$h3def{SST_TEXTURE_S16_SHIFT});

$h3def{SST_TEXTURE_LOD32_SHIFT} =   18;
$h3def{SST_TEXTURE_LOD32} =         (0xF<<$h3def{SST_TEXTURE_LOD32_SHIFT});
$h3def{SST_TEXTURE_T32_SHIFT} =     10;
$h3def{SST_TEXTURE_T32} =           (0xFF<<$h3def{SST_TEXTURE_T32_SHIFT});
$h3def{SST_TEXTURE_S32_SHIFT} =     2;
$h3def{SST_TEXTURE_S32} =           (0xFF<<$h3def{SST_TEXTURE_S32_SHIFT});

#For new 2048x2048 textures

#For compressed textures, the s & t index microtiles, not individual texels
$h3def{SST_TEXTURE_BIG_LOD4_COMPRESSED_SHIFT}=  19;
$h3def{SST_TEXTURE_BIG_LOD4_COMPRESSED}=        (0xF<<$h3def{SST_TEXTURE_BIG_LOD4_COMPRESSED_SHIFT});
$h3def{SST_TEXTURE_BIG_T4_COMPRESSED_SHIFT}=    11;
$h3def{SST_TEXTURE_BIG_T4_COMPRESSED}=           (0xFF<<$h3def{SST_TEXTURE_BIG_T4_COMPRESSED_SHIFT});
$h3def{SST_TEXTURE_BIG_S4_COMPRESSED_SHIFT}=     4;
$h3def{SST_TEXTURE_BIG_S4_COMPRESSED}=          (0x7F<<$h3def{SST_TEXTURE_BIG_S4_COMPRESSED_SHIFT});
$h3def{SST_TEXTURE_BIG_WORDSELECT4_COMPRESSED_SHIFT}=  2;
$h3def{SST_TEXTURE_BIG_WORDSELECT4_COMPRESSED}=        (0x3<<$h3def{SST_TEXTURE_BIG_WORDSELECT4_COMPRESSED_SHIFT});
  
$h3def{SST_TEXTURE_BIG_LOD8_COMPRESSED_SHIFT}=  20;
$h3def{SST_TEXTURE_BIG_LOD8_COMPRESSED}=        (0xF<<$h3def{SST_TEXTURE_BIG_LOD8_COMPRESSED_SHIFT});
$h3def{SST_TEXTURE_BIG_T8_COMPRESSED_SHIFT}=    12;
$h3def{SST_TEXTURE_BIG_T8_COMPRESSED}=          (0xFF<<$h3def{SST_TEXTURE_BIG_T8_COMPRESSED_SHIFT});
$h3def{SST_TEXTURE_BIG_S8_COMPRESSED_SHIFT}=    4;
$h3def{SST_TEXTURE_BIG_S8_COMPRESSED}=          (0xFF<<$h3def{SST_TEXTURE_BIG_S8_COMPRESSED_SHIFT});
$h3def{SST_TEXTURE_BIG_WORDSELECT8_COMPRESSED_SHIFT}=  2;
$h3def{SST_TEXTURE_BIG_WORDSELECT8_COMPRESSED}=        (0x3<<$h3def{SST_TEXTURE_BIG_WORDSELECT8_COMPRESSED_SHIFT});

$h3def{SST_TEXTURE_BIG_LOD8_SHIFT}=  20;
$h3def{SST_TEXTURE_BIG_LOD8}=        (0xF<<$h3def{SST_TEXTURE_BIG_LOD8_SHIFT});
$h3def{SST_TEXTURE_BIG_T8_SHIFT}=    10;
$h3def{SST_TEXTURE_BIG_T8}=          (0x3FF<<$h3def{SST_TEXTURE_BIG_T8_SHIFT});
$h3def{SST_TEXTURE_BIG_S8_SHIFT}=    0;
$h3def{SST_TEXTURE_BIG_S8}=          (0x3FF<<$h3def{SST_TEXTURE_BIG_S8_SHIFT});

$h3def{SST_TEXTURE_BIG_LOD16_SHIFT}=  21;
$h3def{SST_TEXTURE_BIG_LOD16}=        (0xF<<$h3def{SST_TEXTURE_BIG_LOD16_SHIFT});
$h3def{SST_TEXTURE_BIG_T16_SHIFT}=    11;
$h3def{SST_TEXTURE_BIG_T16}=          (0x3FF<<$h3def{SST_TEXTURE_BIG_T16_SHIFT});
$h3def{SST_TEXTURE_BIG_S16_SHIFT}=    1;
$h3def{SST_TEXTURE_BIG_S16}=          (0x3FF<<$h3def{SST_TEXTURE_BIG_S16_SHIFT});

$h3def{SST_TEXTURE_BIG_LOD32_SHIFT}=  22;
$h3def{SST_TEXTURE_BIG_LOD32}=        (0xF<<$h3def{SST_TEXTURE_BIG_LOD32_SHIFT});
$h3def{SST_TEXTURE_BIG_T32_SHIFT}=    12;
$h3def{SST_TEXTURE_BIG_T32}=          (0x3FF<<$h3def{SST_TEXTURE_BIG_T32_SHIFT});
$h3def{SST_TEXTURE_BIG_S32_SHIFT}=    2;
$h3def{SST_TEXTURE_BIG_S32}=          (0x3FF<<$h3def{SST_TEXTURE_BIG_S32_SHIFT});

#-------------------- PLL Registers --------------------
#-------------------- pllCtrl0,1,2
$PLLCTRL{SST_PLL_K_SHIFT}=				0;
$PLLCTRL{SST_PLL_K}=				(0x3<<$PLLCTRL{SST_PLL_K_SHIFT});
$PLLCTRL{SST_PLL_M_SHIFT}=				2;
$PLLCTRL{SST_PLL_M}=				(0x3F<<$PLLCTRL{SST_PLL_M_SHIFT});
$PLLCTRL{SST_PLL_N_SHIFT}=				8;
$PLLCTRL{SST_PLL_N}=				(0xFF<<$PLLCTRL{SST_PLL_N_SHIFT});
$PLLCTRL{SST_PLL_TEST}=				BIT(16);

#-------------------- DAC Registers --------------------
#-------------------- dacMode
$DACMODE{SST_DAC_MODE_2X}=				BIT(0);
$DACMODE{SST_DAC_DPMS_ON_VSYNC}=			BIT(1);
$DACMODE{SST_DAC_FORCE_VSYNC}=			        BIT(2);
$DACMODE{SST_DAC_DPMS_ON_HSYNC}=			BIT(3);
$DACMODE{SST_DAC_FORCE_HSYNC}=			        BIT(4);

#-------------------- dacAddr
$DACADDR{SST_DAC_ADDR_SHIFT}=			0;
$DACADDR{SST_DAC_ADDR}=			(0xFF<<$DACADDR{SST_DAC_ADDR_SHIFT});

#-------------------- dacData
$DACDATA{SST_DAC_DATA_BLUE_SHIFT}= 		0;
$DACDATA{SST_DAC_DATA_BLUE}=			(0xFF<<$DACDATA{SST_DAC_DATA_BLUE_SHIFT});
$DACDATA{SST_DAC_DATA_GREEN_SHIFT}= 		8;
$DACDATA{SST_DAC_DATA_GREEN}=			(0xFF<<$DACDATA{SST_DAC_DATA_GREEN_SHIFT});
$DACDATA{SST_DAC_DATA_RED_SHIFT}=  		16;
$DACDATA{SST_DAC_DATA_RED}= 			(0xFF<<$DACDATA{SST_DAC_DATA_RED_SHIFT});


#------------------- SST Video ----------------------------
#------------------- vidProcCfg
$VIDPROCCFG{SST_VIDEO_PROCESSOR_EN} =                 BIT(0);
$VIDPROCCFG{SST_CURSOR_MODE_SHIFT} =                  1;
$VIDPROCCFG{SST_CURSOR_MODE} =                        (1<<$VIDPROCCFG{SST_CURSOR_MODE_SHIFT});
	$SST_CURSOR_MODE{SST_CURSOR_X11} =                 (1<<$VIDPROCCFG{SST_CURSOR_MODE_SHIFT});
	$SST_CURSOR_MODE{SST_CURSOR_MICROSOFT} =           (0<<$VIDPROCCFG{SST_CURSOR_MODE_SHIFT});
$VIDPROCCFG{SST_OVERLAY_STEREO_EN} =                  BIT(2);
$VIDPROCCFG{SST_INTERLACED_EN} =                      BIT(3);
$VIDPROCCFG{SST_HALF_MODE} =                          BIT(4);
$VIDPROCCFG{SST_CHROMA_EN} =                          BIT(5);
$VIDPROCCFG{SST_CHROMA_INVERT} =                      BIT(6);
$VIDPROCCFG{SST_DESKTOP_EN} =                         BIT(7);
$VIDPROCCFG{SST_OVERLAY_EN} =                         BIT(8);
$VIDPROCCFG{SST_VIDEOIN_AS_OVERLAY} =                 BIT(9);
$VIDPROCCFG{SST_DESKTOP_CLUT_BYPASS} =                BIT(10);
$VIDPROCCFG{SST_OVERLAY_CLUT_BYPASS} =                BIT(11);
$VIDPROCCFG{SST_DESKTOP_CLUT_SELECT} =                BIT(12);
$VIDPROCCFG{SST_OVERLAY_CLUT_SELECT} =                BIT(13);
$VIDPROCCFG{SST_OVERLAY_HORIZ_SCALE_EN} =             BIT(14);
$VIDPROCCFG{SST_OVERLAY_VERT_SCALE_EN} =              BIT(15);
$VIDPROCCFG{SST_OVERLAY_FILTER_MODE_SHIFT} =          16;
$VIDPROCCFG{SST_OVERLAY_FILTER_MODE} =                (3<<$VIDPROCCFG{SST_OVERLAY_FILTER_MODE_SHIFT});
	$SST_OVERLAY_FILTER_MODE{SST_OVERLAY_FILTER_POINT} =       (0<<$VIDPROCCFG{SST_OVERLAY_FILTER_MODE_SHIFT})    ;
	$SST_OVERLAY_FILTER_MODE{SST_OVERLAY_FILTER_2X2} =         (1<<$VIDPROCCFG{SST_OVERLAY_FILTER_MODE_SHIFT});
	$SST_OVERLAY_FILTER_MODE{SST_OVERLAY_FILTER_4X4} =         (2<<$VIDPROCCFG{SST_OVERLAY_FILTER_MODE_SHIFT});
	$SST_OVERLAY_FILTER_MODE{SST_OVERLAY_FILTER_BILINEAR} =    (3<<$VIDPROCCFG{SST_OVERLAY_FILTER_MODE_SHIFT});
$VIDPROCCFG{SST_DESKTOP_PIXEL_FORMAT_SHIFT} =         18;
$VIDPROCCFG{SST_DESKTOP_PIXEL_FORMAT} =               (7<<$VIDPROCCFG{SST_DESKTOP_PIXEL_FORMAT_SHIFT});
	$SST_DESKTOP_PIXEL_FORMAT{SST_DESKTOP_PIXEL_PAL8} =         (0<<$VIDPROCCFG{SST_DESKTOP_PIXEL_FORMAT_SHIFT});
	$SST_DESKTOP_PIXEL_FORMAT{SST_DESKTOP_PIXEL_RGB565} =       (1<<$VIDPROCCFG{SST_DESKTOP_PIXEL_FORMAT_SHIFT});
	$SST_DESKTOP_PIXEL_FORMAT{SST_DESKTOP_PIXEL_RGB24} =        (2<<$VIDPROCCFG{SST_DESKTOP_PIXEL_FORMAT_SHIFT});
	$SST_DESKTOP_PIXEL_FORMAT{SST_DESKTOP_PIXEL_RGB32} =        (3<<$VIDPROCCFG{SST_DESKTOP_PIXEL_FORMAT_SHIFT});
$VIDPROCCFG{SST_OVERLAY_PIXEL_FORMAT_SHIFT} =         21;
$VIDPROCCFG{SST_OVERLAY_PIXEL_FORMAT} =               (7<<$VIDPROCCFG{SST_OVERLAY_PIXEL_FORMAT_SHIFT});
	$SST_OVERLAY_PIXEL_FORMAT{SST_OVERLAY_PIXEL_RGB565U} =      (1<<$VIDPROCCFG{SST_OVERLAY_PIXEL_FORMAT_SHIFT});
	$SST_OVERLAY_PIXEL_FORMAT{SST_OVERLAY_PIXEL_YUV411} =       (4<<$VIDPROCCFG{SST_OVERLAY_PIXEL_FORMAT_SHIFT});
	$SST_OVERLAY_PIXEL_FORMAT{SST_OVERLAY_PIXEL_YUYV422} =      (5<<$VIDPROCCFG{SST_OVERLAY_PIXEL_FORMAT_SHIFT});
	$SST_OVERLAY_PIXEL_FORMAT{SST_OVERLAY_PIXEL_UYVY422} =      (6<<$VIDPROCCFG{SST_OVERLAY_PIXEL_FORMAT_SHIFT});
	$SST_OVERLAY_PIXEL_FORMAT{SST_OVERLAY_PIXEL_RGB565D} =      (7<<$VIDPROCCFG{SST_OVERLAY_PIXEL_FORMAT_SHIFT});
$VIDPROCCFG{SST_DESKTOP_TILED_EN} =                   BIT(24);
$VIDPROCCFG{SST_OVERLAY_TILED_EN} =                   BIT(25);
$VIDPROCCFG{SST_VIDEO_2X_MODE_EN} =                   BIT(26);
$VIDPROCCFG{SST_CURSOR_EN} =                          BIT(27);
$VIDPROCCFG{SST_DISABLE_DESKTOP_MEM_OPT} =            BIT(28);
$VIDPROCCFG{SST_DISABLE_OVERLAY_MEM_OPT} =            BIT(29);
$VIDPROCCFG{SST_OVERLAY_STRIDE_ADJUST} =              BIT(30);
$VIDPROCCFG{SST_OVERLAY_DEINTERLACE_EN} =             BIT(31);

$h3def{SST_CURSOR_X_SHIFT} =                     0;
$h3def{SST_CURSOR_X} =                           (0x7FF<<$h3def{SST_CURSOR_X_SHIFT});
$h3def{SST_CURSOR_Y_SHIFT} =                     16;
$h3def{SST_CURSOR_Y} =                           (0x7FF<<$h3def{SST_CURSOR_Y_SHIFT});

$h3def{SST_CURSOR_BLUE_SHIFT} =                  0;
$h3def{SST_CURSOR_BLUE} =                        (0xFF<<$h3def{SST_CURSOR_BLUE_SHIFT});
$h3def{SST_CURSOR_RED_SHIFT} =                   8;
$h3def{SST_CURSOR_RED} =                         (0xFF<<$h3def{SST_CURSOR_RED_SHIFT});
$h3def{SST_CURSOR_GREEN_SHIFT} =                 16;
$h3def{SST_CURSOR_GREEN} =                       (0xFF<<$h3def{SST_CURSOR_GREEN_SHIFT});

#------------------- vidInFormat

$VIDINFORMAT{SST_VIDEOIN_PIXEL_FORMAT_SHIFT} =         1;
$VIDINFORMAT{SST_VIDEOIN_PIXEL_FORMAT} =               (7<<$VIDINFORMAT{SST_VIDEOIN_FORMAT_SHIFT});
	$SST_VIDEOIN_PIXEL_FORMAT{SST_VIDEOIN_PIXEL_YUV411} =       (4<<$VIDINFORMAT{SST_VIDEOIN_FORMAT_SHIFT});
	$SST_VIDEOIN_PIXEL_FORMAT{SST_VIDEOIN_PIXEL_YUYV422} =      (5<<$VIDINFORMAT{SST_VIDEOIN_FORMAT_SHIFT});
	$SST_VIDEOIN_PIXEL_FORMAT{SST_VIDEOIN_PIXEL_UYUV422} =      (6<<$VIDINFORMAT{SST_VIDEOIN_FORMAT_SHIFT});
$VIDINFORMAT{SST_VIDEOIN_DEINTERLACE_EN} =             BIT(4);
$VIDINFORMAT{SST_VIDEOIN_VSYNC_POLARITY} =             BIT(5);
$VIDINFORMAT{SST_VIDEOIN_HSYNC_POLARITY} =             BIT(6);
$VIDINFORMAT{SST_VIDEOIN_VACTIVE_POLARITY} =           BIT(7);
$VIDINFORMAT{SST_VIDEOIN_G4_FOR_POSEDGE} =             BIT(8);
$VIDINFORMAT{SST_VIDEOIN_BUFFERING_MODE_SHIFT} =       9;
$VIDINFORMAT{SST_VIDEOIN_BUFFERING_MODE} =             (3<<$VIDINFORMAT{SST_VIDEOIN_BUFFERING_MODE_SHIFT});
	$SST_VIDEOIN_BUFFERING_MODE{SST_VIDEOIN_SINGLE_BUFFERING} =   (0<<$VIDINFORMAT{SST_VIDEOIN_BUFFERING_MODE_SHIFT});
	$SST_VIDEOIN_BUFFERING_MODE{SST_VIDEOIN_DOUBLE_BUFFERING} =   (1<<$VIDINFORMAT{SST_VIDEOIN_BUFFERING_MODE_SHIFT});
	$SST_VIDEOIN_BUFFERING_MODE{SST_VIDEOIN_TRIPLE_BUFFERING} =   (2<<$VIDINFORMAT{SST_VIDEOIN_BUFFERING_MODE_SHIFT});
$VIDINFORMAT{SST_VIDEOIN_TILED_EN} =                   BIT(11);
$VIDINFORMAT{SST_TVOUT_VSYNC_POLARITY_LOW} =           BIT(12);
$VIDINFORMAT{SST_TVOUT_HSYNC_POLARITY_LOW} =           BIT(13);
$VIDINFORMAT{SST_VIDEOIN_INTERFACE_SHIFT} =            14;
$VIDINFORMAT{SST_VIDEOIN_INTERFACE} =                  (3<<$VIDINFORMAT{SST_VIDEOIN_INTERFACE_SHIFT});
	$SST_VIDEOIN_INTERFACE{SST_VIDEOIN_INTERFACE_OFF} =      (0<<$VIDINFORMAT{SST_VIDEOIN_INTERFACE_SHIFT});
	$SST_VIDEOIN_INTERFACE{SST_VIDEOIN_VMI_ENABLE} =         (1<<$VIDINFORMAT{SST_VIDEOIN_INTERFACE_SHIFT});
	$SST_VIDEOIN_INTERFACE{SST_VIDEOIN_TVOUT_ENABLE} =       (2<<$VIDINFORMAT{SST_VIDEOIN_INTERFACE_SHIFT});
$VIDINFORMAT{SST_VIDEOIN_GENLOCK_ENABLE} =             BIT(16);
$VIDINFORMAT{SST_VIDEOIN_NOT_USE_VGA_TIMING} =         BIT(17);
$VIDINFORMAT{SST_VIDEOIN_GENLOCK_SOURCE_SHIFT} =       18;
$VIDINFORMAT{SST_VIDEOIN_GENLOCK_SOURCE} =             (1<<$VIDINFORMAT{SST_VIDEOIN_GENLOCK_SOURCE_SHIFT});
        $SST_VIDEOIN_GENLOCK_SOURCE{SST_VIDEOIN_GENLOCK_SOURCE_TV} =  (1<<$VIDINFORMAT{SST_VIDEOIN_GENLOCK_SOURCE_SHIFT});
        $SST_VIDEOIN_GENLOCK_SOURCE{SST_VIDEOIN_GENLOCK_SOURCE_VMI} = (0<<$VIDINFORMAT{SST_VIDEOIN_GENLOCK_SOURCE_SHIFT});
$VIDINFORMAT{SST_VIDEOIN_DATA_CAPTURE_EN} =            BIT(19);
$VIDINFORMAT{SST_VIDEOIN_HORIZ_DECIM_EN} =             BIT(20);
$VIDINFORMAT{SST_VIDEOIN_VERT_DECIM_EN} =              BIT(21);

#------------------- vidInStatusCurrentLine

$VIDINSTATUSCURRENTLINE{SST_VIDEOIN_CURRENT_LINE_SHIFT} =     0;
$VIDINSTATUSCURRENTLINE{SST_VIDEOIN_CURRENT_LINE} =           (SST_MASK(11) << $VIDINSTATUSCURRENTLINE{SST_VIDEOIN_CURRENT_LINE_SHIFT});
$VIDINSTATUSCURRENTLINE{SST_VIDEOIN_LAST_FIELD} =             BIT(16);
$VIDINSTATUSCURRENTLINE{SST_VIDEOIN_LAST_BUFFER_SHIFT} =      17;
$VIDINSTATUSCURRENTLINE{SST_VIDEOIN_LAST_BUFFER} =            (3<<$VIDINSTATUSCURRENTLINE{SST_VIDEOIN_LAST_BUFFER_SHIFT});
        $SST_VIDEOIN_LAST_BUFFER{SST_VIDEOIN_VMI_BUFFER_0} =  (0<<$VIDINSTATUSCURRENTLINE{SST_VIDEOIN_LAST_BUFFER_SHIFT});
        $SST_VIDEOIN_LAST_BUFFER{SST_VIDEOIN_VMI_BUFFER_1} =  (1<<$VIDINSTATUSCURRENTLINE{SST_VIDEOIN_LAST_BUFFER_SHIFT});
        $SST_VIDEOIN_LAST_BUFFER{SST_VIDEOIN_VMI_BUFFER_2} =  (2<<$VIDINSTATUSCURRENTLINE{SST_VIDEOIN_LAST_BUFFER_SHIFT});
        $SST_VIDEOIN_LAST_BUFFER{SST_VIDEOIN_VMI_NO_BUFFER} = (3<<$VIDINSTATUSCURRENTLINE{SST_VIDEOIN_LAST_BUFFER_SHIFT});

#------------------- vidSerialParallelPort

$VIDSERIALPARALLELPORT{SST_SERPAR_VMI_EN} =            BIT(0);
$VIDSERIALPARALLELPORT{SST_SERPAR_VMI_CS_N} =          BIT(1);
$VIDSERIALPARALLELPORT{SST_SERPAR_VMI_DS_N} =          BIT(2);
$VIDSERIALPARALLELPORT{SST_SERPAR_VMI_RD_N} =          BIT(2);
$VIDSERIALPARALLELPORT{SST_SERPAR_VMI_RW_N} =          BIT(3);
$VIDSERIALPARALLELPORT{SST_SERPAR_VMI_DTACK_N} =       BIT(4);
$VIDSERIALPARALLELPORT{SST_SERPAR_VMI_RDY} =           BIT(4);
$VIDSERIALPARALLELPORT{SST_SERPAR_VMI_DATA_OE_N} =     BIT(5);
$VIDSERIALPARALLELPORT{SST_SERPAR_VMI_DATA_SHIFT} =    6;
$VIDSERIALPARALLELPORT{SST_SERPAR_VMI_DATA} =          (SST_MASK(8)<<SST_SERPAR_VMI_DATA_SHIFT);
$VIDSERIALPARALLELPORT{SST_SERPAR_VMI_ADDR_SHIFT} =    14;
$VIDSERIALPARALLELPORT{SST_SERPAR_VMI_ADDR} =          (SST_MASK(4)<<SST_SERPAR_VMI_ADDR_SHIFT);
$VIDSERIALPARALLELPORT{SST_SERPAR_DDC_EN} =            BIT(18);
$VIDSERIALPARALLELPORT{SST_SERPAR_DDC_DCK_OUT} =       BIT(19);
$VIDSERIALPARALLELPORT{SST_SERPAR_DDC_DDA_OUT} =       BIT(20);
$VIDSERIALPARALLELPORT{SST_SERPAR_DDC_DCK_IN} =        BIT(21);
$VIDSERIALPARALLELPORT{SST_SERPAR_DDC_DDA_IN} =        BIT(22);
$VIDSERIALPARALLELPORT{SST_SERPAR_I2C_EN} =            BIT(23);
$VIDSERIALPARALLELPORT{SST_SERPAR_I2C_SCK_OUT} =       BIT(24);
$VIDSERIALPARALLELPORT{SST_SERPAR_I2C_DSA_OUT} =       BIT(25);
$VIDSERIALPARALLELPORT{SST_SERPAR_I2C_SCK_IN} =        BIT(26);
$VIDSERIALPARALLELPORT{SST_SERPAR_I2C_DSA_IN} =        BIT(27);
$VIDSERIALPARALLELPORT{SST_SERPAR_VMI_RESET_N} =       BIT(28);
$VIDSERIALPARALLELPORT{SST_SERPAR_GPIO_1} =            BIT(29);
$VIDSERIALPARALLELPORT{SST_SERPAR_GPIO_2} =            BIT(30);
$VIDSERIALPARALLELPORT{TVOUT_RESET_N} =                BIT(31);

#------------------- vidInXDecimDeltas, vidInYDecimDeltas

$h3def{SST_VIDEOIN_DECIM_POS_DELTA_SHIFT} =      0;
$h3def{SST_VIDEOIN_DECIM_POS_DELTA} =            (0xFFF<<$h3def{SST_VIDEOIN_DECIM_POS_DELTA_SHIFT});
$h3def{SST_VIDEOIN_DECIM_NEG_DELTA_SHIFT} =      16;
$h3def{SST_VIDEOIN_DECIM_NEG_DELTA} =            (0xFFF<<$h3def{SST_VIDEOIN_DECIM_NEG_DELTA_SHIFT});

#------------------- vidInDecimInitErrs

$h3def{SST_VIDEOIN_X_DECIM_ERROR_SHIFT} =        0;
$h3def{SST_VIDEOIN_X_DECIM_ERROR} =              (0x1FFF<<$h3def{SST_VIDEOIN_X_DECIM_ERROR_SHIFT});
$h3def{SST_VIDEOIN_Y_DECIM_ERROR_SHIFT} =        16;
$h3def{SST_VIDEOIN_Y_DECIM_ERROR} =              (0x1FFF<<$h3def{SST_VIDEOIN_Y_DECIM_ERROR_SHIFT});

# skipping vidPixelBufThold
# skipping vidChromaKeyMin/Max

#-------------vidScreenSize
 
$VIDSCREENSIZE{SST_VIDEO_SCREEN_WIDTH_SHIFT} =           0;
$VIDSCREENSIZE{SST_VIDEO_SCREEN_WIDTH} =                 (0xFFF<<$VIDSCREENSIZE{SST_VIDEO_SCREEN_WIDTH_SHIFT});
$VIDSCREENSIZE{SST_VIDEO_SCREEN_HEIGHT_SHIFT} =          12;
$VIDSCREENSIZE{SST_VIDEO_SCREEN_HEIGHT} =                (0xFFF<<$VIDSCREENSIZE{SST_VIDEO_SCREEN_HEIGHT_SHIFT});

#-------------vidOverlayStartCoords, vidOverlayEndCoords
$h3def{SST_OVERLAY_X_SHIFT} =                    0;
$h3def{SST_OVERLAY_X} =                          (0xFFF<<$h3def{SST_OVERLAY_X_SHIFT});
$h3def{SST_OVERLAY_Y_SHIFT} =                    12;
$h3def{SST_OVERLAY_Y} =                          (0xFFF<<$h3def{SST_OVERLAY_Y_SHIFT});
$h3def{SST_OVERLAY_XADJ_SHIFT} =                 24;
$h3def{SST_OVERLAY_XADJ} =                       (3<<$h3def{SST_OVERLAY_XADJ_SHIFT});
$h3def{SST_OVERLAY_YADJ_SHIFT} =                 26;
$h3def{SST_OVERLAY_YADJ} =                       (3<<$h3def{SST_OVERLAY_YADJ_SHIFT});

#------------vidOverlayDudxOffsetSrcWidth, vidOverlayDvDy
$h3def{SST_OVERLAY_DELTA} =                      SST_MASK(20);
$h3def{SST_OVERLAY_DELTA_OFFSET} =               SST_MASK(19);
$h3def{SST_OVERLAY_FETCH_SIZE_SHIFT} =           19;
$h3def{SST_OVERLAY_FETCH_SIZE} =                 (0x1FFF<<$h3def{SST_OVERLAY_FETCH_SIZE_SHIFT});

$h3def{SST_VIDEO_START_ADDR_SHIFT} =             0;
$h3def{SST_VIDEO_START_ADDR} =                   (0xFFFFFF<<$h3def{SST_VIDEO_START_ADDR_SHIFT});
$h3def{SST_DESKTOP_STRIDE_SHIFT} =               0;
$h3def{SST_DESKTOP_LINEAR_STRIDE} =              (0x3FFF<<$h3def{SST_DESKTOP_STRIDE_SHIFT});
$h3def{SST_DESKTOP_TILE_STRIDE} =                (0x7F<<$h3def{SST_DESKTOP_STRIDE_SHIFT});
$h3def{SST_OVERLAY_STRIDE_SHIFT} =               16;
$h3def{SST_OVERLAY_LINEAR_STRIDE} =              (0x3FFF<<$h3def{SST_OVERLAY_STRIDE_SHIFT});
$h3def{SST_OVERLAY_TILE_STRIDE} =                (0x7F<<$h3def{SST_OVERLAY_STRIDE_SHIFT});

#------------vidInStride
$h3def{SST_VIDEO_IN_STRIDE_SHIFT} =              0;
$h3def{SST_VIDEO_IN_LINEAR_STRIDE} =             (0x3FFF<<$h3def{SST_VIDEO_IN_STRIDE_SHIFT});
$h3def{SST_VIDEO_IN_TILE_STRIDE} =               (0x7F<<$h3def{SST_VIDEO_IN_STRIDE_SHIFT});

#----------------- SST *Init*  ----------------------------

# pciInit0
$PCIINIT0{SST_PCI_STALL_ENABLE} =              BIT(0);
$PCIINIT0{SST_PCI_LOWTHRESH_SHIFT} =           2;
$PCIINIT0{SST_PCI_LOWTHRESH} =                 (SST_MASK(4) << $PCIINIT0{SST_PCI_LOWTHRESH_SHIFT});;
$PCIINIT0{SST_PCI_HARDCODE_BASE} =             BIT(7);
$PCIINIT0{SST_PCI_READ_WS} =                   BIT(8);
$PCIINIT0{SST_PCI_WRITE_WS} =                  BIT(9);
$PCIINIT0{SST_PCI_DISABLE_IO} =                BIT(11);
$PCIINIT0{SST_PCI_DISABLE_MEM} =               BIT(12);
$PCIINIT0{SST_PCI_RETRY_INTERVAL_SHIFT} =      13;
$PCIINIT0{SST_PCI_RETRY_INTERVAL} =            (SST_MASK(5) << $PCIINIT0{SST_PCI_RETRY_INTERVAL_SHIFT});;
$PCIINIT0{SST_PCI_INTERRUPT_ENABLE} =          BIT(18);
$PCIINIT0{SST_PCI_TIMEOUT_ENABLE} =            BIT(19);
$PCIINIT0{SST_PCI_FIFO_READ_THRESHOLD_SHIFT} = 20;
$PCIINIT0{SST_PCI_FIFO_READ_THRESHOLD} =       (SST_MASK(6) << $PCIINIT0{SST_FIFO_READ_THRESHOLD_SHIFT});
$PCIINIT0{SST_PCI_CMD_FB_HI_PRIORITY} =        BIT(26);
$PCIINIT0{SST_PCI_TO_LFB_WR_TIMEOUT_SHIFT} =   27;
$PCIINIT0{SST_PCI_TO_LFB_WR_TIMEOUT} =         (SST_MASK(2) << $PCIINIT0{SST_PCI_TO_LFB_WR_TIMEOUT_SHIFT});

#sipMonitor
$SIPMONITOR{SST_SIPROCESS_OSC_CNTR}=             0xFFFF;
$SIPMONITOR{SST_SIPROCESS_PCI_CNTR_SHIFT}=       16;
$SIPMONITOR{SST_SIPROCESS_PCI_CNTR}=             (0xFFF<<$SIPMONITOR{SST_SIPROCESS_PCI_CNTR_SHIFT});
$SIPMONITOR{SST_SIPROCESS_OSC_CNTR_RESET_N}=     0;
$SIPMONITOR{SST_SIPROCESS_OSC_CNTR_RUN}=         BIT(28);
$SIPMONITOR{SST_SIPROCESS_OSC_NAND_SEL}=         0;
$SIPMONITOR{SST_SIPROCESS_OSC_NOR_SEL}=          BIT(29);
$SIPMONITOR{SST_SIPROCESS_OSC_FORCE_ENABLE}=     BIT(30);


# lfbMemoryConfig
$LFBMEMORYCONFIG{SST_RAW_LFB_TILE_BEGIN_PAGE_SHIFT} = 	0;
$LFBMEMORYCONFIG{SST_RAW_LFB_TILE_BEGIN_PAGE} = 		(0x1FFF<<$LFBMEMORYCONFIG{SST_RAW_LFB_TILE_BEGIN_PAGE_SHIFT});
$LFBMEMORYCONFIG{SST_RAW_LFB_ADDR_STRIDE_SHIFT} =    	13;
$LFBMEMORYCONFIG{SST_RAW_LFB_ADDR_STRIDE} =          	(0x7<<$LFBMEMORYCONFIG{SST_RAW_LFB_ADDR_STRIDE_SHIFT});
	$SST_RAW_LFB_ADDR_STRIDE{SST_RAW_LFB_ADDR_STRIDE_1K} =	(0x0<<$LFBMEMORYCONFIG{SST_RAW_LFB_ADDR_STRIDE_SHIFT});
	$SST_RAW_LFB_ADDR_STRIDE{SST_RAW_LFB_ADDR_STRIDE_2K} =	(0x1<<$LFBMEMORYCONFIG{SST_RAW_LFB_ADDR_STRIDE_SHIFT});
	$SST_RAW_LFB_ADDR_STRIDE{SST_RAW_LFB_ADDR_STRIDE_4K} =	(0x2<<$LFBMEMORYCONFIG{SST_RAW_LFB_ADDR_STRIDE_SHIFT});
	$SST_RAW_LFB_ADDR_STRIDE{SST_RAW_LFB_ADDR_STRIDE_8K} =	(0x3<<$LFBMEMORYCONFIG{SST_RAW_LFB_ADDR_STRIDE_SHIFT});
	$SST_RAW_LFB_ADDR_STRIDE{SST_RAW_LFB_ADDR_STRIDE_16K} =	(0x4<<$LFBMEMORYCONFIG{SST_RAW_LFB_ADDR_STRIDE_SHIFT});
	$SST_RAW_LFB_ADDR_STRIDE{SST_RAW_LFB_ADDR_STRIDE_MAX} =	(0x4<<$LFBMEMORYCONFIG{SST_RAW_LFB_ADDR_STRIDE_SHIFT});
$LFBMEMORYCONFIG{SST_RAW_LFB_TILE_STRIDE_SHIFT} =     	16;
$LFBMEMORYCONFIG{SST_RAW_LFB_TILE_STRIDE} =           	(0x7F<<$LFBMEMORYCONFIG{SST_RAW_LFB_TILE_STRIDE_SHIFT});

$LFBMEMORYCONFIG{SST_RAW_LFB_ADDR_SHIFT} = 			0;
$LFBMEMORYCONFIG{SST_RAW_LFB_ADDR} = 			(0x1FFFFFF<<$LFBMEMORYCONFIG{SST_RAW_LFB_ADDR_SHIFT});

#
#
#lfbMemoryTileCtrl and lfbMemoryTileCompare are not written out here
#
#

# miscInit0
$MISCINIT0{SST_GRX_RESET} =            	        BIT(0);
$MISCINIT0{SST_FBI_FIFO_RESET} =        	BIT(1);
$MISCINIT0{SST_REGISTER_BYTE_SWIZZLE_EN} =      BIT(2);
$MISCINIT0{SST_REGISTER_WORD_SWIZZLE_EN} =      BIT(3);
$MISCINIT0{SST_VIDEO_RESET} =           	BIT(4);
$MISCINIT0{SST_2D_RESET} =              	BIT(5);
$MISCINIT0{SST_MEMORY_TIMING_RESET} =   	BIT(6);
$MISCINIT0{SST_VGA_TIMING_RESET} =   	        BIT(7);
$MISCINIT0{SST_TVOUT_BLANK_DELAY_SHIFT} =   	8;
$MISCINIT0{SST_TVOUT_BLANK_DELAY} =   	        (0x7<<$MISCINIT0{SST_TVOUT_BLANK_DELAY_SHIFT});
$MISCINIT0{SST_TVOUT_SYNC_DELAY_SHIFT} =   	11;
$MISCINIT0{SST_TVOUT_SYNC_DELAY} =   	        (0x7<<$MISCINIT0{SST_TVOUT_SYNC_DELAY_SHIFT});
$MISCINIT0{SST_MONITOR_BLANK_DELAY_SHIFT} =   	14;
$MISCINIT0{SST_MONITOR_BLANK_DELAY} =   	(0x7<<$MISCINIT0{SST_MONITOR_BLANK_DELAY_SHIFT});
$MISCINIT0{SST_YORIGIN_TOP_SHIFT} =     	(18);
$MISCINIT0{SST_YORIGIN_TOP} =           	(0xFFF<<$MISCINIT0{SST_YORIGIN_TOP_SHIFT});
$MISCINIT0{SST_RAWLFB_BYTE_SWIZZLE_EN} =       BIT(30);
$MISCINIT0{SST_RAWLFB_WORD_SWIZZLE_EN} =       BIT(31);

# miscInit1
$MISCINIT1{SST_H3_CLUTINVERT} =		BIT(0);
$MISCINIT1{SST_ALT_REGMAPPING} =       	BIT(5);
$MISCINIT1{SST_DISABLE_TEXTURE} = 		BIT(6);
$MISCINIT1{SST_POWERDOWN_CLUT} = 		BIT(7);
$MISCINIT1{SST_POWERDOWN_DAC} = 		BIT(8);
$MISCINIT1{SST_POWERDOWN_VID_PLL} = 		BIT(9);
$MISCINIT1{SST_POWERDOWN_GRX_PLL} = 		BIT(10);
$MISCINIT1{SST_POWERDOWN_MEM_PLL} = 		BIT(11);
$MISCINIT1{SST_BLOCK_WRITE_THRESH_SHIFT} =  	12;
$MISCINIT1{SST_BLOCK_WRITE_THRESH} =  	 	(0x7<<$MISCINIT1{SST_BLOCK_WRITE_THRESH_SHIFT});
$MISCINIT1{SST_DISABLE_2D_BLOCK_WRITE} =  	BIT(15);
$MISCINIT1{SST_DISABLE_3D_BLOCK_WRITE} =  	BIT(16);
$MISCINIT1{SST_DISABLE_2D_STALL_ON_3D} =  	BIT(17);
$MISCINIT1{SST_DISABLE_3D_STALL_ON_2D} =  	BIT(18);

# dramInit0
$DRAMINIT0{SST_SGRAM_RRD_SHIFT} = 		0;
$DRAMINIT0{SST_SGRAM_RRD} = 			(0x3<<$DRAMINIT0{SST_SGRAM_RRD_SHIFT});
$DRAMINIT0{SST_SGRAM_RCD_SHIFT} = 		2;
$DRAMINIT0{SST_SGRAM_RCD} = 			(0x3<<$DRAMINIT0{SST_SGRAM_RCD_SHIFT});
$DRAMINIT0{SST_SGRAM_RP_SHIFT} = 		4;
$DRAMINIT0{SST_SGRAM_RP} = 			(0x3<<$DRAMINIT0{SST_SGRAM_RP_SHIFT});
$DRAMINIT0{SST_SGRAM_RAS_SHIFT} = 		6;
$DRAMINIT0{SST_SGRAM_RAS} = 			(0xF<<$DRAMINIT0{SST_SGRAM_RAS_SHIFT});
$DRAMINIT0{SST_SGRAM_RC_SHIFT} = 		10;
$DRAMINIT0{SST_SGRAM_RC} = 			(0xF<<$DRAMINIT0{SST_SGRAM_RC_SHIFT});
$DRAMINIT0{SST_SGRAM_CAS_SHIFT} = 		14;
$DRAMINIT0{SST_SGRAM_CAS} = 			(0x3<<$DRAMINIT0{SST_SGRAM_CAS_SHIFT});
$DRAMINIT0{SST_SGRAM_MRS_SHIFT} = 		16;
$DRAMINIT0{SST_SGRAM_MRS} = 			BIT(16);
$DRAMINIT0{SST_SGRAM_RD2DQM} = 		BIT(17);
$DRAMINIT0{SST_SGRAM_BWC} = 			BIT(18);
$DRAMINIT0{SST_SGRAM_WL} = 			BIT(19);
$DRAMINIT0{SST_SGRAM_BWL_SHIFT} = 		20;
$DRAMINIT0{SST_SGRAM_BWL} = 			(0x3<<$DRAMINIT0{SST_SGRAM_BWL_SHIFT});
$DRAMINIT0{SST_SGRAM_RL} = 			BIT(22);
$DRAMINIT0{SST_SGRAM_USE_BST} = 		BIT(23);
$DRAMINIT0{SST_SGRAM_NO_DEAD} = 		BIT(24);
$DRAMINIT0{SST_SGRAM_EN_WPB} = 		BIT(25);
$DRAMINIT0{SST_SGRAM_NUM_CHIPSETS} = 		BIT(26);
$DRAMINIT0{SST_SGRAM_TYPE_SHIFT} = 		27;
$DRAMINIT0{SST_SGRAM_TYPE} = 			(0x7<<$DRAMINIT0{SST_SGRAM_TYPE_SHIFT});
	$SST_SGRAM_TYPE{SST_SGRAM_TYPE_8MBIT}  =	(0x0<<$DRAMINIT0{SST_SGRAM_TYPE_SHIFT});
	$SST_SGRAM_TYPE{SST_SGRAM_TYPE_16MBIT} =	(0x1<<$DRAMINIT0{SST_SGRAM_TYPE_SHIFT});
	$SST_SGRAM_TYPE{SST_SGRAM_TYPE_32MBIT} =	(0x2<<$DRAMINIT0{SST_SGRAM_TYPE_SHIFT});
	$SST_SGRAM_TYPE{SST_SGRAM_TYPE_64MBIT} =	(0x3<<$DRAMINIT0{SST_SGRAM_TYPE_SHIFT});
	$SST_SGRAM_TYPE{SST_SGRAM_TYPE_128MBIT} =	(0x4<<$DRAMINIT0{SST_SGRAM_TYPE_SHIFT});
$DRAMINIT0{SST_MCTL_DRAM_NUMBANKS} =          BIT(30);


# dramInit1
$DRAMINIT1{SST_DRAM_REFRESH_EN} = 			BIT(0);
$DRAMINIT1{SST_DRAM_REFRESH_VALUE_SHIFT} =  	1;
$DRAMINIT1{SST_DRAM_REFRESH_VALUE} = 		(0x1FF<<$DRAMINIT1{SST_DRAM_REFRESH_VALUE_SHIFT});
$DRAMINIT1{SST_VIDEO_OVERRIDE_EN} = 		BIT(10);
$DRAMINIT1{SST_TRIPLE_BUFFER_EN} = 			BIT(11);
$DRAMINIT1{SST_DITHER_PASSTHROUGH} = 		BIT(12);
$DRAMINIT1{SST_SGRAM_CLK_NODELAY} =			BIT(13);
$DRAMINIT1{SST_SGRAM_USE_INV_SAMPLE} =		BIT(14);
$DRAMINIT1{SST_SGRAM_DEL_CLK_INVERT} =		BIT(15);
$DRAMINIT1{SST_SGRAM_CLK_ADJ_SHIFT} =		16;
$DRAMINIT1{SST_SGRAM_CLK_ADJ} =			(0xF<<$DRAMINIT1{SST_SGRAM_CLK_ADJ_SHIFT});
$DRAMINIT1{SST_SGRAM_OFLOP_DEL_ADJ_SHIFT} =		20;
$DRAMINIT1{SST_SGRAM_OFLOP_DEL_ADJ} =		(0xF<<$DRAMINIT1{SST_SGRAM_OFLOP_DEL_ADJ_SHIFT});
$DRAMINIT1{SST_SGRAM_OFLOP_TRANS_LATCH} =		BIT(24);
$DRAMINIT1{SST_MCTL_SHORT_POWER_ON} =		BIT(25);
$DRAMINIT1{SST_MCTL_NO_AGGRESSIVE} =		BIT(26);
$DRAMINIT1{SST_MCTL_PAGEBREAK} =			BIT(27);
$DRAMINIT1{SST_MCTL_TRISTATE_OUTPUTS} =		BIT(28);
$DRAMINIT1{SST_MCTL_NO_VIN_LOCKING} =		BIT(29);
$DRAMINIT1{SST_MCTL_TYPE_SDRAM} =		BIT(30);

# agpInit0
$AGPINIT0{SST_AGP_LTOH} =            		BIT(0);
$AGPINIT0{SST_AGP_RQSIZE_SHIFT} =    		BIT(1);
$AGPINIT0{SST_AGP_RQSIZE} =         		(0x7<<$AGPINIT0{SST_AGP_RQSIZE_SHIFT});
$AGPINIT0{SST_AGP_REQFULL_THRESH_SHIFT} =     4 ;
$AGPINIT0{SST_AGP_REQFULL_THRESH} =           (0x7<<$AGPINIT0{SST_AGP_REQFULL_THRESH_SHIFT});
$AGPINIT0{SST_AGP_READFULL_THRESH_SHIFT} =    7 ;
$AGPINIT0{SST_AGP_READFULL_THRESH} =          (0xF<<$AGPINIT0{SST_AGP_READFULL_THRESH_SHIFT});

# tmuGbeInit
$TMUGBEINIT{SST_TGI_READLOW_THRESH_SHIFT} = 	0;
$TMUGBEINIT{SST_TGI_READLOW_THRESH} =           (SST_MASK(4)<<$TMUGBEINIT{SST_TGI_READLOW_THRESH_SHIFT});
$TMUGBEINIT{SST_TGI_READHIGH_THRESH_SHIFT} =    4 ;
$TMUGBEINIT{SST_TGI_READHIGH_THRESH} = 		(SST_MASK(4)<<$TMUGBEINIT{SST_TGI_READHIGH_THRESH_SHIFT});
$TMUGBEINIT{SST_TGI_PIXELHIGH_THRESH_SHIFT} =   8;
$TMUGBEINIT{SST_TGI_PIXELHIGH_THRESH} =         (SST_MASK(4)<<$TMUGBEINIT{SST_TGI_PIXELHIGH_THRESH_SHIFT});
$TMUGBEINIT{SST_TXC_DISABLE_RDRQ_MAX} =         BIT(12);
$TMUGBEINIT{SST_TXC_USE_MIN_REQ} =              BIT(13);
$TMUGBEINIT{SST_TXC_FORCE_CAM_MISS} =           BIT(14);
$TMUGBEINIT{SST_TVOUT_CLK_INVERT} =             BIT(15);
$TMUGBEINIT{SST_TVOUT_CLK_DELAY_SHIFT} =        16;
$TMUGBEINIT{SST_TVOUT_CLK_DELAY} =              (SST_MASK(4)<<$TMUGBEINIT{SST_TVOUT_CLK_DELAY_SHIFT});

#----------------- fastfillCMD bits --------------------------
$FASTFILLCMD{SST_FASTFILL_DISABLE_DITHER} =      BIT(0);

#----------------- nopCMD bits -------------------------------
$NOPCMD{SST_NOP_RESET_PIXEL_STATS} =        BIT(0);
$NOPCMD{SST_NOP_RESET_TRIANGLE_STATS} =     BIT(1);

#----------------- swapBufferCMD bits ------------------------
$SWAPBUFFERCMD{SST_SWAP_EN_WAIT_ON_VSYNC} =        BIT(0);
$SWAPBUFFERCMD{SST_SWAP_BUFFER_INTERVAL_SHIFT} =   1;
$SWAPBUFFERCMD{SST_SWAP_BUFFER_INTERVAL} =         (0xff << $h3def{SST_SWAP_BUFFER_INTERVAL_SHIFT});;
$SWAPBUFFERCMD{SST_SWAP_DONT_SWAP} =               BIT(9);

#----------------- SST Configuration Registers ---------------
$h3def{SST_MEMBASE_ADDR} =                 0x10;
$h3def{SST_INIT_ENABLE} =                  0x40;
$h3def{SST_BUS_SNOOP0} =                   0x44;
$h3def{SST_BUS_SNOOP1} =                   0x48;
$h3def{SST_CFG_STATUS} =                   0x4c;

#----------------- SST Remapped Registers --------------------
$h3def{SST_DAC_RD_DATA} =                  0x00000218;
$h3def{SST_VIDEO_CHECKSUM} =               0x0000021c;

#----------------- SST initEnable (Config register) bits -----
$h3def{SST_INITWR_EN} =                    BIT(0);
$h3def{SST_PCI_FIFOWR_EN} =                BIT(1);
$h3def{SST_FBIINIT23_REMAP} =              BIT(2);
$h3def{SST_SNOOP0_EN} =                    BIT(4);
$h3def{SST_SNOOP0_IO} =                    BIT(5);
$h3def{SST_SNOOP0_RD} =                    BIT(6);
$h3def{SST_SNOOP1_EN} =                    BIT(7);
$h3def{SST_SNOOP1_IO} =                    BIT(8);
$h3def{SST_SNOOP1_RD} =                    BIT(9);
$h3def{SST_SCANLINE_SLV_OWNPCI} =          BIT(10);
$h3def{SST_SCANLINE_SLI_SLV} =             BIT(11);

#----------------- SST vRetrace bits ---------------------
$h3def{SST_VRETRACE_VSYNC_OFF} =   0xff;

#----------------- SST hSync bits ------------------------
$h3def{SST_HSYNC_ON} =             0xff;
$h3def{SST_HSYNC_OFF_SHIFT} =      16;
$h3def{SST_HSYNC_OFF} =            (0x3ff << $h3def{SST_HSYNCOFF_SHIFT});;

#----------------- SST vSync bits ------------------------
$h3def{SST_VSYNC_ON} =             0xff;
$h3def{SST_VSYNC_OFF_SHIFT} =      16;
$h3def{SST_VSYNC_OFF} =            (0xfff << $h3def{SST_VSYNC_OFF_SHIFT});;

#----------------- SST backPorch bits --------------------
$h3def{SST_HBACKPORCH} =           0xff;
$h3def{SST_VBACKPORCH_SHIFT} =     16;
$h3def{SST_VBACKPORCH} =           (0xff << $h3def{SST_VBACKPORCH_SHIFT});;

#----------------- SST videoDimensionS Register -----------
$h3def{SST_VIDEODIMENSIONS_X} =    0x3ff;
$h3def{SST_VIDEODIMENSIONS_Y_SHIFT} =  16;
$h3def{SST_VIDEODIMENSIONS_Y} =    0x3ff;

#-------------------- cfgSliLfbCtrl --------------------#
$CFGSLILFBCTRL{SST_SLI_LFB_RENDERMASK_SHIFT}=      0;
$CFGSLILFBCTRL{SST_SLI_LFB_RENDERMASK}=            (0xFF << $CFGSLILFBCTRL{SST_SLI_LFB_RENDERMASK_SHIFT});
$CFGSLILFBCTRL{SST_SLI_LFB_COMPAREMASK_SHIFT}=     8;
$CFGSLILFBCTRL{SST_SLI_LFB_COMPAREMASK}=           (0xFF << $CFGSLILFBCTRL{SST_SLI_LFB_COMPAREMASK_SHIFT});
$CFGSLILFBCTRL{SST_SLI_LFB_SCANMASK_SHIFT}=        16;
$CFGSLILFBCTRL{SST_SLI_LFB_SCANMASK}=              (0xFF << $CFGSLILFBCTRL{SST_SLI_LFB_SCANMASK_SHIFT});
$CFGSLILFBCTRL{SST_SLI_LFB_NUMCHIPS_LOG2_SHIFT}=   24;
$CFGSLILFBCTRL{SST_SLI_LFB_NUMCHIPS_LOG2}=         (0x3 << $CFGSLILFBCTRL{SST_SLI_LFB_NUMCHIPS_LOG2_SHIFT});
$CFGSLILFBCTRL{SST_SLI_LFB_CPU_WRITE_ENABLE}=      BIT(26);
$CFGSLILFBCTRL{SST_SLI_LFB_DISPATCH_WRITE_ENABLE}= BIT(27);
$CFGSLILFBCTRL{SST_SLI_LFB_READ_ENABLE}=           BIT(28);


#-------------------- cfgAADepthBufferAperture --------------------#
$CFGAADEPTHBUFFERAPERTURE{SST_AA_DEPTH_BUFFER_APERTURE_BEGIN_SHIFT}=  0;
$CFGAADEPTHBUFFERAPERTURE{SST_AA_DEPTH_BUFFER_APERTURE_BEGIN}=        
    (0x7FFF << $CFGAADEPTHBUFFERAPERTURE{SST_AA_DEPTH_BUFFER_APERTURE_BEGIN_SHIFT});
$CFGAADEPTHBUFFERAPERTURE{SST_AA_DEPTH_BUFFER_APERTURE_END_SHIFT}=    16;
$CFGAADEPTHBUFFERAPERTURE{SST_AA_DEPTH_BUFFER_APERTURE_END}=          
    (0xFFFF << $CFGAADEPTHBUFFERAPERTURE{SST_AA_DEPTH_BUFFER_APERTURE_END_SHIFT});


#-------------------- cfgAALfbCtrl --------------------#
$CFGAALFBCTRL{SST_SECONDARY_BUFFER_BASE_SHIFT}=    4;
$CFGAALFBCTRL{SST_SECONDARY_BUFFER_BASE}=          (0x3FFFFF << $CFGAALFBCTRL{SST_SECONDARY_BUFFER_BASE_SHIFT});
$CFGAALFBCTRL{SST_AA_LFB_CPU_WRITE_ENABLE}=        BIT(26);
$CFGAALFBCTRL{SST_AA_LFB_DISPATCH_WRITE_ENABLE}=   BIT(27);
$CFGAALFBCTRL{SST_AA_LFB_READ_ENABLE}=             BIT(28);
$CFGAALFBCTRL{SST_AA_LFB_READ_FORMAT_SHIFT}=       29;
$CFGAALFBCTRL{SST_AA_LFB_READ_FORMAT}=             (0x3 << $CFGAALFBCTRL{SST_AA_LFB_READ_FORMAT_SHIFT});
$CFGAALFBCTRL{SST_AA_LFB_READ_FORMAT_16BPP}=       (0 << $CFGAALFBCTRL{SST_AA_LFB_READ_FORMAT_SHIFT}); 
$CFGAALFBCTRL{SST_AA_LFB_READ_FORMAT_15BPP}=       (1 << $CFGAALFBCTRL{SST_AA_LFB_READ_FORMAT_SHIFT}); 
$CFGAALFBCTRL{SST_AA_LFB_READ_FORMAT_32BPP}=       (2 << $CFGAALFBCTRL{SST_AA_LFB_READ_FORMAT_SHIFT}); 
$CFGAALFBCTRL{SST_AA_LFB_RD_DIVIDE_BY_FOUR}=       BIT(31);


# Here are the defines for the alternate register mappings
$h3def{r_ALT} =            "r";
$h3def{drdx_ALT} =         "g";
$h3def{drdy_ALT} =         "b";
$h3def{g_ALT} =            "z";
$h3def{dgdx_ALT} =         "a";
$h3def{dgdy_ALT} =         "s";
$h3def{b_ALT} =            "t";
$h3def{dbdx_ALT} =         "w";

$h3def{dbdy_ALT} =         "drdx";
$h3def{z_ALT} =            "dgdx";
$h3def{dzdx_ALT} =         "dbdx";
$h3def{dzdy_ALT} =         "dzdx";
$h3def{a_ALT} =            "dadx";
$h3def{dadx_ALT} =         "dsdx";
$h3def{dady_ALT} =         "dtdx";
$h3def{s_ALT} =            "dwdx";

$h3def{dsdx_ALT} =         "drdy";
$h3def{dsdy_ALT} =         "dgdy";
$h3def{t_ALT} =            "dbdy";
$h3def{dtdx_ALT} =         "dzdy";
$h3def{dtdy_ALT} =         "dady";
$h3def{w_ALT} =            "dsdy";
$h3def{dwdx_ALT} =         "dtdy";
$h3def{dwdy_ALT} =         "dwdy";

$h3def{Fr_ALT} =           "Fr";
$h3def{Fdrdx_ALT} =        "Fg";
$h3def{Fdrdy_ALT} =        "Fb";
$h3def{Fg_ALT} =           "Fz";
$h3def{Fdgdx_ALT} =        "Fa";
$h3def{Fdgdy_ALT} =        "Fs";
$h3def{Fb_ALT} =           "Ft";
$h3def{Fdbdx_ALT} =        "Fw";

$h3def{Fdbdy_ALT} =        "Fdrdx";
$h3def{Fz_ALT} =           "Fdgdx";
$h3def{Fdzdx_ALT} =        "Fdbdx";
$h3def{Fdzdy_ALT} =        "Fdzdx";
$h3def{Fa_ALT} =           "Fdadx";
$h3def{Fdadx_ALT} =        "Fdsdx";
$h3def{Fdady_ALT} =        "Fdtdx";
$h3def{Fs_ALT} =           "Fdwdx";

$h3def{Fdsdx_ALT} =        "Fdrdy";
$h3def{Fdsdy_ALT} =        "Fdgdy";
$h3def{Ft_ALT} =           "Fdbdy";
$h3def{Fdtdx_ALT} =        "Fdzdy";
$h3def{Fdtdy_ALT} =        "Fdady";
$h3def{Fw_ALT} =           "Fdsdy";
$h3def{Fdwdx_ALT} =        "Fdtdy";
$h3def{Fdwdy_ALT} =        "Fdwdy";

#----------------- useful addressing macros -----------------------
# return pointer to SST at specified WRAP, CHIP, or TREX
sub SST_WRAP	{ my($sst,$n) = @_; return (($n)*0x4000+($sst)); }
sub SST_CHIP	{ my($sst,$n) = @_; return (($n)*0x400+($sst)); }
sub SST_TMU	{ my($sst,$n) = @_; return ((0x800<<($n))+($sst)); }
sub SST_TREX	{ my($sst,$n) = @_; return SST_TMU($sst,$n); }

# offsets from the base of memBaseAddr0
$h3def{SST_IO_OFFSET} =            0x000000;
$h3def{SST_CMDAGP_OFFSET} = 	0x080000;
$h3def{SST_2D_OFFSET} = 		0x100000;
$h3def{SST_3D_OFFSET} = 		0x200000;
$h3def{SST_3D_ALT_OFFSET} = 	0x400000;
$h3def{SST_TEX_OFFSET} = 		0x600000;
$h3def{SST_RESERVED_OFFSET} = 	0x800000;
$h3def{SST_YUV_OFFSET} = 		0xC00000;
$h3def{SST_LFB_OFFSET} = 	        0x1000000;

# NOTE: this assumes that memBaseAddr1 is mapped such that it is
# immediately after the 32MB pointed to by memBaseAddr0
$h3def{SST_RAW_LFB_OFFSET} = 	0x2000000;

sub SST_IS_IO_ADDR	{ my($a) = @_; return ( ($a) >= $h3def{SST_IO_OFFSET}       && ($a) < $h3def{SST_CMDAGP_OFFSET} ); }
sub SST_IS_CMDAGP_ADDR	{ my($a) = @_; return ( ($a) >= $h3def{SST_CMDAGP_OFFSET}   && ($a) < $h3def{SST_2D_OFFSET} ); }
sub SST_IS_2D_ADDR	{ my($a) = @_; return ( ($a) >= $h3def{SST_2D_OFFSET}       && ($a) < $h3def{SST_3D_OFFSET} ); }
sub SST_IS_3D_ADDR	{ my($a) = @_; return ( ($a) >= $h3def{SST_3D_OFFSET}       && ($a) < $h3def{SST_3D_ALT_OFFSET} ); }
sub SST_IS_3D_ALT_ADDR	{ my($a) = @_; return ( ($a) >= $h3def{SST_3D_ALT_OFFSET}   && ($a) < $h3def{SST_TEX_OFFSET} ); }
sub SST_IS_TEX_ADDR	{ my($a) = @_; return ( ($a) >= $h3def{SST_TEX_OFFSET}      && ($a) < $h3def{SST_RESERVED_OFFSET} ); }
sub SST_IS_RESERVED_ADDR{ my($a) = @_; return ( ($a) >= $h3def{SST_RESERVED_OFFSET} && ($a) < $h3def{SST_YUV_OFFSET} ); }
sub SST_IS_YUV_ADDR	{ my($a) = @_; return ( ($a) >= $h3def{SST_YUV_OFFSET}      && ($a) < $h3def{SST_LFB_OFFSET} ); }
sub SST_IS_LFB_ADDR	{ my($a) = @_; return ( ($a) >= $h3def{SST_LFB_OFFSET}      && ($a) < $h3def{SST_RAW_LFB_OFFSET} ); }
sub SST_IS_RAW_LFB_ADDR	{ my($a) = @_; return ( ($a) >= $h3def{SST_RAW_LFB_OFFSET} ); }

sub SST_IS_REGISTER_ADDR{ my($a) = @_; return ( ($a) >= $h3def{SST_IO_OFFSET}       && ($a) < $h3def{SST_TEX_OFFSET} ); }

sub SST_BASE_ADDRESS	{ my($sst) = @_; return $sst-$h3def{SST_3D_OFFSET}; }
sub SST_IO_ADDRESS	{ my($sst) = @_; return $h3def{SST_IO_OFFSET}+SST_BASE_ADDRESS($sst); }
sub SST_CMDAGP_ADDRESS	{ my($sst) = @_; return $h3def{SST_CMDAGP_OFFSET}+SST_BASE_ADDRESS($sst); }
sub SST_GUI_ADDRESS	{ my($sst) = @_; return $h3def{SST_2D_OFFSET}+SST_BASE_ADDRESS($sst); }
sub SST_TEX_ADDRESS	{ my($sst) = @_; return $h3def{SST_TEX_OFFSET}+SST_BASE_ADDRESS($sst); }
sub SST_LFB_ADDRESS	{ my($sst) = @_; return $h3def{SST_LFB_OFFSET}+SST_BASE_ADDRESS($sst); }
sub SST_YUV_ADDRESS	{ my($sst) = @_; return $h3def{SST_YUV_OFFSET}+SST_BASE_ADDRESS($sst); }

#sub SST_IS_MOVECMD	{ my($iaddr) = @_; return ($iaddr >= ($h3def{SST_CMDAGP_OFFSET} + AGPREQSIZE) && $iaddr <= ($h3def{SST_CMDAGP_OFFSET} + MOVECMD)); }



# SET macros for FBI
#$h3def{SET_FBI(d,s)} =     SET (*(&(d)+0x100),s);
#$h3def{SET_FBIF(d,s)} =    SETF(*(&(d)+0x100),s);

# SET macros for TMU0
#$h3def{SET_0(d,s)} =       SET (*(&(d)+0x200),s);
#$h3def{SET_0F(d,s)} =      SETF(*(&(d)+0x200),s);

# SET macros for FBI+TMU0
#$h3def{SET_FBI_0(d,s)} =   SET (*(&(d)+0x300),s);
#$h3def{SET_FBI_0F(d,s)} =  SETF(*(&(d)+0x300),s);

# SET macros for TMU1
#$h3def{SET_1(d,s)} =       SET (*(&(d)+0x400),s);
#$h3def{SET_1F(d,s)} =      SETF(*(&(d)+0x400),s);

# SET macros for FBI+TMU1
#$h3def{SET_FBI_1(d,s)} =   SET (*(&(d)+0x500),s);
#$h3def{SET_FBI_1F(d,s)} =  SETF(*(&(d)+0x500),s);

################################################################################
################################################################################
##
## from $/devel/h3/incsrc/h3gdefs.h
##
################################################################################
################################################################################

#
# Copyright (c) 1997, 3Dfx Interactive, Inc.
# All Rights Reserved.
#
# This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
# the contents of this file may not be disclosed to third parties, copied or
# duplicated in any form, in whole or in part, without the prior written
# permission of 3Dfx Interactive, Inc.
#
# RESTRICTED RIGHTS LEGEND:
# Use, duplication or disclosure by the Government is subject to restrictions
# as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
# and Computer Software clause at DFARS 252.227-7013, and/or in similar or
# successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished  -
# rights reserved under the Copyright Laws of the United States.
#
# $Revision: 17 $
# $Date: 9/09/97 5:45p $
#

$h3def{SSTG_XY_SIZE} = 13;

#----------------- SSTG command bits ---------------------------
$COMMAND{SSTG_COMMAND_SHIFT} =      0;
$COMMAND{SSTG_COMMAND} =            (0xF<<$COMMAND{SSTG_COMMAND_SHIFT});
       $SSTG_COMMAND{SSTG_NOP} =                 (0<<$COMMAND{SSTG_COMMAND_SHIFT});
       $SSTG_COMMAND{SSTG_BLT} =                 (1<<$COMMAND{SSTG_COMMAND_SHIFT});
       $SSTG_COMMAND{SSTG_STRETCH_BLT} =         (2<<$COMMAND{SSTG_COMMAND_SHIFT});
       $SSTG_COMMAND{SSTG_HOST_BLT} =            (3<<$COMMAND{SSTG_COMMAND_SHIFT});
       $SSTG_COMMAND{SSTG_HOST_STRETCH_BLT} =    (4<<$COMMAND{SSTG_COMMAND_SHIFT});
       $SSTG_COMMAND{SSTG_RECTFILL} =            (5<<$COMMAND{SSTG_COMMAND_SHIFT});
       $SSTG_COMMAND{SSTG_LINE} =                (6<<$COMMAND{SSTG_COMMAND_SHIFT});
       $SSTG_COMMAND{SSTG_POLYLINE} =            (7<<$COMMAND{SSTG_COMMAND_SHIFT});
       $SSTG_COMMAND{SSTG_POLYFILL} =            (8<<$COMMAND{SSTG_COMMAND_SHIFT});
#if COLORTRANSLUT
$COMMAND{SSTG_EN_CLUT88} =          BIT(7);
#endif
$COMMAND{SSTG_GO} =                 BIT(8);
$COMMAND{SSTG_REVERSIBLE} =         BIT(9);
$COMMAND{SSTG_UPDATE_DSTX} =        BIT(10);
$COMMAND{SSTG_UPDATE_DSTY} =        BIT(11);
$COMMAND{SSTG_EN_LINESTIPPLE} =     BIT(12);
$COMMAND{SSTG_MONO_PATTERN} =       BIT(13);
$COMMAND{SSTG_XDIR} =               BIT(14);
$COMMAND{SSTG_YDIR} =               BIT(15);
$COMMAND{SSTG_TRANSPARENT} =        BIT(16);
$COMMAND{SSTG_X_PATOFFSET_SHIFT} =	17;
$COMMAND{SSTG_X_PATOFFSET} =		(0x7<<$COMMAND{SSTG_X_PATOFFSET_SHIFT});
$COMMAND{SSTG_Y_PATOFFSET_SHIFT} =	20;
$COMMAND{SSTG_Y_PATOFFSET} =		(0x7<<$COMMAND{SSTG_Y_PATOFFSET_SHIFT});
$COMMAND{SSTG_CLIPSELECT} =         BIT(23);
$COMMAND{SSTG_ROP0_SHIFT} =		24;
$COMMAND{SSTG_ROP0} =		(0xFF<<$COMMAND{SSTG_ROP0_SHIFT});

#----------------- $h3def{SSTG} commandEx bits ---------------------------
$COMMANDEX{SSTG_EN_SRC_COLORKEY_EX} =	BIT(0);
$COMMANDEX{SSTG_EN_DST_COLORKEY_EX} =	BIT(1);
$COMMANDEX{SSTG_WAIT_FOR_VSYNC_EX} =	BIT(2);
$COMMANDEX{SSTG_PAT_FORCE_ROW0} =	BIT(3);

#----------------- $h3def{SSTG} src,dstBaseAddr bits ---------------------------
$h3def{SSTG_IS_TILED} =	BIT(31);
$h3def{SSTG_BASEADDR_SHIFT} = 0;
$h3def{SSTG_BASEADDR} = (0xFFFFFF << $h3def{SSTG_BASEADDR_SHIFT});

#----------------- $h3def{SSTG} srcFormat bits ---------------------------
$SRCFORMAT{SSTG_SRC_STRIDE_SHIFT} =	0;
	$SRCFORMAT{SSTG_SRC_LINEAR_STRIDE} =	(0x3FFF << $SRCFORMAT{SSTG_SRC_STRIDE_SHIFT});
	$SRCFORMAT{SSTG_SRC_TILE_STRIDE} =	(0x7F << $SRCFORMAT{SSTG_SRC_STRIDE_SHIFT});
$SRCFORMAT{SSTG_SRC_FORMAT_SHIFT} =	16;
$SRCFORMAT{SSTG_SRC_FORMAT} =	(0xF << $SRCFORMAT{SSTG_SRC_FORMAT_SHIFT});
	$SSTG_SRC_FORMAT{SSTG_PIXFMT_1BPP} =	(0x0 << $SRCFORMAT{SSTG_SRC_FORMAT_SHIFT});
	$SSTG_SRC_FORMAT{SSTG_PIXFMT_8BPP} =	(0x1 << $SRCFORMAT{SSTG_SRC_FORMAT_SHIFT});
	$SSTG_SRC_FORMAT{SSTG_PIXFMT_15BPP} =	(0x2 << $SRCFORMAT{SSTG_SRC_FORMAT_SHIFT});
	$SSTG_SRC_FORMAT{SSTG_PIXFMT_16BPP} =	(0x3 << $SRCFORMAT{SSTG_SRC_FORMAT_SHIFT});
	$SSTG_SRC_FORMAT{SSTG_PIXFMT_24BPP} =	(0x4 << $SRCFORMAT{SSTG_SRC_FORMAT_SHIFT});
	$SSTG_SRC_FORMAT{SSTG_PIXFMT_32BPP} =	(0x5 << $SRCFORMAT{SSTG_SRC_FORMAT_SHIFT});
	$SSTG_SRC_FORMAT{SSTG_PIXFMT_422YUV} =	(0x8 << $SRCFORMAT{SSTG_SRC_FORMAT_SHIFT});
	$SSTG_SRC_FORMAT{SSTG_PIXFMT_422UYV} =	(0x9 << $SRCFORMAT{SSTG_SRC_FORMAT_SHIFT});
$SRCFORMAT{SSTG_HOST_BYTE_SWIZZLE} =	BIT(20);
$SRCFORMAT{SSTG_HOST_WORD_SWIZZLE} =	BIT(21);
$SRCFORMAT{SSTG_SRC_PACK_SHIFT} =	22;
$SRCFORMAT{SSTG_SRC_PACK} =	(0x3 << $SRCFORMAT{SSTG_SRC_PACK_SHIFT});
	$SSTG_SRC_PACK{SSTG_SRC_PACK_SRC} = (0x0 << $SRCFORMAT{SSTG_SRC_PACK_SHIFT});
	$SSTG_SRC_PACK{SSTG_SRC_PACK_8} =	(0x1 << $SRCFORMAT{SSTG_SRC_PACK_SHIFT});
	$SSTG_SRC_PACK{SSTG_SRC_PACK_16} =	(0x2 << $SRCFORMAT{SSTG_SRC_PACK_SHIFT});
	$SSTG_SRC_PACK{SSTG_SRC_PACK_32} =	(0x3 << $SRCFORMAT{SSTG_SRC_PACK_SHIFT});

#----------------- $h3def{SSTG} dstFormat bits ---------------------------
$DSTFORMAT{SSTG_DST_STRIDE_SHIFT} =	0;
	$DSTFORMAT{SSTG_DST_LINEAR_STRIDE} =	(0x3FFF << $DSTFORMAT{SSTG_DST_STRIDE_SHIFT});
	$DSTFORMAT{SSTG_DST_TILE_STRIDE} =	(0x7F << $DSTFORMAT{SSTG_DST_STRIDE_SHIFT});
$DSTFORMAT{SSTG_DST_FORMAT_SHIFT} =	16;
$DSTFORMAT{SSTG_DST_FORMAT} =	(0x7 << $DSTFORMAT{SSTG_DST_FORMAT_SHIFT});

	$SSTG_DST_FORMAT{SSTG_PIXFMT_1BPP} =	(0x0 << $DSTFORMAT{SSTG_SRC_FORMAT_SHIFT});
	$SSTG_DST_FORMAT{SSTG_PIXFMT_8BPP} =	(0x1 << $DSTFORMAT{SSTG_SRC_FORMAT_SHIFT});
	$SSTG_DST_FORMAT{SSTG_PIXFMT_15BPP} =	(0x2 << $DSTFORMAT{SSTG_SRC_FORMAT_SHIFT});
	$SSTG_DST_FORMAT{SSTG_PIXFMT_16BPP} =	(0x3 << $DSTFORMAT{SSTG_SRC_FORMAT_SHIFT});
	$SSTG_DST_FORMAT{SSTG_PIXFMT_24BPP} =	(0x4 << $DSTFORMAT{SSTG_SRC_FORMAT_SHIFT});
	$SSTG_DST_FORMAT{SSTG_PIXFMT_32BPP} =	(0x5 << $DSTFORMAT{SSTG_SRC_FORMAT_SHIFT});
	$SSTG_DST_FORMAT{SSTG_PIXFMT_422YUV} =	(0x8 << $DSTFORMAT{SSTG_SRC_FORMAT_SHIFT});
	$SSTG_DST_FORMAT{SSTG_PIXFMT_422UYV} =	(0x9 << $DSTFORMAT{SSTG_SRC_FORMAT_SHIFT});
#ifdef YUV411
#411 has been removed from the spec as a blit source format -KMW
#$h3def{SSTG_PIXFMT_411YUV} =	(0xA << $h3def{SSTG_SRC_FORMAT_SHIFT});
#endif

if ( $SRCFORMAT{SSTG_SRC_FORMAT_SHIFT} != $DSTFORMAT{SSTG_DST_FORMAT_SHIFT} ) {
    die "SSTG_SRC_FORMAT_SHIFT and SSTG_DST_FORMAT_SHIFT are not equal";
}

#----------------- $h3def{SSTG} lineStyle bits ---------------------------
$LINESTYLE{SSTG_LSREPEAT_SHIFT} =	0;
$LINESTYLE{SSTG_LSREPEAT} =		(0xFF << $LINESTYLE{SSTG_LSREPEAT_SHIFT});
$LINESTYLE{SSTG_LSSIZE_SHIFT} =	8;
$LINESTYLE{SSTG_LSSIZE} =		(0x1F << $LINESTYLE{SSTG_LSSIZE_SHIFT});
$LINESTYLE{SSTG_LSPOS_FRAC_SHIFT} =	16;
$LINESTYLE{SSTG_LSPOS_FRAC} =		(0xFF << $LINESTYLE{SSTG_LSPOS_FRAC_SHIFT});
$LINESTYLE{SSTG_LSPOS_INT_SHIFT} =	24;
$LINESTYLE{SSTG_LSPOS_INT} =		(0x1F << $LINESTYLE{SSTG_LSPOS_INT_SHIFT});

#----------------- $h3def{SSTG} rop defines ---------------------------
$h3def{SSTG_ROP_ZERO} =	0x00;
$h3def{SSTG_ROP_NOR} =	0x11;
$h3def{SSTG_ROP_ANDI} =	0x22;
$h3def{SSTG_ROP_NSRC} =	0x33;
$h3def{SSTG_ROP_ANDR} =	0x44;
$h3def{SSTG_ROP_NDST} =	0x55;
$h3def{SSTG_ROP_XOR} =	0x66;
$h3def{SSTG_ROP_NAND} =	0x77;
$h3def{SSTG_ROP_AND} =	0x88;
$h3def{SSTG_ROP_XNOR} =	0x99;
$h3def{SSTG_ROP_DST} =	0xaa;
$h3def{SSTG_ROP_ORI} =	0xbb;
$h3def{SSTG_ROP_SRC} =	0xcc;
$h3def{SSTG_ROP_ORR} =	0xdd;
$h3def{SSTG_ROP_OR} =	0xee;
$h3def{SSTG_ROP_ONE} =	0xff;

# $h3def{SSTG} ternary rop defines (named rop3's that use pattern)
$h3def{SSTG_ROP_PATINVERT} =	0x5a;
$h3def{SSTG_ROP_MERGECOPY} =	0xC0;
$h3def{SSTG_ROP_PATCOPY} =	0xF0;
$h3def{SSTG_ROP_PATPAINT} =	0xFB;

sub SSTG_ISBINARYROP	{ my($rop) = @_; return (($rop>>4)&0xF) == ($rop&0xF); }


#-----------------------  CMDFIFO Section ---------------------------
#-----------------------  CMDFIFO Section ---------------------------
#-----------------------  CMDFIFO Section ---------------------------

#----------------- $h3def{SST} cmdFifo*.baseSize bits ---------------------------
$CMD0_BASESIZE{SST_CMDFIFO_SIZE} =		0xFF;
$CMD0_BASESIZE{SST_EN_CMDFIFO} =			BIT(8);
$CMD0_BASESIZE{SST_CMDFIFO_AGP} =			BIT(9);
$CMD0_BASESIZE{SST_CMDFIFO_DISABLE_HOLES} =	BIT(10);

$CMD1_BASESIZE{SST_CMDFIFO_SIZE} =		0xFF;
$CMD1_BASESIZE{SST_EN_CMDFIFO} =			BIT(8);
$CMD1_BASESIZE{SST_CMDFIFO_AGP} =			BIT(9);
$CMD1_BASESIZE{SST_CMDFIFO_DISABLE_HOLES} =	BIT(10);

# $h3def{SST} COMMAND PACKET defines
$h3def{SSTCP_PKT_SIZE} =	3;
$h3def{SSTCP_PKT} =	SST_MASK($h3def{SSTCP_PKT_SIZE});
$h3def{SSTCP_PKT0} =	0;
$h3def{SSTCP_PKT1} =	1;
$h3def{SSTCP_PKT2} =	2;
$h3def{SSTCP_PKT3} =	3;
$h3def{SSTCP_PKT4} =	4;
$h3def{SSTCP_PKT5} =	5;
$h3def{SSTCP_PKT6} =	6;
$h3def{SSTCP_PKT7} =	7;

$h3def{SSTCP_BOGUS_WORDS_SHIFT} =	29;
$h3def{SSTCP_BOGUS_WORDS} =	(7 << $h3def{SSTCP_BOGUS_WORDS_SHIFT});

# packet 0 defines
$h3def{SSTCP_PKT0_FUNC_SHIFT} =	3;
$h3def{SSTCP_PKT0_FUNC} =		(7 << $h3def{SSTCP_PKT0_FUNC_SHIFT});
$h3def{SSTCP_PKT0_NOP} =		((0<<$h3def{SSTCP_PKT0_FUNC_SHIFT}) | $h3def{SSTCP_PKT0});
$h3def{SSTCP_PKT0_JSR} =		((1<<$h3def{SSTCP_PKT0_FUNC_SHIFT}) | $h3def{SSTCP_PKT0});
$h3def{SSTCP_PKT0_RET} =		((2<<$h3def{SSTCP_PKT0_FUNC_SHIFT}) | $h3def{SSTCP_PKT0});
$h3def{SSTCP_PKT0_JMP_LOCAL} =	((3<<$h3def{SSTCP_PKT0_FUNC_SHIFT}) | $h3def{SSTCP_PKT0});
$h3def{SSTCP_PKT0_JMP_AGP} =	((4<<$h3def{SSTCP_PKT0_FUNC_SHIFT}) | $h3def{SSTCP_PKT0});
$h3def{SSTCP_PKT0_ADDR_SHIFT} =	6;
$h3def{SSTCP_PKT0_ADDR} =		(0x7FFFFF<<$h3def{SSTCP_PKT0_ADDR_SHIFT});

# packet 1 defines
$h3def{SSTCP_REGBASE_SHIFT} =	$h3def{SSTCP_PKT_SIZE};
$h3def{SSTCP_REGBASE} =		((0x3FF)<<$h3def{SSTCP_REGBASE_SHIFT});
$h3def{SSTCP_PKT1_2D} =		BIT(14);
$h3def{SSTCP_INC} =		BIT(15);
$h3def{SSTCP_PKT1_NWORDS_SHIFT} =	16;
$h3def{SSTCP_PKT1_NWORDS} =	(0xFFFF<<$h3def{SSTCP_PKT1_NWORDS_SHIFT});

# packet 2 defines
$h3def{SSTCP_PKT2_MASK_SHIFT} =	$h3def{SSTCP_PKT_SIZE};
$h3def{SSTCP_PKT2_MASK} =		(0x1FFFFFFF<<$h3def{SSTCP_PKT2_MASK_SHIFT});

# packet 3 defines
$h3def{SSTCP_PKT3_CMD_SHIFT} =	$h3def{SSTCP_PKT_SIZE};
$h3def{SSTCP_PKT3_CMD} =		(0x7<<$h3def{SSTCP_PKT3_CMD_SHIFT});
  $h3def{SSTCP_PKT3_BDDBDD} =	(0<<$h3def{SSTCP_PKT3_CMD_SHIFT});
  $h3def{SSTCP_PKT3_BDDDDD} =	(1<<$h3def{SSTCP_PKT3_CMD_SHIFT});
  $h3def{SSTCP_PKT3_DDDDDD} =	(2<<$h3def{SSTCP_PKT3_CMD_SHIFT});
$h3def{SSTCP_PKT3_NUMVERTEX_SHIFT} = ($h3def{SSTCP_PKT_SIZE}+3);
$h3def{SSTCP_PKT3_NUMVERTEX} =	(0xF << $h3def{SSTCP_PKT3_NUMVERTEX_SHIFT});
$h3def{SSTCP_PKT3_PMASK_SHIFT} =	($h3def{SSTCP_PKT_SIZE}+3+4);
$h3def{SSTCP_PKT3_PMASK} =	(0xFFF<<$h3def{SSTCP_PKT3_PMASK_SHIFT});
$h3def{SSTCP_PKT3_SMODE_SHIFT} =	($h3def{SSTCP_PKT3_PMASK_SHIFT}+12);
$h3def{SSTCP_PKT3_SMODE} =	(0x3F<<$h3def{SSTCP_PKT3_SMODE_SHIFT});
$h3def{SSTCP_PKT3_PACKEDCOLOR} =	BIT(28);

# packet 4 defines
$h3def{SSTCP_REGBASE_SHIFT} =	$h3def{SSTCP_PKT_SIZE};
$h3def{SSTCP_REGBASE} =		((0x3FF)<<$h3def{SSTCP_REGBASE_SHIFT});
$h3def{SSTCP_PKT4_2D} =		BIT(14);
$h3def{SSTCP_PKT4_MASK_SHIFT} =	15;
$h3def{SSTCP_PKT4_MASK} =		(0x3FFF<<$h3def{SSTCP_PKT4_MASK_SHIFT});

# packet 5 defines
$h3def{SSTCP_PKT5_NWORDS_SHIFT} =	3;
$h3def{SSTCP_PKT5_NWORDS} =	(0x7FFFF<<$h3def{SSTCP_PKT5_NWORDS_SHIFT});
$h3def{SSTCP_PKT5_BYTEN_WN_SHIFT} = 22;
$h3def{SSTCP_PKT5_BYTEN_WN} =	(0xF<<$h3def{SSTCP_PKT5_BYTEN_WN_SHIFT});
$h3def{SSTCP_PKT5_BYTEN_W2_SHIFT} = 26;
$h3def{SSTCP_PKT5_BYTEN_W2} =	(0xF<<$h3def{SSTCP_PKT5_BYTEN_W2_SHIFT});
$h3def{SSTCP_PKT5_SPACE_SHIFT} =	30;
$h3def{SSTCP_PKT5_SPACE} =	(0x3<<$h3def{SSTCP_PKT5_SPACE_SHIFT});
$h3def{SSTCP_PKT5_LFB} =		(0x0<<$h3def{SSTCP_PKT5_SPACE_SHIFT});
$h3def{SSTCP_PKT5_YUV} =		(0x1<<$h3def{SSTCP_PKT5_SPACE_SHIFT});
$h3def{SSTCP_PKT5_3DLFB} =	(0x2<<$h3def{SSTCP_PKT5_SPACE_SHIFT});
$h3def{SSTCP_PKT5_TEXPORT} =	(0x3<<$h3def{SSTCP_PKT5_SPACE_SHIFT});
$h3def{SSTCP_PKT5_BASEADDR} =	0x1FFFFFF;

# packet 6 defines
$h3def{SSTCP_PKT6_SPACE_SHIFT} =  3;
$h3def{SSTCP_PKT6_SPACE} =        (0x3<<$h3def{SSTCP_PKT6_SPACE_SHIFT});
$h3def{SSTCP_PKT6_LFB} =          (0x0<<$h3def{SSTCP_PKT6_SPACE_SHIFT});
$h3def{SSTCP_PKT6_YUV} =          (0x1<<$h3def{SSTCP_PKT6_SPACE_SHIFT});
$h3def{SSTCP_PKT6_3DLFB} =        (0x2<<$h3def{SSTCP_PKT6_SPACE_SHIFT});
$h3def{SSTCP_PKT6_TEXPORT} =      (0x3<<$h3def{SSTCP_PKT6_SPACE_SHIFT});
$h3def{SSTCP_PKT6_NBYTES_SHIFT} = 5;
$h3def{SSTCP_PKT6_NBYTES} =       (SST_MASK(20) << $h3def{SSTCP_PKT6_NBYTES_SHIFT});
$h3def{SSTCP_PKT6_SRC_BASELOW} =            SST_MASK(32);
$h3def{SSTCP_PKT6_SRC_WIDTH} =              SST_MASK(14);
$h3def{SSTCP_PKT6_SRC_STRIDE_SHIFT} =       14;
$h3def{SSTCP_PKT6_SRC_STRIDE} =             (SST_MASK(14) <<$h3def{SSTCP_PKT6_SRC_STRIDE_SHIFT});
$h3def{SSTCP_PKT6_SRC_BASEHIGH_SHIFT} =     28;
$h3def{SSTCP_PKT6_SRC_BASEHIGH} =           (SST_MASK(4) <<$h3def{SSTCP_PKT6_SRC_BASEHIGH_SHIFT});
$h3def{SSTCP_PKT6_FRAME_BUFFER_OFFSET} =    SST_MASK(26);
$h3def{SSTCP_PKT6_DST_STRIDE} =             SST_MASK(15);


# move command
$h3def{SST_AGP_REQUEST_SIZE} =        SST_MASK(20);
$h3def{SST_AGP_MOVE_BASELOW} =        $h3def{SSTCP_PKT6_SRC_BASELOW};
$h3def{SST_AGP_SRC_WIDTH} =           $h3def{SSTCP_PKT6_SRC_WIDTH};
$h3def{SST_AGP_SRC_STRIDE_SHIFT} =    $h3def{SSTCP_PKT6_SRC_STRIDE_SHIFT};
$h3def{SST_AGP_SRC_STRIDE} =          $h3def{SSTCP_PKT6_SRC_STRIDE};
$h3def{SST_AGP_SRC_BASEHIGH_SHIFT} =  $h3def{SSTCP_PKT6_SRC_BASEHIGH_SHIFT};
$h3def{SST_AGP_SRC_BASEHIGH} =        $h3def{SSTCP_PKT6_SRC_BASEHIGH};
$h3def{SST_AGP_FRAME_BUFFER_OFFSET} = $h3def{SSTCP_PKT6_FRAME_BUFFER_OFFSET};
$h3def{SST_AGP_DSTSTRIDE} =           $h3def{SSTCP_PKT6_DST_STRIDE};

$h3def{SST_AGPMOVE_SPACE_SHIFT} =      $h3def{SSTCP_PKT6_SPACE_SHIFT};
$h3def{SST_AGPMOVE_SPACE} =            (0x3<<$h3def{SST_AGPMOVE_SPACE_SHIFT});
$h3def{SST_AGPMOVE_LFB} =              (0x0<<$h3def{SST_AGPMOVE_SPACE_SHIFT});
$h3def{SST_AGPMOVE_YUV} =              (0x1<<$h3def{SST_AGPMOVE_SPACE_SHIFT});
$h3def{SST_AGPMOVE_3DLFB} =            (0x2<<$h3def{SST_AGPMOVE_SPACE_SHIFT});
$h3def{SST_AGPMOVE_TEXPORT} =          (0x3<<$h3def{SST_AGPMOVE_SPACE_SHIFT});
$h3def{SST_AGPMOVE_CMDID} =            BIT(5);
$h3def{SST_AGPMOVE_CMDID_SHIFT} =	     5;
$h3def{SST_AGPMOVE_CMDID0} =           0;
$h3def{SST_AGPMOVE_CMDID1} =           BIT(5);

$h3def{SSTCP_SPACE} =                  0x3;
$h3def{SSTCP_LFB_SPACE} =              0x0;
$h3def{SSTCP_YUV_SPACE} =              0x1;
$h3def{SSTCP_3DLFB_SPACE} =            0x2;
$h3def{SSTCP_TEXPORT_SPACE} =          0x3;

################################################################################
################################################################################
##
## from common/src/tests/diags/h3cfg.h
##
################################################################################
################################################################################

$h3def{CAPID_SHIFT} =     0;
$h3def{CAPID_MASK} =      0xFF;
$h3def{NEXTPTR_SHIFT} =    8;
$h3def{NEXTPTR_MASK} =     (0xFF << $h3def{NEXTPTR_SHIFT});

# bit definitions of config space registers
$h3def{PCISTATUS_CAPLIST} =        BIT(4);

$h3def{AGP_ID} =                   0x2;

$h3def{AGPSTATUS_RATE} =           0x3;
$h3def{AGPSTATUS_RES0} =           (SST_MASK(3) << 2);
$h3def{AGPSTATUS_4G} =             BIT(5);
$h3def{AGPSTATUS_RES1} =           (SST_MASK(3) << 6);
$h3def{AGPSTATUS_SBA} =            BIT(9);
$h3def{AGPSTATUS_RES2} =           (SST_MASK(14) << 10);
$h3def{AGPSTATUS_RQDEPTH_SHIFT} =  24;
$h3def{AGPSTATUS_RQDEPTH} =        (SST_MASK(8) << $h3def{AGPSTATUS_RQDEPTH_SHIFT});

$h3def{AGPCMD_RATE} =              0x7;
$h3def{AGPCMD_RES0} =              (SST_MASK(3) << 2);
$h3def{AGPCMD_4G} =             BIT(5);
$h3def{AGPCMD_RES1} =              (SST_MASK(2) << 6);
$h3def{AGPCMD_ENABLE} =            BIT(8);
$h3def{AGPCMD_SBAENABLE} =         BIT(9);
$h3def{AGPCMD_RES2} =              (SST_MASK(14) << 10);
$h3def{AGPCMD_RQDEPTH_SHIFT} =     24;
$h3def{AGPCMD_RQDEPTH} =           (SST_MASK(8) << $h3def{AGPCMD_RQDEPTH_SHIFT});

# PCISTATUS

$PCISTATUS{PCISTATUS_NEW_CAP}             = BIT(4);
$PCISTATUS{PCISTATUS_66_CAP}              = BIT(5);
$PCISTATUS{PCISTATUS_UDF_SUPPORT}         = BIT(6);
$PCISTATUS{PCISTATUS_FASTBACK_CAP}        = BIT(7);
$PCISTATUS{PCISTATUS_DATA_PARITY_RPT}     = BIT(8);
$PCISTATUS{PCISTATUS_DEV_SEL_TIM_SHIFT}   = 9;
$PCISTATUS{PCISTATUS_DEV_SEL_TIM}         = (SST_MASK(2) << $PCISTATUS{PCISTATUS_DEV_SEL_TIM_SHIFT});
$PCISTATUS{PCISTATUS_SENT_TARGET_ABORT}   = BIT(11);
$PCISTATUS{PCISTATUS_RCVD_TARGET_ABORT}   = BIT(12);
$PCISTATUS{PCISTATUS_RCVD_MASTER_ABORT}   = BIT(13);
$PCISTATUS{PCISTATUS_SENT_SYSTEM_ERROR}   = BIT(14);
$PCISTATUS{PCISTATUS_DETECT_PARITY_ERROR} = BIT(15);

################################################################################
################################################################################
##
## other
##
################################################################################
################################################################################

$h3def{SST_PLL_K_SHIFT} = 0;
$h3def{SST_PLL_K} = (0x3<<$h3def{SST_PLL_K_SHIFT});
$h3def{SST_PLL_M_SHIFT} = 2;
$h3def{SST_PLL_M} = (0x3F<<$h3def{SST_PLL_M_SHIFT});
$h3def{SST_PLL_N_SHIFT} = 8;
$h3def{SST_PLL_N} = (0xFF<<$h3def{SST_PLL_N_SHIFT});
$h3def{SST_PLL_TEST} = BIT(16);

################################################################################
################################################################################
################################################################################

%h3field = (
	    LFBMODE		=>	{ %LFBMODE },
	    STATUS		=>	{ %STATUS },
	    TEXTUREMODE		=>	{ %TEXTUREMODE },
	    TLOD		=>	{ %TLOD },
	    TDETAIL		=>	{ %TDETAIL },
	    FBZMODE		=>	{ %FBZMODE },
	    FBZCOLORPATH	=>	{ %FBZCOLORPATH },
	    FOGMODE		=>	{ %FOGMODE },
	    ALPHACOLR		=>	{ %ALPHAMODE },
	    ZACOLOR		=>	{ %ZACOLOR },
	    CHROMARANCE		=>	{ %CHROMARANGE },
	    CLIPLEFTRIGHT	=>	{ %CLIPLEFTRIGHT },
	    CLIPBOTTOMTOP	=>	{ %CLIPBOTTOMTOP },
	    CLIPLEFTRIGHT1	=>	{ %CLIPLEFTRIGHT1 },
	    CLIPBOTTOMTOP1	=>	{ %CLIPBOTTOMTOP1 },
	    SSETUPMODE		=>	{ %SSETUPMODE },
	    AUXBUFFERADDR	=>	{ %AUXBUFFERADDR },
	    AUXBUFFERSTRIDE	=>	{ %AUXBUFFERSTRIDE },
	    COLBUFFERADDR	=>	{ %COLBUFFERADDR },
	    COLBUFFERSTRIDE	=>	{ %COLBUFFERSTRIDE },
	    TEXBASEADDR		=>	{ %TEXBASEADDR },
	    VIDPROCCFG		=>	{ %VIDPROCCFG },
	    VIDINFORMAT         =>      { %VIDINFORMAT },
	    VIDINSTATUSCURRENTLINE =>   { %VIDINSTATUSCURRENTLINE },
	    VIDSERIALPARALLELPORT =>    { %VIDSERIALPARALLELPORT },
	    PCIINIT0		=>	{ %PCIINIT0 },
	    LFBMEMORYCONFIG	=>	{ %LFBMEMORYCONFIG },
	    MISCINIT0		=>	{ %MISCINIT0 },
	    MISCINIT1		=>	{ %MISCINIT1 },
	    DRAMINIT0		=>	{ %DRAMINIT0 },
	    DRAMINIT1		=>	{ %DRAMINIT1 },
	    AGPINIT0		=>	{ %AGPINIT0 },
	    TMUGBEINIT		=>	{ %TMUGBEINIT },
	    FASTFILLCMD		=>	{ %FASTFILLCMD },
	    SWAPBUFFERCMD	=>	{ %SWAPBUFFERCMD },
	    COMMAND		=>	{ %COMMAND },
	    COMMANDEX		=>	{ %COMMANDEX },
	    SRCFORMAT		=>	{ %SRCFORMAT },
	    DSTFORMAT		=>	{ %DSTFORMAT },
	    LINESTYLE		=>	{ %LINESTYLE },
	    CMD0_BASESIZE	=>	{ %CMD0_BASESIZE },
	    CMD1_BASESIZE	=>	{ %CMD1_BASESIZE },
	    PCISTATUS           =>      { %PCISTATUS }
	    );

%h3fieldv = (
	     SST_LFB_FORMAT		=>	{ %SST_LFB_FORMAT },
	     SST_LFB_READBUFSELECT	=>	{ %SST_LFB_READBUFSELECT },
	     SST_LFB_RGBALANES		=>	{ %SST_LFB_RGBALANES},
	     SST_TFORMAT		=>	{ %SST_TFORMAT },
	     SST_TC_MSELECT		=>	{ %SST_TC_MSELECT },
	     SST_TCA_MSELECT		=>	{ %SST_TCA_MSELECT },
	     SST_ZFUNC			=>	{ %SST_ZFUNC },
	     SST_RGBSELECT		=>	{ %SST_RGBSELECT },
	     SST_ASELECT		=>	{ %SST_ASELECT },
	     SST_CC_MSELECT		=>	{ %SST_CC_MSELECT },
	     SST_CCA_MSELECT		=>	{ %SST_CCA_MSELECT },
	     SST_ALPHAFUNC		=>	{ %SST_ALPHAFUNC },
	     SST_CURSOR_MODE		=>	{ %SST_CURSOR_MODE },
	     SST_OVERLAY_FILTER_MODE	=>	{ %SST_OVERLAY_FILTER_MODE },
	     SST_DESKTOP_PIXEL_FORMAT	=>	{ %SST_DESKTOP_PIXEL_FORMAT },
	     SST_OVERLAY_PIXEL_FORMAT	=>	{ %SST_OVERLAY_PIXEL_FORMAT },
	     SST_VIDEOIN_PIXEL_FORMAT	=>	{ %SST_VIDEOIN_PIXEL_FORMAT },
	     SST_VIDEOIN_BUFFERING_MODE	=>	{ %SST_VIDEOIN_BUFFERING_MODE },
	     SST_VIDEOIN_INTERFACE	=>	{ %SST_VIDEOIN_INTERFACE },
	     SST_VIDEOIN_GENLOCK_SOURCE =>      { %SST_VIDEOIN_GENLOCK_SOURCE },
	     SST_RAW_LFB_ADDR_STRIDE	=>	{ %SST_RAW_LFB_ADDR_STRIDE },
	     SST_SGRAM_TYPE		=>	{ %SST_SGRAM_TYPE },
	     SSTG_COMMAND		=>	{ %SSTG_COMMAND },
	     SSTG_SRC_FORMAT		=>	{ %SSTG_SRC_FORMAT },
	     SSTG_SRC_PACK		=>	{ %SSTG_SRC_PACK },
	     SSTG_DST_FORMAT		=>	{ %SSTG_DST_FORMAT },
);

# load field definitions into %h3def
foreach $reg ( keys %h3field ) {
    addToDefs(\%{ $h3field{$reg} });
}

# load field-value definitions into %h3def
foreach $reg ( keys %h3fieldv ) {
    addToDefs(\%{ $h3fieldv{$reg} });
}

#
# addToDefs(\%href)
#
# load a hash table's entries into %h3def
#
sub addToDefs {
    my($href) = @_;
    foreach $key ( keys %$href ) {
	$h3def{$key} = $href->{$key};
#	printf "adding %s => 0x%x\n",$key,$h3def{$key};
    }
}

#
# PARSE($regname,$data)
#
# parse $data into the fields defined for $regname and display the results
#
sub PARSE {
    my($regname,$data) = @_;
    my($name, $val, $mask, $shift, $found, $maskall);
    my(%lines, $aline, $k);
    my(%h);
    
    %h = %{$h3field{$regname}};
    printf "%s = 0x%x\n",$regname,$data;

    if ( exists $h3field{$regname} ) {
	$maskall = 0;

	# Loop thru fields, sorted from lsb's to msb's. This sort sub is
	# numerically sorting the keys based on their values (which are bit masks).
	foreach $name ( sort {$h3field{$regname}{$a} <=> $h3field{$regname}{$b} } keys %{$h3field{$regname}}  ) {

	    # mask off field
	    $mask = $h3field{$regname}{$name};
	    $maskall |= $mask;
	    $val = $data & $mask;

	    # shift field into least significant bits
	    for ( $shift=0; $shift<32; $shift++ ) {
		last if ( ($mask>>$shift) & 0x1 );
	    }
	    $val >>= $shift;
	    $mask >>= $shift;

	    # display all multibit fields and nonzero single-bit fields
	    if ( ($val || $mask>1) && $name !~ /SHIFT/ ) {
		printf "\t%-30s\t0x%x",$name,$val;
		if ( exists $h3fieldv{$name} ) {
		    $found = 0;
		    # loop thru names for field values
		    while ( ($token,$tval) = each %{$h3fieldv{$name}} ) {
			( $found = 1, print " ($token)" ) if ( ($tval>>$shift) == $val );
		    }
		    print " (undefined)" if (!$found);
		}
		print "\n";
	    }
	}
	printf "\t*** invalid bits = 0x%x ***\n",$data&(~$maskall) if ($data&(~$maskall));
    }
}


1;
